/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the appearance of an object.
 * <p>
 *
 * Attribute classes are used to control various visual details about the
 * object being rendered. Although a single instance of this class can take
 * all the attribute types at once, you'll need to consider what this object
 * is being used for. Most of the time you should only ever need to use one of
 * Point/Line/Polygon attributes.
 * <p>
 *
 * <b>Note:</b> If you have semi-transparent objects that need blending, you
 * do <i>not</i> need provide an instance of
 * {@link org.j3d.aviatrix3d.BlendAttributes}. Aviatrix3D will
 * internally handle the correct settings that you need, so there is no need
 * to provide your own. Only use it when you want to blend objects with
 * something different than the standard blending setup for non-transparent
 * and transparent objects as it will remove this object from the normal
 * transparency sorting processing and into a separate state.
 *
 * @author Alan Hudson, Justin Couch
 * @version $Revision: 1.30.2.1 $
 */
public class Appearance extends NodeComponent
    implements RenderableObject
{
    /** Listing of 32 texture IDs as Integers for the render() calls */
    private static final Integer[] TEX_IDS;

    /** Texture ID for only a single texture */
    private static final Integer SINGLE_TEXTURE;

    /** The material properites */
    private Material material;

    /** The texture properties */
    private TextureUnit[] textureUnits;

    /** Current number of valid textures */
    private int numTextures;

    /** The shader used by the appearance */
    private Shader shader;

    /** Attributes for rendering polygons */
    private PolygonAttributes polyAttr;

    /** Attributes for rendering lines */
    private LineAttributes lineAttr;

    /** Attributes for rendering points */
    private PointAttributes pointAttr;

    /** Attributes for blend management */
    private BlendAttributes blendAttr;

    /** Visibility flag for this object. If not visible, it's not rendered */
    private boolean visible;

    /**
     * Static initialiser creates the texture IDs
     */
    static
    {
        TEX_IDS = new Integer[32];
        for(int i = 0; i < 32; i++)
            TEX_IDS[i] = new Integer(GL.GL_TEXTURE0 + i);

        SINGLE_TEXTURE = new Integer(-1);
    }

    /**
     * The default constructor.
     */
    public Appearance()
    {
        visible = true;
    }

    //---------------------------------------------------------------
    // Methods defined by SceneGraphObject
    //---------------------------------------------------------------

    /**
     * Set the scenegraph update handler for this node.  It will notify
     * all its children of the value. A null value will clear the current
     * handler.
     *
     * @param handler The instance to use as a handler
     */
    protected void setUpdateHandler(NodeUpdateHandler handler)
    {
        super.setUpdateHandler(handler);

        if(material != null)
            material.setUpdateHandler(handler);

        if(blendAttr != null)
            blendAttr.setUpdateHandler(handler);

        if(polyAttr != null)
            polyAttr.setUpdateHandler(handler);

        if(lineAttr != null)
            lineAttr.setUpdateHandler(handler);

        if(pointAttr != null)
            pointAttr.setUpdateHandler(handler);

        for(int i = 0; i < numTextures; i++)
            textureUnits[i].setUpdateHandler(handler);

        if(shader != null)
            shader.setUpdateHandler(updateHandler);
    }

    /**
     * Notification that this object is live now. Overridden to make sure that
     * the live state of the nodes represents the same state as the parent
     * scene graph.
     *
     * @param state true if this should be marked as live now
     */
    protected void setLive(boolean state)
    {
        if(state)
            liveCount++;
        else if(liveCount > 0)
            liveCount--;

        if((liveCount == 0) || !alive)
        {
            super.setLive(state);

            if(numTextures != 0)
            {
                for(int i = 0; i < numTextures; i++)
                    textureUnits[i].setLive(state);
            }

            if(polyAttr != null)
                polyAttr.setLive(state);

            if(lineAttr != null)
                lineAttr.setLive(state);

            if(pointAttr != null)
                pointAttr.setLive(state);

            if(blendAttr != null)
                blendAttr.setLive(state);

            if(material != null)
                material.setLive(state);

            if(shader != null)
                shader.setLive(state);
        }
    }

    /**
     * Check to see if this node is the same reference as the passed node that
     * is a parent of this node. This is the downwards check to ensure that
     * there is no cyclic scene graph structures at the point where someone
     * adds a node to the scenegraph. When the reference and this are the
     * same, an exception is generated. Since each class may have different
     * lists of child node setups, this should be overriden by any class that
     * can take children, and have the call passed along to the children.
     *
     * @param parent The reference to check against this class
     * @throws CyclicSceneGraphStructureException Equal parent and child
     */
    protected void checkForCyclicChild(SceneGraphObject parent)
        throws CyclicSceneGraphStructureException
    {
        if(parent == this)
            throw new CyclicSceneGraphStructureException();

        for(int i = 0; i < numTextures; i++)
            textureUnits[i].checkForCyclicChild(parent);

        // Don't bother with anything else for now. That's just wasted CPU cycles.
    }

    //---------------------------------------------------------------
    // Methods defined by RenderableObject
    //---------------------------------------------------------------

    /**
     * Issue ogl commands needed for this component.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void render(GL gl, GLU glu)
    {
        if(numTextures == 1)
        {
            gl.glPushAttrib(GL.GL_TEXTURE_BIT);
            textureUnits[0].render(gl, glu, SINGLE_TEXTURE);
        }
        else if(numTextures > 1)
        {
            gl.glPushAttrib(GL.GL_TEXTURE_BIT);
            for(int i = 0; i < numTextures ; i++)
                textureUnits[i].render(gl, glu, TEX_IDS[i]);
        }

        if(blendAttr != null)
            blendAttr.render(gl, glu);

        if(polyAttr != null)
            polyAttr.render(gl, glu);

        if(lineAttr != null)
            lineAttr.render(gl, glu);

        if(pointAttr != null)
            pointAttr.render(gl, glu);

        if(material != null)
        {
            gl.glPushAttrib(GL.GL_LIGHTING_BIT);
            material.render(gl, glu);
        }

        if(shader != null)
            shader.render(gl, glu);
    }

    /**
     * Restore all openGL state.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void postRender(GL gl, GLU glu)
    {
        if(shader != null)
            shader.postRender(gl, glu);

        if(material != null)
        {
            material.postRender(gl, glu);
            gl.glPopAttrib();
        }

        if(pointAttr != null)
            pointAttr.postRender(gl, glu);

        if(lineAttr != null)
            lineAttr.postRender(gl, glu);

        if(polyAttr != null)
            polyAttr.postRender(gl, glu);

        if(blendAttr != null)
            blendAttr.postRender(gl, glu);

        if(numTextures == 1)
        {
            textureUnits[0].postRender(gl, glu, SINGLE_TEXTURE);
            gl.glPopAttrib();
        }
        else if(numTextures > 1)
        {
            for(int i = numTextures - 1; i >= 0; i--)
                textureUnits[i].postRender(gl, glu, TEX_IDS[i]);

            gl.glPopAttrib();
        }
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Set the visibility state of any geometry associated with this instance.
     * This can only be set during the data-changed callback as it only effects
     * visual state and not rendering state. An object that is not visible is
     * still valid for the bounds, which means it can be picked etc. That is,
     * making something invisible does not effect the bounds of the parent
     * objects.
     *
     * @param state true to make the object visible
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setVisible(boolean state)
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        visible = state;
    }

    /**
     * Check to see if this appearance is making the geometry visible or
     * not.
     *
     * @return true when the geometry is visible
     */
    public boolean isVisible()
    {
        return visible;
    }

    /**
     * Set the material to use. Null will clear the material.
     *
     * @param mat The new material
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setMaterial(Material mat)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(material != null)
            material.setLive(false);

        material = mat;

        if(material != null)
        {
            material.setUpdateHandler(updateHandler);
            material.setLive(alive);
        }
    }

    /**
     * Get the current material in use.
     *
     * @return The current material instance or null
     */
    public Material getMaterial()
    {
        return material;
    }

    /**
     * Set the polygon rendering attributes to use. Null will clear the
     * current attributes.
     *
     * @param attr The new attributes or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setPolygonAttributes(PolygonAttributes attr)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(polyAttr != null)
            polyAttr.setLive(false);

        polyAttr = attr;

        if(polyAttr != null)
        {
            polyAttr.setUpdateHandler(updateHandler);
            polyAttr.setLive(alive);
        }
    }

    /**
     * Get the current polygon rendering attributes in use.
     *
     * @return The current attributes instance or null
     */
    public PolygonAttributes getPolygonAttributes()
    {
        return polyAttr;
    }

    /**
     * Set the line rendering attributes to use. Null will clear the
     * current attributes.
     *
     * @param attr The new attributes or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setLineAttributes(LineAttributes attr)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(lineAttr != null)
            lineAttr.setLive(false);

        lineAttr = attr;

        if(lineAttr != null)
        {
            lineAttr.setUpdateHandler(updateHandler);
            lineAttr.setLive(alive);
        }
    }

    /**
     * Get the current line rendering attributes in use.
     *
     * @return The current attributes instance or null
     */
    public LineAttributes getLineAttributes()
    {
        return lineAttr;
    }

    /**
     * Set the point rendering attributes to use. Null will clear the
     * current attributes.
     *
     * @param attr The new attributes or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setPointAttributes(PointAttributes attr)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(pointAttr != null)
            pointAttr.setLive(false);

        pointAttr = attr;

        if(pointAttr != null)
        {
            pointAttr.setUpdateHandler(updateHandler);
            pointAttr.setLive(alive);
        }
    }

    /**
     * Get the current pointgon rendering attributes in use.
     *
     * @return The current attributes instance or null
     */
    public PointAttributes getPointAttributes()
    {
        return pointAttr;
    }

    /**
     * Set the blend rendering attributes to use. Null will clear the
     * current attributes. Setting a value will disable state sorting for
     * transparency on this object.
     *
     * @param attr The new attributes or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setBlendAttributes(BlendAttributes attr)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(blendAttr != null)
            blendAttr.setLive(false);

        blendAttr = attr;

        if(blendAttr != null)
        {
            blendAttr.setUpdateHandler(updateHandler);
            blendAttr.setLive(alive);
        }
    }

    /**
     * Get the current blend rendering attributes in use.
     *
     * @return The current attributes instance or null
     */
    public BlendAttributes getBlendAttributes()
    {
        return blendAttr;
    }

    /**
     * Set the shader to use. Null will clear the current shader.
     *
     * @param s The new shader instance to use or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setShader(Shader s)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(shader != null)
            shader.setLive(false);

        shader = s;

        if(shader != null)
        {
            shader.setUpdateHandler(updateHandler);
            shader.setLive(alive);
        }
    }

    /**
     * Get the current shader in use.
     *
     * @return The current shader instance or null
     */
    public Shader getShader()
    {
        return shader;
    }

    /**
     * Set the texture units to use.  Null will disable texturing.
     *
     * @param texUnits The new Texture lists to set
     * @param num The number of valid entries in the array
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTextureUnits(TextureUnit[] texUnits, int num)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if((textureUnits == null) || (textureUnits.length < num))
            textureUnits = new TextureUnit[num];

        // clear out remaining pointers
        for(int i = numTextures; i < num; i++)
            textureUnits[i] = null;

        for(int i = 0; i < num; i++)
        {
            textureUnits[i] = texUnits[i];
            textureUnits[i].setUpdateHandler(updateHandler);
        }

        numTextures = num;
    }

    /**
     * Get the current number of texture units that are valid
     *
     * @return a positive number
     */
    public int numTextureUnits()
    {
        return numTextures;
    }

    /**
     * Get the texture units used by the material.
     *
     * @param texUnits An array to copy the texture units into
     * @throws IllegalArgumentException The provided array is not big enough
     *   to hold all the texture units
     */
    public void getTextureUnits(TextureUnit[] texUnits)
    {
        if(texUnits.length < numTextures)
            throw new IllegalArgumentException("Array size is too small");

        // TODO: Revist using System.arraycopy as average number of texUnits goes up
        for(int i = 0; i < numTextures; i++) {
            texUnits[i] = textureUnits[i];
        }
    }
}
