/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
// None

// Local imports
// None

/**
 * Class for passing the detailed rendering information about a clip plane
 * through the pipeline.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.4 $
 */
public class ClipDetails
{
    /** The hash of this object */
    private int hash;

    /** The lead node instance to be rendered. Will be either Shape or Clip */
    private ClipPlane clipPlane;

    /** The transform from the root of the scene graph to here */
    private float[] transform;

    /**
     * Construct a default instance with just the transform initialised
     */
    public ClipDetails()
    {
        transform = new float[16];
        hash = 0;
    }

    //---------------------------------------------------------------
    // Methods defined by Comparable
    //---------------------------------------------------------------

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     * <p>
     *
     * Derived classes should extend this one to add the extra comparisons
     * needed.
     *
     * @param o The object to be compared
     * @return -1, 0 or 1 depending on order
     * @throws ClassCastException The specified object's type prevents it from
     *    being compared to this Object
     */
    public int compareTo(Object o)
        throws ClassCastException
    {
        ClipDetails cd = (ClipDetails)o;
        return compareTo(cd);
    }

    //---------------------------------------------------------------
    // Methods defined by Object
    //---------------------------------------------------------------

    /**
     * Compare the given details to this one to see if they are equal. Equality
     * is defined as pointing to the same clipPlane source, with the same
     * transformation value.
     *
     * @param o The object to compare against
     * @return true if these represent identical objects
     */
    public boolean equals(Object o)
    {
        if(!(o instanceof ClipDetails))
            return false;

        ClipDetails ld = (ClipDetails)o;

        if(clipPlane != ld.clipPlane)
            return false;

        for(int i = 0; i < 16; i++)
            if(transform[i] != ld.transform[i])
                return false;

        return true;
    }

    /**
     * Calculate the hashcode for this object.
     *
     */
    public int hashCode()
    {
        return hash;
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Update the details with a new set of values.
     *
     * @param cp The clipPlane instance to be used
     * @param tx The transform array to copy
     */
    public void update(ClipPlane cp, float[] tx)
    {
        clipPlane = cp;
        System.arraycopy(tx, 0, transform, 0, 16);

        // Now regenerate the hash code
        long h = 0;

        for(int i = 0; i < 16; i++)
            h = 31 * h + Float.floatToIntBits(transform[i]);

        h += clipPlane.hashCode();

        hash = (int)(h & 0xFFFFFFFF);
    }

    /**
     * Return the current transform array. Don't play with the return value.
     *
     * @return The current transform matrix array
     */
    public float[] getTransform()
    {
        return transform;
    }

    /**
     * Get the current clipPlane instance in use.
     *
     * @return A reference to the current clipPlane
     */
    public ClipPlane getClipPlane()
    {
        return clipPlane;
    }

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     *
     * @param cd The details to be compared
     * @return -1, 0 or 1 depending on order
     */
    public int compareTo(ClipDetails cd)
    {
        if(cd == this)
            return 0;

        if(cd == null)
            return 1;

        if(clipPlane != cd.clipPlane)
        {
            if(clipPlane == null)
               return -1;
            else if(cd.clipPlane == null)
                return 1;

            int res = clipPlane.compareTo(cd.clipPlane);
            if(res != 0)
                return res;
        }

        for(int i = 0; i < 16; i++)
        {
            if(transform[i] != cd.transform[i])
                return transform[i] < cd.transform[i] ? -1 : 1;
        }

        return 0;
    }
}
