/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import java.util.ArrayList;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import java.nio.ByteBuffer;
import net.java.games.joal.AL;

import org.j3d.util.MatrixUtils;

// Local imports
// None

/**
 * A ConeSound class which emits a conical sound in one direction.
 * <p>
 * By default this class is mimics a PointSound.  You must provide
 * parameters for the direction and cone angles to make it directional.
 *
 * The sound will attenuate by distance based on the refDistance
 * and maxDistance parameters.
 * <p>
 *
 * @author Alan Hudson
 * @version $Revision: 1.5 $
 */
public class ConeSound extends PointSound
{
    /** The OpenAL source */
    private int source;

    /** PointSound Position */
    private Point3f position;

    /** Scratch point */
    private Point3f tmpPoint;

    /** Scratch matrix val */
    private Matrix4f tmpMatrix;

    /** The sound direction */
    private float[] direction;

    /** The cones inner angle in degrees */
    private float innerAngle;

    /** The cones outer angle in degrees */
    private float outerAngle;

    /**
     * Creates a sound.
     */
    public ConeSound()
    {
        position = new Point3f();
        tmpPoint = new Point3f();
        tmpMatrix = new Matrix4f();

        refDistance = 0;
        maxDistance = Float.MAX_VALUE;
        innerAngle = 360;
        outerAngle = 360;
        direction = new float[] {0,0,0};
        rolloffFactor = 1;
    }

    /**
     * Set the direction.
     *
     * @param dir The direction to emit sound.
     */
    public void setRefDistance(float[] dir)
    {
        if(alive && updateHandler != null && !updateHandler.isBoundsWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        direction[0] = dir[0];
        direction[1] = dir[1];
        direction[2] = dir[2];
    }

    /**
     * Set the inner angle for the cone.
     *
     * @param angle The inner angle
     */
    public void setInnerAngle(float angle)
    {
        if(alive && updateHandler != null && !updateHandler.isBoundsWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        innerAngle = angle;
    }

    /**
     * Set the outer angle for the cone.
     *
     * @param angle The outer angle
     */
    public void setOuterAngle(float angle)
    {
        if(alive && updateHandler != null && !updateHandler.isBoundsWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        outerAngle = angle;
    }

    /**
     * This method is called to render this node.  All openGL commands needed
     * to render the node should be executed.  Any transformations needed
     * should be added to the transformation stack premultiplied.  This
     * method must be re-entrant as it can be called from multiple
     * places at once.
     *
     * @param al The al context to render with
     * @param transform The transformation stack to this node
     */
    public void render(AL al, Matrix4f transform)
    {
        transform.transform(position,tmpPoint);

        if (dataChanged)
        {
            buffer = soundSource.getBufferId(al, seq);
            if (buffer == -1)
                return;

            dataChanged = false;
            playChanged = true;

            int[] sources = new int[1];

            // Bind buffer with a source.
            al.alGenSources(1, sources);
            source = sources[0];

            if (al.alGetError() != AL.AL_NO_ERROR)
                System.out.println("Error creating OpenAL Source");

            al.alSourcei(source, AL.AL_BUFFER, buffer);
            al.alSourcef(source, AL.AL_REFERENCE_DISTANCE, refDistance);
            al.alSourcef(source, AL.AL_ROLLOFF_FACTOR, rolloffFactor);
            al.alSourcef(source, AL.AL_MAX_DISTANCE, maxDistance);
            al.alSourcef(source, AL.AL_CONE_INNER_ANGLE, innerAngle);
            al.alSourcef(source, AL.AL_CONE_OUTER_ANGLE, outerAngle);
            al.alSourcefv(source, AL.AL_DIRECTION, direction);
            al.alSourcef(source, AL.AL_GAIN, 1.0f);
            al.alSourcei(source, AL.AL_LOOPING, loop ? 1 : 0);
            al.alSourcef(source, AL.AL_PITCH, pitch);

            al.alSource3f(source, AL.AL_POSITION, tmpPoint.x, tmpPoint.y,tmpPoint.z);

            if (al.alGetError() != AL.AL_NO_ERROR)
                System.out.println("Error setting OpenAL Source params");

        }

        if (paramsChanged)
        {
            al.alSourcef(source, AL.AL_PITCH, pitch);
            al.alSourcei(source, AL.AL_LOOPING, loop ? 1 : 0);

            paramsChanged = false;
        }

        al.alSource3f(source, AL.AL_POSITION, tmpPoint.x, tmpPoint.y,tmpPoint.z);

        if (playChanged)
        {
            if (playing && paused)
            {
                al.alSourcePlay(source);
            }

            if (playing)
            {
                if (paused)
                    al.alSourcePause(source);
                else {
                    al.alSourcePlay(source);
                }
            } else
                al.alSourceStop(source);

            playChanged = false;
        } else if (playing)
        {
            int state = al.alGetSourcei(source, AL.AL_SOURCE_STATE);
            if (state == AL.AL_STOPPED)
            {
                playing = false;
            }
        }
    }

    /*
     * This method is called after a node has been rendered.  This method
     * must be re-entrant.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void postRender(AL al)
    {
    }

    /**
     * State check to see whether the sound is enabled.
     * @return true if the sound has something to render
     */
    public boolean isEnabled()
    {
        return true;
    }
}
