/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
// None

/**
 * Interface representing the output of a render pipeline.
 * <p>
 *
 * The output may be any of the traditional types: pBuffer, screen or memory
 * or any non-traditional type like haptic devices, network streams etc.
 *
 *
 * @author Justin Couch
 * @version $Revision: 1.17 $
 */
public interface DrawableSurface
{
    /**
     * The surface will render only non-stereo projection (traditional
     * monoscopic projection policy).
     */
    public static final int NO_STEREO = 0;

    /**
     * The stereo rendering type to be used should use quad buffered,
     * alternate frame rendering. In this type it renders to the left back
     * buffer, then right back buffer and then swaps both. This is useful if
     * you have HMDs which have two separate windows for the rendering.
     */
    public static final int QUAD_BUFFER_STEREO = 1;

    /**
     * The stereo rendering type draws to alternate eyes on each frame. Used if
     * you have shutter glasses where the user's vision is restricted to one
     * eye or the other, but not both at the same time. Only uses a single
     * double buffer to render alternate eye points
     */
    public static final int ALTERNATE_FRAME_STEREO = 2;

    /**
     * The stereo rendering uses two canvases - one for each eye. The canvases
     * are placed either side by side or vertically stacked, depending on the
     * rendering options requested.
     */
    public static final int TWO_CANVAS_STEREO = 3;

    /**
     * Check to see whether this surface supports stereo rendering. As this is
     * not known until after initialisation, this method will return false
     * until it can determine whether or not stereo is available.
     *
     * @return true Stereo support is currently available
     */
    public boolean isStereoAvailable();

    /**
     * Check to see whether this surface supports Quad buffer stereo rendering.
     * Quadbuffers uses the GL_BACK_LEFT and GL_BACK_RIGHT for rendering pairs
     * rather than drawing alternate frames to the same window.
     * <p>
     * As this is not known until after initialisation, this method will return
     * false until it can determine whether or not stereo is available.
     *
     * @return true Stereo support is currently available
     */
    public boolean isQuadStereoAvailable();

    /**
     * Set the rendering policy used when handling stereo. The policy must be
     * one of the _STEREO constants defined in this interface.
     *
     * @param policy The policy to currently use
     * @throws IllegalArgumentException The policy type is not one of the legal
     *    selections.
     */
    public void setStereoRenderingPolicy(int policy);

    /**
     * Get the current stereo rendering policy in use. If not explicitly set by
     * the user, then it will default to <code>NO_STEREO</code>.
     *
     * @return One of the *_STEREO values
     */
    public int getStereoRenderingPolicy();

    /**
     * Set the eye separation value when rendering stereo, defined as the
     * distance from the center axis to one eye. The default value is 0.33 for
     * most applications. The absolute value of the separation is always used.
     *
     * @param sep The amount of eye separation
     */
    public void setStereoEyeSeparation(float sep);

    /**
     * Get the current eye separation value, defined as the distance from the
     * center axis to one eye. If we are in no-stereo mode then this will
     * return zero.
     *
     * @return sep The amount of eye separation
     */
    public float getStereoEyeSeparation();

    /**
     * Tell the surface to use the given thread as the rendering control thread.
     * Used when you want to clock the renderer to the rest of the pipeline,
     * rather than letting it sit in it's own loop. Due to
     * bugs in JOGL, we have to start the internal rendering threads after the
     * containing window has become visible. See JOGL issue #54 for more details
     * on this bug.
     *
     * @param th The render control thread instance to use or null
     */
    public void setRenderControlThread(Thread th);

    /**
     * Set the background colour that this surface should be cleared to before
     * the drawing step. Colours range from 0 to 1 in the normal manner.
     *
     * @param r The red component of the background clear colour
     * @param g The green component of the background clear colour
     * @param b The blue component of the background clear colour
     * @param a The alpha component of the background clear colour
     */
    public void setClearColor(float r, float g, float b, float a);

    /**
     * An object has been detected as being deleted (no internal or
     * external references to it left) and it needs to be cleaned up. The
     * next oppourtunity this drawable and any sub-drawables get, this object
     * should have it's <code>cleanup()</code> method called and then the
     * reference to it should be discarded.
     *
     * @param obj The instance to be queued for deletion
     */
    public void queueDeletedObject(DeletableSceneGraphObject obj);

    /**
     * Queue up a collection of shader objects for processing on the next
     * frame. These processing requests are for either initialisation or log
     * handling, not for rendering.
     *
     * @param initList The shaders needing initialisation
     * @param numInit The number of shaders needing initialisation
     * @param logList The shaders needing log fetching
     * @param numLog The number of shaders needing log fetching
     */
    public void queueShaderRequests(ShaderSceneGraphObject[] initList,
                                    int numInit,
                                    ShaderSceneGraphObject[] logList,
                                    int numLog);

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param commands The list of drawable surfaces to render
     * @param numValid The number of valid items in the array
     */
    public void setDrawableObjects(RenderInstructions[] commands,
                                   int numValid);

    /**
     * Instruct the surface to draw the collected set of nodes now. The
     * registered view environment is used to draw to this surface. If no
     * view is registered, the surface is cleared and then this call is
     * exited. The drawing surface does not swap the buffers at this point.
     */
    public void draw();

    /**
     * Swap the buffers now if the surface supports multiple buffer drawing.
     * For surfaces that don't support multiple buffers, this does nothing.
     */
    public void swap();

    /**
     * Get the underlying object that this surface is rendered to. If it is a
     * screen display device, the surface can be one of AWT Component or
     * Swing JComponent. An off-screen buffer would be a form of AWT Image etc.
     *
     * @return The drawable surface representation
     */
    public Object getSurfaceObject();

    /**
     * Instruct this surface that you have finished with the resources needed
     * and to dispose all rendering resources.
     */
    public void dispose();

    /**
     * Get the surface to VWorld transformation matrix.
     *
     * @param matrix The matrix to copy into
     */
    public void getSurfaceToVWorld(Matrix4f matrix);

    /**
     * Convert a pixel location to surface coordinates.
     *
     * @param x The X coordinate
     * @param y The Y coordinate
     * @param position The converted position.  It must be preallocated.
     */
    public void getPixelLocationInSurface(float x, float y, Point3f position);

    /**
     * Get the Center Eye position in surface coordinates.
     *
     * @param position The current eye position.  It must be preallocated.
     */
    public void getCenterEyeInSurface(Point3f position);
}
