/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
// None

// Local imports
// None

/**
 * Class for passing the detailed rendering information about a light source.
 * through the pipeline.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.7 $
 */
public class LightDetails
{
    /** The hash of this object */
    private int hash;

    /** The lead node instance to be rendered. Will be either Shape or Light */
    private Light light;

    /** The transform from the root of the scene graph to here */
    private float[] transform;

    /**
     * Construct a default instance with just the transform initialised
     */
    public LightDetails()
    {
        transform = new float[16];
        hash = 0;
    }

    //---------------------------------------------------------------
    // Methods defined by Comparable
    //---------------------------------------------------------------

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     * <p>
     *
     * Derived classes should extend this one to add the extra comparisons
     * needed.
     *
     * @param o The object to be compared
     * @return -1, 0 or 1 depending on order
     * @throws ClassCastException The specified object's type prevents it from
     *    being compared to this Object
     */
    public int compareTo(Object o)
        throws ClassCastException
    {
        LightDetails ld = (LightDetails)o;
        return compareTo(ld);
    }

    //---------------------------------------------------------------
    // Methods defined by Object
    //---------------------------------------------------------------

    /**
     * Compare the given details to this one to see if they are equal. Equality
     * is defined as pointing to the same light source, with the same
     * transformation value.
     *
     * @param o The object to compare against
     * @return true if these represent identical objects
     */
    public boolean equals(Object o)
    {
        if(!(o instanceof LightDetails))
            return false;

        LightDetails ld = (LightDetails)o;

        if(light != ld.light)
            return false;
        float[] t = ld.transform;

        if (transform[0] != t[0] ||
            transform[1] != t[1] ||
            transform[2] != t[2] ||
            transform[3] != t[3] ||
            transform[4] != t[4] ||
            transform[5] != t[5] ||
            transform[6] != t[6] ||
            transform[7] != t[7] ||
            transform[8] != t[8] ||
            transform[9] != t[9] ||
            transform[10] != t[10] ||
            transform[11] != t[11] ||
            transform[12] != t[12] ||
            transform[13] != t[13] ||
            transform[14] != t[14] ||
            transform[15] != t[15])
            return false;

        return true;
    }

    /**
     * Calculate the hashcode for this object.
     *
     */
    public int hashCode()
    {
        return hash;
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Update the details with a new set of values.
     *
     * @param l The light instance to be used
     * @param tx The transform array to copy
     */
    public void update(Light l, float[] tx)
    {
        light = l;
        System.arraycopy(tx, 0, transform, 0, 16);

        // Now regenerate the hash code
        float h = 0;

        // Original code, below is faster but is it as good?
/*
        for(int i = 0; i < 16; i++)
            h = 31 * h + Float.floatToIntBits(transform[i]);
*/
        // Fastest so far!
        h = 31 * transform[0] + transform[1];
        h = 31 * h + transform[2];
        h = 31 * h + transform[3];
        h = 31 * h + transform[4];
        h = 31 * h + transform[5];
        h = 31 * h + transform[6];
        h = 31 * h + transform[7];
        h = 31 * h + transform[8];
        h = 31 * h + transform[9];
        h = 31 * h + transform[10];
        h = 31 * h + transform[11];
        h = 31 * h + transform[12];
        h = 31 * h + transform[13];
        h = 31 * h + transform[14];
        h = 31 * h + transform[15];

        long ans = Float.floatToIntBits(h) + light.hashCode();

        hash = (int)(ans & 0xFFFFFFFF);
    }

    /**
     * Return the current transform array. Don't play with the return value.
     *
     * @return The current transform matrix array
     */
    public float[] getTransform()
    {
        return transform;
    }

    /**
     * Get the current light instance in use.
     *
     * @return A reference to the current light
     */
    public Light getLight()
    {
        return light;
    }

    /**
     * Compares this object with the specified object for order. Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     *
     * @param ld The details to be compared
     * @return -1, 0 or 1 depending on order
     */
    public final int compareTo(LightDetails ld)
    {
        if(ld == this)
            return 0;

        if(ld == null)
            return 1;

        if(light != ld.light)
        {
            if(light == null)
               return -1;
            else if(ld.light == null)
                return 1;

            int res = light.compareTo(ld.light);
            if(res != 0)
                return res;
        }

        float[] t = ld.transform;

        if (transform[0] != t[0])
            return transform[0] < t[0] ? -1 : 1;
        if (transform[1] != t[1])
            return transform[1] < t[1] ? -1 : 1;
        if (transform[2] != t[2])
            return transform[2] < t[2] ? -1 : 1;
        if (transform[3] != t[3])
            return transform[3] < t[3] ? -1 : 1;
        if (transform[4] != t[4])
            return transform[4] < t[4] ? -1 : 1;
        if (transform[5] != t[5])
            return transform[5] < t[5] ? -1 : 1;
        if (transform[6] != t[6])
            return transform[6] < t[6] ? -1 : 1;
        if (transform[7] != t[7])
            return transform[7] < t[7] ? -1 : 1;
        if (transform[8] != t[8])
            return transform[8] < t[8] ? -1 : 1;
        if (transform[9] != t[9])
            return transform[9] < t[9] ? -1 : 1;
        if (transform[10] != t[10])
            return transform[10] < t[10] ? -1 : 1;
        if (transform[11] != t[11])
            return transform[11] < t[11] ? -1 : 1;
        if (transform[12] != t[12])
            return transform[12] < t[12] ? -1 : 1;
        if (transform[13] != t[13])
            return transform[13] < t[13] ? -1 : 1;
        if (transform[14] != t[14])
            return transform[14] < t[14] ? -1 : 1;
        if (transform[15] != t[15])
            return transform[15] < t[15] ? -1 : 1;

        return 0;
    }
}
