/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
// None

// Local imports
// None

/**
 * Abstract representation of a piece of code that wants to manage the node
 * update process internally to the scene graph.
 * <p>
 *
 * This interface is never directly called by user-land code. It's used to
 * abstract the process of maintaining a node as part of the scene graph, and
 * the implementation of the management system behind it. The management system
 * is responsible for coordinating and marshalling the user code into correct
 * timing for feeding directly to the rendering pipeline and beyond. As such,
 * it is responsible for preventing user updates at inappropriate times, and
 * also keeping track of what has been requested to update.
 * <p>
 *
 * Methods here provide both timing information, and ways of registering
 * objects for further processing. When a node implementation needs to know if
 * it is acceptable to make or allow certain changes, then these methods can be
 * queried to provide the appropriate guidance.
 *
 * @author Justin Couch
 * @version $Revision: 1.14 $
 */
public interface NodeUpdateHandler
{
    /**
     * Check to see if writing to the bounding information of the node is
     * permitted currently.
     *
     * @param src The object that is requesting the check
     * @return true if the end user can write, false if not
     */
    public boolean isBoundsWritePermitted(Object src);

    /**
     * Check to see if writing to the data information of the node is
     * permitted currently.
     *
     * @param src The object that is requesting the check
     * @return true if the end user can write, false if not
     */
    public boolean isDataWritePermitted(Object src);

    /**
     * Check to see if picking is permitted currently.
     *
     * @return true if the end user can pick, false if not
     */
    public boolean isPickingPermitted();

    /**
     * Notify the handler that you have updates to the SG that might alter
     * a node's bounds.
     *
     * @param l The change requestor
     * @param src The object that is passing this listener through.
     * @param intL Internal listener for making callbacks at a later time
     *    to propogate the bounds changes.
     * @return Was the notification accepted.  Duplicates will return false.
     * @throws InvalidListenerSetTimingException If called when the node called
     *    during one of the bounds/data changed callbacks
     */
    public boolean boundsChanged(NodeUpdateListener l,
                                 Object src,
                                 InternalNodeUpdateListener intL)
        throws InvalidListenerSetTimingException;

    /**
     * Notify the handler that you have updates to the SG that will not
     * alter a node's bounds.
     *
     * @param l The change requestor
     * @param src The object that is passing this listener through.
     * @throws InvalidListenerSetTimingException If called when the node called
     *    during one of the bounds/data changed callbacks
     */
    public void dataChanged(NodeUpdateListener l, Object src)
        throws InvalidListenerSetTimingException;

    /**
     * The object provided is no longer needed by its parent. Put this on
     * a queue for objects that could potentially be deleted next frame. Used
     * for any scene graph object type that must be explicitly cleaned up
     * during the OpenGL render cycle, such as textures and pBuffers.
     *
     * @param obj The object that is registering itself for deletion
     */
    public void objectRemoved(DeletableSceneGraphObject obj);

    /**
     * Get the picking handler so that we can do some picking operations.
     *
     * @return the current instance of the picking system
     */
    public PickingHandler getPickHandler();

    /**
     * The shader object passed requires an initialisation be performed. Queue
     * the shader up for processing now.
     *
     * @param shader The shader instance to queue
     * @param updateResponse true if this is being made as a response to a node's
     *   setUpdateHandler() method
     */
    public void shaderRequiresInit(ShaderSceneGraphObject shader,
                                   boolean updateResponse);

    /**
     * The shader object passed requires updating the log info. Queue
     * the shader up for processing now so that at the next oppourtunity it
     * can call glGetLogInfoARB.
     *
     * @param shader The shader instance to queue
     * @param updateResponse true if this is being made as a response to a node's
     *   setUpdateHandler() method
     */
    public void shaderRequiresLogInfo(ShaderSceneGraphObject shader,
                                      boolean updateResponse);
}
