/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
// None

// Local imports
// None

/**
 * A marker interface that represents a single complete rendering pipeline.
 * <p>
 *
 * A pipeline represents all of the drawing steps that may be accomplished
 * within a rendering cycle - culling, sorting and drawing. While an end-user
 * may wish to directly call the methods on this interface directly to control
 * their own rendering, it is recommended that a dedicated pipeline manager be
 * used for this task.
 * <p>
 *
 * If the pipeline does not have a drawable surface registered, it will still
 * complete all the steps up to that point. If no scene is registered, no
 * functionality is performed - render() will return immediately.
 *
 * @author Justin Couch
 * @version $Revision: 1.10 $
 */
public interface RenderPipeline
{
    /**
     * Register a drawing surface that this pipeline will send its output to.
     * Setting a null value will remove the current drawable surface.
     *
     * @param surface The surface instance to use or replace
     */
    public void setDrawableSurface(DrawableSurface surface);

    /**
     * Get the currently registered drawable surface instance. If none is set,
     * return null.
     *
     * @return The currently set surface instance or null
     */
    public DrawableSurface getDrawableSurface();

    /**
     * Start the pipeline functioning now. All steps will be called and this
     * method will not return until all are completed. Only the swap
     * function is not called in this time as we need to have this called
     * separately if the system is running parallel pipelines.
     */
    public void render();

    /**
     * Cause the surface to redraw the next frame only, with no processing of
     * the pipeline. This is typically an optimisation step when nothing has
     * changed in user land, so there's no processing that needs to be done.
     * Skip the processing and tell the drawable surface to render again what
     * it already has set from the previous frame.
     */
    public void displayOnly();

    /**
     * Instruct the drawable at the end of this pipeline to swap the buffers
     * now.
     */
    public void swapBuffers();

    /**
     * Set the root of the scene graph to be used by this pipeline. A value of
     * null will remove the scene from being rendered, causing the pipeline to
     * function as a no-op when rendered.
     * <p>
     *
     * If layers are currently set, and a non-null scene is provided, this will
     * remove the layers and use the scene instead. The layers will be removed
     * completely.
     *
     * @param scene The scene instance to render
     */
    public void setRenderableScene(Scene scene);

    /**
     * Set the set of layers to be used by this pipeline. Providing an argument
     * of a zero number of layers remove the layers from being rendered,
     * causing the pipeline to function as a no-op when rendered.
     * <p>
     *
     * If a scene is currently set, and a non-zero number of layers is
     * provided, this will remove the scene and use the layers instead.
     *
     * @param layers The collection of layers, in order, to render
     * @param numLayers The number of valid layers to use
     */
    public void setRenderableLayers(Layer[] layers, int numLayers);

    /**
     * Set the eyepoint offset from the centre position. This is used to model
     * offset view frustums, such as multiple displays or a powerwall. Positive
     * values will move the eyepoint to the right, up and back, relative to the
     * screen's center, effectively moving the screen to the left, down and
     * away from the user.
     *
     * @param x The x axis offset
     * @param y The y axis offset
     * @param z The z axis offset
     */
    public void setEyePointOffset(float x, float y, float z);

    /**
     * Set the orientation of this screen relative to the user's normal view
     * direction. The normal orientation of the screen is along the negative
     * Z axis. This method provides and axis-angle reorientation of that
     * direction to one that is facing the screen. Typically this will just
     * involve a rotation around the Y axis of some amount (45 and 90 deg
     * being the most common used in walls and caves).
     *
     * @param x The x axis component
     * @param y The y axis component
     * @param z The z axis component
     * @param a The angle to rotate around the axis in radians
     */
    public void setScreenOrientation(float x, float y, float z, float a);

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt();
}
