/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
// None

// Local imports
// None

/**
 * A marker interface that represents a class capable of managing
 * the complete rendering pipeline.
 * <p>
 *
 * A manager is used to handle a system-specific rendering technique. The goal
 * is to manage the contained pipeline(s) in a way that is most efficient to
 * the hardware provided. Thus, it is expected there will be many different
 * types of managers to suit the many hardware configurations available.
 * <p>
 *
 * Example implementations of the pipeline manager would be one that handles
 * all the pipelines with simultaneous threads, each pinned to a particular
 * CPU/Graphics pipe that the machine has. Another implementation may hold all
 * the pipelines for sequential evaluation piping the output from one into the
 * input for another (eg for handling dynamic cubic environment maps).
 * <p>
 *
 * Basic common methods are provided for all implementations to use. It is
 * expected that implementations will add additional technique-specific
 * extension methods to the basic features.
 * <p>
 *
 * <b>Controlling Rendering Updates</b>
 * <p>
 *
 * Almost all applications will wish to control the rate at which rendering
 * occurs. Typically this is just so that the rest of the application and
 * user's system remains interactive (the default setup is to use 100% CPU to
 * run as fast as possible). Alternatively the application may have a very
 * small rate of change, or wish to run in an offscreen-rendering mode (eg
 * server-based for web-delivery to an end user).
 * <p>
 *
 * The basic frame rate is controlled by the {@link #setMinimumFrameInterval(int)}
 * method. This provides the minimum time, in milliseconds, between sequential
 * frame renderings. If the rendering of a single frame takes longer than this
 * time, then the next frame will begin as soon as possible. (Note, we do not
 * currently provide any mechanism for guaranteeing a specific frame rate,
 * only a maximum framerate). For example, setting a value of 20 will result
 * in a maximum of 50 frames per second.
 * <p>
 *
 * To return the system to an unregulated frame rate, after having set the
 * value previously, call <code>setMinimumFrameInterval()</code> method with a
 * value of 0.
 * <p>
 *
 * If you wish to render on demand, then set the render manager to disabled
 * it is disabled by default on construction). Then, whenever you need to
 * repaint the surface, call the {@link #renderOnce()} method.
 *
 *
 * @author Justin Couch
 * @version $Revision: 1.14 $
 */
public interface RenderPipelineManager
{
    /**
     * Tell render to start or stop rendering. If currently running, it
     * will wait until all the pipelines have completed their current cycle
     * and will then halt.
     *
     * @param state True if to enable rendering
     */
    public void setEnabled(boolean state);

    /**
     * Get the current render state of the manager.
     *
     * @return true if the manager is currently running
     */
    public boolean isEnabled();

    /**
     * Force a single render of all pipelines now. Ignores the enabled and
     * cycle time settings to cause a single render at this point in time.
     * If a render is currently in progress, an exception is generated. If
     * this method is called, the application observer is not called
     * beforehand. It assumes that you'll be making the scene graph updates
     * yourself before calling this method.
     *
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void renderOnce()
        throws IllegalStateException;

    /**
     * Request that the manager perform a full scene render pass and update,
     * ignoring any usual optimisations that it may take. For performance
     * reasons, a render manager may elect to not run or update some portions
     * of the scene graph. This method requests that the next frame only should
     * ignore those optimisations and process the full scene graph.
     * <p>
     *
     * This method will work both in automated rendering and with the
     * {@link #renderOnce()} method.
     */
    public void requestFullSceneRender();

    /**
     * Set the scene for this manager. Setting a value of <code>null</code>
     * will remove the currently set scene. If this is set while a current
     * collection of layers is set, then the layers will be cleared.
     *
     * @param scene The new scene
     */
    public void setScene(Scene scene);

    /**
     * Fetch the currently set scene. If layers are set, then this will return
     * null.
     *
     * @return The current scene, or null
     */
    public Scene getScene();

    /**
     * Set the set of layers for this manager. Setting a value of
     * <code>null</code> will remove the currently set of layers. If this is
     * set while a current scene is set, then the scene will be cleared. Layers
     * are presented in depth order - layers[0] is rendered before layers[1]
     * etc.
     *
     * @param layers The collection of layers, in order, to render
     * @param numLayers The number of valid layers to use
     */
    public void setLayers(Layer[] layers, int numLayers);

    /**
     * Get the number of layers that are currently set. If no layers are set,
     * or a scene is set, this will return zero.
     *
     * @return a value greater than or equal to zero
     */
    public int numLayers();

    /**
     * Fetch the current layers that are set. The values will be copied into
     * the user-provided array. That array must be at least
     * {@link #numLayers()} in length. If not, this method does nothing (the
     * provided array will be unchanged).
     *
     * @param layers An array to copy the values into
     */
    public void getLayers(Layer[] layers);

    /**
     * Set the minimum duty cycle of the render manager. This is the type in
     * milliseconds that should be the minimum between frames and can be used
     * to throttle the rendering loop to a maximum frame rate should other
     * systems require CPU time. This can be changed at any time.
     * <p>
     *
     * Setting a value of zero will return the render to an un-regulated state
     * where the code will attempt to render as fast as possible (ie guaranteed
     * 100% CPU usage).
     *
     * @param cycleTime The minimum time in milliseconds between frames or zero
     */
    public void setMinimumFrameInterval(int cycleTime);

    /**
     * Fetch the currently set duty cycle value.
     *
     * @return The duty cycle time, in milliseconds
     */
    public int getMinimumFrameInterval();

    /**
     * Add a pipeline to be rendered to the manager. A duplicate registration
     * or null value is ignored.
     *
     * @param pipe The new pipe instance to be added
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void addPipeline(RenderPipeline pipe)
        throws IllegalStateException;

    /**
     * Remove an already registered pipeline from the manager. A or null value
     * or one that is not currently registered is ignored.
     *
     * @param pipe The pipe instance to be removed
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void removePipeline(RenderPipeline pipe)
        throws IllegalStateException;

    /**
     * Add a pipeline to be rendered to the manager. A duplicate registration
     * or null value is ignored.
     *
     * @param pipe The new pipe instance to be added
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void addAudioPipeline(AudioPipeline pipe)
        throws IllegalStateException;

    /**
     * Remove an already registered pipeline from the manager. A or null value
     * or one that is not currently registered is ignored.
     *
     * @param pipe The pipe instance to be removed
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void removeAudioPipeline(AudioPipeline pipe)
        throws IllegalStateException;

    /**
     * Register an observer that can be used to know when the application is
     * safe to update the scene graph. A value of null will remove the
     * currently set value.
     *
     * @param obs The observer instance to use
     */
    public void setApplicationObserver(ApplicationUpdateObserver obs);

    /**
     * Disable the internal shutdown hook system. It will be up to the calling
     * application to make sure the {@link #shutdown()} method is called to
     * turn off the OpenGL rendering system. If it does not, there is a good
     * possibility of a crash of the system.
     * <p>
     *
     * If the internal shutdown is disabled, then the shutdown callback of the
     * <code>ApplicationUpdateObserver</code> will not be called.
     */
    public void disableInternalShutdown();

    /**
     * Notification to shutdown the internals of the renderer because the
     * application is about to exit.
     */
    public void shutdown();
}
