/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Local imports
import java.util.HashMap;

import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.ByteOrder;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// External imports
// None

/**
 * Background node that renders a list of user-provided Shape3D instances.
 * <p>
 *
 * Backgrounds are rendered as the first item but do not interact with the
 * normal geometry in the rendering process. Typically, backgrounds are
 * rendered in a fixed volume (a unit box or sphere is the most common) with
 * depthbuffer reads and writes disabled. Ordinary geometry is then drawn over
 * the top.  Backgrounds must fit within clipping planes of [0.1,1].
 *
 * Rendering is performed in the order the nodes are added to the instance.
 * As such
 *
 * This will set the background colour to a single colour for the entire
 * viewport. If used, this will override the setClearColor() on
 * {@link org.j3d.aviatrix3d.DrawableSurface}.
 *
 * @author Justin Couch
 * @version $Revision: 1.8 $
 */
public class ShapeBackground extends Background
{
    /** Base colour of the fog */
    private float[] color;

    /** Collection of shapes that define this background */
    private Shape3D[] geometry;

    /** The last shape on the list */
    private int numGeometry;

    /**
     * Constructs a background node for a base colour of black.
     */
    public ShapeBackground()
    {
        color = new float[4];
        color[3] = 1;
        numGeometry = 0;
        geometry = new Shape3D[5];
    }

    /**
     * Construct a background node for a user-provided colour
     */
    public ShapeBackground(float[] c)
    {
        this();
        color[0] = c[0];
        color[1] = c[1];
        color[2] = c[2];
        color[3] = c[3];
    }

    //----------------------------------------------------------
    // Methods defined by RenderableObject
    //----------------------------------------------------------

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void render(GL gl, GLU glu)
    {
        gl.glClearColor(color[0], color[1], color[2], color[3]);
        gl.glClear(GL.GL_COLOR_BUFFER_BIT);

        for(int i = 0; i < numGeometry; i++)
        {
            if(!geometry[i].isVisible())
                continue;

            geometry[i].render(gl, glu);
            geometry[i].postRender(gl, glu);
        }
    }

    /**
     * Restore all openGL state to the given drawable.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void postRender(GL gl, GLU glu)
    {
    }

    //---------------------------------------------------------------
    // Methods defined by SceneGraphObject
    //---------------------------------------------------------------

    /**
     * Set the scenegraph update handler for this node.  It will notify
     * all its children of the value. A null value will clear the current
     * handler.
     *
     * @param handler The instance to use as a handler
     */
    protected void setUpdateHandler(NodeUpdateHandler handler)
    {
        super.setUpdateHandler(handler);

        for(int i=0; i < numGeometry; i++)
        {
            if(geometry[i] != null)
                geometry[i].setUpdateHandler(handler);
        }
    }

    /**
     * Notification that this object is live now.
     *
     * @param state true to set this to live
     */
    protected void setLive(boolean state)
    {
        for(int i = 0; i < numGeometry; i++)
        {
            if(geometry[i] != null)
                geometry[i].setLive(state);
        }

        // Call this after, that way the bounds are recalculated here with
        // the correct bounds of all the children set up.
        super.setLive(state);
    }

    //----------------------------------------------------------
    // Local methods
    //----------------------------------------------------------

    /**
     * Change the colour to the new colour. Colour takes RGBA.
     *
     * @param c The colour to copy in
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setColor(float[] c)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        color[0] = c[0];
        color[1] = c[1];
        color[2] = c[2];
        color[3] = c[3];
    }

    /**
     * Change the colour to the new colour. Colour takes RGBA.
     *
     * @param r The red colour component to use
     * @param g The green colour component to use
     * @param b The blue colour component to use
     * @param a The alpha colour component to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setColor(float r, float g, float b, float a)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        color[0] = r;
        color[1] = g;
        color[2] = b;
        color[3] = a;
    }

    /**
     * Get the current drawing colour
     *
     * @param c An array of length 4 or more to copy the colour to
     */
    public void getColor(float[] c)
    {
        c[0] = color[0];
        c[1] = color[1];
        c[2] = color[2];
        c[3] = color[3];
    }

    /**
     * Add a shape to be rendered to the end of the listing.
     *
     * @param shape The object instance to be rendered
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void addShape(Shape3D shape)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(shape == null)
            return;

        if(numGeometry == geometry.length)
        {
            Shape3D[] tmp = new Shape3D[numGeometry + 5];
            System.arraycopy(geometry, 0, tmp, 0, numGeometry);
            geometry = tmp;
        }

        geometry[numGeometry++] = shape;

        if(shape.isLive() != alive)
            shape.setLive(alive);

        shape.setParent(this);
        shape.setUpdateHandler(updateHandler);
    }

    /**
     * Remove the shape at the given index position. If the index is
     * out of range, then this will generate an exception.
     *
     * @param idx The index of the shape to be fetched
     * @return The shape at the given index
     */
    public Shape3D removeShape(int idx)
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(idx < 0 || idx >= numGeometry)
            throw new IllegalArgumentException("idx Is out of bounds");

        Shape3D shape = geometry[idx];

        if(geometry[idx] != null)
        {
            geometry[idx].setLive(false);
            geometry[idx].removeParent(this);
        }

        if(idx == (numGeometry - 1))
            geometry[idx] = null;
        else
            System.arraycopy(geometry,
                             idx,
                             geometry,
                             idx + 1,
                             numGeometry - idx - 1);

        numGeometry--;

        return shape;
    }
    /**
     * Change the shape at the given index. A shape can only be replaced by
     * another shape. If you wish to delete it, then use the remove method,
     * not this one with a null value. A null value is illegal here.
     *
     * @param idx The index of the shape to be fetched
     * @param shape The object instance to be rendered
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setShape(Shape3D shape, int idx)
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(shape == null)
            throw new IllegalArgumentException("Null shape provided");

        if(idx < 0 || idx >= numGeometry)
            throw new IllegalArgumentException("idx Is out of bounds");

        if(geometry[idx] != null)
        {
            geometry[idx].setLive(false);
            geometry[idx].removeParent(this);
        }

        geometry[idx] = shape;

        if(shape.isLive() != alive)
            shape.setLive(alive);

        shape.setParent(this);
        shape.setUpdateHandler(updateHandler);
    }

    /**
     * Get the current the shape and the given index position. If the index is
     * out of range, then this will generate an exception.
     *
     * @param idx The index of the shape to be fetched
     * @return The shape at the given index
     */
    public Shape3D getShape(int idx)
    {
        if(idx < 0 || idx >= numGeometry)
            throw new IllegalArgumentException("idx Is out of bounds");

        return geometry[idx];
    }
}
