/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import java.util.ArrayList;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import java.nio.ByteBuffer;
import net.java.games.joal.AL;

// Local imports
import org.j3d.aviatrix3d.iutil.AudioUpdateListener;

/**
 * A Sound class represents all sound emiting nodes in the system.
 * <p>
 *
 * @author Alan Hudson
 * @version $Revision: 1.6 $
 */
public abstract class Sound extends Leaf
   implements AudioRenderable, AudioUpdateListener
{
    /** The source feeding this sound */
    protected AudioComponent soundSource;

    /** The format of this sound.  Defined in AudioSource */
    protected int format;

    /** The frequency of the samples */
    protected int freq;

    /** Whether to loop this sample */
    protected boolean loop;

    /** The buffer created for this sound */
    protected int buffer;

    /** The sounds pitch */
    protected float pitch;

    /** The current seq */
    protected int seq;

    /** Has the sound data  changed */
    protected boolean dataChanged;

    /** Has the play state changed */
    protected boolean playChanged;

    /** Has the sound params */
    protected boolean paramsChanged;

    /** Is the sound playing */
    protected boolean playing;

    /** Is the sound paused */
    protected boolean paused;

    /**
     * Creates a sound.
     */
    public Sound()
    {
        dataChanged = false;
        playChanged = false;
        paramsChanged = false;
        playing = false;
        paused = false;
    }

    public void setAudioSource(AudioComponent src)
    {
        if(alive && updateHandler != null && !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        soundSource = src;

        this.format = src.getFormat();
        this.freq = src.getFrequency();
        this.loop = src.getLoop();
        this.pitch = src.getPitch();

        src.addUpdateListener(this);

        dataChanged = true;
    }

    /**
     * Notification that the audio has updated a section of the sample.
     * This is generic for all audio sources correctly.
     *
     * @param format The format of the samples
     * @param frequency The frequency of the samples
     * @param seq The sample seq.  Seq 0 means complete file.
     *            Streamed sources start at 1
     */
    public void audioUpdated(int format, int frequency, int seq)
    {

        this.format = format;
        this.freq = frequency;

        // Seq ignored for now
        this.seq = seq;

        dataChanged = true;
    }

    /**
     * Notification that the audio's parameters have changed.
     *
     * @param loop Whether to loop this sample
     * @param pitch The pitch to play at.
     */
    public void paramsUpdated(boolean loop, float pitch) {
        this.pitch = pitch;
        this.loop = loop;

        paramsChanged = true;
    }

    /**
     * Start a sound playing.
     */
    public void startSound() {
        if (!playing || paused) {
            playChanged = true;
            playing = true;
            paused = false;
        }
    }

    /**
     * Stop a sound playing.
     */
    public void stopSound() {
        if (playing) {
            playChanged = true;
            playing = false;
            paused = false;
        }
    }

    /**
     * Pause a sound playing.
     */
    public void pauseSound() {
        if (playing && !paused) {
            playChanged = true;
            paused = true;
        }
    }

    /**
     * Is this sound currently playing.
     *
     * @return Whether the sound is playing.
     */
    public boolean isPlaying() {
        return playing;
    }

    /**
     * Is this sound currently paused;
     *
     * @return Whether the sound is paused.
     */
    public boolean isPaused() {
        return paused;
    }

    /**
     * This method is called to render this node.  All openGL commands needed
     * to render the node should be executed.  Any transformations needed
     * should be added to the transformation stack premultiplied.  This
     * method must be re-entrant as it can be called from multiple
     * places at once.
     *
     * @param al The al context to render with
     * @param transform The transformation stack to this node
     */
    public void render(AL al, Matrix4f transform)
    {
    }

    /*
     * This method is called after a node has been rendered.  This method
     * must be re-entrant.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void postRender(AL al)
    {
    }

    /**
     * Mark this node as having dirty bounds due to one of it's children having
     * their bounds changed.
     */
    protected void markBoundsDirty()
    {
    }

    /**
     * Internal method to recalculate the implicit bounds of this Node. By
     * default the bounds are a point sphere, so derived classes should
     * override this method with something better.
     */
    protected void recomputeBounds()
    {
    }

    /**
     * Request a recomputation of the bounds of this object. If this object is
     * not currently live, you can request a recompute of the bounds to get the
     * most current values. If this node is currently live, then the request is
     * ignored.
     * <p>
     * This will request the geometry to recompute the bounds. If the geometry
     * is found to be live during this process, it will not update, and thus
     * the value used will be the last updated (ie from the previous frame it
     * was processed).
     */
    public void requestBoundsUpdate()
    {
    }

    /**
     * Get the currently set bounds for this object. If no explicit bounds have
     * been set, then an implicit set of bounds is returned based on the
     * current scene graph state.
     *
     * @return The current bounds of this object
     */
    public BoundingVolume getBounds()
    {
        return INVALID_BOUNDS;
    }

    /**
     * State check to see whether the sound is enabled.
     * @return true if the sound has something to render
     */
    public boolean isEnabled()
    {
        return false;
    }

    /**
     * Set the scenegraph update handler for this node.  It will notify
     * all its children of the value. A null value will clear the current
     * handler.
     *
     * @param handler The instance to use as a handler
     */
    protected void setUpdateHandler(NodeUpdateHandler handler)
    {
        super.setUpdateHandler(handler);
    }

    /**
     * Notification that this object is live now. Overridden to make sure that
     * the live state of the nodes represents the same state as the parent
     * scene graph.
     *
     * @param state true if this should be marked as live now
     */
    protected void setLive(boolean state)
    {
        // This will also force the geometry recompute of the bounds.
        super.setLive(state);
    }

    /**
     * Check to see if this node is the same reference as the passed node that
     * is a parent of this node. This is the downwards check to ensure that
     * there is no cyclic scene graph structures at the point where someone
     * adds a node to the scenegraph. When the reference and this are the
     * same, an exception is generated. Since each class may have different
     * lists of child node setups, this should be overriden by any class that
     * can take children, and have the call passed along to the children.
     *
     * @param parent The reference to check against this class
     * @throws CyclicSceneGraphStructureException Equal parent and child
     */
    protected void checkForCyclicChild(SceneGraphObject parent)
        throws CyclicSceneGraphStructureException
    {
        if(parent == this)
            throw new CyclicSceneGraphStructureException();
    }
}
