/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import java.awt.Rectangle;

// Local imports
// None

/**
 * Abstract representation of the physical environment setup used to connect
 * a virtual Viewpoint object to the real one that is rendered on a drawable
 * surface.
 * <p>
 *
 *
 * @author Justin Couch
 * @version $Revision: 1.10 $
 */
public class ViewEnvironment
{
    /** Message when the projection type provided is invalid */
    private static final String INVALID_PROJECTION_MSG =
        "Invalid projection type provided. Must be one of " +
        "ORTHOGRAPHIC_PROJECTION or PERSPECTIVE_PROJECTION";

    private static final String INVALID_FOV_MSG =
        "The field of view angle provided is invalid. Must be between 0 and 180";

    /** Invalid clip distance. It is negative */
    private static final String NEGATIVE_CLIP_MSG =
        "The clip distance provided is negative ";

    /** The far clip distance is less than the near clip distance */
    private static final String INVERT_CLIP_MSG =
        "The near clip plane is greater than the far clip plane";

    /** The projection type is perspective mode */
    public static final int PERSPECTIVE_PROJECTION = 1;

    /** The projection type is perspective mode */
    public static final int ORTHOGRAPHIC_PROJECTION = 2;

    /** The stereo setting for this environment */
    private boolean useStereo;

    /** The perspective/orthographic setting for this environment */
    private int projectionType;

    /** The field of view used */
    private double fov;

    /** Current near clip plane setting */
    private double nearClip;

    /** Current far clip plane setting */
    private double farClip;

    /** The aspectRatio ratio */
    private double aspectRatio;

    /** If defined, the explicit viewport to use */
    private Rectangle viewportSize;

    /**
     * Create a default instance of this class with stereo false,
     * perspective projection and field of view set to 45 degrees.
     */
    public ViewEnvironment()
    {
        useStereo = false;
        projectionType = PERSPECTIVE_PROJECTION;
        fov = 45;

        nearClip = 0.01;
        farClip = 1000;
        aspectRatio = 1;
    }

    /**
     * Set the stereo flag used for this environment.
     *
     * @param stereo True if stereo should be rendered
     */
    public void setStereoEnabled(boolean stereo)
    {
        useStereo = stereo;
    }

    /**
     * Check to see if stereo has been enabled for this environment.
     *
     * @return true if stereo rendering is to be used
     */
    public boolean getStereoEnabled()
    {
        return useStereo;
    }

    /**
     * Set the perspective projection flag used for this environment.
     *
     * @param type One of ORTHOGRAPHIC_PROJECTION or PERSPECTIVE_PROJECTION
     * @throws IllegalArgumentException The type is not valid
     */
    public void setProjectionType(int type)
        throws IllegalArgumentException
    {
        if(type < PERSPECTIVE_PROJECTION || type > ORTHOGRAPHIC_PROJECTION)
            throw new IllegalArgumentException(INVALID_PROJECTION_MSG);

        projectionType = type;
    }

    /**
     * Check to see if stereo has been enabled for this environment.
     *
     * @return true if stereo rendering is to be used
     */
    public int getProjectionType()
    {
        return projectionType;
    }

    /**
     * Set the field of view to be used. The value supplied must be in degrees.
     *
     * @param angle The angle in degress
     * @throws IllegalArgumentException The angle is less than or equal to zero
     */
    public void setFieldOfView(double angle)
        throws IllegalArgumentException
    {
        if((angle < 0) || (angle > 180))
            throw new IllegalArgumentException(INVALID_FOV_MSG);

        fov = angle;
    }

    /**
     * Get the currently set field of view. The field of view is specified in
     * degrees.
     *
     * @return A value 0 <= x <= 180;
     */
    public double getFieldOfView()
    {
        return fov;
    }

    /**
     * Set aspect ratio, which is the ratio of window Width / Height.
     * An aspect ratio of <= 0 will be calculated from the current
     * window or viewport dimensions.
     *
     * @param aspect The new aspectRatio ratio.
     */
    public void setAspectRatio(double aspect)
    {
        aspectRatio = aspect;
    }

    /**
     * Get the currently set aspect ratio.  Width / Height.
     *
     * @return The aspect ratio
     */
    public double getAspectRatio()
    {
        return aspectRatio;
    }

    /**
     * Set an explicit viewport dimensions. If set, this will be the explicit
     * viewport size to use when rendering to the surface. If an aspect ratio
     * is not set, then this will be used to define the implicit ratio. If
     * set to null, the system should use the entire surface space. If no
     * aspect ratio is set ( <= 0) this can be used to imply the view frustum.
     * <p>
     * Note that if explicit coordinates are provided, the end user application
     * should also add window listener events to the surface and adjust the
     * values accordingly.
     *
     * @param size The new dimensions to use or null to clear
     */
    public void setViewport(Rectangle size)
    {
        if(size == null)
            viewportSize = null;
        else
        {
            if(viewportSize == null)
                viewportSize = new Rectangle();

            viewportSize.setBounds(size);
        }
    }

    /**
     * Get the currently set dimensions of the viewport. If none are currently
     * set, this will return null, otherwise it will return the object
     * representing the complete size;
     *
     * @return The current viewport size, or null
     */
    public Rectangle getViewport()
    {
        return viewportSize;
    }

    /**
     * Set the near clipping distance to be used by the application.
     *
     * @param d The distance to set the near clip plane to
     */
    public void setNearClipDistance(double d)
    {
        if(d < 0)
            throw new IllegalArgumentException(NEGATIVE_CLIP_MSG);

        if(d >= farClip)
            throw new IllegalArgumentException(INVERT_CLIP_MSG);

        nearClip = d;
    }

    /**
     * Get the current setting of the far clip plane.
     *
     * @return The current value, which is less than the far plane
     */
    public double getNearClipDistance()
    {
        return nearClip;
    }

    /**
     * Set the near clipping distance to be used by the application.
     *
     * @param near The distance to set the near clip plane
     * @param far The distance to the far clip plane
     */
    public void setClipDistance(double near, double far)
    {
        if(near < 0)
            throw new IllegalArgumentException(NEGATIVE_CLIP_MSG);

        if(far < 0)
            throw new IllegalArgumentException(NEGATIVE_CLIP_MSG);

        if(near >= far)
            throw new IllegalArgumentException(INVERT_CLIP_MSG);

        nearClip = near;
        farClip = far;
    }

    /**
     * Set the far clipping distance to be used by the application.
     *
     * @param d The distance to set the near clip plane to
     */
    public void setFarClipDistance(double d)
    {
        if(d < 0)
            throw new IllegalArgumentException(NEGATIVE_CLIP_MSG);

        if(d <= nearClip)
            throw new IllegalArgumentException(INVERT_CLIP_MSG);

        farClip = d;
    }

    /**
     * Get the current setting of the far clip plane.
     *
     * @return The current value, which is greater than the near plane
     */
    public double getFarClipDistance()
    {
        return farClip;
    }

    /**
     * From the current view setup of FoV, near and far clipping distances and
     * the aspectRatio ratio, generate the 6 parameters that describe a view
     * frustum. These parameters are what could be used as arguments to the
     * glFrustum call. The parameter order is:
     * [x min, x max, y min, y max, z near, z far]
     *
     * @param viewFrustum An array at least 6 in length for the values generated
     */
    public void generateViewFrustum(double[] viewFrustum)
    {
        double xmin = 0;
        double xmax = 0;
        double ymin = 0;
        double ymax = 0;

        ymax = nearClip * Math.tan(fov * Math.PI / 360.0);
        ymin = -ymax;

        if(aspectRatio <= 0)
        {
            if(viewportSize != null)
            {
                double ar = (float)viewportSize.width / viewportSize.height;
                xmin = ymin * ar;
                xmax = ymax * ar;
            }
            else
            {
                // Needs to be looked at. Assume a 1:1 ratio for now.
System.out.println("Implied aspect ratio and no screen size set");
                xmin = ymin;
                xmax = ymax;
            }
        }
        else
        {
            xmin = ymin * aspectRatio;
            xmax = ymax * aspectRatio;
        }

        viewFrustum[0] = xmin;
        viewFrustum[1] = xmax;
        viewFrustum[2] = ymin;
        viewFrustum[3] = ymax;
        viewFrustum[4] = nearClip;
        viewFrustum[5] = farClip;
    }
}
