/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import java.util.HashMap;
//import org.web3d.vecmath.Matrix4f;
import javax.vecmath.Matrix4f;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * A viewpoint into the scene.
 * <p>
 *
 * Viewpoints have their own implicit directional light (a headlight) that
 * is controlled separately to the normal in-scenegraph lights. Lights are
 * scoped by default and thus having the viewpoint also have a light following
 * it around is a pain to deal with, you can select on directly on the
 * viewpoint itself. This light is always pointing down the Z axis in the
 * local coordinate system.
 * <p>
 *
 * In addition to the headlight, a separate ambient light colour may be
 * provided. Although this is provided, note that whether it has any effect is
 * dependent on whether lighting is enabled on the object being rendered.
 *
 * @author Justin Couch
 * @version $Revision: 1.20.2.2 $
 */
public class Viewpoint extends Leaf
{
    /** Location of the local headlight when directional */
    private static final float[] DIR_LOCATION = { 0, 0, 1, 0 };

    /** Location of the local headlight when spot */
    private static final float[] SPOT_LOCATION = { 0, 0, 1, 1 };

    /** OpenGL default value for ambient lights. */
    private static final float[] DEFAULT_AMBIENT = { 0.2f, 0.2f, 0.2f, 1 };

    /** Flag indicating the headlight usage state */
    private boolean useHeadlight;

    /**
     * Flag indicating if the headlight should be a spot (false) or
     * directional light source.
     */
    private boolean useDirectionalLight;

    /** Flag indicating the global ambient usage state */
    private boolean useGlobalAmbient;

    /** The global ambient light color values */
    private float[] globalColor;

    /**
     * The default constructor where the headlight is implicitly turned off,
     * and the default light type is directional.
     */
    public Viewpoint()
    {
        useHeadlight = false;
        useDirectionalLight = true;
        useGlobalAmbient = false;
    }

    //----------------------------------------------------------
    // Methods defined by RenderableObject
    //----------------------------------------------------------

    /**
     * This method is called to render this node.  All openGL commands needed
     * to render the node should be executed.  Any transformations needed
     * should be added to the transformation stack premultiplied.  This
     * method must be re-entrant as it can be called from multiple
     * places at once.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void render(GL gl, GLU glu)
    {
        if(useHeadlight)
        {
            gl.glEnable(GL.GL_LIGHT0);

            if(!useDirectionalLight)
            {
                gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, SPOT_LOCATION);
                gl.glLightf(GL.GL_LIGHT0, GL.GL_SPOT_CUTOFF, 45);
            }
            else
                gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, DIR_LOCATION);
        }

        if(useGlobalAmbient)
            gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, globalColor);
    }

    /*
     * This method is called after a node has been rendered.  This method
     * must be re-entrant.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void postRender(GL gl, GLU glu)
    {
        if(useHeadlight)
            gl.glDisable(gl.GL_LIGHT0);

        if(useGlobalAmbient)
            gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, DEFAULT_AMBIENT);
    }

    //----------------------------------------------------------
    // Methods defined by Node
    //----------------------------------------------------------

    /**
     * Mark this node as having dirty bounds due to one of it's children having
     * their bounds changed.
     */
    protected void markBoundsDirty()
    {
        // Ignored as the bounds of a viewpoint are not valid. It is considered
        // a point in space.
    }

    //----------------------------------------------------------
    // Local Methods
    //----------------------------------------------------------

    /**
     * Get the current state of the headlight usage.
     *
     * @return true if the headlight is currently active
     */
    public boolean isHeadlightEnabled()
    {
        return useHeadlight;
    }

    /**
     * Turn the local headlight on/off.
     *
     * @param state true to turn the light on
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setHeadlightEnabled(boolean state)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);


        useHeadlight = state;
    }

    /**
     * Turn the global ambient lighting setting on or off.
     *
     * @param state true to turn the light on
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setGlobalAmbientLightEnabled(boolean state)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        useGlobalAmbient = state;

        // if we don't have it yet, create one and initialise it to the OpenGL
        // default values.
        if(globalColor == null)
            globalColor = new float[] { 0.2f, 0.2f, 0.2f, 1 };
    }

    /**
     * Get the current state of the global ambient light usage.
     *
     * @return true if the ambient light is currently active
     */
    public boolean isGlobalAmbientLightEnabled()
    {
        return useGlobalAmbient;
    }

    /**
     * Set the ambient colour to the new value. Colour must be in the range
     * [0, 1] otherwise an exception is generated.
     *
     * @param col The new colour value to use
     * @throws IllegalArgumentException The colour value is out of range
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setGlobalAmbientColor(float[] col)
        throws IllegalArgumentException, InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(col[0] < 0 || col[0] > 1)
            throw new IllegalArgumentException("Red component out of range");

        if(col[1] < 0 || col[1] > 1)
            throw new IllegalArgumentException("Green component out of range");

        if(col[2] < 0 || col[2] > 1)
            throw new IllegalArgumentException("Blue component out of range");

        globalColor[0] = col[0];
        globalColor[1] = col[1];
        globalColor[2] = col[2];
    }

    /**
     * Retrieve the current colour value from the light.
     *
     * @param col An array to copy the colour value into
     */
    public void getGlobalAmbientColor(float[] col)
    {
        col[0] = globalColor[0];
        col[1] = globalColor[1];
        col[2] = globalColor[2];
    }

    /**
     * Get the current state of the headlight type.
     *
     * @return true if the headlight is currently a directional light
     */
    public boolean isDirectionalLight()
    {
        return useDirectionalLight;
    }

    /**
     * Change the style of the headlight between directional and spot types.
     *
     * @param state true to turn the light to directional, false for spot
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setHeadlightType(boolean state)
        throws InvalidWriteTimingException
    {
        if(isLive() && updateHandler != null &&
           !updateHandler.isDataWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        useDirectionalLight = state;
    }
}
