/*****************************************************************************
 *                        Yumetech, Inc Copyright (c) 2004
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.audio;

// External imports
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.util.HashMap;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

import net.java.games.joal.*;
import net.java.games.joal.util.ALut;
import net.java.games.joal.OpenALException;

// Local imports
import org.j3d.aviatrix3d.*;
import org.j3d.util.MatrixUtils;

import org.j3d.util.IntHashMap;

/**
 * Implementation of the most basic audio device, supporting the minimal
 * number of features using OpenAL.
 * <p>
 *
 * @author Alan Hudson
 * @version $Revision: 1.5 $
 */
public class OpenALAudioDevice
    implements AudioDevice
{
    /** Default orientation that we set the listener to every frame */
    private static final float[] DEFAULT_ORIENTATION =
        { 0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 0.0f };

    /** The initial size of the children list */
    private static final int LIST_START_SIZE = 20;

    /** Number of items in the renderable list */
    private int numRenderables;

    /** Flag to say that init has been called on the canvas */
    private boolean initComplete;

    /** The AudioLibary to render to */
    private AL al;

    /** The Audio device context in use */
    private ALC.Context context;

    /** The audio device itself */
    private ALC.Device device;

    private AudioInstructions commands;

    /** Matrix Utilities to invert matrices */
    private MatrixUtils matrixUtils;

    /** Scratch matrix val */
    private Matrix4f tmpMatrix;

    /** Listener Position */
    private Point3f listenerPoint;

    /** Scratch point */
    private Point3f tmpPoint;

    /** OpenAL initialization failed */
    private boolean initFailed;

    /** Request that the current drawing terminate immediately. App closing */
    private boolean terminate;

    /** The largest buffer/source id generated */
    private int lastId;

    /**
     * Construct a surface that requires the given set of capabilities.
     */
    public OpenALAudioDevice()
    {

        matrixUtils = new MatrixUtils();
        tmpMatrix = new Matrix4f();
        tmpPoint = new Point3f();
        listenerPoint = new Point3f(0,0,0);

        terminate = false;
        initComplete = false;
        initFailed = false;

        lastId = 0;
   }

    //---------------------------------------------------------------
    // Methods defined by AudioDevice
    //---------------------------------------------------------------

    /**
     * An object has been detected as being deleted (no internal or
     * external references to it left) and it needs to be cleaned up. The
     * next oppourtunity this drawable and any sub-drawables get, this object
     * should have it's <code>cleanup()</code> method called and then the
     * reference to it should be discarded.
     *
     * @param obj The instance to queue
     */
    public void queueDeletedObject(DeletableSceneGraphObject obj)
    {
        System.out.println("Audio deletions unsupported atm");
    }

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param commands The list of drawable surfaces to render
     */
    public void setDrawableObjects(AudioInstructions commands)
    {
        this.commands = commands;
    }

    /**
     * Instruct the surface to draw the collected set of nodes now. The
     * registered view environment is used to draw to this surface. If no
     * view is registered, the surface is cleared and then this call is
     * exited. The drawing surface does not swap the buffers at this point.
     */
    public void draw()
    {
        AudioInstructions localCommands = commands;
        int len = localCommands.numValid;

        if(initFailed || len == 0)
            return;

        if (len > lastId)
            lastId = len;

        // tell the draw lock that it's ok to run now, so long as it's not called
        // before the canvas has completed initialisation.
        if(!initComplete)
        {
            try
            {
                al = ALFactory.getAL();

                ALFactory.initialize();
                ALC alc = ALFactory.getALC();
                device = alc.alcOpenDevice(null);

                context = alc.alcCreateContext(device, null);
                alc.alcMakeContextCurrent(context);
                al.alGetError();

                al.alDistanceModel(AL.AL_INVERSE_DISTANCE_CLAMPED);

                initComplete = true;
            }
            catch (OpenALException oae)
            {
                System.out.println("OpenAL initialisation error " +
                                   oae.getMessage());
                oae.printStackTrace();
                initFailed = true;
            }
        }
        else
        {
            if (len == 0)
                return;

            try
            {
                ALC alc = ALFactory.getALC();
                alc.alcMakeContextCurrent(context);
                Sound obj;
                AudioDetails[] details = localCommands.renderList;

                // Update Listener
                matrixUtils.inverse(localCommands.renderData.viewTransform, tmpMatrix);
                tmpMatrix.transform(listenerPoint,tmpPoint);

                al.alListener3f(AL.AL_POSITION, tmpPoint.x, tmpPoint.y, tmpPoint.z);
                al.alListenerfv(AL.AL_ORIENTATION, DEFAULT_ORIENTATION);

                for(int i = 0; i < len && !terminate; i++)
                {
                    switch(localCommands.renderOps[i])
                    {
                        case RenderOp.START_RENDER:
                            obj = (Sound)details[i].renderable;
                            obj.render(al, details[i].transform);
                            break;

                        case RenderOp.STOP_RENDER:
                            obj = (Sound)details[i].renderable;
                            obj.postRender(al);
                            break;
                        default:
                            System.out.println("Unknown OP in AudioDevice");

                    }
                }

                alc.alcFreeCurrentContext();

                if(terminate)
                    return;
            }
            catch(OpenALException oae)
            {
                System.out.println("OpenAL exception!");
                oae.printStackTrace();
            }
        }
    }

    /**
     * Get the underlying object that this surface is rendered to. If it is a
     * screen display device, the surface can be one of AWT Component or
     * Swing JComponent. An off-screen buffer would be a form of AWT Image etc.
     *
     * @return The drawable surface representation
     */
    public Object getSurfaceObject()
    {
        return null;
    }

    /**
     * Instruct this surface that you have finished with the resources needed
     * and to dispose all rendering resources.
     */
    public void dispose()
    {
        // OpenAL has some shutdown code to do this.  This code occasionally
        // dies, so lets not run it for now.  We should cleanup all
        // sources and buffers though.
        terminate = true;


        if (initComplete) {
            try {

                ALC alc = ALFactory.getALC();
                alc.alcMakeContextCurrent(context);

                // TODO: Assume id's are issued sequentially.  Really need
                // a better way todo this.  OpenGL cleans this up for us, make joal do it?
                int[] buffer = new int[lastId+1];

                for(int i=0; i < lastId+1; i++) {
                    buffer[i] = i;
                }

                al.alDeleteBuffers(buffer.length, buffer);
                al.alDeleteSources(buffer.length, buffer);
                alc.alcFreeCurrentContext();


                alc.alcMakeContextCurrent(null);

                alc.alcDestroyContext(context);

                alc.alcCloseDevice(device);
            } catch(OpenALException oal) {
                oal.printStackTrace();
            }
        }
    }

    /**
     * Tell the surface to use the given thread as the rendering control thread.
     * Used when you want to clock the renderer to the rest of the pipeline,
     * rather than letting it sit in it's own loop. Due to
     * bugs in JOGL, we have to start the internal rendering threads after the
     * containing window has become visible. See JOGL issue #54 for more details
     * on this bug.
     *
     * @param th The render control thread instance to use or null
     */
    public void setRenderControlThread(Thread th)
    {
        // TODO: Should we call makeContextCurrent?
/*
        if(canvas.getRenderingThread() == th)
            return;

//        canvas.setRenderingThread(th);
        canvas.setNoAutoRedrawMode(th != null);
*/
    }

}
