/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
// None

// Local imports
import org.j3d.aviatrix3d.*;

/**
 * The default implementation of the audio pipeline usable by most
 * applications.
 * <p>
 * This implementation is targeted towards single threaded architectures.
 * After setting the stages, the render command will not return until
 * everything is complete.
 * <p>
 *
 * @author Alan Hudson
 * @version $Revision: 1.4 $
 */
public class DefaultAudioPipeline implements AudioPipeline
{
    /** The culling stage to be used */
    private AudioCullStage culler;

    /** The sorting stage to be used */
    private AudioSortStage sorter;

    /** Surface to draw to */
    private AudioDevice device;

    /** The listener to pass cull to sort */
    private AudioCullToSingleSortListener ctsListener;

    /** The listener to pass sort to the drawable listener */
    private AudioSortToSingleDeviceListener stdListener;

    /** Root of the scene this pipeline manages */
    private Scene sceneRoot;

    /**
     * Create an instance of the pipeline with nothing registered.
     */
    public DefaultAudioPipeline()
    {
        ctsListener = new AudioCullToSingleSortListener();
        stdListener = new AudioSortToSingleDeviceListener();
    }

    /**
     * Construct a pipeline with the sort and cull stages provided.
     *
     * @param ss The sort stage instance to use
     * @param cs The cull stage instance to use
     */
    public DefaultAudioPipeline(AudioCullStage cs, AudioSortStage ss)
    {
        this();

        culler = cs;
        sorter = ss;

        if(cs != null)
            cs.setCulledAudioReceiver(ctsListener);

        if(ss != null)
        {
            ctsListener.setSorter(ss);
            ss.setSortedAudioReceiver(stdListener);
        }
    }

    //---------------------------------------------------------------
    // Methods defined by AudioPipeline
    //---------------------------------------------------------------

    /**
     * Register a drawing surface that this pipeline will send its output to.
     * Setting a null value will remove the current drawable surface.
     *
     * @param device The audio output device instance to use or replace
     */
    public void setAudioDevice(AudioDevice device)
    {
        stdListener.setDevice(device);

        this.device = device;
    }

    /**
     * Get the currently registered drawable device instance. If none is set,
     * return null.
     *
     * @return The currently set surface instance or null
     */
    public AudioDevice getAudioDevice()
    {
        return device;
    }

    /**
     * Start the pipeline functioning now. All steps will be called and this
     * method will not return until all are completed and the surface has
     * swapped.
     */
    public void render()
    {
        if(sceneRoot == null)
            return;

        if(culler != null)
        {
            culler.cull(sceneRoot, 0);

            // then draw after the cull is complete
            if(device != null)
                device.draw();
        }
    }

    /**
     * Cause the surface to redraw the next frame only, with no processing of
     * the pipeline. This is typically an optimisation step when nothing has
     * changed in user land, so there's no processing that needs to be done.
     * Skip the processing and tell the drawable surface to render again what
     * it already has set from the previous frame.
     */
    public void displayOnly()
    {
        if((sceneRoot != null) && (device != null))
            device.draw();
    }

    /**
     * Set the root of the scene graph to be used by this pipeline. A value of
     * null will remove the scene from being rendered, causing the pipeline to
     * function as a no-op when rendered.
     *
     * @param scene The scene to be rendered
     */
    public void setRenderableScene(Scene scene)
    {
        sceneRoot = scene;
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        if(culler != null)
            culler.halt();

        if(sorter != null)
            sorter.halt();
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Set the sorter instance to be used. If the instance is null, the current
     * sorter is removed.
     *
     * @param ss The sorter instance to use or null
     */
    public void setSorter(AudioSortStage ss)
    {
        ctsListener.setSorter(ss);

        if(sorter != null)
            sorter.setSortedAudioReceiver(null);

        if(ss != null)
            ss.setSortedAudioReceiver(stdListener);

        sorter = ss;
    }

    /**
     * Set the cull instance to be used. If the instance is null, the current
     * culler is removed.
     *
     * @param cs The cull instance to use or null
     */
    public void setCuller(AudioCullStage cs)
    {
        if(culler != null)
            culler.setCulledAudioReceiver(null);

        if(cs != null)
        {
            cs.setCulledAudioReceiver(ctsListener);
        }

        culler = cs;
    }
}
