/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.util.ArrayList;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.MatrixUtils;

/**
 * The default internal implementation of the pick handling system
 * that handles batch picking requests.
 * <p>
 *
 * Used for processing batch requests. The difference between this and mainline
 * picking is that instead of recursing down the scene graph with a single
 * request, each level is processed for all requests that are still valid before
 * descending to the next level. This allows for a lot of optimisation where
 * only a single traversal of the tree is needed for all pick requests.
 *
 * @author Justin Couch
 * @version $Revision: 1.7 $
 */
class DefaultBatchPickingHandler
{
    /** The initial size of the children list */
    private static final int LIST_START_SIZE = 32;

    /** The increment size of the list if it gets overflowed */
    private static final int LIST_INCREMENT = 8;

    /**
     * Current active depth flags for deciding whether to continue down this
     * stent path or it has already been terminated due to non-intersection
     * at this point or some higher level.
     */
    private int[] activePicks;

    /** Path from the root of the scene graph to the stent place */
    private Node[] pickPath;

    /** Path of transforms from the root of the scene to the stent place */
    private Matrix4f[] transformPath;

    /** Flags to say with of the transforms are stently valid */
    private boolean[] validTransform;

    /** Last index of valid items on the path */
    private int lastPathIndex;

    /** Distance for the closest object found to date */
    private float closestDistance;

    /**
     * The working vector for start and finish locations or direction. The
     * first dimension corresponds to the activePicks index ie Which of the
     * requests in the batch are we stently looking at. The second dimension
     * is the pick depth down the transform heirarchy. This will correspond to
     * the depth down the scene graph we're at (lastPathIndex). The third
     * dimension is the temporary values for the stent depth.
     */
    private float[][][] start;
    private float[][][] end;
    private float[][][] extraData;

    /** Place to fetch the intersection point data for the line picking */
    private float[] vertexPickData;

    /** The matrix to set everything in and then invert it. */
    private Matrix4f vworldMatrix;

    /** The matrix containing the local inverted form. */
    private Matrix4f invertedMatrix;

    /** Working vector for interacting with the world matrix */
    private Vector4f wkVec;

    /** Working vector for interacting with the world matrix */
    private Vector3f wkNormal;

    /** Temp value for holding the frustum planes */
    private Vector4f[][] frustumPlanes;

    /** Matrix utility code for doing inversions */
    private MatrixUtils matrixUtils;

    /**
     * Initialise a new instance of the pick handler.
     */
    public DefaultBatchPickingHandler()
    {
        vertexPickData = new float[3];
        pickPath = new Node[LIST_START_SIZE];
        transformPath = new Matrix4f[LIST_START_SIZE];
        validTransform = new boolean[LIST_START_SIZE];

        for(int i = 0; i < LIST_START_SIZE; i++)
            transformPath[i] = new Matrix4f();

        vworldMatrix = new Matrix4f();
        invertedMatrix = new Matrix4f();
        wkVec = new Vector4f();
        wkVec.w = 1;

        wkNormal = new Vector3f();

        // Only assign activePicks because the first time through the call to
        // here we're going to reallocate everything anyway in resizeForBatch.
        activePicks = new int[0];

        matrixUtils = new MatrixUtils();
    }

    //---------------------------------------------------------------
    // Misc Internal methods
    //---------------------------------------------------------------


    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given set of requests. Assumes
     * there has been some earlier processing to weed out the situations where
     * calling this class would be useless.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @param numRequests The number of valid pick requests to process
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    void processPick(PickableObject root, PickRequest[] req, int numRequests)
        throws NotPickableException
    {
        resizeForBatch(numRequests);
        lastPathIndex = 0;
        transformPath[0].setIdentity();
        boolean keep_going = false;

        for(int i = 0; i < numRequests; i++)
        {
            activePicks[i] = 0;
            req[i].pickCount = 0;

            if(!(root instanceof Node))
            {
System.out.println("batchPick not handling non-node root");
                continue;
            }

            Node n_root = (Node)root;
            BoundingVolume bounds = n_root.getBounds();

            // Sort out the basic scene graph path info first before descending
            // into the quagmire.
            switch(req[i].pickSortType)
            {
                case PickRequest.SORT_ALL:
                case PickRequest.SORT_ORDERED:
                    if(!(req[i].foundPaths instanceof ArrayList))
                        req[i].foundPaths = new ArrayList();
                    break;

                case PickRequest.SORT_ANY:
                case PickRequest.SORT_CLOSEST:
                    if(!(req[i].foundPaths instanceof SceneGraphPath))
                        req[i].foundPaths = new SceneGraphPath();
                    break;
            }

            float[] st = start[i][0];
            float[] ed = end[i][0];
            float[] data = extraData[i][0];

            switch(req[i].pickGeometryType)
            {
                case PickRequest.PICK_POINT:
                    if(bounds.checkIntersectionPoint(req[i].origin))
                    {
                        st[0] = req[i].origin[0];
                        st[1] = req[i].origin[1];
                        st[2] = req[i].origin[2];

                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                case PickRequest.PICK_RAY:
                    if(bounds.checkIntersectionRay(req[i].origin,
                                                   req[i].destination))
                    {
                        st[0] = req[i].origin[0];
                        st[1] = req[i].origin[1];
                        st[2] = req[i].origin[2];

                        ed[0] = req[i].destination[0];
                        ed[1] = req[i].destination[1];
                        ed[2] = req[i].destination[2];

                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                case PickRequest.PICK_LINE_SEGMENT:
                    if(bounds.checkIntersectionSegment(req[i].origin,
                                                       req[i].destination))
                    {
                        st[0] = req[i].origin[0];
                        st[1] = req[i].origin[1];
                        st[2] = req[i].origin[2];

                        ed[0] = req[i].destination[0];
                        ed[1] = req[i].destination[1];
                        ed[2] = req[i].destination[2];

                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                case PickRequest.PICK_CYLINDER:
                case PickRequest.PICK_CYLINDER_SEGMENT:
                    float x = req[i].origin[0] - req[i].destination[0];
                    float y = req[i].origin[1] - req[i].destination[1];
                    float z = req[i].origin[2] - req[i].destination[2];

                    float height = (float)Math.sqrt(x * x + y * y + z * z);

                    if(height == 0)
                        return;

                    // Start will be the center and end will be the axis vector
                    st[0] = (req[i].origin[0] + req[i].destination[0]) * 0.5f;
                    st[1] = (req[i].origin[1] + req[i].destination[0]) * 0.5f;
                    st[2] = (req[i].origin[2] + req[i].destination[0]) * 0.5f;

                    ed[0] = x;
                    ed[1] = y;
                    ed[2] = z;

                    float radius = req[i].additionalData;

                    if(bounds.checkIntersectionCylinder(st, ed, radius, height))
                    {
                        data[0] = radius;
                        data[1] = height;

                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                case PickRequest.PICK_CONE:
                case PickRequest.PICK_CONE_SEGMENT:
                    if(bounds.checkIntersectionCone(req[i].origin,
                                                    req[i].destination,
                                                    req[i].additionalData))
                    {
                        st[0] = req[i].origin[0];
                        st[1] = req[i].origin[1];
                        st[2] = req[i].origin[2];

                        ed[0] = req[i].destination[0];
                        ed[1] = req[i].destination[1];
                        ed[2] = req[i].destination[2];

                        data[0] = req[i].additionalData;

                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                case PickRequest.PICK_BOX:
                    if(bounds.checkIntersectionBox(req[i].origin,
                                                   req[i].destination))
                    {
                        st[0] = req[i].origin[0];
                        st[1] = req[i].origin[1];
                        st[2] = req[i].origin[2];

                        ed[0] = req[i].destination[0];
                        ed[1] = req[i].destination[1];
                        ed[2] = req[i].destination[2];

                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                case PickRequest.PICK_FRUSTUM:
                    frustumPlanes[i][0].x = req[i].origin[0];
                    frustumPlanes[i][0].y = req[i].origin[1];
                    frustumPlanes[i][0].z = req[i].origin[2];
                    frustumPlanes[i][0].w = req[i].origin[3];

                    frustumPlanes[i][1].x = req[i].origin[4];
                    frustumPlanes[i][1].y = req[i].origin[5];
                    frustumPlanes[i][1].z = req[i].origin[6];
                    frustumPlanes[i][1].w = req[i].origin[7];

                    frustumPlanes[i][2].x = req[i].origin[8];
                    frustumPlanes[i][2].y = req[i].origin[9];
                    frustumPlanes[i][2].z = req[i].origin[10];
                    frustumPlanes[i][2].w = req[i].origin[11];

                    frustumPlanes[i][3].x = req[i].origin[12];
                    frustumPlanes[i][3].y = req[i].origin[13];
                    frustumPlanes[i][3].z = req[i].origin[14];
                    frustumPlanes[i][3].w = req[i].origin[15];

                    frustumPlanes[i][4].x = req[i].origin[16];
                    frustumPlanes[i][4].y = req[i].origin[17];
                    frustumPlanes[i][4].z = req[i].origin[18];
                    frustumPlanes[i][4].w = req[i].origin[19];

                    frustumPlanes[i][5].x = req[i].origin[20];
                    frustumPlanes[i][5].y = req[i].origin[21];
                    frustumPlanes[i][5].z = req[i].origin[22];
                    frustumPlanes[i][5].w = req[i].origin[23];

                    if(bounds.checkIntersectionFrustum(frustumPlanes[i],
                                                       transformPath[lastPathIndex])
                       == BoundingVolume.FRUSTUM_ALLOUT)
                    {
                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                case PickRequest.PICK_SPHERE:
                    if(bounds.checkIntersectionSphere(req[i].origin,
                                                      req[i].additionalData))
                    {
                        st[0] = req[i].origin[0];
                        st[1] = req[i].origin[1];
                        st[2] = req[i].origin[2];

                        data[0] = req[i].additionalData;

                        activePicks[i]++;
                        keep_going = true;
                    }
                    break;

                default:
    System.out.println("No picking type has been specified");
            }
        }

        if(keep_going)
            descendTree((Node)root, req, numRequests);
    }

    /**
     * Descend to the next level of the scene graph tree.
     *
     * @param parent The parent node that is to be descended into
     * @param req The list of picks to be made, starting at this object
     * @param numRequests The number of valid pick requests to process
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    private void descendTree(Node parent, PickRequest[] req, int numRequests)
    {
        BoundingVolume bounds = parent.getBounds();
        Node target_node = parent;
        int num_kids = 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            target_node = bg.getProxyGeometry();
        }

        // Early bug out if needed.

        if(target_node instanceof Group)
        {
            Group g = (Group)target_node;
            num_kids = g.numRenderableChildren();

            if(num_kids == 0)
                return;
        }
        else if(target_node instanceof SharedNode)
        {
            descendSharedTree((SharedNode)target_node, req, numRequests);
            return;
        }
        else if(target_node instanceof Shape3D)
        {
            testShape((Shape3D)target_node, req, numRequests);
            return;
        }

        // If the parent is a transform, we need to take the current values
        // and transform them across to the new items.
        if(target_node instanceof TransformGroup)
        {
            Matrix4f tx = transformPath[lastPathIndex];
            ((TransformGroup)target_node).getTransform(tx);

            matrixUtils.inverse(tx, invertedMatrix);
            float scale = invertedMatrix.getScale();

            // multiply the transform matrix through to the current level.
            tx.mul(transformPath[lastPathIndex - 1], tx);
            validTransform[lastPathIndex] = true;

            for(int i = 0; i < numRequests; i++)
            {
                if(activePicks[i] < lastPathIndex)
                    continue;

                switch(req[i].pickGeometryType)
                {
                    // do nothing else
                    case PickRequest.PICK_POINT:
                        transform(invertedMatrix,
                                  start[i][lastPathIndex],
                                  start[i][lastPathIndex + 1]);
                        break;

                    case PickRequest.PICK_RAY:
                        transform(invertedMatrix,
                                  start[i][lastPathIndex],
                                  start[i][lastPathIndex + 1]);
                        transformNormal(invertedMatrix,
                                        end[i][lastPathIndex],
                                        end[i][lastPathIndex + 1]);
                        break;

                    case PickRequest.PICK_LINE_SEGMENT:
                        transform(invertedMatrix,
                                  start[i][lastPathIndex],
                                  start[i][lastPathIndex + 1]);
                        transform(invertedMatrix,
                                  end[i][lastPathIndex],
                                  end[i][lastPathIndex + 1]);
                        break;

                    case PickRequest.PICK_CYLINDER:
                    case PickRequest.PICK_CYLINDER_SEGMENT:
                        transform(invertedMatrix,
                                  start[i][lastPathIndex],
                                  start[i][lastPathIndex + 1]);
                        transformNormal(invertedMatrix,
                                        end[i][lastPathIndex],
                                        end[i][lastPathIndex + 1]);

                        extraData[i][lastPathIndex + 1][0] =
                            scale * extraData[i][lastPathIndex][0];
                        extraData[i][lastPathIndex + 1][1] =
                            scale * extraData[i][lastPathIndex][1];
                        break;

                    case PickRequest.PICK_CONE:
                    case PickRequest.PICK_CONE_SEGMENT:
                        transform(invertedMatrix,
                                  start[i][lastPathIndex],
                                  start[i][lastPathIndex + 1]);
                        transformNormal(invertedMatrix,
                                        end[i][lastPathIndex],
                                        end[i][lastPathIndex + 1]);

                        extraData[i][lastPathIndex + 1][0] =
                            scale * extraData[i][lastPathIndex][0];
                        break;

                    case PickRequest.PICK_BOX:
                        transform(invertedMatrix,
                                  start[i][lastPathIndex],
                                  start[i][lastPathIndex + 1]);
                        transform(invertedMatrix,
                                  end[i][lastPathIndex],
                                  end[i][lastPathIndex + 1]);
                        break;

                    case PickRequest.PICK_FRUSTUM:
                        // do nothing here for frustum picking
                        break;

                    case PickRequest.PICK_SPHERE:
                        transform(invertedMatrix,
                                  start[i][lastPathIndex],
                                  start[i][lastPathIndex + 1]);

                        extraData[i][lastPathIndex + 1][0] =
                            scale * extraData[i][lastPathIndex][0];
                        break;

                    default:
                }
            }
        }
        else
        {
            // don't bother setting to identity, just mark it as non-valid.
            // Assignment is a waste of CPU cycles. Then copy over the details
            // from the level above.
            validTransform[lastPathIndex] = false;
            if(lastPathIndex == 0)
                transformPath[lastPathIndex].setIdentity();
            else
                transformPath[lastPathIndex].set(transformPath[lastPathIndex - 1]);

            for(int i = 0; i < numRequests; i++)
            {
                if(activePicks[i] < lastPathIndex)
                    continue;

                start[i][lastPathIndex + 1][0] = start[i][lastPathIndex][0];
                start[i][lastPathIndex + 1][1] = start[i][lastPathIndex][1];
                start[i][lastPathIndex + 1][2] = start[i][lastPathIndex][2];

                end[i][lastPathIndex + 1][0] = end[i][lastPathIndex][0];
                end[i][lastPathIndex + 1][1] = end[i][lastPathIndex][1];
                end[i][lastPathIndex + 1][2] = end[i][lastPathIndex][2];

                extraData[i][lastPathIndex + 1][0] = extraData[i][lastPathIndex][0];
                extraData[i][lastPathIndex + 1][1] = extraData[i][lastPathIndex][1];
            }
        }

        Group g = (Group)target_node;
        Node[] kids = g.getRenderableChild();
        boolean keep_going;

        pickPath[lastPathIndex] = g;
        lastPathIndex++;

        for(int i = 0; i < num_kids; i++)
        {
            keep_going = false;

            if(kids[i] == null)
                continue;

            bounds = kids[i].getBounds();

            for(int j = 0; j < numRequests; j++)
            {
                if(activePicks[j] < lastPathIndex)
                    continue;

                switch(req[j].pickGeometryType)
                {
                    // do nothing else
                    case PickRequest.PICK_POINT:
                        if(bounds.checkIntersectionPoint(start[j][lastPathIndex]))
                        {
                            keep_going = true;
                            activePicks[j]++;
                        }
                        break;

                    case PickRequest.PICK_RAY:
                        if(bounds.checkIntersectionRay(start[j][lastPathIndex],
                                                       end[j][lastPathIndex]))
                        {
                            keep_going = true;
                            activePicks[j]++;
                        }
                        break;

                    case PickRequest.PICK_LINE_SEGMENT:
                        if(bounds.checkIntersectionSegment(start[j][lastPathIndex],
                                                           end[j][lastPathIndex]))
                        {
                            keep_going = true;
                            activePicks[j]++;
                        }
                        break;

                    case PickRequest.PICK_CYLINDER:
                    case PickRequest.PICK_CYLINDER_SEGMENT:
                        if(bounds.checkIntersectionCylinder(start[j][lastPathIndex],
                                                            end[j][lastPathIndex],
                                                            extraData[j][lastPathIndex][0],
                                                            extraData[j][lastPathIndex][1]))
                        {
                            keep_going = true;
                            activePicks[j]++;
                        }
                        break;

                    case PickRequest.PICK_CONE:
                    case PickRequest.PICK_CONE_SEGMENT:
                        if(bounds.checkIntersectionCone(start[j][lastPathIndex],
                                                        end[j][lastPathIndex],
                                                        extraData[j][lastPathIndex][0]))
                        {
                            keep_going = true;
                            activePicks[j]++;
                        }
                        break;

                    case PickRequest.PICK_BOX:
                        if(bounds.checkIntersectionBox(start[j][lastPathIndex],
                                                       end[j][lastPathIndex]))
                        {
                            keep_going = true;
                                activePicks[j]++;
                        }
                        break;

                    case PickRequest.PICK_FRUSTUM:
                        if(bounds.checkIntersectionFrustum(frustumPlanes[j],
                                                           transformPath[lastPathIndex])
                           == BoundingVolume.FRUSTUM_ALLOUT)
                        {
                            keep_going = true;
                            activePicks[j]++;
                        }
                        break;

                    case PickRequest.PICK_SPHERE:
                        if(bounds.checkIntersectionSphere(start[j][lastPathIndex],
                                                         extraData[j][lastPathIndex][0]))
                        {
                            keep_going = true;
                            activePicks[j]++;
                        }
                        break;

                    default:
                }
            }

            if(keep_going)
            {
                descendTree(kids[i], req, numRequests);

                // decrement matching picks
                for(int j = 0; j < numRequests; j++)
                {
                    if(activePicks[j] > lastPathIndex)
                        activePicks[j]--;
                }
            }
        }

        lastPathIndex--;
        pickPath[lastPathIndex] = null;
    }

    /**
     * We've hit a shared node. Descend down it's children until we find a real
     * node instance again.
     *
     * @param parent The shared node that is to be descended
     * @param req The list of picks to be made, starting at this object
     * @param numRequests The number of valid pick requests to process
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    private void descendSharedTree(SharedNode parent,
                                   PickRequest[] req,
                                   int numRequests)
    {
        int pick_mask = parent.getPickMask();
        boolean keep_going = false;

        for(int i = 0; i < numRequests; i++)
        {
            if((pick_mask & req[i].pickType) != 0)
            {
                keep_going = true;
                break;
            }
        }

        if(!keep_going)
            return;

        Node child = parent.getChild();

        if(child instanceof Group)
        {
            descendTree((Group)child, req, numRequests);
        }
        else if(child instanceof SharedNode)
        {
            descendSharedTree((SharedNode)child, req, numRequests);
        }
        else if(child instanceof Shape3D)
        {
            testShape((Shape3D)child, req, numRequests);
        }
    }

    /**
     * Do the final checks on a shape to see if it can be intersected by the
     * geometry.
     *
     * @param target The shape that is to be tested
     * @param req The list of picks to be made, starting at this object
     * @param numRequests The number of valid pick requests to process
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    private void testShape(Shape3D target, PickRequest[] req, int numRequests)
    {
        int pick_mask = target.getPickMask();
        boolean intersect;
        BoundingVolume bounds = target.getBounds();

        for(int i = 0; i < numRequests; i++)
        {
            // Only process this if we have an intersection for the parent,
            // there is the right pick type to match what we're looking for,
            // or we are not in pick_any mode with a match already.
            if((activePicks[i] < lastPathIndex) ||
               ((pick_mask & req[i].pickType) == 0) ||
                ((req[i].pickSortType == PickRequest.SORT_ANY) &&
                 (req[i].pickCount == 1)))
                continue;

            intersect = false;

            switch(req[i].pickGeometryType)
            {
                // do nothing else
                case PickRequest.PICK_POINT:
                    intersect =
                        bounds.checkIntersectionPoint(start[i][lastPathIndex]);
                    break;

                case PickRequest.PICK_RAY:
                    if(bounds.checkIntersectionRay(start[i][lastPathIndex],
                                                   end[i][lastPathIndex]))
                    {
                        if(req[i].useGeometry)
                        {
                            Geometry geom = target.getGeometry();
                            boolean sort =
                                req[i].pickSortType == PickRequest.SORT_CLOSEST ||
                                req[i].pickSortType == PickRequest.SORT_ORDERED;


                            intersect = geom.pickLineRay(start[i][lastPathIndex],
                                                         end[i][lastPathIndex],
                                                         sort,
                                                         vertexPickData,
                                                         0);

                            // TODO:
                            // Need to have something here that checks distance
                            // and updates only when required.
                            intersect = true;
                        }
                        else
                            intersect = true;
                    }
                    break;

                case PickRequest.PICK_LINE_SEGMENT:
                    if(bounds.checkIntersectionSegment(start[i][lastPathIndex],
                                                       end[i][lastPathIndex]))
                    {
                        if(req[i].useGeometry)
                        {
                            Geometry geom = target.getGeometry();
                            boolean sort =
                                req[i].pickSortType == PickRequest.SORT_CLOSEST ||
                                req[i].pickSortType == PickRequest.SORT_ORDERED;


                            intersect = geom.pickLineSegment(start[i][lastPathIndex],
                                                             end[i][lastPathIndex],
                                                             sort,
                                                             vertexPickData,
                                                             0);
                            // TODO:
                            // Need to have something here that checks distance
                            // and updates only when required.
                            intersect = true;
                        }
                        else
                            intersect = true;
                    }
                    break;

                case PickRequest.PICK_CYLINDER:
                case PickRequest.PICK_CYLINDER_SEGMENT:
                    intersect =
                        bounds.checkIntersectionCylinder(start[i][lastPathIndex],
                                                         end[i][lastPathIndex],
                                                         extraData[i][lastPathIndex][0],
                                                         extraData[i][lastPathIndex][1]);
                    break;

                case PickRequest.PICK_CONE:
                case PickRequest.PICK_CONE_SEGMENT:
                    intersect =
                        bounds.checkIntersectionCone(start[i][lastPathIndex],
                                                     end[i][lastPathIndex],
                                                     extraData[i][lastPathIndex][0]);
                    break;

                case PickRequest.PICK_BOX:
                    intersect =
                        bounds.checkIntersectionBox(start[i][lastPathIndex],
                                                    end[i][lastPathIndex]);
                    break;

                case PickRequest.PICK_FRUSTUM:
                    intersect =
                        (bounds.checkIntersectionFrustum(frustumPlanes[i],
                                                         transformPath[lastPathIndex])
                           == BoundingVolume.FRUSTUM_ALLOUT);
                    break;

                case PickRequest.PICK_SPHERE:
                    intersect =
                        bounds.checkIntersectionSphere(start[i][lastPathIndex],
                                                       extraData[i][lastPathIndex][0]);
                    break;

                default:
            }

            // TODO:
            // Optimise this heavily. We should only need to calculate all the
            // vworld transforms etc just once per call to this method, not for
            // every iteration of the loop. Need some flags at the top ofthe
            // method to indicate if these have already been calculated.
            if(intersect)
            {
                resizePath();
                pickPath[lastPathIndex] = target;
                validTransform[lastPathIndex] = false;
                lastPathIndex++;

                SceneGraphPath p = null;

                switch(req[i].pickSortType)
                {
                    case PickRequest.SORT_ALL:
                    case PickRequest.SORT_ORDERED:
                        ArrayList paths = (ArrayList)req[i].foundPaths;
                        if(req[i].pickCount >= paths.size())
                        {
                            p = new SceneGraphPath();
                            paths.add(p);
                        }
                        else
                            p = (SceneGraphPath)paths.get(req[i].pickCount);

                        break;

                    case PickRequest.SORT_ANY:
                    case PickRequest.SORT_CLOSEST:
                        p = (SceneGraphPath)req[i].foundPaths;
                        break;
                }

                // Note that the transform path is already pre-multiplied
                // through here. Just and invert is all that needs to be done.
                if(req[i].generateVWorldMatrix)
                    matrixUtils.inverse(transformPath[lastPathIndex - 1],
                                        vworldMatrix);
                else
                    vworldMatrix.setIdentity();

                p.updatePath(pickPath,
                             lastPathIndex,
                             transformPath[lastPathIndex - 1],
                             vworldMatrix);

                req[i].pickCount++;
                lastPathIndex--;
            }
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizePath()
    {
        if(lastPathIndex == pickPath.length)
        {
            int old_size = pickPath.length;
            int new_size = old_size + LIST_INCREMENT;

            Node[] tmp_nodes = new Node[new_size];

            System.arraycopy(pickPath, 0, tmp_nodes, 0, old_size);

            pickPath = tmp_nodes;

            Matrix4f[] tmp_tx = new Matrix4f[new_size];
            System.arraycopy(transformPath, 0, tmp_tx, 0, old_size);
            transformPath = tmp_tx;

            for(int i = old_size; i < new_size; i++)
                transformPath[i] = new Matrix4f();

            boolean[] tmp_flags = new boolean[new_size];
            System.arraycopy(validTransform, 0, tmp_flags, 0, old_size);
            validTransform = tmp_flags;
        }
    }

    /**
     * Resize all the top level arrays that depend on the number of items in
     * the batch picking request. Only called at the start of the request.
     *
     * @param numBatch The number of valid items in the request
     */
    private void resizeForBatch(int numBatch)
    {
        if(numBatch <= activePicks.length)
            return;

        // Just wipe out the old ones with new values. No point copying over
        // as this should happen vary rarely - probably only once in the
        // lifecycle of the application.
        activePicks = new int[numBatch];

        int depth = lastPathIndex == 0 ? LIST_START_SIZE : lastPathIndex;
        start = new float[numBatch][depth][3];
        end = new float[numBatch][depth][3];
        extraData = new float[numBatch][depth][2];

        frustumPlanes = new Vector4f[numBatch][6];

        for(int i = 0; i < numBatch; i++)
        {
            frustumPlanes[i][0] = new Vector4f();
            frustumPlanes[i][1] = new Vector4f();
            frustumPlanes[i][2] = new Vector4f();
            frustumPlanes[i][3] = new Vector4f();
            frustumPlanes[i][4] = new Vector4f();
            frustumPlanes[i][5] = new Vector4f();
        }
    }

    /**
     * Transform the location vector by the matrix. Location includes the
     * translation offset.
     *
     * @param mat The matrix to do the transformation with
     * @param vec The vector to be changed
     * @param out The vector to put the updated values in
     */
    private void transform(Matrix4f mat, float[] vec, float[] out)
    {
        float x = vec[0];
        float y = vec[1];
        float z = vec[2];

        out[0] = x * mat.m00 + y * mat.m01 + z * mat.m02 + mat.m03;
        out[1] = x * mat.m10 + y * mat.m11 + z * mat.m12 + mat.m13;
        out[2] = x * mat.m20 + y * mat.m21 + z * mat.m22 + mat.m23;
    }

    /**
     * Transform the normal vector by the matrix. The normal vector does not
     * include the translation offset.
     *
     * @param mat The matrix to do the transformation with
     * @param vec The vector to be changed
     * @param out The vector to put the updated values in
     */
    private void transformNormal(Matrix4f mat, float[] vec, float[] out)
    {
        float x = vec[0];
        float y = vec[1];
        float z = vec[2];

        out[0] = x * mat.m00 + y * mat.m01 + z * mat.m02;
        out[1] = x * mat.m10 + y * mat.m11 + z * mat.m12;
        out[2] = x * mat.m20 + y * mat.m21 + z * mat.m22;
    }
}
