/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.util.HashMap;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.MatrixUtils;

/**
 * Implementation of the sort stage that separates out the transparent and
 * non transparent objects, and then arranges them using depth sorting.
 * <p>
 *
 * The depth sorting algorithm used is to first assemble all the transparen
 * objects into a list. Calculate their distance from the screen using the
 * center of the bounds as the world space position. Normalise the depth to
 * an integer between 0 and 32K (16bit depth buffer equivalent) where 0 is the
 * near clipping plane and 32K is the far clipping plane. With the distance
 * values, perform a counting-sort operation to sort the objects into the
 * correct order with deepest first.
 *
 *
 * @author Justin Couch
 * @version $Revision: 1.13 $
 */
public class DepthSortedTransparencyStage implements SortStage
{
    /** Message when the number of depth bits <= 0 */
    private static final String INVALID_DEPTH_MSG =
        "The depthBits must be greater than or equal to 1";

    /** The initial size of the instruction list */
    private static final int LIST_START_SIZE = 1;

    /** The initial size of the transparent list */
    private static final int TRANSPARENT_LIST_SIZE = 200;

    /** Default accuracy bits. Matches most common video cards */
    private static final int DEFAULT_DEPTH_BITS = 16;

    /**
     * The accuracy of the depth quantisation. This will be a power of two
     * value that is used to setup the depth list for input to the counting
     * sort.
     */
    private final int depthAccuracy;

    /** Receiver for the output of this stage */
    private SortedGeometryReceiver receiver;

    /** Output array for passing on to the receiver */
    private RenderInstructions[] commandList;

    /** Temporary array for holding the transparent objects */
    private CullOutputDetails[] transparentList;

    /** Temporary array for holding the transparent objects prior to sorting */
    private CullOutputDetails[] unsortedTransparentList;

    /** Temporary array for holding the transparent object depths */
    private float[] depthList1;

    /** Temporary array for holding the quantized depths */
    private int[] depthList2;

    /** The count list used during sorting */
    private int[] countList;

    /** Temp array for reading texture units from the appearance */
    private TextureUnit[] textureTmp;

    /** A semi-unique ID counter used for assigning light IDs. May wrap */
    private int lastGlobalId;

    /** Map of the first occurance of a scene parent to it's render instruction */
    private HashMap instructionMap;

    /** Temp matrix used to hold the local to vworld values */
    private Matrix4f modelMatrix;

    /** Temp matrix used to hold inverted camera matrix */
    private Matrix4f cameraMatrix;

    /** Temporary point location while working out the camera-space depth */
    private Point3f wkPoint;

    /** Temporary array for fetching the center position */
    private float[] center;

    /** Matrix utility code for doing inversions */
    private MatrixUtils matrixUtils;

    /** Flag indicating a shutdown of the current processing is requested */
    private boolean terminate;

    /**
     * Create an empty sorting stage that assumes just a single renderable
     * output. The default list size is 1 and the number of depth bits is
     * 16.
     */
    public DepthSortedTransparencyStage()
    {
        this(LIST_START_SIZE, DEFAULT_DEPTH_BITS);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen. The number of depth bits is set to 16.
     *
     * @param numSurfaces The number of surfaces to start the internal lists
     *     sizes at.
     */
    public DepthSortedTransparencyStage(int numSurfaces)
    {
        this(numSurfaces, DEFAULT_DEPTH_BITS);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen.
     *
     * @param numSurfaces The number of surfaces to start the internal lists
     *     sizes at.
     * @param depthBits The number of bits of depth precision to quantize the
     *     depth sorting to
     * @throws IllegalArgumentException Depth bits is zero or negative
     */
    public DepthSortedTransparencyStage(int numSurfaces, int depthBits)
    {
        if(depthBits < 1)
            throw new IllegalArgumentException(INVALID_DEPTH_MSG);

        int depth = 1;

        for(int i = 1; i < depthBits; i++)
            depth <<= 1;

        depthAccuracy = depth;

        transparentList = new CullOutputDetails[TRANSPARENT_LIST_SIZE];
        unsortedTransparentList = new CullOutputDetails[TRANSPARENT_LIST_SIZE];
        depthList1 = new float[TRANSPARENT_LIST_SIZE];
        depthList2 = new int[TRANSPARENT_LIST_SIZE];
        countList = new int[depthAccuracy];

        textureTmp = new TextureUnit[32];
        instructionMap = new HashMap();
        wkPoint = new Point3f();
        modelMatrix = new Matrix4f();
        cameraMatrix = new Matrix4f();
        center = new float[3];

        lastGlobalId = 0;

        commandList = new RenderInstructions[numSurfaces];

        for(int i = 0; i < numSurfaces; i++)
            commandList[i] = new RenderInstructions();

        matrixUtils = new MatrixUtils();
    }

    /**
     * Sort the listing of nodes in the given array. Do not return until the
     * sort has been completed. If the sceneParent is the main scene graph and
     * not an offscreen texture, then that spot should be set to null in the list.
     *
     * @param data External rendering environment information
     * @param nodes List of processed nodes based on the scene they came from
     * @param numNodes Number of nodes in each scene
     * @param sceneParents Parent nodes that hold the subscene in the main scene graph
     * @param numScenes The number of valid scenes to process
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void sort(RenderEnvironmentData[] data,
                     CullOutputDetails[][] nodes,
                     int[] numNodes,
                     SceneGraphObject[][] sceneParents,
                     int numScenes,
                     int buffer)
    {
        // Buffer ID is not used currently.

        terminate = false;

        if(commandList.length < numScenes)
        {
            RenderInstructions[] tmp = new RenderInstructions[numScenes];

            System.arraycopy(commandList, 0, tmp, 0, commandList.length);

            for(int i = commandList.length; i < numScenes; i++)
                tmp[i] = new RenderInstructions();

            commandList = tmp;
        }

        // While doing these loops, we need to find the main scene and ensure
        // that it is the last item in the output list.
        int main_idx = 0;
        int curr_scene = 0;

        for(int i = 0; i < numScenes && !terminate; i++)
        {
            if(sceneParents[i][0] == null)
            {
                main_idx = i;
                continue;
            }

            commandList[curr_scene].pbuffer =
                (PBufferTextureSource)sceneParents[i][0];
            commandList[curr_scene].parentSource =
                (PBufferTextureSource)sceneParents[i][1];
            commandList[curr_scene].renderData = data[i];

            RenderInstructions original =
                (RenderInstructions)instructionMap.get(sceneParents[i][0]);

            if(original != null)
            {
                commandList[curr_scene].copyOf = original;
            }
            else
            {
                commandList[curr_scene].copyOf = null;
                sortSingle(data[i].viewTransform,
                           nodes[i],
                           numNodes[i],
                           commandList[curr_scene]);

                instructionMap.put(sceneParents[i][0],
                                   commandList[curr_scene]);
            }

            curr_scene++;
        }

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        commandList[curr_scene].pbuffer = null;
        commandList[curr_scene].renderData = data[main_idx];

        sortSingle(data[main_idx].viewTransform,
                   nodes[main_idx],
                   numNodes[main_idx],
                   commandList[curr_scene]);

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        curr_scene++;

        if(receiver != null)
            receiver.sortedOutput(commandList, numScenes);

        instructionMap.clear();
    }


    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setSortedGeometryReceiver(SortedGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        terminate = true;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Sort a single set of nodes into the output details and place in the
     * provided RenderInstructions instance.
     *
     * @param instr Instruction instant to put the details into
     */
    private void sortSingle(Matrix4f camera,
                            CullOutputDetails[] nodes,
                            int numNodes,
                            RenderInstructions instr)
    {
        // check to see the array is big enough. The +2 is to accomodate the
        // start and stop transparency commands.
        int req_size = 2 + numNodes << 1;
        for(int i = 0; i < numNodes; i++)
            req_size += nodes[i].numLights <<1;

        if(instr.renderList.length < req_size)
        {
            RenderDetails[] tmp = new RenderDetails[req_size];

            System.arraycopy(instr.renderList,
                             0,
                             tmp,
                             0,
                             instr.renderList.length);

            for(int i = instr.renderList.length; i < req_size; i++)
                tmp[i] = new RenderDetails();

            instr.renderList = tmp;
            instr.renderOps = new int[req_size];
        }

        if(transparentList.length < req_size)
        {
            transparentList = new CullOutputDetails[req_size];
            unsortedTransparentList = new CullOutputDetails[req_size];
            depthList1 = new float[req_size];
            depthList2 = new int[req_size];
        }

        int idx = 0;
        int trans = 0;

        for(int i = 0; i < numNodes && !terminate; i++)
        {
            // First check the geometry and so forth for something that
            // has an alpha channel. Need to check everything - geometry
            // colour values, textures and materials. Anything that says it
            // may have alpha is passed off onto the transparent list.
            // Now process the node type.
            if(nodes[i].renderable instanceof Shape3D)
            {
                Shape3D shape = (Shape3D)nodes[i].renderable;

                Appearance app = shape.getAppearance();

                if(app != null)
                {
                    // Anything that has explicit blend ops is not counted in
                    // the normal transparency sorting. We'll dump it back into
                    // the non-transparent bucket, not transparent.
                    BlendAttributes blender = app.getBlendAttributes();
                    if(blender != null)
                    {
                        unsortedTransparentList[trans++] = nodes[i];
                        continue;
                    }

                    Material mat = app.getMaterial();
                    if(mat != null && mat.getTransparency() != 1)
                    {
                        unsortedTransparentList[trans++] = nodes[i];
                        continue;
                    }

                    int num_tex = app.numTextureUnits();
                    if(num_tex != 0)
                    {
                        boolean alpha_found = false;

                        app.getTextureUnits(textureTmp);
                        for(int j = 0; j < num_tex; j++)
                        {
                            Texture tex = textureTmp[j].getTexture();
                            if(tex == null)
                                continue;

                            // Any of the following formats, this is is added
                            // to the transparency collection. Break out of the
                            // inner loop by messing with the loop index.
                            switch(tex.getFormat())
                            {
                                case Texture.FORMAT_ALPHA:
                                case Texture.FORMAT_INTENSITY_ALPHA:
                                case Texture.FORMAT_RGBA:
                                    unsortedTransparentList[trans++] = nodes[i];
                                    alpha_found = true;
                                    j = num_tex;
                                    break;

                            }
                        }

                        if(alpha_found)
                            continue;
                    }
                }

                Geometry geom = shape.getGeometry();
                if(geom.hasTransparency())
                {
                    unsortedTransparentList[trans++] = nodes[i];
                    continue;
                }
            }
            else if(nodes[i].renderable instanceof CustomRenderable)
            {
                if(((CustomRenderable)nodes[i].renderable).hasTransparency())
                {
                    unsortedTransparentList[trans++] = nodes[i];
                    continue;
                }
            }
            else
            {
                System.out.println("Non-shape node in sorter " +
                                   nodes[i].renderable);
                continue;
            }

            idx = appendOutputShape(idx, nodes[i], instr);
        }

        // Now process the transparency listing. If there is no transparent
        // objects, then exit now.
        switch(trans)
        {
            case 0:
                instr.numValid = idx;
                return;

            case 1:
                instr.renderOps[idx++] = RenderOp.START_TRANSPARENT;

                idx = appendOutputShape(idx,
                                        unsortedTransparentList[0],
                                        instr);

                // clear it from the list
                unsortedTransparentList[0] = null;

                instr.renderOps[idx++] = RenderOp.STOP_TRANSPARENT;
                instr.numValid = idx;
                return;
        }

        // Time to work out the distance values.
        float max_depth = Float.NEGATIVE_INFINITY;
        float min_depth = Float.POSITIVE_INFINITY;

        matrixUtils.inverse(camera, cameraMatrix);

        for(int i = 0; i < trans && !terminate; i++)
        {
            // transform the center of the bounds first to world space and then
            // into camera space to work out the depth.
            modelMatrix.set(unsortedTransparentList[i].transform);
            Shape3D shape = (Shape3D)unsortedTransparentList[i].renderable;
            BoundingVolume bounds = shape.getBounds();
            bounds.getCenter(center);

            wkPoint.x = center[0];
            wkPoint.y = center[1];
            wkPoint.z = center[2];

            modelMatrix.transform(wkPoint);
            cameraMatrix.transform(wkPoint);

            depthList1[i] = wkPoint.z;

            if(max_depth < wkPoint.z)
                max_depth = wkPoint.z;

            if(min_depth > wkPoint.z)
                min_depth = wkPoint.z;
        }


        // Now work out the range of values available and how to quantize those
        // to the given accuracy. Take 1 from the depth accuracy so that the max
        // depth will always end up at one less than the final array size - to
        // avoid AAOB exceptions.
        float depth_range = max_depth - min_depth;
        float quanta = (depthAccuracy - 1)/ depth_range;

        // Change all the values from floats to quantised ints
        for(int i = 0; i < trans; i++)
            depthList2[i] = (int)((depthList1[i] - min_depth) * quanta);

        if(terminate)
            return;

        // Now perform the counting sort on them. Initialise all the counts to
        // 0 again
        for(int i = 0; i < depthAccuracy; i++)
            countList[i] = 0;

        if(terminate)
            return;

        for(int j = 0; j < trans; j++)
            countList[depthList2[j]]++;

        if(terminate)
            return;

        // countList now contains the number of elements equal to i
        for(int i = 1; i < depthAccuracy; i++)
            countList[i] += countList[i - 1];

        if(terminate)
            return;

        for(int j = trans; --j >= 0 && !terminate; )
        {
            int dl = depthList2[j];
            transparentList[countList[dl] - 1] = unsortedTransparentList[j];
            countList[dl]--;
        }

        instr.renderOps[idx++] = RenderOp.START_TRANSPARENT;

        for(int i = 0; i < trans && !terminate; i++)
        {
            idx = appendOutputShape(idx, transparentList[i], instr);

            // clear it from the list
            transparentList[i] = null;
        }

        instr.renderOps[idx++] = RenderOp.STOP_TRANSPARENT;
        instr.numValid = idx;
    }

    /**
     * Convenience method to add a Shape3D to the list at the given index point.
     *
     * @param idx The starting place in the instruction list
     * @param node The node instance to add
     * @param instr List of instructions to add the node to
     * @return The update idx value taking into account the new nodes
     */
    private int appendOutputShape(int idx,
                                 CullOutputDetails node,
                                 RenderInstructions instr)
    {
        if(node.numLights != 0)
        {
            for(int j = 0; j < node.numLights; j++)
            {
                LightDetails ld = node.lights[j];
                instr.renderList[idx].renderable = ld.getLight();

                System.arraycopy(ld.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_LIGHT;
                idx++;
            }
        }

        if(node.numClipPlanes != 0)
        {
            for(int j = 0; j < node.numClipPlanes; j++)
            {
                ClipDetails cd = node.clipPlanes[j];
                instr.renderList[idx].renderable = cd.getClipPlane();

                System.arraycopy(cd.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_CLIP_PLANE;
                idx++;
            }
        }

        int fog_id = 0;

        if(node.localFog != null)
        {
            fog_id = lastGlobalId++;
            instr.renderList[idx].id = fog_id;
            instr.renderList[idx].renderable = node.localFog;
            instr.renderOps[idx] = RenderOp.START_FOG;
            idx++;
        }

        // Now process the node type.
        // For this simple one, just use the basic render command.
        // Something more complex would pull the shape apart and do
        // state/depth/transparency sorting in this section.
        if(node.renderable instanceof Shape3D)
        {
            instr.renderList[idx].renderable = node.renderable;

            instr.renderList[idx].transform[0] = node.transform.m00;
            instr.renderList[idx].transform[1] = node.transform.m10;
            instr.renderList[idx].transform[2] = node.transform.m20;
            instr.renderList[idx].transform[3] = node.transform.m30;

            instr.renderList[idx].transform[4] = node.transform.m01;
            instr.renderList[idx].transform[5] = node.transform.m11;
            instr.renderList[idx].transform[6] = node.transform.m21;
            instr.renderList[idx].transform[7] = node.transform.m31;

            instr.renderList[idx].transform[8] = node.transform.m02;
            instr.renderList[idx].transform[9] = node.transform.m12;
            instr.renderList[idx].transform[10] = node.transform.m22;
            instr.renderList[idx].transform[11] = node.transform.m32;

            instr.renderList[idx].transform[12] = node.transform.m03;
            instr.renderList[idx].transform[13] = node.transform.m13;
            instr.renderList[idx].transform[14] = node.transform.m23;
            instr.renderList[idx].transform[15] = node.transform.m33;

            instr.renderOps[idx] = RenderOp.START_RENDER;
            idx++;

            instr.renderList[idx].renderable = node.renderable;
            instr.renderOps[idx] = RenderOp.STOP_RENDER;
            idx++;
        }
        else if(node.renderable instanceof CustomRenderable)
        {
            instr.renderList[idx].renderable = node.renderable;
            instr.renderList[idx].instructions = node.customData;

            Matrix4f tx = node.transform;

            instr.renderList[idx].transform[0] = tx.m00;
            instr.renderList[idx].transform[1] = tx.m10;
            instr.renderList[idx].transform[2] = tx.m20;
            instr.renderList[idx].transform[3] = tx.m30;

            instr.renderList[idx].transform[4] = tx.m01;
            instr.renderList[idx].transform[5] = tx.m11;
            instr.renderList[idx].transform[6] = tx.m21;
            instr.renderList[idx].transform[7] = tx.m31;

            instr.renderList[idx].transform[8] = tx.m02;
            instr.renderList[idx].transform[9] = tx.m12;
            instr.renderList[idx].transform[10] = tx.m22;
            instr.renderList[idx].transform[11] = tx.m32;

            instr.renderList[idx].transform[12] = tx.m03;
            instr.renderList[idx].transform[13] = tx.m13;
            instr.renderList[idx].transform[14] = tx.m23;
            instr.renderList[idx].transform[15] = tx.m33;

            instr.renderOps[idx] = RenderOp.RENDER_CUSTOM;
            idx++;
        }

        if(node.localFog != null)
        {
            instr.renderList[idx].id = fog_id;
            instr.renderList[idx].renderable = node.localFog;
            instr.renderOps[idx] = RenderOp.STOP_FOG;
            idx++;
        }

        if(node.numClipPlanes != 0)
        {
            for(int j = node.numClipPlanes - 1; j >= 0; j--)
            {
                ClipDetails cd = node.clipPlanes[j];
                instr.renderList[idx].renderable = cd.getClipPlane();

                instr.renderList[idx].id = lastGlobalId - (1 + j);

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_CLIP_PLANE;
                idx++;
            }
        }

        if(node.numLights != 0)
        {
            for(int j = node.numLights - 1; j >= 0; j--)
            {
                LightDetails ld = node.lights[j];
                instr.renderList[idx].renderable = ld.getLight();

                instr.renderList[idx].id = lastGlobalId - (1 + j);

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;
            }
        }

        return idx;
    }
}
