/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.lang.ref.SoftReference;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.security.AccessController;
import java.security.PrivilegedAction;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

/**
 * Implementation of the {@link RenderPipelineManager} that uses separate
 * threads for each of the pipelines that it is managing, allowing it to run
 * multiple output surfaces simultaneously (eg Powerwall, CAVE etc).
 * <p>
 *
 * By default the manager does not start of enabled. An explicit enable call
 * will be needed to kick the rendering process off.
 * <p>
 *
 * Change List processing will process bounds changed items before data
 * changed.
 * <p>
 * <b>Note:</b><p>
 * Layers are not implemented yet.
 *
 * @author Justin Couch
 * @version $Revision: 1.27.2.1 $
 */
public class MultiThreadRenderManager
    implements Runnable,
               NodeUpdateHandler,
               RenderPipelineManager,
               PipelineStateObserver
{
    private static final String ACTIVE_RENDERING_MSG =
        "You cannot make this call now because the system is currently " +
        "rendering. To make this call, first disable the rendering. ";

    /** The initial size of the children list */
    private static final int CHANGELIST_START_SIZE = 200;

    /** The increment size of the list if it gets overflowed */
    private static final int CHANGELIST_INCREMENT = 100;

    /** The initial size of the shader init/log lists */
    private static final int SHADERLIST_START_SIZE = 64;

    /** The increment size of the init/log list if it gets overflowed */
    private static final int SHADERLIST_INCREMENT = 32;

    /** The Scene Graph root */
    private WorldRootGroup root;

    /** The main scene containing all the data we need. */
    private Scene worldScene;

    /** The change requestors for data changed sets */
    private NodeUpdateListener[] dataChangeList;

    /** The change requestors for bounds changed sets */
    private NodeUpdateListener[] boundsChangeList;

    /** The source objects for data changed sets */
    private Object[] dataSourceList;

    /** The source objects for bounds changed sets */
    private Object[] boundsSourceList;

    /** The internal requestors for bounds changed sets */
    private InternalNodeUpdateListener[] boundsInternalList;

    /** The current place to add change requestors */
    private int lastDataChangeItem;

    /** The current place to add change requestors */
    private int lastBoundsChangeItem;

    /** HashSet to determine duplicates in the change List */
    private HashSet dataChangeListenerSet;

    /** HashSet to determine duplicates in the change List */
    private HashSet dataChangeSrcSet;

    /** HashSet to determine duplicates in the change List */
    private HashSet boundsChangeListenerSet;

    /** HashSet to determine duplicates in the change List */
    private HashSet boundsChangeSrcSet;

    /** The change requestors for data changed sets */
    private ShaderSceneGraphObject[] shaderInitList;

    /** The change requestors for bounds changed sets */
    private ShaderSceneGraphObject[] shaderLogList;

    /** The current place to add shader init requestors */
    private int lastShaderInitItem;

    /** The current place to add shader log requestors */
    private int lastShaderLogItem;

    /** Set of shaders that require init processing */
    private HashSet shaderInitSet;

    /** Set of shaders that require log fetch processing */
    private HashSet shaderLogSet;

    /** The collection of pipeleines to manage */
    private RenderPipeline[] pipeline;

    /** The collection of pipeleines to manage */
    private AudioPipeline[] audioPipeline;

    /** Threads that manage a pipeline. Thread corresponds to pipeline index */
    private PipelineThread[] pipelineThread;

    /** Threads that manage a pipeline. Thread corresponds to pipeline index */
    private AudioPipelineThread[] audioPipelineThread;

    /** Number of valid pipelines to process */
    private int numPipelines;

    /** Number of valid audio pipelines to process */
    private int numAudioPipelines;

    /** Picking handler for managing the picking */
    private PickingHandler pickHandler;

    /** The minimum frame cycle time */
    private int minimumCycleTime;

    /** Current enabled state */
    private boolean enabled;

    /** The thread that contains this runnable */
    private Thread runtimeThread;

    /** Flag that the runtime thread should be terminated at next chance */
    private boolean terminate;

    /** The external observer for keeping updates in check */
    private ApplicationUpdateObserver observer;

    /** Queue for holding deleted textures */
    private ReferenceQueue deletionQueue;

    /** Flag for the writable state used by the NodeUpdateHandler query */
    private boolean pickPermitted;

    /** The current object that is having the update callback called */
    private Object writableBoundsObject;

    /** The current object that is having the update callback called */
    private Object writableDataObject;

    /**
     * Lock object to prevent frame callbacks from finishing before we've
     * called all the other threads this frame.
     */
    private Object frameFinishLock;

    /** Per-frame counter used to track when a frame is complete */
    private int completedFrameCount;

    /** Object to synchronise between the various rendering threads */
    private Object renderWaitLock;

    /** Are we processing the change list */
    private boolean processing;

    /**
     * Flag indicating that the scene has changed and it should be completely
     * rendered and not just skip to the display part.
     */
    private boolean sceneChanged;

    /** The thread used to handle the system shutdown hook */
    private Thread shutdownThread;

    /**
     * Construct a new render manager with no pipelines or renderers
     * registered. Starts by allocating space for 2 render pipes and
     * 1 audio pipe to be handled.
     */
    public MultiThreadRenderManager()
    {
        this(2,1);
    }

    /**
     * Construct a new render manager with no pipelines or renderers
     * registered, but with internal arrays setup for the given number of
     * pipelines to be added.
     *
     * @param numPipes The initial number of pipelines to prepare
     * @param numAudioPipes The initial number of audio pipelines to prepare
     */
    public MultiThreadRenderManager(int numPipes, int numAudioPipes)
    {
        enabled = false;
        terminate = false;
        processing = false;

        pickPermitted = false;
        sceneChanged = false;

        pipeline = new RenderPipeline[numPipes];
        pipelineThread = new PipelineThread[numPipes];

        audioPipeline = new AudioPipeline[numAudioPipes];
        audioPipelineThread = new AudioPipelineThread[numAudioPipes];

        renderWaitLock = new Object();
        frameFinishLock = new Object();

        dataChangeList = new NodeUpdateListener[CHANGELIST_START_SIZE];
        dataSourceList = new Object[CHANGELIST_START_SIZE];
        dataChangeListenerSet = new HashSet(CHANGELIST_START_SIZE);
        dataChangeSrcSet = new HashSet(CHANGELIST_START_SIZE);
        boundsChangeList = new NodeUpdateListener[CHANGELIST_START_SIZE];
        boundsInternalList =
            new InternalNodeUpdateListener[CHANGELIST_START_SIZE];
        boundsSourceList = new Object[CHANGELIST_START_SIZE];
        boundsChangeListenerSet = new HashSet(CHANGELIST_START_SIZE);
        boundsChangeSrcSet = new HashSet(CHANGELIST_START_SIZE);
        shaderInitList = new ShaderSceneGraphObject[SHADERLIST_START_SIZE];
        shaderInitSet = new HashSet(SHADERLIST_START_SIZE);
        shaderLogList = new ShaderSceneGraphObject[SHADERLIST_START_SIZE];
        shaderLogSet = new HashSet(SHADERLIST_START_SIZE);

        pickHandler = new DefaultPickingHandler();
        deletionQueue = new ReferenceQueue();

        lastDataChangeItem = 0;
        lastBoundsChangeItem = 0;
        lastShaderInitItem = 0;
        lastShaderLogItem = 0;

        minimumCycleTime = 0;
        completedFrameCount = 0;

        root = new WorldRootGroup();
        root.addChild(null);
        root.setUpdateHandler(this);
        root.setLive(true);

        shutdownThread = new ShutdownThread(this);
        AccessController.doPrivileged(
            new PrivilegedAction()
            {
                public Object run()
                {
                        Runtime rt = Runtime.getRuntime();
                        rt.addShutdownHook(shutdownThread);
                        return null;
                }
            }
        );
    }

    //---------------------------------------------------------------
    // Methods defined by RenderPipelineManager
    //---------------------------------------------------------------

    /**
     * Tell render to start or stop rendering. If currently running, it
     * will wait until all the pipelines have completed their current cycle
     * and will then halt.
     *
     * @param state True if to enable rendering
     */
    public void setEnabled(boolean state)
    {
        if(enabled == state)
            return;

        if(state)
        {
            enabled = true;
            terminate = false;
            pickPermitted = false;

            if(runtimeThread == null)
            {
                ThreadPrivilegedAction tpa = new ThreadPrivilegedAction(this);
                runtimeThread = tpa.getThread();
            }
        }
        else
        {
            enabled = false;
            terminate = true;
            pickPermitted = true;

            for(int i = 0; i < numPipelines; i++)
                pipelineThread[i].halt();

            for(int i = 0; i < numAudioPipelines; i++)
                audioPipelineThread[i].halt();
        }
    }

    /**
     * Get the current render state of the manager.
     *
     * @return true if the manager is currently running
     */
    public boolean isEnabled()
    {
        return enabled;
    }

    /**
     * Force a single render of all pipelines now. Ignores the enabled and
     * cycle time settings to cause a single render at this point in time.
     * If a render is currently in progress, an exception is generated
     *
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void renderOnce()
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        processChangeList();
        processShaderLists();

        for(int i = 0; i < numPipelines; i++)
            pipelineThread[i].render();

        for(int i = 0; i < numAudioPipelines; i++)
            audioPipelineThread[i].render();

        for(int i = 0; i < numPipelines; i++)
            pipeline[i].swapBuffers();
    }

    /**
     * Request that the manager perform a full scene render pass and update,
     * ignoring any usual optimisations that it may take. For performance
     * reasons, a render manager may elect to not run or update some portions
     * of the scene graph. This method requests that the next frame only should
     * ignore those optimisations and process the full scene graph.
     * <p>
     *
     * This method will work both in automated rendering and with the
     * {@link #renderOnce()} method.
     */
    public void requestFullSceneRender()
    {
        sceneChanged = true;
    }

    /**
     * Set the scene for this manager.  null will remove
     * the set scene.
     *
     * @param scene The new scene
     */
    public void setScene(Scene scene)
    {
        Group geom = null;

        if(scene != null)
            geom = scene.getRenderedGeometry();

        worldScene = scene;
        writableBoundsObject = root;
        root.setChild(geom, 0);
        writableBoundsObject = null;

        for(int i = 0; i < numPipelines; i++)
            pipeline[i].setRenderableScene(scene);

        for(int i = 0; i < numAudioPipelines; i++)
            audioPipeline[i].setRenderableScene(scene);

       sceneChanged = true;
    }

    /**
     * Fetch the currently set scene. If layers are set, then this will return
     * null.
     *
     * @return The current scene, or null
     */
    public Scene getScene()
    {
        return worldScene;
    }

    /**
     * Set the set of layers for this manager. Setting a value of
     * <code>null</code> will remove the currently set of layers. If this is
     * set while a current scene is set, then the scene will be cleared. Layers
     * are presented in depth order - layers[0] is rendered before layers[1]
     * etc.
     *
     * @param layers The collection of layers, in order, to render
     * @param numLayers The number of valid layers to use
     */
    public void setLayers(Layer[] layers, int numLayers)
    {
    }

    /**
     * Get the number of layers that are currently set. If no layers are set,
     * or a scene is set, this will return zero.
     *
     * @return a value greater than or equal to zero
     */
    public int numLayers()
    {
        return 0;
    }

    /**
     * Fetch the current layers that are set. The values will be copied into
     * the user-provided array. That array must be at least
     * {@link #numLayers()} in length. If not, this method does nothing (the
     * provided array will be unchanged).
     *
     * @param layers An array to copy the values into
     */
    public void getLayers(Layer[] layers)
    {
    }

    /**
     * Set the minimum duty cycle of the render manager. This is the type in
     * milliseconds that should be the minimum between frames and can be used
     * to throttle the rendering loop to a maximum frame rate should other
     * systems require CPU time.
     *
     * @param cycleTime The minimum time in milliseconds between frames
     */
    public void setMinimumFrameInterval(int cycleTime)
    {
        minimumCycleTime = cycleTime;
    }

    /**
     * Fetch the currently set duty cycle value.
     *
     * @return The duty cycle time, in milliseconds
     */
    public int getMinimumFrameInterval()
    {
        return minimumCycleTime;
    }

    /**
     * Add a pipeline to be rendered to the manager. A duplicate registration
     * or null value is ignored. For this implementation, a call to this method
     * will replace the existing pipeline rather than add an additional one.
     *
     * @param pipe The new pipe instance to be added
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void addPipeline(RenderPipeline pipe)
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        if(pipe == null)
            return;

        if(numPipelines == pipeline.length)
        {
            RenderPipeline[] tmp1 = new RenderPipeline[numPipelines + 2];
            System.arraycopy(pipeline, 0, tmp1, 0, numPipelines);
            pipeline = tmp1;

            PipelineThread[] tmp2 = new PipelineThread[numPipelines + 2];
            System.arraycopy(pipelineThread, 0, tmp2, 0, numPipelines);
            pipelineThread = tmp2;
        }

        pipe.setRenderableScene(worldScene);

        pipeline[numPipelines] = pipe;
        pipelineThread[numPipelines] = new PipelineThread(pipe);
        pipelineThread[numPipelines].setStateObserver(this);

        numPipelines++;
    }

    /**
     * Add a audio pipeline to be rendered to the manager. A duplicate registration
     * or null value is ignored. For this implementation, a call to this method
     * will replace the existing pipeline rather than add an additional one.
     *
     * @param pipe The new pipe instance to be added
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void addAudioPipeline(AudioPipeline pipe)
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        if(pipe == null)
            return;

        if(numAudioPipelines == audioPipeline.length)
        {
            AudioPipeline[] tmp1 = new AudioPipeline[numAudioPipelines + 2];
            System.arraycopy(audioPipeline, 0, tmp1, 0, numAudioPipelines);
            audioPipeline = tmp1;

            AudioPipelineThread[] tmp2 = new AudioPipelineThread[numAudioPipelines + 2];
            System.arraycopy(audioPipelineThread, 0, tmp2, 0, numAudioPipelines);
            audioPipelineThread = tmp2;
        }

        pipe.setRenderableScene(worldScene);

        audioPipeline[numAudioPipelines] = pipe;
        audioPipelineThread[numAudioPipelines] = new AudioPipelineThread(pipe);
        audioPipelineThread[numAudioPipelines].setStateObserver(this);

        numAudioPipelines++;
    }

    /**
     * Remove an already registered pipeline from the manager. A or null value
     * or one that is not currently registered is ignored. For this
     * implementation, a call to this method will remove the existing pipeline
     * if it is the same instance rather than remove an additional one.
     *
     * @param pipe The pipe instance to be removed
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void removePipeline(RenderPipeline pipe)
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        for(int i = 0; i < numPipelines; i++)
        {
            if(pipeline[i] == pipe)
            {
                pipelineThread[i].shutdown();
                pipelineThread[i].setStateObserver(null);

                System.arraycopy(pipeline,
                                 i + 1,
                                 pipeline,
                                 i,
                                 numPipelines - i - 1);

                System.arraycopy(pipelineThread,
                                 i + 1,
                                 pipelineThread,
                                 i,
                                 numPipelines - i - 1);

                numPipelines--;
                pipe.setRenderableScene(null);

                break;
            }
        }
    }

    /**
     * Remove an already registered pipeline from the manager. A or null value
     * or one that is not currently registered is ignored. For this
     * implementation, a call to this method will remove the existing pipeline
     * if it is the same instance rather than remove an additional one.
     *
     * @param pipe The pipe instance to be removed
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void removeAudioPipeline(AudioPipeline pipe)
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        for(int i = 0; i < numAudioPipelines; i++)
        {
            if(audioPipeline[i] == pipe)
            {
                audioPipelineThread[i].shutdown();
                audioPipelineThread[i].setStateObserver(null);

                System.arraycopy(audioPipeline,
                                 i + 1,
                                 pipeline,
                                 i,
                                 numAudioPipelines - i - 1);

                System.arraycopy(audioPipelineThread,
                                 i + 1,
                                 audioPipelineThread,
                                 i,
                                 numAudioPipelines - i - 1);

                numAudioPipelines--;
                pipe.setRenderableScene(null);

                break;
            }
        }
    }

    /**
     * Register an observer that can be used to know when the application is
     * safe to update the scene graph. A value of null will remove the
     * currently set value.
     *
     * @param obs The observer instance to use
     */
    public void setApplicationObserver(ApplicationUpdateObserver obs)
    {
        observer = obs;
    }

    /**
     * Disable the internal shutdown hook system. It will be up to the calling
     * application to make sure the {@link #shutdown()} method is called to
     * turn off the OpenGL rendering system. If it does not, there is a good
     * possibility of a crash of the system.
     * <p>
     *
     * If the internal shutdown is disabled, then the shutdown callback of the
     * <code>ApplicationUpdateObserver</code> will not be called.
     */
    public void disableInternalShutdown()
    {
        if(shutdownThread != null)
        {
            AccessController.doPrivileged(
                new PrivilegedAction()
                {
                    public Object run()
                    {
                        Runtime rt = Runtime.getRuntime();
                        rt.removeShutdownHook(shutdownThread);
                        return null;
                    }
                }
            );

            shutdownThread = null;
        }
    }

    /**
     * Notification to shutdown the internals of the renderer because the
     * application is about to exit.
     */
    public void shutdown()
    {
        // If this has already been called once, ignore it. Most of the
        // variables will have been nulled out by now.
        if(terminate || !enabled)
            return;

        terminate = true;

        setEnabled(false);

        // If we have a shutdown thread then that means
        if((shutdownThread != null) && (observer != null))
        {
            try
            {
                observer.appShutdown();
            }
            catch(Exception e)
            {
                System.out.println("Error in user callback for shutdown " +
                                   e.getMessage());
                e.printStackTrace();
            }

            observer = null;
        }
    }

    //---------------------------------------------------------------
    // Methods defined by Runnable
    //---------------------------------------------------------------

    /**
     * Run method used to synchronise the internal rendering state and the
     * external state of the canvas. Should never be called directly.
     */
    public void run()
    {
        for(int i = 0; i < numPipelines; i++)
        {
            if(!pipelineThread[i].isAlive())
                pipelineThread[i].start();
        }

        for(int i = 0; i < numAudioPipelines; i++)
        {
            if(!audioPipelineThread[i].isAlive())
                audioPipelineThread[i].start();
        }

        while(!terminate)
        {
            // Throttle the apps runtime
            long start_time = System.currentTimeMillis();

            if(!terminate && observer != null)
            {
                pickPermitted = true;
                observer.updateSceneGraph();
                pickPermitted = false;
            }

            if(terminate)
                break;

            // Look out for any deleted objects and hand them on if needed.
            Reference ref;

            while((ref = deletionQueue.poll()) != null)
            {
                DeletableSceneGraphObject dead =
                    (DeletableSceneGraphObject)ref.get();

                for(int i = 0; i < numPipelines; i++)
                    pipelineThread[i].queueDeletedObject(dead);
            }

            if(terminate)
                break;

            processShaderLists();

            if(terminate)
                break;

            synchronized(frameFinishLock)
            {
                // Can we skip the processing if nothing changed last frame?
                if((lastDataChangeItem == 0) && (lastBoundsChangeItem == 0) &&
                   !sceneChanged)
                {
                    for(int i = 0; i < numPipelines; i++)
                        pipelineThread[i].displayOnly();

                    for(int i = 0; i < numAudioPipelines; i++)
                        audioPipelineThread[i].displayOnly();
                }
                else
                {
                    processChangeList();

                    if(terminate)
                        break;

                    sceneChanged = false;

                    for(int i = 0; i < numPipelines; i++)
                        pipelineThread[i].render();

                    for(int i = 0; i < numAudioPipelines; i++)
                        audioPipelineThread[i].render();
                }

                if(terminate)
                    break;
            }

            // Hang while we wait for notification that all the rendering
            // threads have finished before swapping all the buffers.
            if((numPipelines + numAudioPipelines) != 0)
            {
                try
                {
                    synchronized(renderWaitLock)
                    {
                        renderWaitLock.wait();
                    }
                }
                catch(InterruptedException ie)
                {
                }
            }

            if(terminate)
                break;

            long now = System.currentTimeMillis();
            long diff = now - start_time;

            if(diff < minimumCycleTime)
            {
                try
                {
                    Thread.sleep(minimumCycleTime - diff);
                }
                catch(InterruptedException ie)
                {
                }
            }

            Thread.yield();
        }

        runtimeThread = null;
    }

    //---------------------------------------------------------------
    // Methods defined by PipelineStateObserver
    //---------------------------------------------------------------

    /**
     * Notification that the frame state has finished rendering.
     */
    public synchronized void frameFinished()
    {
        synchronized(frameFinishLock)
        {
            if(++completedFrameCount >= (numPipelines + numAudioPipelines))
            {
                synchronized(renderWaitLock)
                {
                    completedFrameCount = 0;
                    renderWaitLock.notify();
                }
            }
        }
    }

    //---------------------------------------------------------------
    // Methods defined by NodeUpdateHandler
    //---------------------------------------------------------------

    /**
     * Check to see if writing to the node is permitted currently.
     *
     * @param src The object that is requesting the check
     * @return true if the end user can write, false if not
     */
    public boolean isDataWritePermitted(Object src)
    {
        return (src == writableDataObject) || !enabled;
    }

    /**
     * Check to see if writing to the node is permitted currently.
     *
     * @param src The object that is requesting the check
     * @return true if the end user can write, false if not
     */
    public boolean isBoundsWritePermitted(Object src)
    {
        return (src == writableBoundsObject) || !enabled;
    }

    /**
     * Check to see if picking is permitted currently.
     *
     * @return true if the end user can pick, false if not
     */
    public boolean isPickingPermitted()
    {
        return pickPermitted;
    }

    /**
     * Notify the handler that you have updates to the SG that might alter
     * a node's bounds.
     *
     * @param l The change requestor
     * @param src The object that is passing this listener through.
     * @param intL Internal listener for making callbacks at a later time
     *    to propogate the bounds changes.
     * @throws InvalidListenerSetTimingException If called when the node called
     *    during one of the bounds/data changed callbacks
     */
    public synchronized boolean boundsChanged(NodeUpdateListener l,
                                              Object src,
                                              InternalNodeUpdateListener intL)
        throws InvalidListenerSetTimingException
    {
        if(processing)
            throw new InvalidListenerSetTimingException("boundsChanged cannot be called in NodeUpdateListener");

        // Check for duplicates
        if(boundsChangeSrcSet.contains(src) && boundsChangeListenerSet.contains(l)) {
            return false;
        }

        boundsChangeListenerSet.add(l);
        boundsChangeSrcSet.add(src);

        // Add to a single ChangeList for now
        resizeBoundsChangeList();
        boundsChangeList[lastBoundsChangeItem] = l;
        boundsInternalList[lastBoundsChangeItem] = intL;
        boundsSourceList[lastBoundsChangeItem] = src;
        lastBoundsChangeItem++;

        return true;
    }

    /**
     * Notify the handler that you have updates to the SG that will not
     * alter a node's bounds.
     *
     * @param l The change requestor
     * @param src The object that is passing this listener through.
     * @throws InvalidListenerSetTimingException If called when the node called
     *    during one of the bounds/data changed callbacks
     */
    public synchronized void dataChanged(NodeUpdateListener l, Object src)
        throws InvalidListenerSetTimingException
    {
        if(processing)
            throw new InvalidListenerSetTimingException("dataChanged cannot be called in NodeUpdateListener.  Need to fix for threaded content.");

        // Check for duplicates
        if(dataChangeSrcSet.contains(src) &&
           dataChangeListenerSet.contains(l))
            return;

        dataChangeListenerSet.add(l);
        dataChangeSrcSet.add(src);

        // Add to a single ChangeList for now
        resizeDataChangeList();
        dataChangeList[lastDataChangeItem] = l;
        dataSourceList[lastDataChangeItem] = src;
        lastDataChangeItem++;
    }

    /**
     * The shader object passed requires an initialisation be performed. Queue
     * the shader up for processing now.
     *
     * @param shader The shader instance to queue
     * @param updateResponse true if this is being made as a response to a node's
     *   setUpdateHandler() method
     */
    public void shaderRequiresInit(ShaderSceneGraphObject shader,
                                   boolean updateResponse)
    {
        if(processing && !updateResponse)
            throw new RuntimeException("shaderRequiresInit cannot be called in NodeUpdateListener");

        // Check for duplicates
        if(shaderInitSet.contains(shader))
            return;

        shaderInitSet.add(shader);

        resizeShaderInitList();
        shaderInitList[lastShaderInitItem] = shader;
        lastShaderInitItem++;
    }

    /**
     * The shader object passed requires updating the log info. Queue
     * the shader up for processing now so that at the next oppourtunity it
     * can call glGetLogInfoARB.
     *
     * @param shader The shader instance to queue
     * @param updateResponse true if this is being made as a response to a node's
     *   setUpdateHandler() method
     */
    public void shaderRequiresLogInfo(ShaderSceneGraphObject shader,
                                      boolean updateResponse)
    {
        if(processing && !updateResponse)
            throw new RuntimeException("shaderRequiresLog cannot be called in NodeUpdateListener");

        // Check for duplicates
        if(shaderLogSet.contains(shader))
            return;

        shaderLogSet.add(shader);

        resizeShaderLogList();
        shaderLogList[lastShaderLogItem] = shader;
        lastShaderLogItem++;
    }

    /**
     * The object provided is no longer needed by its parent. Put this on
     * a queue for objects that could potentially be deleted next frame. Used
     * for any scene graph object type that must be explicitly cleaned up
     * during the OpenGL render cycle, such as textures and pBuffers.
     *
     * @param obj The object that is registering itself for deletion
     */
    public void objectRemoved(DeletableSceneGraphObject obj)
    {
        new SoftReference(obj, deletionQueue);
    }

    /**
     * Get the picking handler so that we can do some picking operations.
     *
     * @return the current instance of the picking system
     */
    public PickingHandler getPickHandler()
    {
        return pickHandler;
    }

    //---------------------------------------------------------------
    // Misc Internal methods
    //---------------------------------------------------------------

    /**
     * Process the changeList now.
     */
    private void processChangeList()
    {
        processing = true;

        for(int i = 0; i < lastBoundsChangeItem; i++)
        {
            writableBoundsObject = boundsSourceList[i];
            boundsChangeList[i].updateNodeBoundsChanges(boundsSourceList[i]);
            boundsChangeList[i] = null;
            boundsSourceList[i] = null;
        }

        writableBoundsObject = null;

        // Now go through and update all the bounds
        for(int i = 0; i < lastBoundsChangeItem; i++)
        {
            boundsInternalList[i].updateBoundsAndNotify();
            boundsInternalList[i] = null;
        }

        lastBoundsChangeItem = 0;
        boundsChangeSrcSet.clear();
        boundsChangeListenerSet.clear();

        for(int i = 0; i < lastDataChangeItem; i++)
        {
            writableDataObject = dataSourceList[i];
            dataChangeList[i].updateNodeDataChanges(dataSourceList[i]);
            dataChangeList[i] = null;
            dataSourceList[i] = null;
        }

        writableDataObject = null;
        lastDataChangeItem = 0;
        dataChangeSrcSet.clear();
        dataChangeListenerSet.clear();

        processing = false;
    }

    /**
     * Process the shader lists and pass them on to the drawable for processing
     * next frame.
     */
    private void processShaderLists()
    {
        if((lastShaderInitItem == 0) && (lastShaderLogItem == 0))
            return;

        for(int i = 0; i < numPipelines; i++)
        {
            pipelineThread[i].queueShaderRequests(shaderInitList,
                                                  lastShaderInitItem,
                                                  shaderLogList,
                                                  lastShaderLogItem);
        }

        // now clear out the array.
        for(int i = 0; i < lastShaderInitItem; i++)
            shaderInitList[0] = null;

        for(int i = 0; i < lastShaderLogItem; i++)
            shaderLogList[0] = null;

        lastShaderInitItem = 0;
        lastShaderLogItem = 0;

        shaderInitSet.clear();
        shaderLogSet.clear();
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeDataChangeList()
    {
        if((lastDataChangeItem + 1) == dataChangeList.length)
        {
            int old_size = dataChangeList.length;
            int new_size = old_size + CHANGELIST_INCREMENT;

            NodeUpdateListener[] tmp_nodes = new NodeUpdateListener[new_size];
            Object[] tmp_src = new Object[new_size];

            System.arraycopy(dataChangeList, 0, tmp_nodes, 0, old_size);
            System.arraycopy(dataSourceList, 0, tmp_src, 0, old_size);

            dataChangeList = tmp_nodes;
            dataSourceList = tmp_src;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeBoundsChangeList()
    {
        if((lastBoundsChangeItem + 1) == boundsChangeList.length)
        {
            int old_size = boundsChangeList.length;
            int new_size = old_size + CHANGELIST_INCREMENT;

            NodeUpdateListener[] tmp_nodes = new NodeUpdateListener[new_size];
            InternalNodeUpdateListener[] tmp_int =
                new InternalNodeUpdateListener[new_size];
            Object[] tmp_src = new Object[new_size];

            System.arraycopy(boundsChangeList, 0, tmp_nodes, 0, old_size);
            System.arraycopy(boundsSourceList, 0, tmp_src, 0, old_size);
            System.arraycopy(boundsInternalList, 0, tmp_int, 0, old_size);

            boundsChangeList = tmp_nodes;
            boundsSourceList = tmp_src;
            boundsInternalList = tmp_int;
        }
    }


    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeShaderInitList()
    {
        if((lastShaderInitItem + 1) == shaderInitList.length)
        {
            int old_size = shaderInitList.length;
            int new_size = old_size + SHADERLIST_INCREMENT;

            ShaderSceneGraphObject[] tmp_nodes =
                new ShaderSceneGraphObject[new_size];

            System.arraycopy(shaderInitList, 0, tmp_nodes, 0, old_size);

            shaderInitList = tmp_nodes;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeShaderLogList()
    {
        if((lastShaderLogItem + 1) == shaderLogList.length)
        {
            int old_size = shaderLogList.length;
            int new_size = old_size + SHADERLIST_INCREMENT;

            ShaderSceneGraphObject[] tmp_nodes =
                new ShaderSceneGraphObject[new_size];

            System.arraycopy(shaderLogList, 0, tmp_nodes, 0, old_size);

            shaderLogList = tmp_nodes;
        }
    }
}
