/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.awt.Rectangle;

import javax.vecmath.Matrix4f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

/**
 * Handles the scenegraph maintenance and culling operations.
 * <p>
 *
 * The culling phase generates a list of nodes to render.
 * <p>
 *
 * Since this is a null culling stage, and thus no culling is performed, if the
 * scene graph contains instances of Cullable, they are ignored and traversal
 * stops at that point.
 * <p>
 *
 * <b>Note:</b><p>
 * Layers are not implemented yet.
 *
 * @author Alan Hudson
 * @version $Revision: 1.3 $
 */
public class NullAudioCullStage implements AudioCullStage
{
    private static final String SHARED_VP_MSG =
        "The path from the root of the scene graph to the viewpoint contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    /** Initial scene depth used for the transform stack */
    private static final int TRANSFORM_DEPTH_SIZE = 64;

    /** The increment size of the stack if it gets overflowed */
    private static final int STACK_INCREMENT = 32;

    /** The initial size of the children list */
    private static final int LIST_START_DEPTH = 200;

    /** The increment size of the list if it gets overflowed */
    private static final int LIST_INCREMENT = 20;

    /** The list of culled nodes to send to the next stage */
    private CullAudioOutputDetails[][] outputCullList;

    /** List of valid culled nodes as we process them */
    private CullAudioOutputDetails[][] validCullList;

    /** List that is being used to fill values into */
    private CullAudioOutputDetails[] workCullList;

    /** list of culled nodes list sizes to send to the next stage */
    private int[] outputCullSize;

    /** List of environment data as we process them */
    private AudioEnvironmentData[] validEnvData;

    /** list of environment data to send to the next stage */
    private AudioEnvironmentData[] outputEnvData;

    /** List of culled nodes list sizes as we process them */
    private int[] validCullSize;

    /** Index to the next output list in the validCullList */
    private int lastOutputList;

    /** A stack used to control the depth of the transform tree */
    private Matrix4f[] transformStack;

    /** Index to the next place to add items in the transformStack */
    private int lastTxStack;

    /** Handler for the output */
    private CulledAudioReceiver receiver;

    /** Path to the current viewpoint */
    private TransformGroup[] currentViewpointPath;

    /** Matrix used for pre-computing the view stack */
    private Matrix4f viewMatrix1;

    /** Matrix used for pre-computing the view stack */
    private Matrix4f viewMatrix2;

    /** Flag indicating a shutdown of the current processing is requested */
    private boolean terminate;

    /**
     * Create a basic instance of this class with the list initial internal
     * setup for the given number of renderable surfaces. The size is just an
     * initial esstimate, and is used for optimisation purposes to prevent
     * frequent array reallocations internally. As such, the number does not
     * have to be perfect, just good enough.
     *
     */
    public NullAudioCullStage()
    {
        int numDevices = 1;
        outputCullList = new CullAudioOutputDetails[numDevices][];
        validCullList = new CullAudioOutputDetails[numDevices][LIST_START_DEPTH];
        outputEnvData = new AudioEnvironmentData[numDevices];
        validEnvData = new AudioEnvironmentData[numDevices];
        outputCullSize = new int[numDevices];
        validCullSize = new int[numDevices];

        currentViewpointPath = new TransformGroup[TRANSFORM_DEPTH_SIZE];
        viewMatrix1 = new Matrix4f();
        viewMatrix2 = new Matrix4f();

        // populate the list with valid instances...
        for(int i = 0; i < numDevices; i++)
            validEnvData[i] = new AudioEnvironmentData();

        // populate the list with valid instances...
        for(int i = 0; i < numDevices; i++)
            for(int j = 0; j < LIST_START_DEPTH; j++)
                validCullList[i][j] = new CullAudioOutputDetails();

        transformStack = new Matrix4f[TRANSFORM_DEPTH_SIZE];

        // populate the list with valid instances...
        for(int i = 0; i < TRANSFORM_DEPTH_SIZE; i++)
            transformStack[i] = new Matrix4f();
    }

    //---------------------------------------------------------------
    // Methods defined by CullStage
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param scene The scene instance to cull
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void cull(Scene scene, int buffer)
    {
        // Buffer ID is not used currently.
        terminate = false;
        lastOutputList = 0;
        lastTxStack = 0;

        workCullList = validCullList[0];

        Node node = scene.getRenderedGeometry();
        fillEnvData(scene, validEnvData[0]);

        if(node instanceof Group)
        {
            // set the top matrix stack
            if(node instanceof TransformGroup)
                ((TransformGroup)node).getTransform(transformStack[0]);
            else
                transformStack[0].setIdentity();

            validCullSize[0] = findAllNodes((Group)node, 0);
            validCullList[0] = workCullList;
            lastOutputList++;
        }
        else if(node instanceof SharedNode)
        {
            // set the top matrix stack back to identity
            transformStack[0].setIdentity();
            validCullSize[0] = findNextNode((SharedNode)node,  0);
            validCullList[0] = workCullList;
            lastOutputList++;
        }
        else if((node instanceof Sound) && ((Sound)node).isEnabled())
        {
            workCullList[0].renderable = (Sound)node;
            workCullList[0].transform.setIdentity();

            validCullSize[0] = 1;
            validCullList[0] = workCullList;
            lastOutputList++;
        }


        // It is likely in this case that we're doing single-scene
        // rendering and will be doing it all the time. Possibly optimise
        // this by avoiding the copy altogether, but not 100% this won't
        // break in some corner case right now.

        outputCullList[0] = validCullList[0];
        outputCullSize[0] = validCullSize[0];
        outputEnvData[0]  = validEnvData[0];

        if((lastOutputList != 0) && (receiver != null) && !terminate)
            receiver.culledOutput(outputEnvData[0],
                                  outputCullList[0],
                                  outputCullSize[0]);
    }

    /**
     * Update and cull the scenegraph defined by a set of layers. This
     * generates an ordered list of nodes to render. It will not return until
     * the culling is complete.
     *
     * @param layers The collection of layers, in order, to render
     * @param numLayers The number of valid layers to use
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void cull(Layer[] layers, int numLayers, int buffer)
    {
    }

    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setCulledAudioReceiver(CulledAudioReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        terminate = true;
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph of a single node without doing any other
     * processing. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param node The root node to start the cull from
     */
    private int cullSingle(Group node)
    {
        // set the top matrix stack back to identity
        transformStack[0].setIdentity();

        if(node instanceof TransformGroup)
            ((TransformGroup)node).getTransform(transformStack[0]);

        return findAllNodes((Group)node, 0);
    }

    /**
     * Recursive walk of the tree to find all the renderable nodes.
     *
     * @param group The parent node to walk into
     * @param cullEndIndex The current last item on the cull list
     * @return The index of the last item on the cull list
     */
    private int findAllNodes(Group group, int cullEndIndex)
    {
        if(terminate)
            return 0;

        Node[] kids  = group.getRenderableChild();
        int size = group.numRenderableChildren();

        int ret_val = cullEndIndex;

        for(int i = 0; i < size; i++)
        {
            if(kids[i] instanceof Group)
            {
                // If a tg, push the new TX onto the stack
                boolean is_tx = (kids[i] instanceof TransformGroup);

                if(is_tx)
                {
                    resizeStack();
                    TransformGroup tg = (TransformGroup)kids[i];
                    tg.getTransform(transformStack[lastTxStack + 1]);

                    transformStack[lastTxStack + 1].mul(transformStack[lastTxStack],
                                                      transformStack[lastTxStack + 1]);
                    lastTxStack++;
                }

                ret_val = findAllNodes((Group)kids[i], ret_val);

                // Now pop the stacks

                if(is_tx)
                    lastTxStack--;
            }
            else if(kids[i] instanceof Sound)
            {
                // Check the visibility state and ignore if not visible.
                if(!((Sound)kids[i]).isEnabled())
                    continue;

                resizeCullList(ret_val);
                workCullList[ret_val].renderable = (Sound)kids[i];

                Matrix4f mat = transformStack[lastTxStack];

                // Transpose the matrix in place as it is being copied
                workCullList[ret_val].transform.set(mat);


                ret_val++;
            }
            else if(kids[i] instanceof SharedNode)
            {
                ret_val = findNextNode((SharedNode)kids[i],  ret_val);
            }
        }

        return ret_val;
    }

    /**
     * From the given SharedNode instance keep finding a non-sharedNode
     * instance, before continuing on the recursion
     *
     * @param root The parent node to walk into
     * @param cullEndIndex The current last item on the cull list
     * @return The index of the last item on the cull list
     */
    private int findNextNode(SharedNode root, int cullEndIndex)
    {
        if(terminate)
            return 0;

        int ret_val = cullEndIndex;

        Node current = root.getChild();

        while(current != null && current instanceof SharedNode)
            current = ((SharedNode)current).getChild();

        // Now just continue on normally. If the result of the above was a
        // null then it just falls off the end of this if/else naturally.
        if(current instanceof Group)
        {
            // If a tg, push the new TX onto the stack
            boolean is_tx = (current instanceof TransformGroup);

            if(is_tx)
            {
                resizeStack();
                TransformGroup tg = (TransformGroup)current;
                tg.getTransform(transformStack[lastTxStack + 1]);

                transformStack[lastTxStack + 1].mul(transformStack[lastTxStack],
                                                  transformStack[lastTxStack + 1]);
                lastTxStack++;
            }

            ret_val = findAllNodes((Group)current, cullEndIndex);

            // Now pop the stack.
            if(is_tx)
                lastTxStack--;
        }
        else if(current instanceof Sound)
        {
            // Check the visibility state and ignore if not visible.
            if(!((Sound)current).isEnabled())
                return ret_val;

            resizeCullList(cullEndIndex);
            workCullList[cullEndIndex].renderable = (Sound)current;

            Matrix4f mat = transformStack[lastTxStack];

            // Transpose the matrix in place as it is being copied
            workCullList[cullEndIndex].transform.set(mat);

            cullEndIndex++;

            ret_val++;
        }

        return ret_val;
    }

    /**
     * Take the scene and fill in a RenderEnvironmentData instance.
     *
     * @param scene The scene to take data from
     * @param envData Data instance to copy it to
     */
    private void fillEnvData(Scene scene, AudioEnvironmentData envData)
    {
        ViewEnvironment view = scene.getViewEnvironment();

        envData.userData = scene.getUserData();
        envData.viewpoint = scene.getActiveView();

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = envData.viewpoint.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if((parent instanceof SharedGroup) ||
               (parent instanceof SharedNode))
                throw new IllegalArgumentException(SHARED_VP_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        if(currentViewpointPath.length < path_depth)
            currentViewpointPath = new TransformGroup[path_depth];

        // start at the VP and work our way upwards.
        parent = envData.viewpoint.getParent();
        int last_tx = 0;
        while(parent != null)
        {
            if(parent instanceof TransformGroup)
                currentViewpointPath[last_tx++] = (TransformGroup)parent;

            parent = parent.getParent();
        }

        viewMatrix2.setIdentity();

        if(last_tx != 0)
        {
            for(int i = last_tx - 1; i >= 0; i--)
            {
                currentViewpointPath[i].getTransform(viewMatrix1);
                viewMatrix2.mul(viewMatrix1);
            }
        }

        envData.viewTransform.set(viewMatrix2);
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeCullList(int cur_size)
    {
        if((cur_size + 1) == workCullList.length)
        {
            int old_size = workCullList.length;
            int new_size = old_size + LIST_INCREMENT;

            CullAudioOutputDetails[] tmp_nodes = new CullAudioOutputDetails[new_size];

            System.arraycopy(workCullList, 0, tmp_nodes, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_nodes[i] = new CullAudioOutputDetails();

            workCullList = tmp_nodes;
        }
    }

    /**
     * Resize the transform stack if needed. Marked as final in order to
     * encourage the compiler to inline the code for faster execution
     */
    private final void resizeStack()
    {
        if((lastTxStack + 1) == transformStack.length)
        {
            int old_size = transformStack.length;
            int new_size = old_size + STACK_INCREMENT;

            Matrix4f[] tmp_matrix = new Matrix4f[new_size];

            System.arraycopy(transformStack, 0, tmp_matrix, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_matrix[i] = new Matrix4f();

            transformStack = tmp_matrix;
        }
    }
}
