/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.awt.Rectangle;

import javax.vecmath.Matrix4f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

/**
 * Handles the scenegraph maintenance and culling operations.
 * <p>
 *
 * The culling phase generates a list of nodes to render.
 * A future optimization will sort the render list by OGL state.
 * <p>
 *
 * By default, the implementation will walk into shape3Ds looking for any
 * offscreen textures to be rendered. If you know that you do not have any
 * in the scene, then you can set an internal flag to not look for them, thus
 * achieving a simple performance boost.
 * <p>
 *
 * Since this is a null culling stage, and thus no culling is performed, if the
 * scene graph contains instances of Cullable, they are ignored and traversal
 * stops at that point.
 * <p>
 *
 * <b>Note:</b><p>
 * Layers are not implemented yet.
 *
 * @author Justin Couch
 * @version $Revision: 1.39 $
 */
public class NullCullStage implements CullStage
{
    private static final String SHARED_VP_MSG =
        "The path from the root of the scene graph to the viewpoint contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    private static final String SHARED_BG_MSG =
        "The path from the root of the scene graph to the background contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    /** Initial scene depth used for the transform stack */
    private static final int TRANSFORM_DEPTH_SIZE = 64;

    /** The increment size of the stack if it gets overflowed */
    private static final int STACK_INCREMENT = 32;

    /** The initial size of the children list */
    private static final int LIST_START_DEPTH = 200;

    /** Initial size of the cull list for the number of offscreen surfaces */
    private static final int LIST_START_LENGTH = 1;

    /** The increment size of the list if it gets overflowed */
    private static final int LIST_INCREMENT = 20;

    /** The initial size of the light list */
    private static final int LIGHT_START_SIZE = 32;

    /** The increment size of the list if it gets overflowed */
    private static final int LIGHT_INCREMENT = 8;

    /** If set, recurse into Shape3Ds looking for offscreen textures */
    private boolean checkOffscreens;

    /** Buffer list of sub scenes that still need processing */
    private PBufferTextureSource[] pendingSubscenes;

    /** Buffer list of sub scenes that still need processing */
    private PBufferTextureSource[] pendingParents;

    /** Index of the last subscene in the list */
    private int lastSubscene;

    /**
     * List of final subscenes to send to the final stage. First index is
     * the direct owner of the scene contents. The second index is the
     * scene parent of the scene included (needed for pBuffer GL context
     * handling at render time). If this second one is null, then the parent
     * is the main canvas that is being rendered to.
     */
    private SceneGraphObject[][] outputSceneParents;

    /** List of valid subscene parents as we process them */
    private SceneGraphObject[][] validSceneParents;

    /** The list of culled nodes to send to the next stage */
    private CullOutputDetails[][] outputCullList;

    /** List of valid culled nodes as we process them */
    private CullOutputDetails[][] validCullList;

    /** List that is being used to fill values into */
    private CullOutputDetails[] workCullList;

    /** list of culled nodes list sizes to send to the next stage */
    private int[] outputCullSize;

    /** List of environment data as we process them */
    private RenderEnvironmentData[] validEnvData;

    /** list of environment data to send to the next stage */
    private RenderEnvironmentData[] outputEnvData;

    /** List of culled nodes list sizes as we process them */
    private int[] validCullSize;

    /** Index to the next output list in the validCullList */
    private int lastOutputList;

    /** A stack used to control the depth of the transform tree */
    private Matrix4f[] transformStack;

    /** Index to the next place to add items in the transformStack */
    private int lastTxStack;

    /** A stack used to control the depth of the fog nodes down the tree */
    private Fog[] fogStack;

    /** Index to the next place to add items in the fogStack */
    private int lastFogStack;

    /** The list of lights nodes currently valid while traversing. */
    private Light[] lightList;

    /** The list of light transforms currently valid while traversing */
    private float[][] lightTxList;

    /** Index to the next place to add items in the lightList */
    private int lastLight;

    /** The list of clip plane nodes currently valid while traversing. */
    private ClipPlane[] clipList;

    /** The list of clip transforms currently valid while traversing */
    private float[][] clipTxList;

    /** Index to the next place to add items in the clipList */
    private int lastClip;

    /** Temp used to get texture units when looking for offscreen textures. */
    private TextureUnit[] textureUnits;

    /** Collection of offscreen textures we've found this frame */
    private HashSet processedPBufferTextures;

    /** Collection of textures we've found for the current parent texture */
    private HashSet visitedNodes;

    /** Handler for the output */
    private CulledGeometryReceiver receiver;

    /** The active parent scene of the scene we are processing. */
    private PBufferTextureSource activeParent;

    /** Path to the current viewpoint */
    private TransformGroup[] currentViewpointPath;

    /** Matrix used for pre-computing the view stack */
    private Matrix4f viewMatrix1;

    /** Matrix used for pre-computing the view stack */
    private Matrix4f viewMatrix2;

    /** Rotation matrix for the orientation provided by user */
    private Matrix4f orientationMatrix;

    /** Storage variable for the screen orientation values */
    private float[] screenOrientation;

    /** Storage variable for the eye offset values */
    private float[] eyePoint;

    /** Flag to say explicit screen orientation values have been provided */
    private boolean useOrientation;

    /** Flag to say explicit eyepoint values have been provided */
    private boolean useEyePoint;

    /** Flag indicating a shutdown of the current processing is requested */
    private boolean terminate;

    /**
     * Create a basic instance of this class with the list assuming there are
     * no off-screen buffers in use for the initial internal setup.
     */
    public NullCullStage()
    {
        this(LIST_START_LENGTH);
    }

    /**
     * Create a basic instance of this class with the list initial internal
     * setup for the given number of renderable surfaces. The size is just an
     * initial esstimate, and is used for optimisation purposes to prevent
     * frequent array reallocations internally. As such, the number does not
     * have to be perfect, just good enough.
     *
     * @param numSurfaces Total number of surfaces to prepare rendering for
     */
    public NullCullStage(int numSurfaces)
    {
        outputCullList = new CullOutputDetails[numSurfaces][];
        validCullList = new CullOutputDetails[numSurfaces][LIST_START_DEPTH];
        outputSceneParents = new SceneGraphObject[numSurfaces][2];
        validSceneParents = new SceneGraphObject[numSurfaces][2];
        pendingSubscenes = new PBufferTextureSource[numSurfaces];
        pendingParents = new PBufferTextureSource[numSurfaces];
        outputEnvData = new RenderEnvironmentData[numSurfaces];
        validEnvData = new RenderEnvironmentData[numSurfaces];
        outputCullSize = new int[numSurfaces];
        validCullSize = new int[numSurfaces];

        textureUnits = new TextureUnit[32];
        processedPBufferTextures = new HashSet();
        visitedNodes = new HashSet();

        currentViewpointPath = new TransformGroup[TRANSFORM_DEPTH_SIZE];
        viewMatrix1 = new Matrix4f();
        viewMatrix2 = new Matrix4f();
        orientationMatrix = new Matrix4f();
        orientationMatrix.setIdentity();

        screenOrientation = new float[4];
        eyePoint = new float[3];

        useEyePoint = false;
        useOrientation = false;

        // populate the list with valid instances...
        for(int i = 0; i < numSurfaces; i++)
            validEnvData[i] = new RenderEnvironmentData();

        // populate the list with valid instances...
        for(int i = 0; i < numSurfaces; i++)
            for(int j = 0; j < LIST_START_DEPTH; j++)
                validCullList[i][j] = new CullOutputDetails();

        transformStack = new Matrix4f[TRANSFORM_DEPTH_SIZE];

        // populate the list with valid instances...
        for(int i = 0; i < TRANSFORM_DEPTH_SIZE; i++)
            transformStack[i] = new Matrix4f();

        fogStack = new Fog[TRANSFORM_DEPTH_SIZE];

        lightList = new Light[LIGHT_START_SIZE];
        lightTxList = new float[LIGHT_START_SIZE][16];

        clipList = new ClipPlane[LIGHT_START_SIZE];
        clipTxList = new float[LIGHT_START_SIZE][16];

        checkOffscreens = true;
    }

    //---------------------------------------------------------------
    // Methods defined by CullStage
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param scene The scene instance to cull
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void cull(Scene scene, int buffer)
    {
        // Buffer ID is not used currently.
        terminate = false;
        lastOutputList = 0;
        lastLight = 0;
        lastClip = 0;
        lastTxStack = 0;
        lastFogStack = 0;
        lastSubscene = 0;
        activeParent = null;

        workCullList = validCullList[0];
        validSceneParents[0][0] = null;
        validSceneParents[0][1] = null;

        Node node = scene.getRenderedGeometry();
        fillEnvData(scene, validEnvData[0]);

        if(terminate)
            return;

        if(node instanceof Group)
        {
            // set the top matrix stack
            if(node instanceof TransformGroup)
                ((TransformGroup)node).getTransform(transformStack[0]);
            else
                transformStack[0].setIdentity();

            validCullSize[0] = findAllNodes((Group)node, 0);
            validCullList[0] = workCullList;
            lastOutputList++;
        }
        else if(node instanceof SharedNode)
        {
            // set the top matrix stack back to identity
            transformStack[0].setIdentity();
            validCullSize[0] = findNextNode((SharedNode)node,  0);
            validCullList[0] = workCullList;
            lastOutputList++;
        }
        else if((node instanceof Shape3D) && ((Shape3D)node).isVisible())
        {
            // Walk into the shape and check that we don't have any
            // offscreen textures to render.
            if(checkOffscreens)
                checkForOffscreens((Shape3D)node);

            workCullList[0].renderable = (Shape3D)node;
            workCullList[0].transform.setIdentity();
            workCullList[0].numLights = 0;
            workCullList[0].numClipPlanes = 0;

            validCullSize[0] = 1;
            validCullList[0] = workCullList;
            lastOutputList++;
        }

        if(checkOffscreens)
        {
            // May need to shift inside the loop for cases where textures
            // contain scenes which contain further textures. Play a wait and
            // see game for now.
            resizeOutputLists();

            for(int i = 0; i < lastSubscene && !terminate; i++)
            {
                activeParent = pendingSubscenes[i];
                visitedNodes.clear();

                if(processedPBufferTextures.contains(pendingSubscenes[i]))
                {
                    validSceneParents[lastOutputList][0] =
                        (SceneGraphObject)pendingSubscenes[i];
                    validSceneParents[lastOutputList][1] =
                        (SceneGraphObject)pendingParents[i];
                }
                else
                {
                    Scene sub_scene = pendingSubscenes[i].getRenderedScene();

                    Group grp = sub_scene.getRenderedGeometry();

                    if(grp == null)
                        continue;

                    processedPBufferTextures.add(pendingSubscenes[i]);

                    // Reset these for each separate scene
                    lastLight = 0;
                    lastClip = 0;
                    lastTxStack = 0;
                    lastFogStack = 0;

                    // Since the first output cull list index is used by the
                    workCullList = validCullList[lastOutputList];

                    validSceneParents[lastOutputList][0] =
                        (SceneGraphObject)pendingSubscenes[i];
                    validSceneParents[lastOutputList][1] =
                        (SceneGraphObject)pendingParents[i];

                    fillEnvData(sub_scene, validEnvData[lastOutputList]);

                    validCullSize[lastOutputList] = cullSingle(grp);

                    // Assign it back, just in case the list got resized.
                    validCullList[lastOutputList] = workCullList;
                }

                lastOutputList++;
            }

            // Now reverse the list since we're doing a breadth-first search
            // for the subscenes, we end up with the most nested subscene last.
            // This is bad. So now reverse the order of the list so that the
            // most nested collection is first to be rendered and the main
            // screen is the last item.
            if(outputCullList.length < lastOutputList)
            {
                outputCullList = new CullOutputDetails[lastOutputList][];
                outputCullSize = new int[lastOutputList];
                outputEnvData = new RenderEnvironmentData[lastOutputList];
                outputSceneParents = new SceneGraphObject[lastOutputList][2];
            }

            for(int i = 0; i < lastOutputList; i++)
            {
                outputCullList[i] = validCullList[lastOutputList - i - 1];
                outputCullSize[i] = validCullSize[lastOutputList - i - 1];
                outputEnvData[i]  = validEnvData[lastOutputList - i - 1];
                outputSceneParents[i][0] =
                    validSceneParents[lastOutputList - i - 1][0];
                outputSceneParents[i][1] =
                    validSceneParents[lastOutputList - i - 1][1];
            }
        }
        else
        {
            // It is likely in this case that we're doing single-scene
            // rendering and will be doing it all the time. Possibly optimise
            // this by avoiding the copy altogether, but not 100% this won't
            // break in some corner case right now.

            outputCullList[0] = validCullList[0];
            outputCullSize[0] = validCullSize[0];
            outputEnvData[0]  = validEnvData[0];
            outputSceneParents[0][0] = validSceneParents[0][0];
            outputSceneParents[0][1] = validSceneParents[0][1];
        }

        processedPBufferTextures.clear();

        if((lastOutputList != 0) && (receiver != null) && !terminate)
            receiver.culledOutput(outputEnvData,
                                  outputCullList,
                                  outputCullSize,
                                  outputSceneParents,
                                  lastOutputList);
    }

    /**
     * Update and cull the scenegraph defined by a set of layers. This
     * generates an ordered list of nodes to render. It will not return until
     * the culling is complete.
     *
     * @param layers The collection of layers, in order, to render
     * @param numLayers The number of valid layers to use
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void cull(Layer[] layers, int numLayers, int buffer)
    {
    }

    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setCulledGeometryReceiver(CulledGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Set the flag for whether to check for offscreen textures or not. By
     * default, this flag is set to true.
     *
     * @param state true if offscreen textures should be looked for
     */
    public void setOffscreenCheckEnabled(boolean state)
    {
        checkOffscreens = state;
    }

    /**
     * Find out what the current offscreen check state is.
     *
     * @return true if the checking is being performed
     */
    public boolean isOffscreenCheckEnabled()
    {
        return checkOffscreens;
    }

    /**
     * Set the eyepoint offset from the centre position. This is used to model
     * offset view frustums, such as multiple displays or a powerwall. This
     * method will be called with the appropriate values from the
     * RenderPipeline that this culler is inserted into.
     *
     * @param x The x axis offset
     * @param y The y axis offset
     * @param z The z axis offset
     */
    public void setEyePointOffset(float x, float y, float z)
    {
        useEyePoint = true;
        eyePoint[0] = x;
        eyePoint[1] = y;
        eyePoint[2] = z;
    }

    /**
     * Set the orientation of this screen relative to the user's normal view
     * direction. The normal orientation of the screen is along the negative
     * Z axis. This method provides and axis-angle reorientation of that
     * direction to one that is facing the screen. Typically this will just
     * involve a rotation around the Y axis of some amount (45 and 90 deg
     * being the most common used in walls and caves). This method will be
     * called with the appropriate values from the RenderPipeline that this
     * culler is inserted into.
     *
     * @param x The x axis component
     * @param y The y axis component
     * @param z The z axis component
     * @param a The angle to rotate around the axis in radians
     * @throws IllegalArgumentException The length of the axis is zero
     */
    public void setScreenOrientation(float x, float y, float z, float a)
        throws IllegalArgumentException
    {
        float n = x * x + y * y + z * z;
        if(n == 0)
            throw new IllegalArgumentException("Zero length axis vector");

        useOrientation = true;
        screenOrientation[0] = x;
        screenOrientation[1] = y;
        screenOrientation[2] = z;
        screenOrientation[3] = a;

        double rcos = Math.cos(a);
        double rsin = Math.sin(a);

        orientationMatrix.m00 = (float)(rcos + x * x * (1 - rcos));
        orientationMatrix.m01 = (float)(z * rsin + y * x * (1 - rcos));
        orientationMatrix.m02 = (float)(-y * rsin + z * x * (1 - rcos));
        orientationMatrix.m10 = (float)(-z * rsin + x * y * (1 - rcos));
        orientationMatrix.m11 = (float)(rcos + y * y * (1 - rcos));
        orientationMatrix.m12 = (float)(x * rsin + z * y * (1 - rcos));
        orientationMatrix.m20 = (float)(y * rsin + x * z * (1 - rcos));
        orientationMatrix.m21 = (float)(-x * rsin + y * z * (1 - rcos));
        orientationMatrix.m22 = (float)(rcos + z * z * (1 - rcos));
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        terminate = true;
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Update and cull the scenegraph of a single node without doing any other
     * processing. This generates an ordered list
     * of nodes to render. It will not return until the culling is complete.
     *
     * @param node The root node to start the cull from
     */
    private int cullSingle(Group node)
    {
        // set the top matrix stack back to identity
        transformStack[0].setIdentity();

        if(node instanceof TransformGroup)
            ((TransformGroup)node).getTransform(transformStack[0]);

        return findAllNodes((Group)node, 0);
    }

    /**
     * Recursive walk of the tree to find all the renderable nodes.
     *
     * @param group The parent node to walk into
     * @param cullEndIndex The current last item on the cull list
     * @return The index of the last item on the cull list
     */
    private int findAllNodes(Group group, int cullEndIndex)
    {
        if(terminate)
            return 0;

        Node[] kids  = group.getRenderableChild();
        int size = group.numRenderableChildren();

        int num_lights = 0;
        int num_clips = 0;
        int ret_val = cullEndIndex;
        boolean have_local_fog = false;

        // Find all the lights and clip planes at this level first. Place
        // them first into the queue
        for(int i = 0; i < size; i++)
        {
            if(kids[i] instanceof Light)
            {
                Light l = (Light)kids[i];
                if(!l.isEnabled())
                    continue;

                resizeLightList();
                lightList[lastLight] = l;

                Matrix4f mat = transformStack[lastTxStack];

                // Transpose the matrix in place as it is being copied
                lightTxList[lastLight][0] = (float)mat.m00;
                lightTxList[lastLight][1] = (float)mat.m10;
                lightTxList[lastLight][2] = (float)mat.m20;
                lightTxList[lastLight][3] = (float)mat.m30;

                lightTxList[lastLight][4] = (float)mat.m01;
                lightTxList[lastLight][5] = (float)mat.m11;
                lightTxList[lastLight][6] = (float)mat.m21;
                lightTxList[lastLight][7] = (float)mat.m31;

                lightTxList[lastLight][8] = (float)mat.m02;
                lightTxList[lastLight][9] = (float)mat.m12;
                lightTxList[lastLight][10] = (float)mat.m22;
                lightTxList[lastLight][11] = (float)mat.m32;

                lightTxList[lastLight][12] = (float)mat.m03;
                lightTxList[lastLight][13] = (float)mat.m13;
                lightTxList[lastLight][14] = (float)mat.m23;
                lightTxList[lastLight][15] = (float)mat.m33;

                lastLight++;
                num_lights++;
            }
            else if(kids[i] instanceof ClipPlane)
            {
                ClipPlane cp = (ClipPlane)kids[i];
                if(!cp.isEnabled())
                    continue;

                resizeClipList();
                clipList[lastClip] = cp;

                Matrix4f mat = transformStack[lastTxStack];

                // Transpose the matrix in place as it is being copied
                clipTxList[lastClip][0] = (float)mat.m00;
                clipTxList[lastClip][1] = (float)mat.m10;
                clipTxList[lastClip][2] = (float)mat.m20;
                clipTxList[lastClip][3] = (float)mat.m30;

                clipTxList[lastClip][4] = (float)mat.m01;
                clipTxList[lastClip][5] = (float)mat.m11;
                clipTxList[lastClip][6] = (float)mat.m21;
                clipTxList[lastClip][7] = (float)mat.m31;

                clipTxList[lastClip][8] = (float)mat.m02;
                clipTxList[lastClip][9] = (float)mat.m12;
                clipTxList[lastClip][10] = (float)mat.m22;
                clipTxList[lastClip][11] = (float)mat.m32;

                clipTxList[lastClip][12] = (float)mat.m03;
                clipTxList[lastClip][13] = (float)mat.m13;
                clipTxList[lastClip][14] = (float)mat.m23;
                clipTxList[lastClip][15] = (float)mat.m33;

                lastClip++;
                num_clips++;
            }
            else if(kids[i] instanceof Fog)
            {
                Fog fog = (Fog)kids[i];
                if(fog.isEnabled() && !fog.isGlobalOnly() && !have_local_fog)
                {
                    resizeFogStack();

                    have_local_fog = true;
                    lastFogStack++;
                    fogStack[lastFogStack] = fog;
                }
            }
        }

        for(int i = 0; i < size && !terminate; i++)
        {
            if(kids[i] instanceof Group)
            {
                // If a tg, push the new TX onto the stack
                boolean is_tx = (kids[i] instanceof TransformGroup);

                if(is_tx)
                {
                    resizeStack();
                    TransformGroup tg = (TransformGroup)kids[i];
                    tg.getTransform(transformStack[lastTxStack + 1]);

                    transformStack[lastTxStack + 1].mul(transformStack[lastTxStack],
                                                      transformStack[lastTxStack + 1]);
                    lastTxStack++;
                }

                ret_val = findAllNodes((Group)kids[i], ret_val);

                // Now pop the stacks

                if(is_tx)
                    lastTxStack--;
            }
            else if(kids[i] instanceof Shape3D)
            {
                // Check the visibility state and ignore if not visible.
                if(!((Shape3D)kids[i]).isVisible())
                    continue;

                // Walk into the shape and check that we don't have any
                // offscreen textures to render.
                if(checkOffscreens)
                    checkForOffscreens((Shape3D)kids[i]);

                resizeCullList(ret_val);
                workCullList[ret_val].renderable = (Shape3D)kids[i];
                workCullList[ret_val].localFog = fogStack[lastFogStack];

                Matrix4f mat = transformStack[lastTxStack];

                // Transpose the matrix in place as it is being copied
                workCullList[ret_val].transform.set(mat);
                int src_size = (workCullList[ret_val].lights == null) ?
                                0 :
                                workCullList[ret_val].lights.length;

                if(src_size < lastLight)
                {
                    // up the size of the array
                    LightDetails[] tmp = new LightDetails[lastLight];
                    if(src_size != 0)
                    {
                        System.arraycopy(workCullList[ret_val].lights,
                                         0,
                                         tmp,
                                         0,
                                         src_size);
                    }

                    for(int j = src_size; j < lastLight; j++)
                        tmp[j] = new LightDetails();

                    workCullList[ret_val].lights = tmp;
                }

                // copy in the light information.
                LightDetails[] l_tmp = workCullList[ret_val].lights;
                workCullList[ret_val].numLights = lastLight;

                for(int j = 0; j < lastLight; j++)
                    l_tmp[j].update(lightList[j], lightTxList[j]);

                // Same thing again but for clip planes
                src_size = (workCullList[ret_val].clipPlanes == null) ?
                            0 :
                            workCullList[ret_val].clipPlanes.length;

                if(src_size < lastClip)
                {
                    // up the size of the array
                    ClipDetails[] tmp = new ClipDetails[lastClip];
                    if(src_size != 0)
                    {
                        System.arraycopy(workCullList[ret_val].clipPlanes,
                                         0,
                                         tmp,
                                         0,
                                         src_size);
                    }

                    for(int j = src_size; j < lastClip; j++)
                        tmp[j] = new ClipDetails();

                    workCullList[ret_val].clipPlanes = tmp;
                }

                // copy in the light information.
                ClipDetails[] c_tmp = workCullList[ret_val].clipPlanes;
                workCullList[ret_val].numClipPlanes = lastClip;

                for(int j = 0; j < lastClip; j++)
                    c_tmp[j].update(clipList[j], clipTxList[j]);

                ret_val++;
            }
            else if(kids[i] instanceof SharedNode)
            {
                ret_val = findNextNode((SharedNode)kids[i],  ret_val);
            }
        }

        // Pop the valid lights from the stack
        lastLight -= num_lights;
        lastClip -= num_clips;

        if(have_local_fog)
        {
            fogStack[lastFogStack] = null;
            lastFogStack--;
        }

        return ret_val;
    }

    /**
     * From the given SharedNode instance keep finding a non-sharedNode
     * instance, before continuing on the recursion
     *
     * @param root The parent node to walk into
     * @param cullEndIndex The current last item on the cull list
     * @return The index of the last item on the cull list
     */
    private int findNextNode(SharedNode root, int cullEndIndex)
    {
        if(terminate)
            return 0;

        int ret_val = cullEndIndex;

        Node current = root.getChild();

        while(current != null && current instanceof SharedNode)
            current = ((SharedNode)current).getChild();

        // Now just continue on normally. If the result of the above was a
        // null then it just falls off the end of this if/else naturally.
        if(current instanceof Group)
        {
            // If a tg, push the new TX onto the stack
            boolean is_tx = (current instanceof TransformGroup);

            if(is_tx)
            {
                resizeStack();
                TransformGroup tg = (TransformGroup)current;
                tg.getTransform(transformStack[lastTxStack + 1]);

                transformStack[lastTxStack + 1].mul(transformStack[lastTxStack],
                                                  transformStack[lastTxStack + 1]);
                lastTxStack++;
            }

            ret_val = findAllNodes((Group)current, cullEndIndex);

            // Now pop the stack.
            if(is_tx)
                lastTxStack--;
        }
        else if(current instanceof Shape3D)
        {
            // Check the visibility state and ignore if not visible.
            if(!((Shape3D)current).isVisible())
                return ret_val;

            // Walk into the shape and check that we don't have any
            // offscreen textures to render.
            if(checkOffscreens)
                checkForOffscreens((Shape3D)current);

            resizeCullList(cullEndIndex);
            workCullList[cullEndIndex].renderable = (Shape3D)current;
            workCullList[cullEndIndex].localFog = fogStack[lastFogStack];

            Matrix4f mat = transformStack[lastTxStack];

            // Transpose the matrix in place as it is being copied
            workCullList[cullEndIndex].transform.set(mat);

            int src_size = (workCullList[cullEndIndex].lights == null) ?
                            0 :
                            workCullList[cullEndIndex].lights.length;

            if(src_size < lastLight)
            {
                // up the size of the array

                LightDetails[] tmp = new LightDetails[lastLight];
                if(src_size != 0)
                {
                    System.arraycopy(workCullList[cullEndIndex].lights,
                                     0,
                                     tmp,
                                     0,
                                     src_size);
                }

                for(int j = src_size; j < lastLight; j++)
                    tmp[j] = new LightDetails();

                workCullList[cullEndIndex].lights = tmp;
            }

            // copy in the light information.
            LightDetails[] l_tmp = workCullList[cullEndIndex].lights;
            workCullList[cullEndIndex].numLights = lastLight;

            for(int j = 0; j < lastLight; j++)
                l_tmp[j].update(lightList[j], lightTxList[j]);

            // Same again, but with clip information
            src_size = (workCullList[cullEndIndex].clipPlanes == null) ?
                       0 :
                       workCullList[cullEndIndex].clipPlanes.length;

            if(src_size < lastClip)
            {
                // up the size of the array

                ClipDetails[] tmp = new ClipDetails[lastClip];
                if(src_size != 0)
                {
                    System.arraycopy(workCullList[cullEndIndex].clipPlanes,
                                     0,
                                     tmp,
                                     0,
                                     src_size);
                }

                for(int j = src_size; j < lastClip; j++)
                    tmp[j] = new ClipDetails();

                workCullList[cullEndIndex].clipPlanes = tmp;
            }

            // copy in the clip information.
            ClipDetails[] c_tmp = workCullList[cullEndIndex].clipPlanes;
            workCullList[cullEndIndex].numClipPlanes = lastClip;

            for(int j = 0; j < lastClip; j++)
                c_tmp[j].update(clipList[j], clipTxList[j]);

            cullEndIndex++;

            ret_val++;
        }

        return ret_val;
    }

    /**
     * Take the scene and fill in a RenderEnvironmentData instance.
     *
     * @param scene The scene to take data from
     * @param envData Data instance to copy it to
     */
    private void fillEnvData(Scene scene, RenderEnvironmentData envData)
    {
        ViewEnvironment view = scene.getViewEnvironment();

        envData.userData = scene.getUserData();
        envData.effectsProcessor = scene.getRenderEffectsProcessor();
        envData.viewpoint = scene.getActiveView();
        envData.background = scene.getActiveBackground();
        envData.viewEnvironment = view;

        Fog fog = scene.getActiveFog();
        if((fog != null) && fog.isEnabled())
            envData.fog = fog;

        envData.eyeOffset[0] = eyePoint[0];
        envData.eyeOffset[1] = eyePoint[1];
        envData.eyeOffset[2] = eyePoint[2];

        // Calculate frustum parameters based on view environment
        if(view == null)
        {
            double aspect = 1;
            double zNear = 0.1;
            double zFar = 1000;
            double fov = 45;

            double ymax = zNear * Math.tan(fov * Math.PI / 360.0);
            double ymin = -ymax;
            double xmin = ymin * aspect;
            double xmax = ymax * aspect;

            envData.viewFrustum[0] = xmin;
            envData.viewFrustum[1] = xmax;
            envData.viewFrustum[2] = ymin;
            envData.viewFrustum[3] = ymax;
            envData.viewFrustum[4] = zNear;
            envData.viewFrustum[5] = zFar;
        }
        else
        {
            Rectangle viewport = view.getViewport();
            if(viewport == null)
                envData.hasViewport = false;
            else
            {
                envData.hasViewport = true;
                envData.viewport.setBounds(viewport);
            }

            view.generateViewFrustum(envData.viewFrustum);
        }

        if(useEyePoint)
        {
            envData.viewFrustum[0] -= eyePoint[0];
            envData.viewFrustum[1] -= eyePoint[0];
            envData.viewFrustum[2] -= eyePoint[1];
            envData.viewFrustum[3] -= eyePoint[1];
            envData.viewFrustum[4] -= eyePoint[2];
            envData.viewFrustum[5] -= eyePoint[2];
        }


        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = envData.viewpoint.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if((parent instanceof SharedGroup) ||
               (parent instanceof SharedNode))
                throw new IllegalArgumentException(SHARED_VP_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        if(currentViewpointPath.length < path_depth)
            currentViewpointPath = new TransformGroup[path_depth];

        // start at the VP and work our way upwards.
        parent = envData.viewpoint.getParent();
        int last_tx = 0;
        while(parent != null)
        {
            if(parent instanceof TransformGroup)
                currentViewpointPath[last_tx++] = (TransformGroup)parent;

            parent = parent.getParent();
        }

        viewMatrix2.setIdentity();

        if(last_tx != 0)
        {
            for(int i = last_tx - 1; i >= 0; i--)
            {
                currentViewpointPath[i].getTransform(viewMatrix1);
                viewMatrix2.mul(viewMatrix1);
            }
        }

        if(useOrientation)
            viewMatrix2.mul(orientationMatrix);

        envData.viewTransform.set(viewMatrix2);

        // Now do almost the same thing with the background node. Only
        // difference is that this time we deliberately set the translation
        // components to zero.
        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        if(envData.background != null)
        {
            parent = envData.background.getParent();
            path_depth = 0;

            while(parent != null)
            {
                if((parent instanceof SharedGroup) ||
                   (parent instanceof SharedNode))
                    throw new IllegalArgumentException(SHARED_BG_MSG);

                parent = parent.getParent();
                path_depth++;
            }

            // reuse the VP path for this one
            if(currentViewpointPath.length < path_depth)
                currentViewpointPath = new TransformGroup[path_depth];

            // start at the background and work our way upwards.
            parent = envData.background.getParent();
            last_tx = 0;
            while(parent != null)
            {
                if(parent instanceof TransformGroup)
                    currentViewpointPath[last_tx++] = (TransformGroup)parent;

                parent = parent.getParent();
            }

            viewMatrix2.setIdentity();

            if(last_tx != 0)
            {
                for(int i = last_tx - 1; i >= 0; i--)
                {
                    currentViewpointPath[i].getTransform(viewMatrix1);
                    viewMatrix2.mul(viewMatrix1);
                }
            }

            Matrix4f bg_mat = envData.backgroundTransform;

            bg_mat.m00 = viewMatrix2.m00;
            bg_mat.m01 = viewMatrix2.m01;
            bg_mat.m02 = viewMatrix2.m02;
            bg_mat.m03 = viewMatrix2.m03;
            bg_mat.m10 = viewMatrix2.m10;
            bg_mat.m11 = viewMatrix2.m11;
            bg_mat.m12 = viewMatrix2.m12;
            bg_mat.m13 = viewMatrix2.m13;
            bg_mat.m20 = viewMatrix2.m20;
            bg_mat.m21 = viewMatrix2.m21;
            bg_mat.m22 = viewMatrix2.m22;
            bg_mat.m23 = viewMatrix2.m23;
            bg_mat.m30 = 0;
            bg_mat.m31 = 0;
            bg_mat.m32 = 0;
            bg_mat.m33 = viewMatrix2.m33;
        }

        // Create projection matrix
        float left = (float)envData.viewFrustum[0];
        float right = (float)envData.viewFrustum[1];
        float bottom = (float)envData.viewFrustum[2];
        float top = (float)envData.viewFrustum[3];
        float nearval = (float)envData.viewFrustum[4];
        float farval = (float)envData.viewFrustum[5];
        float x, y, z, w;
        float a, b, c, d;

        x = (2.0f * nearval) / (right - left);
        y = (2.0f * nearval) / (top - bottom);
        a = (right + left) / (right - left);
        b = (top + bottom) / (top - bottom);
        c = -(farval + nearval) / ( farval - nearval);
        d = -(2.0f * farval * nearval) / (farval - nearval);

        Matrix4f prj_mat = envData.projectionMatrix;

        prj_mat.m00 = x;
        prj_mat.m01 = 0;
        prj_mat.m02 = a;
        prj_mat.m03 = 0;
        prj_mat.m10 = 0;
        prj_mat.m11 = y;
        prj_mat.m12 = b;
        prj_mat.m13 = 0;
        prj_mat.m20 = 0;
        prj_mat.m21 = 0;
        prj_mat.m22 = c;
        prj_mat.m23 = d;
        prj_mat.m30 = 0;
        prj_mat.m31 = 0;
        prj_mat.m32 = -1;
        prj_mat.m33 = 0;
    }

    /**
     * Check a shape node for the offscreen textures that may be present. If
     * some are found, queue them up for processing.
     *
     * @param shape The object instance to process
     */
    private void checkForOffscreens(Shape3D shape)
    {
        Appearance app = shape.getAppearance();

        if(app == null)
            return;

        int num_units = app.numTextureUnits();
        if(num_units == 0)
            return;

        app.getTextureUnits(textureUnits);

        for(int i =0; i < num_units; i++)
        {
            Texture tex = textureUnits[i].getTexture();

            // Only queue up the texture for repainting if it says it needs a
            // repaint and that we haven't already found this instance in this
            // frame.
            if((tex instanceof PBufferTextureSource) &&
               !visitedNodes.contains(tex) &&
               ((PBufferTextureSource)tex).isRepaintRequired())
            {
                resizeOffscreenList();
                pendingSubscenes[lastSubscene] = (PBufferTextureSource)tex;
                pendingParents[lastSubscene] = activeParent;
                visitedNodes.add(tex);
                lastSubscene++;
            }
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeCullList(int cur_size)
    {
        if((cur_size + 1) == workCullList.length)
        {
            int old_size = workCullList.length;
            int new_size = old_size + LIST_INCREMENT;

            CullOutputDetails[] tmp_nodes = new CullOutputDetails[new_size];

            System.arraycopy(workCullList, 0, tmp_nodes, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_nodes[i] = new CullOutputDetails();

            workCullList = tmp_nodes;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeOutputLists()
    {
        if((lastSubscene + 1) < validCullList.length)
            return;

        int old_size = validCullList.length;
        int new_size = old_size + LIST_INCREMENT;

        CullOutputDetails[][] tmp_nodes =
            new CullOutputDetails[new_size][];

        System.arraycopy(validCullList, 0, tmp_nodes, 0, old_size);

        for(int i = old_size; i < new_size; i++)
        {
            tmp_nodes[i] = new CullOutputDetails[LIST_START_DEPTH];

            for(int j = 0; j < LIST_START_DEPTH; j++)
                tmp_nodes[i][j] = new CullOutputDetails();
        }

        validCullList = tmp_nodes;

        int[] tmp_ints = new int[new_size];
        System.arraycopy(validCullSize, 0, tmp_ints, 0, old_size);

        validCullSize = tmp_ints;

        SceneGraphObject[][] tmp_obj = new SceneGraphObject[new_size][];

        System.arraycopy(validSceneParents, 0, tmp_obj, 0, old_size);

        for(int i = old_size; i < new_size; i++)
            tmp_obj[i] = new SceneGraphObject[2];

        validSceneParents = tmp_obj;

        RenderEnvironmentData[] tmp_env = new RenderEnvironmentData[new_size];
        System.arraycopy(validEnvData, 0, tmp_env, 0, old_size);

         for(int i = old_size; i < new_size; i++)
             tmp_env[i] = new RenderEnvironmentData();

        validEnvData = tmp_env;
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeOffscreenList()
    {
        if((lastSubscene + 1) == pendingSubscenes.length)
        {
            int old_size = pendingSubscenes.length;
            int new_size = old_size + LIST_INCREMENT;

            PBufferTextureSource[] tmp_nodes =
                new PBufferTextureSource[new_size];

            System.arraycopy(pendingSubscenes, 0, tmp_nodes, 0, old_size);

            pendingSubscenes = tmp_nodes;

            tmp_nodes = new PBufferTextureSource[new_size];

            System.arraycopy(pendingParents, 0, tmp_nodes, 0, old_size);

            pendingParents = tmp_nodes;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeLightList()
    {
        if((lastLight + 1) == lightList.length)
        {
            int old_size = lightList.length;
            int new_size = old_size + LIST_INCREMENT;

            Light[] tmp_nodes = new Light[new_size];

            System.arraycopy(lightList, 0, tmp_nodes, 0, old_size);

            lightList = tmp_nodes;

            float[][] tmp_tx  = new float[new_size][];
            System.arraycopy(lightTxList, 0, tmp_tx, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_tx[i] = new float[16];

            lightTxList = tmp_tx;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeClipList()
    {
        if((lastClip + 1) == clipList.length)
        {
            int old_size = clipList.length;
            int new_size = old_size + LIST_INCREMENT;

            ClipPlane[] tmp_nodes = new ClipPlane[new_size];

            System.arraycopy(clipList, 0, tmp_nodes, 0, old_size);

            clipList = tmp_nodes;

            float[][] tmp_tx  = new float[new_size][];
            System.arraycopy(clipTxList, 0, tmp_tx, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_tx[i] = new float[16];

            clipTxList = tmp_tx;
        }
    }

    /**
     * Resize the transform stack if needed. Marked as final in order to
     * encourage the compiler to inline the code for faster execution
     */
    private final void resizeStack()
    {
        if((lastTxStack + 1) == transformStack.length)
        {
            int old_size = transformStack.length;
            int new_size = old_size + STACK_INCREMENT;

            Matrix4f[] tmp_matrix = new Matrix4f[new_size];

            System.arraycopy(transformStack, 0, tmp_matrix, 0, old_size);

            for(int i = old_size; i < new_size; i++)
                tmp_matrix[i] = new Matrix4f();

            transformStack = tmp_matrix;
        }
    }

    /**
     * Resize the fog stack if needed. Marked as final in order to
     * encourage the compiler to inline the code for faster execution
     */
    private final void resizeFogStack()
    {
        if((lastFogStack + 1) == fogStack.length)
        {
            int old_size = fogStack.length;
            int new_size = old_size + STACK_INCREMENT;

            Fog[] tmp_fog = new Fog[new_size];

            System.arraycopy(fogStack, 0, tmp_fog, 0, old_size);

            fogStack = tmp_fog;
        }
    }
}
