/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.util.HashMap;

import javax.vecmath.Matrix4f;

// Local imports
import org.j3d.aviatrix3d.*;

/**
 * Implementation of the sort stage that does nothing.
 * <p>
 *
 * The sort stage just takes the given nodes and expands them into an array
 * renders and then immediately pops the node. No sorting on output is done.
 *
 * @author Justin Couch
 * @version $Revision: 1.20 $
 */
public class NullSortStage implements SortStage
{
    /** The initial size of the instruction list */
    private static final int LIST_START_SIZE = 1;

    /** Receiver for the output of this stage */
    private SortedGeometryReceiver receiver;

    /** Output array for passing on to the receiver */
    private RenderInstructions[] commandList;

    /** A semi-unique ID counter used for assigning light IDs. May wrap */
    private int lastGlobalId;

    /** Map of the first occurance of a scene parent to it's render instruction */
    private HashMap instructionMap;

    /** Flag indicating a shutdown of the current processing is requested */
    private boolean terminate;

    /**
     * Create an empty sorting stage that assumes just a single renderable
     * output.
     */
    public NullSortStage()
    {
        this(LIST_START_SIZE);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen.
     *
     */
    public NullSortStage(int numSurfaces)
    {
        commandList = new RenderInstructions[numSurfaces];
        lastGlobalId = 0;

        for(int i = 0; i < numSurfaces; i++)
            commandList[i] = new RenderInstructions();

        instructionMap = new HashMap();
    }

    /**
     * Sort the listing of nodes in the given array. Do not return until the
     * sort has been completed. If the sceneParent is the main scene graph and
     * not an offscreen texture, then that spot should be set to null in the list.
     *
     * @param data External rendering environment information
     * @param nodes List of processed nodes based on the scene they came from
     * @param numNodes Number of nodes in each scene
     * @param sceneParents Parent nodes that hold the subscene in the main scene graph
     * @param numScenes The number of valid scenes to process
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void sort(RenderEnvironmentData[] data,
                     CullOutputDetails[][] nodes,
                     int[] numNodes,
                     SceneGraphObject[][] sceneParents,
                     int numScenes,
                     int buffer)
    {
        // Buffer ID is not used currently.

        terminate = false;

        if(commandList.length < numScenes)
        {
            RenderInstructions[] tmp = new RenderInstructions[numScenes];

            System.arraycopy(commandList, 0, tmp, 0, commandList.length);

            for(int i = commandList.length; i < numScenes   ; i++)
                tmp[i] = new RenderInstructions();

            commandList = tmp;
        }

        // While doing these loops, we need to find the main scene and ensure
        // that it is the last item in the output list.
        int main_idx = 0;
        int curr_scene = 0;

        for(int i = 0; i < numScenes && !terminate; i++)
        {
            if(sceneParents[i][0] == null)
            {
                main_idx = i;
                continue;
            }

            commandList[curr_scene].pbuffer =
                (PBufferTextureSource)sceneParents[i][0];
            commandList[curr_scene].parentSource =
                (PBufferTextureSource)sceneParents[i][1];
            commandList[curr_scene].renderData = data[i];

            RenderInstructions original =
                (RenderInstructions)instructionMap.get(sceneParents[i][0]);

            if(original != null)
            {
                commandList[curr_scene].copyOf = original;
            }
            else
            {
                commandList[curr_scene].copyOf = null;
                sortSingle(nodes[i], numNodes[i], commandList[curr_scene]);

                instructionMap.put(sceneParents[i][0],
                                   commandList[curr_scene]);
            }

            curr_scene++;
        }

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        commandList[curr_scene].pbuffer = null;
        commandList[curr_scene].renderData = data[main_idx];

        sortSingle(nodes[main_idx], numNodes[main_idx], commandList[curr_scene]);

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        curr_scene++;

        if(receiver != null)
            receiver.sortedOutput(commandList, numScenes);

        instructionMap.clear();
    }


    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setSortedGeometryReceiver(SortedGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        terminate = true;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Sort a single set of nodes into the output details and place in the
     * provided RenderInstructions instance.
     *
     * @param instr Instruction instant to put the details into
     */
    private void sortSingle(CullOutputDetails[] nodes,
                            int numNodes,
                            RenderInstructions instr)
    {
        // check to see the array is big enough

        int req_size = numNodes << 1;
        for(int i = 0; i < numNodes; i++)
        {
            req_size += nodes[i].numLights << 1;
            req_size += nodes[i].numClipPlanes << 1;
        }

        if(instr.renderList.length < req_size)
        {
            RenderDetails[] tmp = new RenderDetails[req_size];

            System.arraycopy(instr.renderList, 0, tmp, 0, instr.renderList.length);

            for(int i = instr.renderList.length; i < req_size; i++)
                tmp[i] = new RenderDetails();

            instr.renderList = tmp;
            instr.renderOps = new int[req_size];
        }

        int idx = 0;
        int fog_id = 0;

        for(int i = 0; i < numNodes && !terminate; i++)
        {
            // First drop all the lights into the queue.
            if(nodes[i].numLights != 0)
            {
                for(int j = 0; j < nodes[i].numLights; j++)
                {
                    LightDetails ld = nodes[i].lights[j];
                    instr.renderList[idx].renderable = ld.getLight();

                    System.arraycopy(ld.getTransform(),
                                     0,
                                     instr.renderList[idx].transform,
                                     0,
                                     16);

                    instr.renderList[idx].id = lastGlobalId++;
                    instr.renderOps[idx] = RenderOp.START_LIGHT;
                    idx++;
                }
            }

            // Next drop all the clip planes into the queue.
            if(nodes[i].numClipPlanes != 0)
            {
                for(int j = 0; j < nodes[i].numClipPlanes; j++)
                {
                    ClipDetails cd = nodes[i].clipPlanes[j];
                    instr.renderList[idx].renderable = cd.getClipPlane();

                    System.arraycopy(cd.getTransform(),
                                     0,
                                     instr.renderList[idx].transform,
                                     0,
                                     16);

                    instr.renderList[idx].id = lastGlobalId++;
                    instr.renderOps[idx] = RenderOp.START_CLIP_PLANE;
                    idx++;
                }
            }

            if(nodes[i].localFog != null)
            {
                fog_id = lastGlobalId++;
                instr.renderList[idx].id = fog_id;
                instr.renderList[idx].renderable = nodes[i].localFog;
                instr.renderOps[idx] = RenderOp.START_FOG;
                idx++;
            }

            // Now process the node type.
            if(nodes[i].renderable instanceof Shape3D)
            {
                // For this simple one, just use the basic render command.
                // Something more complex would pull the shape apart and do
                // state/depth/transparency sorting in this section.

                instr.renderList[idx].renderable = nodes[i].renderable;

                Matrix4f tx = nodes[i].transform;

                instr.renderList[idx].transform[0] = tx.m00;
                instr.renderList[idx].transform[1] = tx.m10;
                instr.renderList[idx].transform[2] = tx.m20;
                instr.renderList[idx].transform[3] = tx.m30;

                instr.renderList[idx].transform[4] = tx.m01;
                instr.renderList[idx].transform[5] = tx.m11;
                instr.renderList[idx].transform[6] = tx.m21;
                instr.renderList[idx].transform[7] = tx.m31;

                instr.renderList[idx].transform[8] = tx.m02;
                instr.renderList[idx].transform[9] = tx.m12;
                instr.renderList[idx].transform[10] = tx.m22;
                instr.renderList[idx].transform[11] = tx.m32;

                instr.renderList[idx].transform[12] = tx.m03;
                instr.renderList[idx].transform[13] = tx.m13;
                instr.renderList[idx].transform[14] = tx.m23;
                instr.renderList[idx].transform[15] = tx.m33;

                instr.renderOps[idx] = RenderOp.START_RENDER;
                idx++;

                instr.renderList[idx].renderable = nodes[i].renderable;
                instr.renderOps[idx] = RenderOp.STOP_RENDER;
                idx++;
            }
            else if(nodes[i].renderable instanceof CustomRenderable)
            {
                instr.renderList[idx].renderable = nodes[i].renderable;
                instr.renderList[idx].instructions = nodes[i].customData;

                Matrix4f tx = nodes[i].transform;

                instr.renderList[idx].transform[0] = tx.m00;
                instr.renderList[idx].transform[1] = tx.m10;
                instr.renderList[idx].transform[2] = tx.m20;
                instr.renderList[idx].transform[3] = tx.m30;

                instr.renderList[idx].transform[4] = tx.m01;
                instr.renderList[idx].transform[5] = tx.m11;
                instr.renderList[idx].transform[6] = tx.m21;
                instr.renderList[idx].transform[7] = tx.m31;

                instr.renderList[idx].transform[8] = tx.m02;
                instr.renderList[idx].transform[9] = tx.m12;
                instr.renderList[idx].transform[10] = tx.m22;
                instr.renderList[idx].transform[11] = tx.m32;

                instr.renderList[idx].transform[12] = tx.m03;
                instr.renderList[idx].transform[13] = tx.m13;
                instr.renderList[idx].transform[14] = tx.m23;
                instr.renderList[idx].transform[15] = tx.m33;

                instr.renderOps[idx] = RenderOp.RENDER_CUSTOM;
                idx++;
            }
            else
            {
                System.out.println("Non-shape node in sorter " + nodes[i].renderable);
            }

            if(nodes[i].localFog != null)
            {
                instr.renderList[idx].id = fog_id;
                instr.renderList[idx].renderable = nodes[i].localFog;
                instr.renderOps[idx] = RenderOp.STOP_FOG;
                idx++;
            }

            // Clean up the clip planess
            if(nodes[i].numClipPlanes != 0)
            {
                for(int j = nodes[i].numClipPlanes - 1; j >= 0; j--)
                {
                    ClipDetails cd = nodes[i].clipPlanes[j];
                    instr.renderList[idx].renderable = cd.getClipPlane();

                    instr.renderList[idx].id = lastGlobalId - (1 + j);

                    // Don't need the transform for the clip stopping. Save
                    // CPU cycles by not copying it.

                    instr.renderOps[idx] = RenderOp.STOP_CLIP_PLANE;
                    idx++;
                }
            }

            // Clean up the lights
            if(nodes[i].numLights != 0)
            {
                for(int j = nodes[i].numLights - 1; j >= 0; j--)
                {
                    LightDetails ld = nodes[i].lights[j];
                    instr.renderList[idx].renderable = ld.getLight();

                    instr.renderList[idx].id = lastGlobalId - (1 + j);

                    // Don't need the transform for the light stopping. Save
                    // CPU cycles by not copying it.

                    instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                    idx++;
                }
            }
        }

        instr.numValid = idx;
    }
}
