/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.util.HashMap;


// Local imports
import org.j3d.aviatrix3d.*;

/**
 * Implementation of the sort stage that separates out the transparent and
 * non transparent objects, but without depth sorting.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.11 $
 */
public class SimpleTransparencySortStage implements SortStage
{
    /** The initial size of the instruction list */
    private static final int LIST_START_SIZE = 1;

    /** The initial size of the transparent list */
    private static final int TRANSPARENT_LIST_SIZE = 200;


    /** Receiver for the output of this stage */
    private SortedGeometryReceiver receiver;

    /** Output array for passing on to the receiver */
    private RenderInstructions[] commandList;

    /** Temporary array for holding the transparent objects */
    private CullOutputDetails[] transparentList;

    /** Temp array for reading texture units from the appearance */
    private TextureUnit[] textureTmp;

    /** A semi-unique ID counter used for assigning light IDs. May wrap */
    private int lastGlobalId;

    /** Map of the first occurance of a scene parent to it's render instruction */
    private HashMap instructionMap;

    /** Flag indicating a shutdown of the current processing is requested */
    private boolean terminate;

    /**
     * Create an empty sorting stage that assumes just a single renderable
     * output.
     */
    public SimpleTransparencySortStage()
    {
        this(LIST_START_SIZE);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen.
     *
     */
    public SimpleTransparencySortStage(int numSurfaces)
    {
        commandList = new RenderInstructions[numSurfaces];
        transparentList = new CullOutputDetails[TRANSPARENT_LIST_SIZE];

        textureTmp = new TextureUnit[32];
        lastGlobalId = 0;

        for(int i = 0; i < numSurfaces; i++)
            commandList[i] = new RenderInstructions();

        instructionMap = new HashMap();
    }

    /**
     * Sort the listing of nodes in the given array. Do not return until the
     * sort has been completed. If the sceneParent is the main scene graph and
     * not an offscreen texture, then that spot should be set to null in the list.
     *
     * @param data External rendering environment information
     * @param nodes List of processed nodes based on the scene they came from
     * @param numNodes Number of nodes in each scene
     * @param sceneParents Parent nodes that hold the subscene in the main scene graph
     * @param numScenes The number of valid scenes to process
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void sort(RenderEnvironmentData[] data,
                     CullOutputDetails[][] nodes,
                     int[] numNodes,
                     SceneGraphObject[][] sceneParents,
                     int numScenes,
                     int buffer)
    {
        // Buffer ID is not used currently.

        terminate = false;

        if(commandList.length < numScenes)
        {
            RenderInstructions[] tmp = new RenderInstructions[numScenes];

            System.arraycopy(commandList, 0, tmp, 0, commandList.length);

            for(int i = commandList.length; i < numScenes; i++)
                tmp[i] = new RenderInstructions();

            commandList = tmp;
        }

        // While doing these loops, we need to find the main scene and ensure
        // that it is the last item in the output list.
        int main_idx = 0;
        int curr_scene = 0;

        for(int i = 0; i < numScenes && !terminate; i++)
        {
            if(sceneParents[i][0] == null)
            {
                main_idx = i;
                continue;
            }

            commandList[curr_scene].pbuffer =
                (PBufferTextureSource)sceneParents[i][0];
            commandList[curr_scene].parentSource =
                (PBufferTextureSource)sceneParents[i][1];
            commandList[curr_scene].renderData = data[i];

            RenderInstructions original =
                (RenderInstructions)instructionMap.get(sceneParents[i][0]);

            if(original != null)
            {
                commandList[curr_scene].copyOf = original;
            }
            else
            {
                commandList[curr_scene].copyOf = null;
                sortSingle(nodes[i], numNodes[i], commandList[curr_scene]);

                instructionMap.put(sceneParents[i][0],
                                   commandList[curr_scene]);
            }

            curr_scene++;
        }

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        commandList[curr_scene].pbuffer = null;
        commandList[curr_scene].renderData = data[main_idx];

        sortSingle(nodes[main_idx],
                   numNodes[main_idx],
                   commandList[curr_scene]);

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        curr_scene++;

        if(receiver != null)
            receiver.sortedOutput(commandList, numScenes);

        instructionMap.clear();
    }


    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setSortedGeometryReceiver(SortedGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        terminate = true;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Sort a single set of nodes into the output details and place in the
     * provided RenderInstructions instance.
     *
     * @param instr Instruction instant to put the details into
     */
    private void sortSingle(CullOutputDetails[] nodes,
                            int numNodes,
                            RenderInstructions instr)
    {
        // check to see the array is big enough. The +2 is to accomodate the
        // start and stop transparency commands.
        int req_size = 2 + numNodes << 1;
        for(int i = 0; i < numNodes; i++)
            req_size += nodes[i].numLights <<1;

        if(instr.renderList.length < req_size)
        {
            RenderDetails[] tmp = new RenderDetails[req_size];

            System.arraycopy(instr.renderList,
                             0,
                             tmp,
                             0,
                             instr.renderList.length);

            for(int i = instr.renderList.length; i < req_size; i++)
                tmp[i] = new RenderDetails();

            instr.renderList = tmp;
            instr.renderOps = new int[req_size];
        }

        if(transparentList.length < req_size)
            transparentList = new CullOutputDetails[req_size];

        int idx = 0;
        int trans = 0;

        for(int i = 0; i < numNodes && !terminate; i++)
        {
            // First check the geometry and so forth for something that
            // has an alpha channel. Need to check everything - geometry
            // colour values, textures and materials. Anything that says it
            // may have alpha is passed off onto the transparent list.
            // Now process the node type.
            if(nodes[i].renderable instanceof Shape3D)
            {
                Shape3D shape = (Shape3D)nodes[i].renderable;

                Appearance app = shape.getAppearance();

                if(app != null)
                {
                    Material mat = app.getMaterial();
                    if(mat != null && mat.getTransparency() != 1)
                    {
                        transparentList[trans++] = nodes[i];
                        continue;
                    }

                    int num_tex = app.numTextureUnits();
                    if(num_tex != 0)
                    {
                        boolean alpha_found = false;

                        app.getTextureUnits(textureTmp);
                        for(int j = 0; j < num_tex; j++)
                        {
                            Texture tex = textureTmp[j].getTexture();
                            if(tex == null)
                                continue;

                            // Any of the following formats, this is is added
                            // to the transparency collection. Break out of the
                            // inner loop by messing with the loop index.
                            switch(tex.getFormat())
                            {
                                case Texture.FORMAT_ALPHA:
                                case Texture.FORMAT_INTENSITY_ALPHA:
                                case Texture.FORMAT_RGBA:
                                    transparentList[trans++] = nodes[i];
                                    alpha_found = true;
                                    j = num_tex;
                                    break;

                            }
                        }

                        if(alpha_found)
                            continue;
                    }
                }

                Geometry geom = shape.getGeometry();
                if(geom.hasTransparency())
                {
                    transparentList[trans++] = nodes[i];
                    continue;
                }
            }
            else if(nodes[i].renderable instanceof CustomRenderable)
            {
                if(((CustomRenderable)nodes[i].renderable).hasTransparency())
                {
                    transparentList[trans++] = nodes[i];
                    continue;
                }
            }
            else
            {
                System.out.println("Non-shape node in sorter " +
                                   nodes[i].renderable);
                continue;
            }

            // Oh, so we have some geometry that is not transparent, Cool. Drop
            // it onto the queue. Start with all the lights:
            idx = appendOutputShape(idx, nodes[i], instr);
        }

        // Now process the transparency listing. If there is no transparent
        // objects, then exit now.
        if(trans == 0)
        {
            instr.numValid = idx;
            return;
        }

        // Basically the same now as the first loop, just running through the
        // transparency listing rather than the node listing.
        instr.renderOps[idx++] = RenderOp.START_TRANSPARENT;

        for(int i = 0; i < trans && !terminate; i++)
        {
            idx = appendOutputShape(idx, transparentList[i], instr);

            // clear it from the list
            transparentList[i] = null;
        }

        instr.renderOps[idx++] = RenderOp.STOP_TRANSPARENT;
        instr.numValid = idx;
    }

    /**
     * Convenience method to add a Shape3D to the list at the given index point.
     *
     * @param idx The starting place in the instruction list
     * @param node The node instance to add
     * @param instr List of instructions to add the node to
     * @return The update idx value taking into account the new nodes
     */
    private int appendOutputShape(int idx,
                                 CullOutputDetails node,
                                 RenderInstructions instr)
    {
        if(node.numLights != 0)
        {
            for(int j = 0; j < node.numLights; j++)
            {
                LightDetails ld = node.lights[j];
                instr.renderList[idx].renderable = ld.getLight();

                System.arraycopy(ld.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_LIGHT;
                idx++;
            }
        }

        if(node.numClipPlanes != 0)
        {
            for(int j = 0; j < node.numClipPlanes; j++)
            {
                ClipDetails cd = node.clipPlanes[j];
                instr.renderList[idx].renderable = cd.getClipPlane();

                System.arraycopy(cd.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_CLIP_PLANE;
                idx++;
            }
        }

        int fog_id = 0;

        if(node.localFog != null)
        {
            fog_id = lastGlobalId++;
            instr.renderList[idx].id = fog_id;
            instr.renderList[idx].renderable = node.localFog;
            instr.renderOps[idx] = RenderOp.START_FOG;
            idx++;
        }

        // Now process the node type.
        // For this simple one, just use the basic render command.
        // Something more complex would pull the shape apart and do
        // state/depth/transparency sorting in this section.
        if(node.renderable instanceof Shape3D)
        {
            instr.renderList[idx].renderable = node.renderable;

            instr.renderList[idx].transform[0] = node.transform.m00;
            instr.renderList[idx].transform[1] = node.transform.m10;
            instr.renderList[idx].transform[2] = node.transform.m20;
            instr.renderList[idx].transform[3] = node.transform.m30;

            instr.renderList[idx].transform[4] = node.transform.m01;
            instr.renderList[idx].transform[5] = node.transform.m11;
            instr.renderList[idx].transform[6] = node.transform.m21;
            instr.renderList[idx].transform[7] = node.transform.m31;

            instr.renderList[idx].transform[8] = node.transform.m02;
            instr.renderList[idx].transform[9] = node.transform.m12;
            instr.renderList[idx].transform[10] = node.transform.m22;
            instr.renderList[idx].transform[11] = node.transform.m32;

            instr.renderList[idx].transform[12] = node.transform.m03;
            instr.renderList[idx].transform[13] = node.transform.m13;
            instr.renderList[idx].transform[14] = node.transform.m23;
            instr.renderList[idx].transform[15] = node.transform.m33;

            instr.renderOps[idx] = RenderOp.START_RENDER;
            idx++;

            instr.renderList[idx].renderable = node.renderable;
            instr.renderOps[idx] = RenderOp.STOP_RENDER;
            idx++;
        }
        else if(node.renderable instanceof CustomRenderable)
        {
            instr.renderList[idx].renderable = node.renderable;
            instr.renderList[idx].instructions = node.customData;

            instr.renderList[idx].transform[0] = node.transform.m00;
            instr.renderList[idx].transform[1] = node.transform.m10;
            instr.renderList[idx].transform[2] = node.transform.m20;
            instr.renderList[idx].transform[3] = node.transform.m30;

            instr.renderList[idx].transform[4] = node.transform.m01;
            instr.renderList[idx].transform[5] = node.transform.m11;
            instr.renderList[idx].transform[6] = node.transform.m21;
            instr.renderList[idx].transform[7] = node.transform.m31;

            instr.renderList[idx].transform[8] = node.transform.m02;
            instr.renderList[idx].transform[9] = node.transform.m12;
            instr.renderList[idx].transform[10] = node.transform.m22;
            instr.renderList[idx].transform[11] = node.transform.m32;

            instr.renderList[idx].transform[12] = node.transform.m03;
            instr.renderList[idx].transform[13] = node.transform.m13;
            instr.renderList[idx].transform[14] = node.transform.m23;
            instr.renderList[idx].transform[15] = node.transform.m33;

            instr.renderOps[idx] = RenderOp.RENDER_CUSTOM;
            idx++;
        }

        if(node.localFog != null)
        {
            instr.renderList[idx].id = fog_id;
            instr.renderList[idx].renderable = node.localFog;
            instr.renderOps[idx] = RenderOp.STOP_FOG;
            idx++;
        }

        if(node.numClipPlanes != 0)
        {
            for(int j = node.numClipPlanes - 1; j >= 0; j--)
            {
                ClipDetails cd = node.clipPlanes[j];
                instr.renderList[idx].renderable = cd.getClipPlane();

                instr.renderList[idx].id = lastGlobalId - (1 + j);

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_CLIP_PLANE;
                idx++;
            }
        }

        if(node.numLights != 0)
        {
            for(int j = node.numLights - 1; j >= 0; j--)
            {
                LightDetails ld = node.lights[j];
                instr.renderList[idx].renderable = ld.getLight();

                instr.renderList[idx].id = lastGlobalId - (1 + j);

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;
            }
        }

        return idx;
    }
}
