/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.util.Arrays;
import java.util.HashMap;

import javax.vecmath.Matrix4f;

import net.java.games.jogl.GL;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

/**
 * Implementation of the sort stage that does nothing.
 * <p>
 *
 * The sort stage just takes the given nodes and expands them into an array
 * renders and then immediately pops the node. No sorting on output is done.
 *
 * @author Justin Couch
 * @version $Revision: 1.20 $
 */
public class StateSortStage implements SortStage
{
    /** The initial size of the instructions list */
    private static final int LIST_START_SIZE = 1;

    /** The size to grow the render list if needed */
    private static final int REALLOC_SIZE = 1024;

    /** The guess at the number of components in a shape */
    private static final int GUESS_NUM_COMPONENTS = 5;

    /** Listing of int to Integer equivalents for texture IDs */
    private static final Integer[] TEX_IDS;

    /** Receiver for the output of this stage */
    private SortedGeometryReceiver receiver;

    /** Output array for passing on to the receiver */
    private RenderInstructions[] commandList;

    /** A semi-unique ID counter used for assigning light IDs. May wrap */
    private int lastGlobalId;

    /** Map of the first occurance of a scene parent to it's render instructions */
    private HashMap instructionMap;

    /** Comparator instance for the state sorting */
    private StateSortComparator stateComparator;

    // Listing of the current items, while we're setting up the state handling.

    /** Temp array for holdin new lights on the object being processed */
    private LightDetails[] newLights;

    /** Temp array for fetching lights objects from the HashSets below. */
    private Object[] lightTmp;

    /** Set of current lights that are being processed */
    private HashSet currentLights;

    /** Set for holding the different between the current and new sets */
    private HashSet oldLights;

    /** Set to hold onto lights that are common between last node and this */
    private HashSet keepLights;

    /** Mapping of the LightDetails instance to the global ID */
    private HashMap lightIdMap;

    /** Temp array for holding new clip planes on the object being processed */
    private ClipDetails[] newClipPlanes;

    /** Temp array for fetching clipPlane objects from the HashSets below. */
    private Object[] clipTmp;

    /** Set of current clip planes that are being processed */
    private HashSet currentClipPlanes;

    /** Set for holding the difference between the current and new sets */
    private HashSet oldClipPlanes;

    /** Set to hold onto clip planes that are common between last node and this */
    private HashSet keepClipPlanes;

    /** Mapping of the ClipDetails instance to the global ID */
    private HashMap clipIdMap;

    /** The currently valid appearance */
    private Appearance currentAppearance;

    /** The currently valid material */
    private Material currentMaterial;

    /** The currently valid polygon attributes */
    private PolygonAttributes currentPolyAttr;

    /** The currently valid line attributes */
    private LineAttributes currentLineAttr;

    /** The currently valid point attributes */
    private PointAttributes currentPointAttr;

    /** The currently valid blend attributes */
    private BlendAttributes currentBlendAttr;

    /** The currently valid GLSlang shader*/
    private Shader currentShader;

    /** The currently valid shader program (GLSLang)*/
    private ShaderProgram currentShaderProgram;

    /** The currently valid shader arguments (GLSLang)*/
    private ShaderArguments currentShaderArgs;

    /** The currently valid vertex shader */
    private VertexShader currentVtxShader;

    /** The currently valid fragment shader */
    private FragmentShader currentFragShader;

    /** The currently valid local fog */
    private Fog currentFog;

    /** temp array for fetching textures from the appearance */
    private TextureUnit[] textureTmp;

    /** Listing of currently valid textures */
    private TextureUnit[] currentTextures;

    /** Number of valid textures currently active */
    private int numTextures;

    /** The requested size for the arrays */
    private int req_size;

    /** Flag indicating a shutdown of the current processing is requested */
    private boolean terminate;

    // Static constructor to initialise some of the constants
    static
    {
        TEX_IDS = new Integer[32];
        for(int i = 0; i < 32; i++)
            TEX_IDS[i] = new Integer(GL.GL_TEXTURE0 + i);
    }

    /**
     * Create an empty sorting stage that assumes just a single renderable
     * output.
     */
    public StateSortStage()
    {
        this(LIST_START_SIZE);
    }

    /**
     * Create an empty sorting stage that initialises the internal structures
     * to assume that there is a minumum number of surfaces, both on and
     * offscreen.
     *
     */
    public StateSortStage(int numSurfaces)
    {
        commandList = new RenderInstructions[numSurfaces];
        lastGlobalId = 0;

        for(int i = 0; i < numSurfaces; i++)
            commandList[i] = new RenderInstructions();

        instructionMap = new HashMap();
        stateComparator = new StateSortComparator();

        currentLights = new HashSet();
        keepLights = new HashSet();
        oldLights = new HashSet();
        lightIdMap = new HashMap();
        lightTmp = new Object[24];
        newLights = new LightDetails[32];

        currentClipPlanes = new HashSet();
        keepClipPlanes = new HashSet();
        oldClipPlanes = new HashSet();
        clipIdMap = new HashMap();
        clipTmp = new Object[24];
        newClipPlanes = new ClipDetails[32];

        // 32 is max textures that OpenGL can support.
        currentTextures = new TextureUnit[32];
        textureTmp = new TextureUnit[32];
    }

    /**
     * Sort the listing of nodes in the given array. Do not return until the
     * sort has been completed. If the sceneParent is the main scene graph and
     * not an offscreen texture, then that spot should be set to null in the list.
     *
     * @param data External rendering environment information
     * @param nodes List of processed nodes based on the scene they came from
     * @param numNodes Number of nodes in each scene
     * @param sceneParents Parent nodes that hold the subscene in the main scene graph
     * @param numScenes The number of valid scenes to process
     * @param buffer The buffer ID to use to output the data to when
     *   multi-threaded
     */
    public void sort(RenderEnvironmentData[] data,
                     CullOutputDetails[][] nodes,
                     int[] numNodes,
                     SceneGraphObject[][] sceneParents,
                     int numScenes,
                     int buffer)
    {
        // Buffer ID is not used currently.

        terminate = false;

        if(commandList.length < numScenes)
        {
            RenderInstructions[] tmp = new RenderInstructions[numScenes];

            System.arraycopy(commandList, 0, tmp, 0, commandList.length);

            for(int i = commandList.length; i < numScenes   ; i++)
                tmp[i] = new RenderInstructions();

            commandList = tmp;
        }

        // While doing these loops, we need to find the main scene and ensure
        // that it is the last item in the output list.
        int main_idx = 0;
        int curr_scene = 0;

        for(int i = 0; i < numScenes && !terminate; i++)
        {
            if(sceneParents[i][0] == null)
            {
                main_idx = i;
                continue;
            }

            commandList[curr_scene].pbuffer =
                (PBufferTextureSource)sceneParents[i][0];
            commandList[curr_scene].parentSource =
                (PBufferTextureSource)sceneParents[i][1];
            commandList[curr_scene].renderData = data[i];

            RenderInstructions original =
                (RenderInstructions)instructionMap.get(sceneParents[i][0]);

            if(original != null)
            {
                commandList[curr_scene].copyOf = original;
            }
            else
            {
                commandList[curr_scene].copyOf = null;
                sortSingle(nodes[i], numNodes[i], commandList[curr_scene]);

                instructionMap.put(sceneParents[i][0],
                                   commandList[curr_scene]);
            }

            curr_scene++;
        }

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        commandList[curr_scene].pbuffer = null;
        commandList[curr_scene].renderData = data[main_idx];

        sortSingle(nodes[main_idx], numNodes[main_idx], commandList[curr_scene]);

        if(terminate)
        {
            instructionMap.clear();
            return;
        }

        curr_scene++;

        if(receiver != null)
            receiver.sortedOutput(commandList, numScenes);

        instructionMap.clear();
    }


    /**
     * Register a reciever for the output of the sorter. If the value is null,
     * it will clear the currently set receiver.
     *
     * @param sgr The receiver instance to add or null
     */
    public void setSortedGeometryReceiver(SortedGeometryReceiver sgr)
    {
        receiver = sgr;
    }

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown or complete scene replacement.
     */
    public void halt()
    {
        terminate = true;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Sort a single set of nodes into the output details and place in the
     * provided RenderInstructions instance.
     *
     * @param instr Instruction instant to put the details into
     */
    private void sortSingle(CullOutputDetails[] nodes,
                            int numNodes,
                            RenderInstructions instr)
    {
        // check to see the array is big enough for the absolute maximum
        // number of items we'd need. The real number is almost always less
        // than this because we have stuff that has duplicated states etc.
        int new_size = 2 + numNodes * GUESS_NUM_COMPONENTS;
        if (new_size > req_size)
            req_size = new_size;

        if(instr.renderList.length < req_size)
        {
            RenderDetails[] tmp = new RenderDetails[req_size];

            System.arraycopy(instr.renderList, 0, tmp, 0, instr.renderList.length);

            for(int i = instr.renderList.length; i < req_size; i++)
                tmp[i] = new RenderDetails();

            instr.renderList = tmp;
            instr.renderOps = new int[req_size];
        }

        Arrays.sort(nodes, 0, numNodes, stateComparator);

        int idx = 0;

        int start = 0;
        boolean done = false;

        while(!done)
        {
            try
            {
                for(int i = start; i < numNodes && !terminate; i++)
                    idx = appendObject(nodes[i], instr, idx);

                done = true;
            }
            catch(ArrayIndexOutOfBoundsException be)
            {
                start = idx;

                realloc(instr);
            }
        }

        done = false;

        while(!done)
        {
            try
            {
                instr.numValid = cleanupObjects(instr, idx);

                done = true;
            }
            catch(ArrayIndexOutOfBoundsException be)
            {
                realloc(instr);
            }
        }
    }

    /**
     * Append a single object onto the existing list. Checks versus existing
     * state and starts and stops as required.
     *
     * @param node The node instance to look at appending
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int appendObject(CullOutputDetails node,
                             RenderInstructions instr,
                             int offset)
    {
        int idx = offset;

        idx = appendLights(node, instr, idx);
        idx = appendClipPlanes(node, instr, idx);
        idx = updateFog(node.localFog, instr, idx);

        // Finally, map the shape instances
        if(node.renderable instanceof Shape3D)
        {
            Shape3D shape = (Shape3D)node.renderable;
            Appearance app = shape.getAppearance();
            Geometry geom = shape.getGeometry();

            if(currentAppearance != app)
            {
                // If the new state is now null, but before it wasn't then
                // we need to clear all the current visual state. Otherwise
                // we have to change the state from the existing one and
                // setup the new state.
                if(app == null)
                {
                    idx = cleanupVisuals(instr, idx);
                }
                else
                {
                    idx = updateMaterial(app, instr, idx);
                    idx = updatePolyAttribs(app, instr, idx);
                    idx = updateLineAttribs(app, instr, idx);
                    idx = updatePointAttribs(app, instr, idx);
                    idx = updateBlendAttribs(app, instr, idx);
                    idx = updateTextures(app, instr, idx);
                    idx = updateShader(app, instr, idx);
                }

                currentAppearance = app;
            }

            instr.renderList[idx].renderable = geom;

            Matrix4f tx = node.transform;

            instr.renderList[idx].transform[0] = tx.m00;
            instr.renderList[idx].transform[1] = tx.m10;
            instr.renderList[idx].transform[2] = tx.m20;
            instr.renderList[idx].transform[3] = tx.m30;

            instr.renderList[idx].transform[4] = tx.m01;
            instr.renderList[idx].transform[5] = tx.m11;
            instr.renderList[idx].transform[6] = tx.m21;
            instr.renderList[idx].transform[7] = tx.m31;

            instr.renderList[idx].transform[8] = tx.m02;
            instr.renderList[idx].transform[9] = tx.m12;
            instr.renderList[idx].transform[10] = tx.m22;
            instr.renderList[idx].transform[11] = tx.m32;

            instr.renderList[idx].transform[12] = tx.m03;
            instr.renderList[idx].transform[13] = tx.m13;
            instr.renderList[idx].transform[14] = tx.m23;
            instr.renderList[idx].transform[15] = tx.m33;

            instr.renderOps[idx] = RenderOp.RENDER_GEOMETRY;
            idx++;
        }
        else if(node.renderable instanceof CustomRenderable)
        {
            // forcefully remove anything left
            idx = cleanupVisuals(instr, idx);

            instr.renderList[idx].renderable = node.renderable;
            instr.renderList[idx].instructions = node.customData;

            Matrix4f tx = node.transform;

            instr.renderList[idx].transform[0] = tx.m00;
            instr.renderList[idx].transform[1] = tx.m10;
            instr.renderList[idx].transform[2] = tx.m20;
            instr.renderList[idx].transform[3] = tx.m30;

            instr.renderList[idx].transform[4] = tx.m01;
            instr.renderList[idx].transform[5] = tx.m11;
            instr.renderList[idx].transform[6] = tx.m21;
            instr.renderList[idx].transform[7] = tx.m31;

            instr.renderList[idx].transform[8] = tx.m02;
            instr.renderList[idx].transform[9] = tx.m12;
            instr.renderList[idx].transform[10] = tx.m22;
            instr.renderList[idx].transform[11] = tx.m32;

            instr.renderList[idx].transform[12] = tx.m03;
            instr.renderList[idx].transform[13] = tx.m13;
            instr.renderList[idx].transform[14] = tx.m23;
            instr.renderList[idx].transform[15] = tx.m33;

            instr.renderOps[idx] = RenderOp.RENDER_CUSTOM;
            idx++;
        }
        else
        {
            System.out.println("Non-shape node in sorter " + node.renderable);
        }

        return idx;
    }

    /**
     * Append the lights onto the existing list. Checks versus existing
     * state and starts and stops as required.
     *
     * @param node The node instance to look at appending
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int appendLights(CullOutputDetails node,
                             RenderInstructions instr,
                             int offset)
    {
        if(node.numLights == 0 && currentLights.size() == 0)
            return offset;

        int idx = offset;

        // First drop all the lights into the queue and turn the state off on
        // the no-longer used ones.
        int new_count = 0;

        for(int i = 0; i < node.numLights; i++)
        {
            LightDetails ld = node.lights[i];

            if(currentLights.contains(ld))
                keepLights.add(ld);
            else
                newLights[new_count++] = ld;
        }

        currentLights.retainAll(keepLights, oldLights);
        keepLights.clear();

        // Unmap the old lights.
        int num_old = oldLights.size();
        if(num_old != 0)
        {
            if(num_old > lightTmp.length)
                lightTmp = new Object[num_old];

            oldLights.toArray(lightTmp);
            oldLights.clear();

            for(int i = 0; i < num_old; i++)
            {
                LightDetails ld = (LightDetails)lightTmp[i];
                Integer l_id = (Integer)lightIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getLight();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;
                lightTmp[i] = null;
            }
        }

        // Map the new lights.
        if(new_count != 0)
        {
            for(int i = 0; i < new_count; i++)
            {
                LightDetails ld = newLights[i];
                currentLights.add(ld);

                instr.renderList[idx].renderable = ld.getLight();

                System.arraycopy(ld.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                lightIdMap.put(ld, new Integer(lastGlobalId));
                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_LIGHT;
                idx++;
                newLights[i] = null;
            }
        }

        return idx;
    }

    /**
     * Append the clip planess onto the existing list. Checks versus existing
     * state and starts and stops as required.
     *
     * @param node The node instance to look at appending
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int appendClipPlanes(CullOutputDetails node,
                                 RenderInstructions instr,
                                 int offset)
    {
        if(node.numClipPlanes == 0 && currentClipPlanes.size() == 0)
            return offset;

        int idx = offset;

        // Drop all the clips into the queue and turn the state off on
        // the no-longer used ones.
        int new_count = 0;

        for(int i = 0; i < node.numClipPlanes; i++)
        {
            ClipDetails ld = node.clipPlanes[i];

            if(currentClipPlanes.contains(ld))
                keepClipPlanes.add(ld);
            else
                newClipPlanes[new_count++] = ld;
        }

        currentClipPlanes.retainAll(keepClipPlanes, oldClipPlanes);
        keepClipPlanes.clear();

        // Unmap the old clips.
        int num_old = oldClipPlanes.size();
        if(num_old != 0)
        {
            if(num_old > clipTmp.length)
                clipTmp = new Object[num_old];

            oldClipPlanes.toArray(clipTmp);
            oldClipPlanes.clear();

            for(int i = 0; i < num_old; i++)
            {
                ClipDetails ld = (ClipDetails)clipTmp[i];
                Integer l_id = (Integer)clipIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getClipPlane();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the clip stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_CLIP_PLANE;
                idx++;
                clipTmp[i] = null;
            }
        }

        // Map the new clips.
        if(new_count != 0)
        {
            for(int i = 0; i < new_count; i++)
            {
                ClipDetails ld = newClipPlanes[i];
                currentClipPlanes.add(ld);

                instr.renderList[idx].renderable = ld.getClipPlane();

                System.arraycopy(ld.getTransform(),
                                 0,
                                 instr.renderList[idx].transform,
                                 0,
                                 16);

                clipIdMap.put(ld, new Integer(lastGlobalId));
                instr.renderList[idx].id = lastGlobalId++;
                instr.renderOps[idx] = RenderOp.START_CLIP_PLANE;
                idx++;
                newClipPlanes[i] = null;
            }
        }

        return idx;
    }

    /**
     * Check on the material setup and change as necessary.
     *
     * @param fog The Fog instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateFog(Fog fog, RenderInstructions instr, int offset)
    {
        int idx = offset;

        if(fog != currentFog)
        {
            if(fog != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!fog.equals(currentFog))
                {
                    // Clear the old state if it is set
                    if(currentFog != null)
                    {
                        instr.renderList[idx].renderable = currentFog;
                        instr.renderOps[idx] = RenderOp.STOP_FOG;
                        idx++;
                    }

                    instr.renderList[idx].renderable = fog;
                    instr.renderOps[idx] = RenderOp.START_FOG;
                    idx++;
                }

                currentFog = fog;
            }
            else
            {
                // The new fogerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentFog;
                instr.renderOps[idx] = RenderOp.STOP_FOG;
                idx++;

                currentFog = null;
            }
        }

        return idx;
    }


    /**
     * Check on the shader setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateShader(Appearance app,
                             RenderInstructions instr,
                             int offset)
    {
        Shader sh = app.getShader();

        if(sh == currentShader)
            return offset;

        int idx = offset;

        if(sh instanceof GLSLangShader)
        {
            // If the old one was GL14, just replace it entirely with the
            // new one.
            if(currentShader instanceof GLSLangShader)
            {
                GLSLangShader gl_sh = (GLSLangShader)sh;
                ShaderProgram prog = gl_sh.getShaderProgram();

                if(prog != null)
                {
                    if(currentShaderProgram != prog)
                    {
                        currentShaderProgram = prog;

                        instr.renderList[idx].renderable = currentShaderProgram;
                        instr.renderOps[idx] = RenderOp.START_SHADER_PROGRAM;
                        idx++;
                    }

                    ShaderArguments args = gl_sh.getShaderArguments();
                    if(currentShaderArgs != args)
                    {
                        currentShaderArgs = args;

                        instr.renderList[idx].renderable = currentShaderArgs;
                        instr.renderOps[idx] = RenderOp.SET_SHADER_ARGS;
                        idx++;
                    }
                }
                else
                {
                    // it's null, so effectively just turn off the old
                    // shader and set the local state here to be as though
                    // no shader is currently set.
                    instr.renderList[idx].renderable = currentShaderProgram;
                    instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                    idx++;

                    currentShader = null;
                    currentShaderProgram = null;
                    currentShaderArgs = null;
                }
            }
            else if(currentShader instanceof GL14Shader)
            {
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                // Now push the new values, but only bother if there is
                // a program to actually run. If not, just ignore it and
                // move on as though there were no state set.
                GLSLangShader gl_sh = (GLSLangShader)sh;
                currentShaderProgram = gl_sh.getShaderProgram();

                if(currentShaderProgram != null)
                {
                    instr.renderList[idx].renderable = currentShaderProgram;
                    instr.renderOps[idx] = RenderOp.START_SHADER_PROGRAM;
                    idx++;

                    currentShaderArgs = gl_sh.getShaderArguments();

                    if(currentShaderArgs != null)
                    {
                        instr.renderList[idx].renderable = currentShaderArgs;
                        instr.renderOps[idx] = RenderOp.SET_SHADER_ARGS;
                        idx++;
                    }

                    currentShader = sh;
                }
            }
            else
            {
                GLSLangShader gl_sh = (GLSLangShader)sh;
                currentShaderProgram = gl_sh.getShaderProgram();

                // Only set a new state if we have something valid to
                // render.
                if(currentShaderProgram != null)
                {
                    instr.renderList[idx].renderable = currentShaderProgram;
                    instr.renderOps[idx] = RenderOp.START_SHADER_PROGRAM;
                    idx++;

                    currentShaderArgs = gl_sh.getShaderArguments();

                    instr.renderList[idx].renderable = currentShaderArgs;
                    instr.renderOps[idx] = RenderOp.SET_SHADER_ARGS;
                    idx++;

                    currentShader = sh;
                }
            }
        }
        else if(sh instanceof GL14Shader)
        {
            if(sh != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!sh.equals(currentShader))
                {
                    // Clear the old state if it is set
                    if(currentShader != null)
                    {
                        if(currentShader instanceof GLSLangShader)
                        {
                            instr.renderList[idx].renderable = currentShader;
                            instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                            idx++;

                            currentShaderProgram = null;
                            currentShaderArgs = null;
                        }
                        else
                        {
                            instr.renderList[idx].renderable = currentShader;
                            instr.renderOps[idx] = RenderOp.STOP_STATE;
                            idx++;
                        }
                    }

                    instr.renderList[idx].renderable = sh;
                    instr.renderOps[idx] = RenderOp.STOP_STATE;
                    idx++;
                }

                currentShader = sh;
            }
            else
            {
                // The new shader is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentShader = null;
            }
        }
        else if(sh != null)
        {
            // A custom shader implementation. Maybe it's Cg or something
            // customised from the extensions area such as h-anim or volume
            // rendering?
            if(!sh.equals(currentShader))
            {
                // Clear the old state if it is set
                if(currentShader != null)
                {
                    if(currentShader instanceof GLSLangShader)
                    {
                        instr.renderList[idx].renderable = currentShader;
                        instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                        idx++;

                        currentShaderProgram = null;
                        currentShaderArgs = null;
                    }
                    else
                    {
                        instr.renderList[idx].renderable = currentShader;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }
                }

                instr.renderList[idx].renderable = sh;
                instr.renderOps[idx] = RenderOp.START_STATE;
                idx++;

                currentShader = sh;
            }
        }
        else
        {
            // incoming shader is null, so clean up anything left
            if(currentShader instanceof GLSLangShader)
            {
                GLSLangShader gl_sh = (GLSLangShader)sh;
                instr.renderList[idx].renderable = currentShaderProgram;
                instr.renderOps[idx] = RenderOp.STOP_SHADER_PROGRAM;
                idx++;

                currentShaderProgram = null;
                currentShaderArgs = null;
                currentShader = null;
            }
            else if(currentShader instanceof GL14Shader)
            {
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentShader = null;
            }
            else if (currentShader != null)
            {
                instr.renderList[idx].renderable = currentShader;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;
            }
        }

        return idx;
    }

    /**
     * Check on the material setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateMaterial(Appearance app,
                               RenderInstructions instr,
                               int offset)
    {
        int idx = offset;

        Material mat = app.getMaterial();
        if(mat != currentMaterial)
        {
            if(mat != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!mat.equals(currentMaterial))
                {
                    // Clear the old state if it is set
                    if(currentMaterial != null)
                    {
                        instr.renderList[idx].renderable = currentMaterial;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = mat;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentMaterial = mat;
            }
            else
            {
                // The new material is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentMaterial;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentMaterial = null;
            }
        }

        return idx;
    }

    /**
     * Check on the polygon attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updatePolyAttribs(Appearance app,
                                  RenderInstructions instr,
                                  int offset)
    {
        int idx = offset;

        PolygonAttributes attr = app.getPolygonAttributes();
        if(attr != currentPolyAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentPolyAttr))
                {
                    // Clear the old state if it is set
                    if(currentPolyAttr != null)
                    {
                        instr.renderList[idx].renderable = currentPolyAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentPolyAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentPolyAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentPolyAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the line attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateLineAttribs(Appearance app,
                                  RenderInstructions instr,
                                  int offset)
    {
        int idx = offset;

        LineAttributes attr = app.getLineAttributes();
        if(attr != currentLineAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentLineAttr))
                {
                    // Clear the old state if it is set
                    if(currentLineAttr != null)
                    {
                        instr.renderList[idx].renderable = currentLineAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentLineAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentLineAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentLineAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the polygon attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updatePointAttribs(Appearance app,
                                  RenderInstructions instr,
                                  int offset)
    {
        int idx = offset;

        PointAttributes attr = app.getPointAttributes();
        if(attr != currentPointAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentPointAttr))
                {
                    // Clear the old state if it is set
                    if(currentPointAttr != null)
                    {
                        instr.renderList[idx].renderable = currentPointAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentPointAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentPointAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentPointAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the blending attributes setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateBlendAttribs(Appearance app,
                                   RenderInstructions instr,
                                   int offset)
    {
        int idx = offset;

        BlendAttributes attr = app.getBlendAttributes();
        if(attr != currentBlendAttr)
        {
            if(attr != null)
            {
                // Something has really changed? If so set the
                // state up to start running now.
                if(!attr.equals(currentBlendAttr))
                {
                    // Clear the old state if it is set
                    if(currentBlendAttr != null)
                    {
                        instr.renderList[idx].renderable = currentBlendAttr;
                        instr.renderOps[idx] = RenderOp.STOP_STATE;
                        idx++;
                    }

                    instr.renderList[idx].renderable = attr;
                    instr.renderOps[idx] = RenderOp.START_STATE;
                    idx++;
                }

                currentBlendAttr = attr;
            }
            else
            {
                // The new attrerial is null, so clear the existing
                // state and leave it empty.
                instr.renderList[idx].renderable = currentBlendAttr;
                instr.renderOps[idx] = RenderOp.STOP_STATE;
                idx++;

                currentBlendAttr = null;
            }
        }

        return idx;
    }

    /**
     * Check on the texture listing setup and change as necessary.
     *
     * @param app The appearance instance to source info from
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int updateTextures(Appearance app,
                               RenderInstructions instr,
                               int offset)
    {
        int idx = offset;

        int num_tex = app.numTextureUnits();

        if(textureTmp.length < num_tex)
            textureTmp = new TextureUnit[num_tex];

        app.getTextureUnits(textureTmp);

        if(num_tex != numTextures)
        {
            if(num_tex == 0)
            {
                // Clear the old textures that were set
                for(int i = 0; i < numTextures; i++)
                {
                    instr.renderList[idx].renderable = currentTextures[i];
                    instr.renderList[idx].instructions = TEX_IDS[i];
                    instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                    idx++;

                    currentTextures[i] = null;
                    textureTmp[i] = null;
                }
            }
            else if(numTextures == 0)
            {
                // There were no old textures, so just list the new ones
                for(int i = 0; i < num_tex; i++)
                {
                    instr.renderList[idx].renderable = textureTmp[i];
                    instr.renderList[idx].instructions = TEX_IDS[i];
                    instr.renderOps[idx] = RenderOp.START_TEXTURE;
                    idx++;

                    currentTextures[i] = textureTmp[i];
                    textureTmp[i] = null;
                }
            }
            else
            {
                int smaller = (numTextures < num_tex) ? numTextures : num_tex;

                for(int i = 0; i < smaller; i++)
                {
                    if(!textureTmp[i].equals(currentTextures[i]))
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                        idx++;

                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.START_TEXTURE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                        textureTmp[i] = null;
                    }
                }

                // Clean up what is left or add the new stuff
                if(numTextures < num_tex)
                {
                    for(int i = smaller; i < num_tex; i++)
                    {
                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.START_TEXTURE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                        textureTmp[i] = null;
                    }
                }
                else
                {
                    for(int i = smaller; i < numTextures; i++)
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                        idx++;

                        currentTextures[i] = null;
                        textureTmp[i] = null;
                    }
                }
            }

            numTextures = num_tex;
        }
        else if(num_tex != 0)
        {
            // same number of texture units.
            // Special case the most common situation.
            if(num_tex == 1)
            {
                if(!textureTmp[0].equals(currentTextures[0]))
                {
                    instr.renderList[idx].renderable = currentTextures[0];
                    instr.renderList[idx].instructions = TEX_IDS[0];
                    instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                    idx++;

                    instr.renderList[idx].renderable = textureTmp[0];
                    instr.renderList[idx].instructions = TEX_IDS[0];
                    instr.renderOps[idx] = RenderOp.START_TEXTURE;
                    idx++;

                    currentTextures[0] = textureTmp[0];
                    textureTmp[0] = null;
                }
            }
            else
            {
                for(int i = 0; i < numTextures; i++)
                {
                    if(!textureTmp[i].equals(currentTextures[i]))
                    {
                        instr.renderList[idx].renderable = currentTextures[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
                        idx++;

                        instr.renderList[idx].renderable = textureTmp[i];
                        instr.renderList[idx].instructions = TEX_IDS[i];
                        instr.renderOps[idx] = RenderOp.START_TEXTURE;
                        idx++;

                        currentTextures[i] = textureTmp[i];
                        textureTmp[i] = null;
                    }
                }
            }
        }

        return idx;
    }

    /**
     * At the end of the frame, we need to make sure that we've turned off all
     * the lights, clip planes and current rendering state that is left. Do
     * that now.
     *
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int cleanupObjects(RenderInstructions instr, int offset)
    {
        int idx = offset;

        int num_lights = currentLights.size();
        if(num_lights != 0)
        {
            if(num_lights > lightTmp.length)
                lightTmp = new Object[num_lights];

            currentLights.toArray(lightTmp);
            currentLights.clear();

            for(int i = 0; i < num_lights; i++)
            {
                LightDetails ld = (LightDetails)lightTmp[i];
                Integer l_id = (Integer)lightIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getLight();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the light stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_LIGHT;
                idx++;

                lightTmp[i] = null;
            }
        }

        // Now the clip planes.
        int num_clips = currentClipPlanes.size();
        if(num_clips != 0)
        {
            if(num_clips > clipTmp.length)
                clipTmp = new Object[num_clips];

            currentClipPlanes.toArray(clipTmp);
            currentClipPlanes.clear();

            for(int i = 0; i < num_clips; i++)
            {
                ClipDetails ld = (ClipDetails)clipTmp[i];
                Integer l_id = (Integer)clipIdMap.remove(ld);

                instr.renderList[idx].renderable = ld.getClipPlane();
                instr.renderList[idx].id = l_id.intValue();

                // Don't need the transform for the clip stopping. Save
                // CPU cycles by not copying it.

                instr.renderOps[idx] = RenderOp.STOP_CLIP_PLANE;
                idx++;

                clipTmp[i] = null;
            }
        }

        return cleanupVisuals(instr, idx);
    }

    /**
     * Convenience method to clear all the current visual information.
     *
     * @param offset The distance into instr array of values to start
     * @param instr Instruction instant to put the details into
     * @return The completed offset into the instr array we finished at
     */
    private int cleanupVisuals(RenderInstructions instr, int offset)
    {
        int idx = offset;

        if(currentMaterial != null)
        {
            instr.renderList[idx].renderable = currentMaterial;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentMaterial = null;
        }

        if(currentShader != null)
        {
            instr.renderList[idx].renderable = currentShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentShader = null;
        }

        if(currentVtxShader != null)
        {
            instr.renderList[idx].renderable = currentVtxShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentVtxShader = null;
        }

        if(currentFragShader != null)
        {
            instr.renderList[idx].renderable = currentFragShader;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentFragShader = null;
        }

        if(currentPolyAttr != null)
        {
            instr.renderList[idx].renderable = currentPolyAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentPolyAttr = null;
        }

        if(currentLineAttr != null)
        {
            instr.renderList[idx].renderable = currentLineAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentLineAttr = null;
        }

        if(currentPointAttr != null)
        {
            instr.renderList[idx].renderable = currentPointAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentPointAttr = null;
        }

        if(currentBlendAttr != null)
        {
            instr.renderList[idx].renderable = currentBlendAttr;
            instr.renderOps[idx] = RenderOp.STOP_STATE;
            idx++;

            currentBlendAttr = null;
        }

        for(int i = 0; i < numTextures; i++)
        {
            instr.renderList[idx].renderable = currentTextures[i];
            instr.renderList[idx].instructions = TEX_IDS[i];
            instr.renderOps[idx] = RenderOp.STOP_TEXTURE;
            idx++;

            currentTextures[i] = null;
        }

        if(currentFog != null)
        {
            instr.renderList[idx].renderable = currentFog;
            instr.renderOps[idx] = RenderOp.STOP_FOG;
            idx++;

            currentFog = null;
        }

        numTextures = 0;
        currentAppearance = null;

        return idx;
    }

    /**
     * Reallocate the renderList and renderOps arrays.
     * Grows by a minimum of REALLOC_SIZE.
     *
     * @param instr The instructionss to resize.
     */
    private void realloc(RenderInstructions instr)
    {
        if(req_size - instr.renderList.length < REALLOC_SIZE)
            req_size += REALLOC_SIZE;

        RenderDetails[] tmp = new RenderDetails[req_size];
        int[] roTmp = new int[req_size];

        System.arraycopy(instr.renderList, 0, tmp, 0, instr.renderList.length);
        System.arraycopy(instr.renderOps, 0, roTmp, 0, instr.renderOps.length);

        for(int i = instr.renderList.length; i < req_size; i++)
            tmp[i] = new RenderDetails();

        instr.renderList = tmp;
        instr.renderOps = roTmp;
    }
}
