/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.surface;

// External imports
import net.java.games.jogl.*;

import java.awt.Rectangle;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

import javax.vecmath.Vector3f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.IntHashMap;
import org.j3d.util.MatrixUtils;

/**
 * Base handling to render for a single output device - be it on-screen or off.
 *
 * @author Alan Hudson
 * @version $Revision: 1.3.2.2 $
 */
public class BaseRenderingProcessor
    implements GLEventListener, Renderer
{
    /** The initial size of the children list */
    protected static final int LIST_START_SIZE = 400;

    /** Initial depth for the number of transforms above a view,fog or bg */
    protected static final int INITIAL_PATH_DEPTH = 64;

    /** Initial size of the deletion queue array */
    protected static final int INITIAL_DELETE_SIZE = 20;

    /** The initial size of the shader init/log lists */
    protected static final int SHADERLIST_START_SIZE = 64;

    /** The invalid shader ID */
    protected static final Integer INVALID_SHADER = new Integer(-1);

    /** Message if the node is not live */
    protected static final String NOT_LIVE_VP_MSG =
        "The viewpoint given is not part of a live scene graph";

    /** Message when we discover the VP is under a shared group/node */
    protected static final String SHARED_VP_MSG =
        "The path from the root of the scene graph to the viewpoint contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    /** Message if the node is not live */
    protected static final String NOT_LIVE_FOG_MSG =
        "The Fog given is not part of a live scene graph";

    /** Message when we discover the fog is under a shared group/node */
    protected static final String SHARED_FOG_MSG =
        "The path from the root of the scene graph to the fog contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    /** Message if the node is not live */
    protected static final String NOT_LIVE_BG_MSG =
        "The background given is not part of a live scene graph";

    /** Message when we discover the background is under a shared group/node */
    protected static final String SHARED_BG_MSG =
        "The path from the root of the scene graph to the Background contains " +
        "a SharedGroup or SharedNode instance. This is not permitted.";

    /** Generic user data that is assigned to this scene */
    protected Object userData;

    /** Pre and Post-render effects processor, if set */
    protected RenderEffectsProcessor effectsProcessor;

    /** The current viewpoint instance */
    protected Viewpoint currentViewpoint;

    /** The current fog instance */
    protected Fog currentFog;

    /** Path to the current fog */
    protected TransformGroup[] currentFogPath;

    /** The current background instance */
    protected Background currentBackground;

    /** Path to the current background */
    protected TransformGroup[] currentBackgroundPath;

    /** The current clear colour */
    protected float[] clearColor;

    /** Current view environment */
    protected ViewEnvironment viewEnvironment;

    /** The default view environment if none set */
    protected ViewEnvironment defaultEnvironment;

    /** The current frustum paramaters */
    protected double[] currentFrustum;

    /** Flag to say that colour needs to be reset this frame */
    protected boolean resetColor;

    /** Local storage of the nodes that need to be rendered */
    protected RenderDetails[] renderableList;

    /** Local storage of the nodes that need to be rendered */
    protected int[] operationList;

    /** Number of items in the renderable list */
    protected int numRenderables;

    /** The current aspect ratio. Reset each time reshape() is called */
    protected double aspectRatio;

    /** Current surface width */
    protected int windowWidth;

    /** Current surface height */
    protected int windowHeight;

    /** If the view environment or one of it's parameters have changed */
    protected boolean viewEnvironmentChanged;

    /** If the viewport has changed, update */
    protected boolean viewportChanged;

    /** Queue for holding deleted textures */
    protected DeletableSceneGraphObject[] deleteQueue;

    /** The number of deletable objects to handle next frame */
    protected int numDeletions;

    /** The change requestors for data changed sets */
    protected ShaderSceneGraphObject[] shaderInitList;

    /** The change requestors for bounds changed sets */
    protected ShaderSceneGraphObject[] shaderLogList;

    /** The current place to add shader init requestors */
    protected int lastShaderInitIdx;

    /** The current place to add shader log requestors */
    protected int lastShaderLogIdx;

    /** The current program ID for GLSLang shaders.  */
    protected Integer currentShaderProgramId;

    /** Stack of the available OGL light IDs that could be assigned to lights */
    protected Integer[] availableLights;

    /** index in the availableLights list of the last available */
    protected int lastLightIdx;

    /** Mapping of the object ID to it's used light ID */
    protected IntHashMap lightIdMap;

    /** Stack of the available OGL clip IDs that could be assigned to clips */
    protected Integer[] availableClips;

    /** index in the availableClips list of the last available */
    protected int lastClipIdx;

    /** Mapping of the object ID to it's used clip ID */
    protected IntHashMap clipIdMap;

    /** Array version of the view matrix in use - direct to OpenGL */
    protected float[] viewMatrix;

    /** Matrix for the background rendering */
    protected float[] backgroundMatrix;

    /** World to local matrix (the camera) matrix */
    protected Matrix4f cameraMatrix;

    /** Combined view and projection matrix */
    protected Matrix4f surfaceMatrix;

    /** The eye offset from the normal position */
    protected float[] eyePoint;

    /** Recalculate aspect ratio on resize. */
    protected boolean recalcAspect;

    /** Matrix utility code for doing inversions */
    protected MatrixUtils matrixUtils;

    /** Flag indicating whether an explicit viewport is provided or not */
    protected boolean explicitViewport;

    /** Data describing the current viewport if explicit values used */
    protected Rectangle currentViewport;

    /** Request that the current drawing terminate immediately. App closing */
    protected boolean terminate;

    /**
     * Construct handler for rendering objects to the main screen.
     */
    BaseRenderingProcessor()
    {
        clearColor = new float[4];
        viewMatrix = new float[16];
        backgroundMatrix = new float[16];
        cameraMatrix = new Matrix4f();
        surfaceMatrix = new Matrix4f();
        currentFrustum = new double[6];
        eyePoint = new float[3];

        defaultEnvironment = new ViewEnvironment();
        viewEnvironment = defaultEnvironment;
        viewEnvironment.generateViewFrustum(currentFrustum);

        aspectRatio = defaultEnvironment.getAspectRatio();

        resetColor = false;
        terminate = false;

        numRenderables = 0;

        currentFogPath = new TransformGroup[INITIAL_PATH_DEPTH];
        currentBackgroundPath = new TransformGroup[INITIAL_PATH_DEPTH];

        lightIdMap = new IntHashMap();
        clipIdMap = new IntHashMap();

        lastLightIdx = 0;
        lastClipIdx = 0;
        lastShaderInitIdx = 0;
        lastShaderLogIdx = 0;
        currentShaderProgramId = INVALID_SHADER;

        recalcAspect = false;

        deleteQueue = new DeletableSceneGraphObject[INITIAL_DELETE_SIZE];
        shaderInitList = new ShaderSceneGraphObject[SHADERLIST_START_SIZE];
        shaderLogList = new ShaderSceneGraphObject[SHADERLIST_START_SIZE];

        matrixUtils = new MatrixUtils();
        explicitViewport = false;
        currentViewport = new Rectangle();
    }

    //---------------------------------------------------------------
    // Methods defined by GLEventListener
    //---------------------------------------------------------------

    /**
     * Called by the drawable immediately after the OpenGL context is
     * initialized; the GLContext has already been made current when
     * this method is called.
     *
     * @param drawable The display context to render to
     */
    public void init(GLDrawable drawable)
    {
//        drawable.setGL(new TraceGL(drawable.getGL(), System.out));
        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        gl.glClearColor(clearColor[0],
                        clearColor[1],
                        clearColor[2],
                        clearColor[3]);

        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glEnable(GL.GL_NORMALIZE);

        gl.glEnable(GL.GL_CULL_FACE);
        gl.glCullFace(GL.GL_BACK);

        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
        gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);
        gl.glHint(GL.GL_FOG_HINT, GL.GL_NICEST);

        updateProjectionMatrix(gl, glu);

        // Find out the number of available lights and clip planes and
        // initialise the arrays to that length.
        int[] num_id = new int[1];

        gl.glGetIntegerv(GL.GL_MAX_CLIP_PLANES, num_id);

        availableClips = new Integer[num_id[0]];
        for(int i = 0; i < num_id[0]; i++)
            availableClips[i] = new Integer(GL.GL_CLIP_PLANE0 + i);

        gl.glGetIntegerv(GL.GL_MAX_LIGHTS, num_id);

        // GL_LIGHT0 is assumed to be used by the current viewpoint
        availableLights = new Integer[num_id[0]];

        for(int i = 0; i < num_id[0]; i++)
            availableLights[i] = new Integer(GL.GL_LIGHT1 + i);
    }

    /**
     * Called by the drawable when the surface resizes itself. Used to
     * reset the viewport dimensions.
     *
     * @param drawable The display context to render to
     */
    public void reshape(GLDrawable drawable,
                        int x,
                        int y,
                        int width,
                        int height)
    {
        windowWidth = width;
        windowHeight = height;
        aspectRatio = (float)width / (float)height;

        if(recalcAspect)
        {
            viewEnvironment.setAspectRatio(aspectRatio);
            viewEnvironment.generateViewFrustum(currentFrustum);
        }

        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        if(!explicitViewport)
            gl.glViewport(0, 0, width, height);

        updateProjectionMatrix(gl, glu);
    }

    /**
     * Called by the drawable when the display mode or the display device
     * associated with the GLDrawable has changed.
     *
     * @param drawable The display context to render to
     */
    public void displayChanged(GLDrawable drawable,
                               boolean modeChanged,
                               boolean deviceChanged)
    {
// Need a way to clean up the old buffer here.
//        if(outputTexture != null)
//        {
//            outputTexture.unregisterBuffer(old_gl);
//            outputTexture.registerBuffer(gl, pBuffer);
//        }
    }

    /**
     * Called by the drawable to perform rendering by the client.
     *
     * @param drawable The display context to render to
     */
    public void display(GLDrawable drawable)
    {
        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        // Process the deleted nodes
        if(numDeletions != 0)
        {
            for(int i = 0; i < numDeletions; i++)
            {
                deleteQueue[i].cleanup(gl, glu);
                deleteQueue[i] = null;
            }

            numDeletions = 0;
        }

        // Process the shader init calls and then log requests
        if(lastShaderInitIdx != 0)
        {
            for(int i = 0; i < lastShaderInitIdx; i++)
            {
                shaderInitList[i].initialize(gl, glu);
                shaderInitList[i] = null;
            }

            lastShaderInitIdx = 0;
        }

        if(lastShaderLogIdx != 0)
        {
            for(int i = 0; i < lastShaderLogIdx; i++)
            {
                shaderLogList[i].fetchLogInfo(gl, glu);
                shaderLogList[i] = null;
            }

            lastShaderLogIdx = 0;
        }

        if(terminate)
            return;

        if(effectsProcessor != null)
            effectsProcessor.preDraw(drawable, userData);

        if(terminate)
            return;

        if(viewportChanged)
        {
            if(explicitViewport)
            {
                gl.glViewport(currentViewport.x,
                              currentViewport.y,
                              currentViewport.width,
                              currentViewport.height);
            }
            else
            {
                gl.glViewport(0, 0, windowWidth, windowHeight);
            }

            viewportChanged = false;
        }

        if(terminate)
            return;

        if(currentBackground != null)
        {
            gl.glMatrixMode(GL.GL_PROJECTION);
            gl.glPushMatrix();
            gl.glLoadIdentity();

            setupBackgroundFrustum(gl, 0.01, 1.1, viewEnvironment.getFieldOfView());

            gl.glClear(GL.GL_DEPTH_BUFFER_BIT);

            gl.glMatrixMode(GL.GL_MODELVIEW);
            gl.glPushMatrix();

            gl.glLoadMatrixf(backgroundMatrix);

            gl.glDepthMask(false);
            gl.glDisable(GL.GL_BLEND);
            gl.glDisable(GL.GL_DEPTH_TEST);

            currentBackground.render(gl, glu);
            currentBackground.postRender(gl, glu);

            gl.glDepthMask(true);
            gl.glEnable(GL.GL_DEPTH_TEST);
            gl.glEnable(GL.GL_BLEND);
            gl.glPopMatrix();


            gl.glMatrixMode(GL.GL_PROJECTION);
            gl.glPopMatrix();
            gl.glMatrixMode(GL.GL_MODELVIEW);

            renderViewpoint(gl, glu);
        }
        else
        {
            if(resetColor)
            {
                gl.glClearColor(clearColor[0],
                                clearColor[1],
                                clearColor[2],
                                clearColor[3]);
            }

            gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

            renderViewpoint(gl, glu);
        }

        if(terminate)
            return;

        boolean fog_active = false;

        if(currentFog != null)
        {
            fog_active = true;
            gl.glEnable(GL.GL_FOG);
            currentFog.render(gl, glu);
        }

        RenderableObject obj;
        RenderableComponent comp;

        for(int i = 0; i < numRenderables && !terminate; i++)
        {
            switch(operationList[i])
            {
                case RenderOp.START_RENDER:
                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_RENDER:
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);
                    gl.glPopMatrix();
                    break;

                case RenderOp.RENDER_GEOMETRY:
                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    gl.glPopMatrix();
                    break;

                case RenderOp.RENDER_CUSTOM:
                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);

                    CustomRenderable cr =
                        (CustomRenderable)renderableList[i].renderable;
                    cr.render(gl, glu, renderableList[i].instructions);
                    gl.glPopMatrix();
                    break;

                case RenderOp.START_STATE:
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_STATE:
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);
                    break;

                case RenderOp.START_LIGHT:
                    // Get the next available light ID

// TODO:
// Fix this so that if we run off the end we can still recover and not disable
// Lighting/Clipping completely for the next frame.
                    if(lastLightIdx >= availableLights.length)
                        continue;

                    Integer l_id = availableLights[lastLightIdx++];

                    lightIdMap.put(renderableList[i].id, l_id);

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, l_id);
                    gl.glPopMatrix();
                    break;

                case RenderOp.STOP_LIGHT:
                    if(lastLightIdx >= availableLights.length)
                        continue;

                    l_id = (Integer)lightIdMap.remove(renderableList[i].id);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.postRender(gl, glu, l_id);
                    availableLights[--lastLightIdx] = l_id;
                    break;

                case RenderOp.START_CLIP_PLANE:
                    // Get the next available clip plane ID

// TODO:
// Fix this so that if we run off the end we can still recover and not disable
// Lighting/Clipping completely for the next frame.
                    if(lastClipIdx >= availableClips.length)
                        continue;

                    Integer c_id = availableClips[lastClipIdx++];
                    clipIdMap.put(renderableList[i].id, c_id);

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, c_id);
                    gl.glPopMatrix();
                    break;

                case RenderOp.STOP_CLIP_PLANE:
                    if(lastClipIdx >= availableClips.length)
                        continue;

                    c_id = (Integer)clipIdMap.remove(renderableList[i].id);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.postRender(gl, glu, c_id);

                    availableClips[--lastClipIdx] = c_id;
                    break;

                case RenderOp.START_TRANSPARENT:
                    gl.glDepthMask(false);
                    gl.glEnable(GL.GL_BLEND);
                    gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                    break;

                case RenderOp.STOP_TRANSPARENT:
                    gl.glDisable(GL.GL_BLEND);
                    gl.glDepthMask(true);
                    break;

                case RenderOp.START_FOG:
                    if(!fog_active)
                    {
                        gl.glEnable(GL.GL_FOG);
                        fog_active = true;
                    }

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_FOG:
                    if(currentFog != null)
                        currentFog.render(gl, glu);
                    else
                    {
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.postRender(gl, glu);
                        fog_active = false;
                        gl.glDisable(GL.GL_FOG);
                    }
                    break;

                case RenderOp.START_SHADER_PROGRAM:
                    ShaderProgram prog =
                        (ShaderProgram)renderableList[i].renderable;

                    if(!prog.isValid(gl))
                    {
                        currentShaderProgramId = INVALID_SHADER;
                        continue;
                    }

// TODO: Optimise this to avoid the allocation. Use IntHashMap for lookup.
                    currentShaderProgramId = new Integer(prog.getProgramId(gl));
                    prog.render(gl, glu);
                    break;

                case RenderOp.STOP_SHADER_PROGRAM:
                    if(currentShaderProgramId == INVALID_SHADER)
                        continue;

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);

                    currentShaderProgramId = INVALID_SHADER;
                    break;

                case RenderOp.SET_SHADER_ARGS:
                    if(currentShaderProgramId == INVALID_SHADER)
                        continue;

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, currentShaderProgramId);
                    break;

                case RenderOp.START_TEXTURE:
                    comp = (RenderableComponent)renderableList[i].renderable;
                    Integer id = (Integer)(renderableList[i].instructions);
                    comp.render(gl, glu, id);
                    break;

                case RenderOp.STOP_TEXTURE:
                    comp = (RenderableComponent)renderableList[i].renderable;
                    id = (Integer)(renderableList[i].instructions);
                    comp.postRender(gl, glu, id);
                    break;
            }
        }

        if(terminate)
            return;

        if(currentFog != null)
        {
            currentFog.postRender(gl, glu);
            gl.glDisable(GL.GL_FOG);
        }

        if(terminate)
            return;

        if(effectsProcessor != null)
            effectsProcessor.postDraw(drawable, userData);

        if(terminate)
            return;

        if(currentViewpoint != null)
            currentViewpoint.postRender(gl, glu);

        if(terminate)
            return;

        gl.glFlush();
    }

    //---------------------------------------------------------------
    // Methods defined by Renderer
    //---------------------------------------------------------------

    /**
     * Force a halt of the current processing. Any processing in progress
     * should exit immediately. Used to abort the current scene processing due
     * to application shutdown.
     */
    public void halt()
    {
        terminate = true;
    }

    /**
     * An object has been detected as being deleted (no internal or
     * external references to it left) and it needs to be cleaned up. The
     * next oppourtunity this drawable and any sub-drawables get, this object
     * should have it's <code>cleanup()</code> method called and then the
     * reference to it should be discarded.
     *
     * @param obj The instance to be queued up as no longer being used
     */
    public void queueDeletedObject(DeletableSceneGraphObject obj)
    {
        if(deleteQueue.length == numDeletions)
        {
            int size = deleteQueue.length;
            DeletableSceneGraphObject[] tmp = new
                DeletableSceneGraphObject[size << 1];

            System.arraycopy(deleteQueue, 0, tmp, 0, size);
            deleteQueue = tmp;
        }

        deleteQueue[numDeletions++] = obj;
    }

    /**
     * Queue up a collection of shader objects for processing on the next
     * frame. These processing requests are for either initialisation or log
     * handling, not for rendering.
     *
     * @param initList The shaders needing initialisation
     * @param numInit The number of shaders needing initialisation
     * @param logList The shaders needing log fetching
     * @param numLog The number of shaders needing log fetching
     */
    public void queueShaderRequests(ShaderSceneGraphObject[] initList,
                                    int numInit,
                                    ShaderSceneGraphObject[] logList,
                                    int numLog)
    {
        if(numInit > shaderInitList.length)
            shaderInitList = new ShaderSceneGraphObject[numInit];

        if(numLog > shaderLogList.length)
            shaderLogList = new ShaderSceneGraphObject[numLog];

        System.arraycopy(initList, 0, shaderInitList, 0, numInit);
        System.arraycopy(logList, 0, shaderLogList, 0, numLog);

        lastShaderInitIdx = numInit;
        lastShaderLogIdx = numLog;
    }

    /**
     * Convenience method to assign up all the environment data. Calls the
     * setActiveX() methods in this class.
     *
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setEnvironmentData(RenderEnvironmentData data)
        throws IllegalArgumentException
    {
        setActiveView(data.viewpoint);
        setActiveBackground(data.background);
        setActiveFog(data.fog);

        if(data.viewEnvironment == null ||
           data.viewEnvironment.getAspectRatio() <= 0)
        {
            recalcAspect = true;
        }

        setViewEnvironment(data.viewEnvironment);
        viewEnvironmentChanged = true;

        if((data.hasViewport != explicitViewport) ||
           (data.hasViewport && !data.viewport.equals(currentViewport)))
        {
            viewportChanged = true;
            explicitViewport = data.hasViewport;
            currentViewport.setBounds(data.viewport);
        }

        setRenderEffects(data.effectsProcessor, data.userData);

        currentFrustum[0] = data.viewFrustum[0];
        currentFrustum[1] = data.viewFrustum[1];
        currentFrustum[2] = data.viewFrustum[2];
        currentFrustum[3] = data.viewFrustum[3];
        currentFrustum[4] = data.viewFrustum[4];
        currentFrustum[5] = data.viewFrustum[5];

        eyePoint[0] = data.eyeOffset[0];
        eyePoint[1] = data.eyeOffset[1];
        eyePoint[2] = data.eyeOffset[2];

        matrixUtils.inverse(data.viewTransform, cameraMatrix);
//System.out.println("VT: " + data.viewTransform);
        // Setup the matrix for the rendering. Transpose in place.
        viewMatrix[0] = cameraMatrix.m00;
        viewMatrix[1] = cameraMatrix.m10;
        viewMatrix[2] = cameraMatrix.m20;
        viewMatrix[3] = cameraMatrix.m30;
        viewMatrix[4] = cameraMatrix.m01;
        viewMatrix[5] = cameraMatrix.m11;
        viewMatrix[6] = cameraMatrix.m21;
        viewMatrix[7] = cameraMatrix.m31;
        viewMatrix[8] = cameraMatrix.m02;
        viewMatrix[9] = cameraMatrix.m12;
        viewMatrix[10] = cameraMatrix.m22;
        viewMatrix[11] = cameraMatrix.m32;
        viewMatrix[12] = cameraMatrix.m03;
        viewMatrix[13] = cameraMatrix.m13;
        viewMatrix[14] = cameraMatrix.m23;
        viewMatrix[15] = cameraMatrix.m33;

        Matrix4f bg_mat = data.backgroundTransform;

        bg_mat.mul(data.viewTransform);

        backgroundMatrix[0] = bg_mat.m00;
        backgroundMatrix[1] = bg_mat.m01;
        backgroundMatrix[2] = bg_mat.m02;
        backgroundMatrix[3] = 0;
        backgroundMatrix[4] = bg_mat.m10;
        backgroundMatrix[5] = bg_mat.m11;
        backgroundMatrix[6] = bg_mat.m12;
        backgroundMatrix[7] = 0;
        backgroundMatrix[8] = bg_mat.m20;
        backgroundMatrix[9] = bg_mat.m21;
        backgroundMatrix[10] = bg_mat.m22;
        backgroundMatrix[11] = 0;
        backgroundMatrix[12] = bg_mat.m30;
        backgroundMatrix[13] = bg_mat.m31;
        backgroundMatrix[14] = bg_mat.m32;
        backgroundMatrix[15] = bg_mat.m33;


        // Reset the matrix with the un-inverted form
        surfaceMatrix.set(data.viewTransform);
    }

    /**
     * Set the viewpoint path that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param vp The instance of the active viewpoint to use
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveView(Viewpoint vp)
        throws IllegalArgumentException
    {
        if(vp == null)
        {
            currentViewpoint = vp;
            return;
        }

        if(!vp.isLive())
            throw new IllegalArgumentException(NOT_LIVE_VP_MSG);

        currentViewpoint = vp;
    }

    /**
     * Set the fog that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param fog The instance of the active fog node
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveFog(Fog fog)
        throws IllegalArgumentException
    {
        if(fog == null)
        {
            currentFog = fog;
            return;
        }

        if(!fog.isLive())
            throw new IllegalArgumentException(NOT_LIVE_FOG_MSG);

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = fog.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if((parent instanceof SharedGroup) ||
               (parent instanceof SharedNode))
                throw new IllegalArgumentException(SHARED_FOG_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        currentFog = fog;

        if(currentFogPath.length < path_depth)
            currentFogPath = new TransformGroup[path_depth];
    }

    /**
     * Set the background path that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param bg The instance of the active background
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveBackground(Background bg)
        throws IllegalArgumentException
    {
        if(bg == null)
        {
            currentBackground = bg;
            return;
        }

        if(!bg.isLive())
            throw new IllegalArgumentException(NOT_LIVE_BG_MSG);

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = bg.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if((parent instanceof SharedGroup) ||
               (parent instanceof SharedNode))
                throw new IllegalArgumentException(SHARED_BG_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        currentBackground = bg;

        if(currentBackgroundPath.length < path_depth)
            currentBackgroundPath = new TransformGroup[path_depth];
    }

    /**
     * Set the view environment that is used to render this surface.
     *
     * @param env The environment instance to use for the render setup
     */
    public void setViewEnvironment(ViewEnvironment env)
    {
        viewEnvironment = env;

        if(env == null)
            viewEnvironment = defaultEnvironment;
        else
        {
            if(viewEnvironment.getAspectRatio() <= 0)
            {
                recalcAspect = true;
                viewEnvironment.setAspectRatio(aspectRatio);
            }
        }
    }

    /**
     * Set the rendering effects processor and the associated user data.  The
     * data object will be passed back to the client code through the callback
     * method when the pre/post render calls are done.
     */
    public void setRenderEffects(RenderEffectsProcessor effects, Object data)
    {
        effectsProcessor = effects;
        userData = data;
    }

    /**
     * Set the background colour that this surface should be cleared to before
     * the drawing step. Colours range from 0 to 1 in the normal manner.
     *
     * @param r The red component of the background clear colour
     * @param g The green component of the background clear colour
     * @param b The blue component of the background clear colour
     * @param a The alpha component of the background clear colour
     */
    public void setClearColor(float r, float g, float b, float a)
    {
        clearColor[0] = r;
        clearColor[1] = g;
        clearColor[2] = b;
        clearColor[3] = a;

        resetColor = true;
    }

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param nodes The list of nodes in sorted order
     * @param renderOps Operation to perform on each node
     * @param numValid The number of valid items in the array
     */
    public void setDrawableObjects(RenderDetails[] nodes,
                                   int[] renderOps,
                                   int numValid)
    {
        renderableList = nodes;
        operationList = renderOps;

        numRenderables = numValid;
    }

    /**
     * Get the surface to VWorld transformation matrix.
     *
     * @param matrix The matrix to copy into  It must be preallocated.
     */
    public void getSurfaceToVWorld(Matrix4f matrix)
    {
        matrix.set(surfaceMatrix);
    }

    /**
     * Convert a pixel location to surface coordinates.
     *
     * @param x The X coordinate
     * @param y The Y coordinate
     * @param position The converted position.  It must be preallocated.
     */
    public void getPixelLocationInSurface(float x, float y, Point3f position)
    {
        if(windowWidth == 0)
            return;

        position.x = (float)((currentFrustum[1] - currentFrustum[0]) *
                     (x / windowWidth - 0.5f));
        position.y = (float) ((currentFrustum[3] - currentFrustum[2]) *
                     ((windowHeight - y) / windowHeight - 0.5f));
        position.z = (float) -currentFrustum[4];
    }

    /**
     * Get the Center Eye position in surface coordinates.
     *
     * @param position The current eye position.  It must be preallocated.
     */
    public void getCenterEyeInSurface(Point3f position)
    {
        position.set(eyePoint);
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Render the viewpoint setup.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    protected void renderViewpoint(GL gl, GLU glu)
    {
        // Will want to check here for the VP being non-live.
        if(currentViewpoint == null)
            return;

        if(viewEnvironmentChanged)
            updateProjectionMatrix(gl, glu);

        gl.glLoadIdentity();

        currentViewpoint.render(gl, glu);
        gl.glMultMatrixf(viewMatrix);
    }

    /**
     * Update the projection matrix.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    protected void updateProjectionMatrix(GL gl, GLU glu)
    {
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();

        switch(viewEnvironment.getProjectionType())
        {
            case ViewEnvironment.PERSPECTIVE_PROJECTION:
                 gl.glFrustum(currentFrustum[0],currentFrustum[1],
                              currentFrustum[2],currentFrustum[3],
                              currentFrustum[4],currentFrustum[5]);
                break;

            case ViewEnvironment.ORTHOGRAPHIC_PROJECTION:
                double near_clip = viewEnvironment.getNearClipDistance();
                double far_clip = viewEnvironment.getFarClipDistance();

                gl.glOrtho(0,
                           windowWidth,
                           0,
                           windowHeight,
                           near_clip,
                           far_clip);
                break;

            default:
                System.out.println("unknown projection type");
        }
        viewEnvironmentChanged = false;

        gl.glMatrixMode(GL.GL_MODELVIEW);
    }

    /**
     * From the current view setup of FoV, near and far clipping distances and
     * the aspectRatio ratio, generate the 6 parameters that describe a view
     * frustum. These parameters are what could be used as arguments to the
     * glFrustum call. The parameter order is:
     * [x min, x max, y min, y max, z near, z far]
     *
     * @param gl The OpenGL graphics context to use
     * @param nearClip The distance to the near clip plane
     * @param farClip The distance to the far clip plane
     * @param fov The field of view to use
     */
    protected void setupBackgroundFrustum(GL gl, double nearClip, double farClip, double fov)
    {
        double xmin = 0;
        double xmax = 0;
        double ymin = 0;
        double ymax = 0;

        ymax = nearClip * Math.tan(fov * Math.PI / 360.0);
        ymin = -ymax;

        if(aspectRatio <= 0)
        {
            System.out.println("Negative aspect ratio in background calc");
        }
        else
        {
            xmin = ymin * aspectRatio;
            xmax = ymax * aspectRatio;
        }


        gl.glFrustum(xmin,xmax,ymin,ymax,nearClip,farClip);
    }
}
