/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.surface;

// External imports
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import net.java.games.jogl.*;

import java.util.HashMap;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.IntHashMap;

/**
 * Implementation of drawable surface with the key mapping defined to allow the
 * GL trace to be debugged for a single frame.
 * <p>
 *
 * To dump a set of GL trace, the 'd' key is mapped to dump the next frame.
 * <p>
 * This implementation of DrawableSurface renders to a normal GLCanvas instance
 * and provides pBuffer support as needed. Stereo support is not provided and
 * all associated methods always indicate negative returns on query about
 * support.
 *
 * @author Justin Couch
 * @version $Revision: 1.12.2.1 $
 */
public class DebugAWTSurface
    implements DrawableSurface, GLEventListener, KeyListener
{
    /** The initial size of the children list */
    private static final int LIST_START_SIZE = 20;

    /** The real canvas that we draw to */
    private GLCanvas canvas;

    /** Rendering handler devoted to the main canvas */
    private DebugRenderingProcessor canvasRenderer;

    /** Local storage of the nodes that need to be rendered */
    private GLPbuffer[] renderableList;

    /** Number of items in the renderable list */
    private int numRenderables;

    /** Flag to say that init has been called on the canvas */
    private boolean initComplete;

    /** List of extensions we need to check on at startup */
    private String[] extensionList;

    /** Number of extension types to check for */
    private int numExtensions;

    /** Maps from a pBuffer instance to the rendererProcessor for it */
    private HashMap rendererMap;

    /** Maps an offscreen texture to it's GLDrawable representation */
    private HashMap glDrawableMap;

    /** Used to fetch clear colour values when setting up pBuffers */
    private float[] colourTmp;

    /** Request that the current drawing terminate immediately. App closing */
    private boolean terminate;

    /**
     * Construct a surface that requires the given set of capabilities.
     *
     * @param caps A set of required capabilities for this canvas.
     */
    public DebugAWTSurface(GLCapabilities caps)
    {
        colourTmp = new float[4];

        // Set up handler for rendering. We add a listener for this
        // class because we need to check if any extensions are currently
        // aviable when it first starts up. After the init method has been
        // called, we unregister this class and let the rendering processor
        // deal with it.
        canvasRenderer = new DebugRenderingProcessor();
        GLDrawableFactory fac = GLDrawableFactory.getFactory();
        canvas = fac.createGLCanvas(caps);
        canvas.setAutoSwapBufferMode(false);
        canvas.setNoAutoRedrawMode(true);
        canvas.addKeyListener(this);

        // Set the canvas first then, "this" so that we don't run afoul of
        // the problems with JOGL Issue #31.
        // https://jogl.dev.java.net/issues/show_bug.cgi?id=31
        canvas.addGLEventListener(canvasRenderer);
        canvas.addGLEventListener(this);

        initComplete = false;
        terminate = false;

        renderableList = new GLPbuffer[LIST_START_SIZE];
        numRenderables = 0;

        numExtensions = 0;

        rendererMap = new HashMap();
        glDrawableMap = new HashMap();
   }

    //------------------------------------------------------------------------
    // Methods for KeyListener events
    //------------------------------------------------------------------------

    /**
     * Notification of a key press event. When the 'd' key is pressed, dump
     * the next frame to stdout.
     *
     * @param evt The key event that caused this method to be called
     */
    public void keyPressed(KeyEvent evt)
    {
        if(evt.getKeyCode() == KeyEvent.VK_D)
        {
            for(int i = 0; i < numRenderables; i++)
            {
                if(renderableList[i] == null)
                    continue;

                DebugRenderingProcessor rp =
                    (DebugRenderingProcessor)rendererMap.get(renderableList[i]);
                rp.traceNextFrame();
            }

            canvasRenderer.traceNextFrame();
        }
    }

    /**
     * Notification of a key release event. Does nothing for this
     * implementation.
     *
     * @param evt The key event that caused this method to be called
     */
    public void keyReleased(KeyEvent evt)
    {
    }

    /**
     * Notification of a key type (press and release) event. This will any one
     * of the key value fields depending on the value.
     *
     * @param evt The key event that caused this method to be called
     */
    public void keyTyped(KeyEvent evt)
    {
    }

    //---------------------------------------------------------------
    // Methods defined by DrawableSurface
    //---------------------------------------------------------------

    /**
     * Check to see whether this surface supports stereo rendering - which is
     * does not. Always returns false.
     *
     * @return false Stereo is not available
     */
    public boolean isStereoAvailable()
    {
        return false;
    }

    /**
     * Check to see whether this surface supports Quad buffer stereo rendering
     * - which it does not. Always returns false for this implementation.
     *
     * @return false The surface does not support stereo at all
     */
    public boolean isQuadStereoAvailable()
    {
        return false;
    }

    /**
     * Set the eye separation value when rendering stereo. The default value is
     * 0.33 for most applications. The absolute value of the separation is
     * always used. Ignored for this implementation.
     *
     * @param sep The amount of eye separation
     */
    public void setStereoEyeSeparation(float sep)
    {
    }

    /**
     * Get the current eye separation value - always returns 0.
     *
     * @return sep The amount of eye separation
     */
    public float getStereoEyeSeparation()
    {
        return 0;
    }

    /**
     * Set the rendering policy used when handling stereo. The policy must be
     * one of the _STEREO constants defined in this interface.
     *
     * @param policy The policy to currently use
     * @throws IllegalArgumentException The policy type is not one of the legal
     *    selections.
     */
    public void setStereoRenderingPolicy(int policy)
    {
    }

    /**
     * Get the current stereo rendering policy in use. If not explicitly set by
     * the user, then it will default to <code>NO_STEREO</code>.
     *
     * @return One of the *_STEREO values
     */
    public int getStereoRenderingPolicy()
    {
        return NO_STEREO;
    }

    /**
     * An object has been detected as being deleted (no internal or
     * external references to it left) and it needs to be cleaned up. The
     * next oppourtunity this drawable and any sub-drawables get, this object
     * should have it's <code>cleanup()</code> method called and then the
     * reference to it should be discarded.
     *
     * @param obj The instance to be queued for deletion
     */
    public void queueDeletedObject(DeletableSceneGraphObject obj)
    {
        canvasRenderer.queueDeletedObject(obj);

        // Take local reference in case the setDrawableObjects decides to
        // update values right now.
        int count = numRenderables;
        GLPbuffer[] surfaces = renderableList;

        if(count > 1)
        {
            for(int i = 0; i < count; i++)
            {
                if(surfaces[i] != null)
                {
                    DebugRenderingProcessor rp =
                        (DebugRenderingProcessor)rendererMap.get(surfaces[i]);
                    rp.queueDeletedObject(obj);
                }
            }
        }
    }

    /**
     * Queue up a collection of shader objects for processing on the next
     * frame. These processing requests are for either initialisation or log
     * handling, not for rendering.
     *
     * @param initList The shaders needing initialisation
     * @param numInit The number of shaders needing initialisation
     * @param logList The shaders needing log fetching
     * @param numLog The number of shaders needing log fetching
     */
    public void queueShaderRequests(ShaderSceneGraphObject[] initList,
                                    int numInit,
                                    ShaderSceneGraphObject[] logList,
                                    int numLog)
    {
        canvasRenderer.queueShaderRequests(initList, numInit, logList, numLog);

        // Take local reference in case the setDrawableObjects decides to
        // update values right now.
        int count = numRenderables;
        GLPbuffer[] surfaces = renderableList;

        if(count > 1)
        {
            for(int i = 0; i < count; i++)
            {
                if(surfaces[i] != null)
                {
                    DebugRenderingProcessor rp =
                        (DebugRenderingProcessor)rendererMap.get(surfaces[i]);
                    rp.queueShaderRequests(initList, numInit, logList, numLog);
                }
            }
        }
    }

    /**
     * Set the background colour that this surface should be cleared to before
     * the drawing step. Colours range from 0 to 1 in the normal manner.
     *
     * @param r The red component of the background clear colour
     * @param g The green component of the background clear colour
     * @param b The blue component of the background clear colour
     * @param a The alpha component of the background clear colour
     */
    public void setClearColor(float r, float g, float b, float a)
    {
        canvasRenderer.setClearColor(r, g, b, a);
    }

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param commands The list of drawable surfaces to render
     * @param numValid The number of valid items in the array
     */
    public void setDrawableObjects(RenderInstructions[] commands,
                                   int numValid)
    {
        if(renderableList.length < numValid)
            renderableList = new GLPbuffer[numValid];
        else
        {
            // Null out what is left of the array of nodes for GC
            for(int i = numValid; i < renderableList.length; i++)
                renderableList[i] = null;
        }

        numRenderables = numValid;

        for(int i = numRenderables; --i >= 0 ; )
        {
            RenderInstructions ri = commands[i];

            // Is this the main canvas or one of the offscreen drawables?
            if(ri.pbuffer != null)
            {
                PBufferTextureSource tex = ri.pbuffer;
                GLPbuffer buf = (GLPbuffer)glDrawableMap.get(tex);

                DebugRenderingProcessor rp = null;

                if(buf == null)
                {
                    // If the parent scene is null, then the parent is the
                    // main canvas. Create the offscreen buffer from the parent
                    // canvas first.
                    GLDrawable parent_ctx;

                    if(ri.parentSource == null)
                        parent_ctx = canvas;
                    else
                        parent_ctx = (GLDrawable)glDrawableMap.get(ri.parentSource);

                    if(!parent_ctx.canCreateOffscreenDrawable())
                        continue;

                    GLCapabilities caps = tex.getGLSetup();
                    int width = tex.getWidth();
                    int height = tex.getHeight();

                    rp = new DebugRenderingProcessor();

                    buf = parent_ctx.createOffscreenDrawable(caps,
                                                             width,
                                                             height);
                    buf.addGLEventListener(rp);

                    tex.registerBuffer(parent_ctx.getGL(), buf);
                    rendererMap.put(buf, rp);
                    glDrawableMap.put(tex, buf);
                }
                else
                {
                    rp = (DebugRenderingProcessor)rendererMap.get(buf);
                }

                renderableList[i] = buf;

                // do we need to go find the original data specification
                // or use the local values?
                if(ri.copyOf != null)
                {
                    ri = ri.copyOf;
                }

                tex.getClearColor(colourTmp);
                rp.setClearColor(colourTmp[0],
                                 colourTmp[1],
                                 colourTmp[2],
                                 colourTmp[3]);

                rp.setEnvironmentData(ri.renderData);
                rp.setDrawableObjects(ri.renderList,
                                      ri.renderOps,
                                      ri.numValid);
            }
            else
            {
                canvasRenderer.setEnvironmentData(ri.renderData);
                canvasRenderer.setDrawableObjects(ri.renderList,
                                                  ri.renderOps,
                                                  ri.numValid);

                renderableList[i] = null;
            }
        }
    }

    /**
     * Instruct the surface to draw the collected set of nodes now. The
     * registered view environment is used to draw to this surface. If no
     * view is registered, the surface is cleared and then this call is
     * exited. The drawing surface does not swap the buffers at this point.
     */
    public void draw()
    {
        // tell the draw lock that it's ok to run now, so long as it's not called
        // before the canvas has completed initialisation.
        if(!initComplete)
            return;

        // Take local reference in case the setDrawableObjects decides to
        // update values right now.
        int count = numRenderables;
        GLPbuffer[] surfaces = renderableList;

        if(count > 1)
        {
            for(int i = 0; i < count; i++)
            {
                if(surfaces[i] != null)
                    surfaces[i].display();
            }
        }

        // Always render the main canvas last.
        canvas.display();
    }

    /**
     * Swap the buffers now if the surface supports multiple buffer drawing.
     * For surfaces that don't support multiple buffers, this does nothing.
     */
    public void swap()
    {
        canvas.swapBuffers();
    }

    /**
     * Get the underlying object that this surface is rendered to. If it is a
     * screen display device, the surface can be one of AWT Component or
     * Swing JComponent. An off-screen buffer would be a form of AWT Image etc.
     *
     * @return The drawable surface representation
     */
    public Object getSurfaceObject()
    {
        return canvas;
    }

    /**
     * Get the surface to VWorld transformation matrix.
     *
     * @param matrix The matrix instance to copy into
     */
    public void getSurfaceToVWorld(Matrix4f matrix)
    {
        canvasRenderer.getSurfaceToVWorld(matrix);
    }

    /**
     * Convert a pixel location to surface coordinates.
     *
     * @param x The X coordinate
     * @param y The Y coordinate
     * @param position The converted position.  It must be preallocated.
     */
    public void getPixelLocationInSurface(float x, float y, Point3f position)
    {
        canvasRenderer.getPixelLocationInSurface(x,y,position);
    }

    /**
     * Get the Center Eye position in surface coordinates.
     *
     * @param position The current eye position.  It must be preallocated.
     */
    public void getCenterEyeInSurface(Point3f position)
    {
        canvasRenderer.getCenterEyeInSurface(position);
    }

    /**
     * Instruct this surface that you have finished with the resources needed
     * and to dispose all rendering resources.
     */
    public void dispose()
    {
        canvasRenderer.halt();

        for(int i = 0; i < numRenderables; i++)
        {
            if(renderableList[i] == null)
                continue;

            DebugRenderingProcessor rp =
                (DebugRenderingProcessor)rendererMap.get(renderableList[i]);

            rp.halt();
        }

        canvas.removeGLEventListener(canvasRenderer);
    }

    /**
     * Tell the surface to use the given thread as the rendering control thread.
     * Used when you want to clock the renderer to the rest of the pipeline,
     * rather than letting it sit in it's own loop. Due to
     * bugs in JOGL, we have to start the internal rendering threads after the
     * containing window has become visible. See JOGL issue #54 for more details
     * on this bug.
     *
     * @param th The render control thread instance to use or null
     */
    public void setRenderControlThread(Thread th)
    {
        if(canvas.getRenderingThread() == th)
            return;

        canvas.setRenderingThread(th);
    }

    //---------------------------------------------------------------
    // Methods defined by GLEventListener
    //---------------------------------------------------------------

    /**
     * Called by the drawable immediately after the OpenGL context is
     * initialized; the GLContext has already been made current when
     * this method is called.
     *
     * @param drawable The display context to render to
     */
    public void init(GLDrawable drawable)
    {
        GL gl = drawable.getGL();

        // Check for extensions:
        for(int i = 0; i < numExtensions; i++)
        {
            if(!gl.isExtensionAvailable(extensionList[i]))
                System.out.println("Extension " + extensionList[i] +
                                   " not available");
        }

        for(int i = 0; i < numRenderables; i++)
        {
            if(renderableList[i] != null)
                renderableList[i].display();
        }

        canvas.removeGLEventListener(this);
        initComplete = true;
    }

    /**
     * Not used in this implementation
     *
     * @param drawable The display context to render to
     */
    public void reshape(GLDrawable drawable,
                        int x,
                        int y,
                        int width,
                        int height)
    {
    }

    /**
     * Not used in this implementation
     *
     * @param drawable The display context to render to
     */
    public void displayChanged(GLDrawable drawable,
                               boolean modeChanged,
                               boolean deviceChanged)
    {
    }

    /**
     * Not used in this implementation
     *
     * @param drawable The display context to render to
     */
    public void display(GLDrawable drawable)
    {
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

   /**
     * Add an extension string to check for at startup.
     *
     * @param glExtensionString String to check for being valid
     */
    public void checkForExtension(String glExtensionString)
    {
        if(extensionList == null || extensionList.length == numExtensions)
        {
            String[] tmp = new String[numExtensions + 5];

            if(extensionList != null)
                System.arraycopy(extensionList, 0, tmp, 0, numExtensions);
            extensionList = tmp;
        }

        extensionList[numExtensions++] = glExtensionString;
    }
}
