/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.surface;

// External imports
import net.java.games.jogl.*;

import java.awt.Rectangle;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

import javax.vecmath.Vector3f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.IntHashMap;
import org.j3d.util.MatrixUtils;

/**
 * Handles the rendering for a single output device - be it on-screen or off.
 * <p>
 * The code expects that everything is set up before each call of the display()
 * callback. It does not handle any recursive rendering requests as that is
 * assumed to have been sorted out before calling this renderer.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.10.2.2 $
 */
class DebugRenderingProcessor extends BaseRenderingProcessor
    implements GLEventListener, Renderer
{
    /** Reusable debug GL context */
    private TraceGL debugGL;

    /** Trigger to dump the next frame using the TraceGL class */
    private boolean dumpNextFrame;

    /**
     * Construct handler for rendering objects to the main screen.
     */
    DebugRenderingProcessor()
    {
        super();

        dumpNextFrame = false;
    }

    //---------------------------------------------------------------
    // Methods defined by GLEventListener
    //---------------------------------------------------------------

    /**
     * Called by the drawable to perform rendering by the client.
     *
     * @param drawable The display context to render to
     */
    public void display(GLDrawable drawable)
    {
        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        // Process the deleted nodes
        if(numDeletions != 0)
        {
            for(int i = 0; i < numDeletions; i++)
            {
                deleteQueue[i].cleanup(gl, glu);
                deleteQueue[i] = null;
            }

            numDeletions = 0;
        }

        // Process the shader init calls and then log requests
        if(lastShaderInitIdx != 0)
        {
            for(int i = 0; i < lastShaderInitIdx; i++)
            {
                shaderInitList[i].initialize(gl, glu);
                shaderInitList[i] = null;
            }

            lastShaderInitIdx = 0;
        }

        if(lastShaderLogIdx != 0)
        {
            for(int i = 0; i < lastShaderLogIdx; i++)
            {
                shaderLogList[i].fetchLogInfo(gl, glu);
                shaderLogList[i] = null;
            }

            lastShaderLogIdx = 0;
        }

        if(terminate)
            return;


        if(effectsProcessor != null)
            effectsProcessor.preDraw(drawable, userData);

        if(terminate)
            return;

        if(dumpNextFrame)
        {
            debugGL = new TraceGL(gl, System.out);

            gl = debugGL;
            // This is going to cause fun with the object deletion queue as it
            // will result in a heap more IDs that will never get cleaned up.
            // Ignore for now as it may not be an issue. The only time this
            // should be called is during debugging, which means the extra
            // wasted resources shouldn't be a problem, I hope!
        }

        if(terminate)
            return;

        if(viewportChanged)
        {
            if(explicitViewport)
            {
                gl.glViewport(currentViewport.x,
                              currentViewport.y,
                              currentViewport.width,
                              currentViewport.height);
            }
            else
            {
                gl.glViewport(0, 0, windowWidth, windowHeight);
            }

            viewportChanged = false;
        }

        if(terminate)
            return;

        if(currentBackground != null)
        {
            gl.glMatrixMode(GL.GL_PROJECTION);
            gl.glPushMatrix();
            gl.glLoadIdentity();

            setupBackgroundFrustum(gl, 0.01, 1.1, viewEnvironment.getFieldOfView());

            gl.glClear(GL.GL_DEPTH_BUFFER_BIT);

            gl.glMatrixMode(GL.GL_MODELVIEW);
            gl.glPushMatrix();

            gl.glLoadMatrixf(backgroundMatrix);

            gl.glDepthMask(false);
            gl.glDisable(GL.GL_BLEND);
            gl.glDisable(GL.GL_DEPTH_TEST);

            currentBackground.render(gl, glu);
            currentBackground.postRender(gl, glu);

            gl.glDepthMask(true);
            gl.glEnable(GL.GL_DEPTH_TEST);
            gl.glEnable(GL.GL_BLEND);
            gl.glPopMatrix();


            gl.glMatrixMode(GL.GL_PROJECTION);
            gl.glPopMatrix();
            gl.glMatrixMode(GL.GL_MODELVIEW);

            renderViewpoint(gl, glu);
        }
        else
        {
            if(resetColor)
            {
                gl.glClearColor(clearColor[0],
                                clearColor[1],
                                clearColor[2],
                                clearColor[3]);
            }

            gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

            renderViewpoint(gl, glu);
        }

        if(terminate)
            return;

        boolean fog_active = false;

        if(currentFog != null)
        {
            fog_active = true;
            gl.glEnable(GL.GL_FOG);
            currentFog.render(gl, glu);
        }

        RenderableObject obj;
        RenderableComponent comp;

        for(int i = 0; i < numRenderables && !terminate; i++)
        {
            switch(operationList[i])
            {
                case RenderOp.START_RENDER:
                    // load the matrix to render
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Start render " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_RENDER:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Stop render " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);
                    gl.glPopMatrix();
                    break;

                case RenderOp.RENDER_GEOMETRY:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Render geometry " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    gl.glPopMatrix();
                    break;

                case RenderOp.RENDER_CUSTOM:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Render custom " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);

                    CustomRenderable cr =
                        (CustomRenderable)renderableList[i].renderable;
                    cr.render(gl, glu, renderableList[i].instructions);
                    gl.glPopMatrix();
                    break;

                case RenderOp.START_STATE:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Start state " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_STATE:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Stop state " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);
                    break;

                case RenderOp.START_LIGHT:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Start light " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    // Get the next available light ID

// TODO:
// Fix this so that if we run off the end we can still recover and not disable
// Lighting/Clipping completely for the next frame.
                    if(lastLightIdx >= availableLights.length)
                        continue;

                    Integer l_id = availableLights[lastLightIdx++];

                    lightIdMap.put(renderableList[i].id, l_id);

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, l_id);
                    gl.glPopMatrix();
                    break;

                case RenderOp.STOP_LIGHT:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Stop light " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    if(lastLightIdx >= availableLights.length)
                        continue;

                    l_id = (Integer)lightIdMap.remove(renderableList[i].id);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.postRender(gl, glu, l_id);
                    availableLights[--lastLightIdx] = l_id;
                    break;

                case RenderOp.START_CLIP_PLANE:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Start clip plane " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    // Get the next available clip plane ID

// TODO:
// Fix this so that if we run off the end we can still recover and not disable
// Lighting/Clipping completely for the next frame.
                    if(lastClipIdx >= availableClips.length)
                        continue;

                    Integer c_id = availableClips[lastClipIdx++];
                    clipIdMap.put(renderableList[i].id, c_id);

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, c_id);
                    gl.glPopMatrix();
                    break;

                case RenderOp.STOP_CLIP_PLANE:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Stop clip plane " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    if(lastClipIdx >= availableClips.length)
                        continue;

                    c_id = (Integer)clipIdMap.remove(renderableList[i].id);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.postRender(gl, glu, c_id);

                    availableClips[--lastClipIdx] = c_id;
                    break;

                case RenderOp.START_TRANSPARENT:
                    if(dumpNextFrame)
                        System.out.println("Start transparency");

                    gl.glDepthMask(false);
                    gl.glEnable(GL.GL_BLEND);
                    gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                    break;

                case RenderOp.STOP_TRANSPARENT:
                    if(dumpNextFrame)
                        System.out.println("Stop transparency");

                    gl.glDisable(GL.GL_BLEND);
                    gl.glDepthMask(true);
                    break;

                case RenderOp.START_FOG:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Start fog " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    if(!fog_active)
                    {
                        gl.glEnable(GL.GL_FOG);
                        fog_active = true;
                    }

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_FOG:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Stop fog " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    if(currentFog != null)
                        currentFog.render(gl, glu);
                    else
                    {
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.postRender(gl, glu);
                        fog_active = false;
                        gl.glDisable(GL.GL_FOG);
                    }
                    break;

                case RenderOp.START_SHADER_PROGRAM:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Start shader program " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    ShaderProgram prog =
                        (ShaderProgram)renderableList[i].renderable;

                    if(!prog.isValid(gl))
                    {
                        currentShaderProgramId = INVALID_SHADER;
                        continue;
                    }

// TODO: Optimise this to avoid the allocation. Use IntHashMap for lookup.
                    currentShaderProgramId = new Integer(prog.getProgramId(gl));
                    prog.render(gl, glu);
                    break;

                case RenderOp.STOP_SHADER_PROGRAM:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Stop shader program " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    if(currentShaderProgramId == INVALID_SHADER)
                        continue;

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);

                    currentShaderProgramId = INVALID_SHADER;
                    break;

                case RenderOp.SET_SHADER_ARGS:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Set shader args " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    if(currentShaderProgramId == INVALID_SHADER)
                        continue;

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, currentShaderProgramId);
                    break;

                case RenderOp.START_TEXTURE:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Start texture " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    comp = (RenderableComponent)renderableList[i].renderable;
                    Integer id = (Integer)(renderableList[i].instructions);
                    comp.render(gl, glu, id);
                    break;

                case RenderOp.STOP_TEXTURE:
                    if(dumpNextFrame) {
                        Renderable s = renderableList[i].renderable;
                        System.out.println("Stop texture " +
                                            s.getClass() +
                                            " 0x" +
                                            Integer.toHexString(s.hashCode()));
                    }

                    comp = (RenderableComponent)renderableList[i].renderable;
                    id = (Integer)(renderableList[i].instructions);
                    comp.postRender(gl, glu, id);
                    break;
            }

            // Also check for GL errors being made
            if(dumpNextFrame)
            {
                int err = gl.glGetError();
                if(err != GL.GL_NO_ERROR)
                {
                    System.out.println("Error: after " + renderableList[i].renderable);
                    System.out.println(glu.gluErrorString(err));
                }
            }

        }

        if(terminate)
            return;

        if(currentFog != null)
        {
            currentFog.postRender(gl, glu);
            gl.glDisable(GL.GL_FOG);
        }

        if(terminate)
            return;

        if(effectsProcessor != null)
            effectsProcessor.postDraw(drawable, userData);

        if(terminate)
            return;

        if(currentViewpoint != null)
            currentViewpoint.postRender(gl, glu);

        if(terminate)
            return;

        gl.glFlush();

        dumpNextFrame = false;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * The next rendered frame, output the GL state to the standard output.
     */
    void traceNextFrame()
    {
        dumpNextFrame = true;
    }

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param nodes The list of nodes in sorted order
     * @param renderOps Operation to perform on each node
     * @param numValid The number of valid items in the array
     */
    public void setDrawableObjects(RenderDetails[] nodes,
                                   int[] renderOps,
                                   int numValid)
    {
        super.setDrawableObjects(nodes, renderOps, numValid);
    }
}
