/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.surface;

// External imports
import net.java.games.jogl.*;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.device.output.elumens.SPI;
import org.j3d.util.IntHashMap;
import org.j3d.util.MatrixUtils;

/**
 * Handles the rendering for a single elumens dome-based output device.
 * <p>
 *
 * The code expects that everything is set up before each call of the display()
 * callback. It does not handle any recursive rendering requests as that is
 * assumed to have been sorted out before calling this renderer.
 *
 * @author Alan Hudson
 * @version $Revision: 1.37.2.1 $
 */
class ElumensRenderingProcessor extends BaseRenderingProcessor
    implements GLEventListener, Renderer
{
    /** The Java wrapper to the native SPI library */
    private SPI spi;

    /** Should we reinitialize SPI */
    private boolean initSpi;

    /** Flag to say that init has been called on the canvas */
    private boolean initComplete;

    /** The new number of channels at next SPI initialize */
    private int newNumChannels;

    /** The number of channels to render each frame. */
    private int numChannels;

    /** Format of the screen arrangements we want for SPI */
    private int spiFormat;

    /** The last near clip plane */
    private double lastNear;

    /** The last far clip plane */
    private double lastFar;

    /** Matrix utility code for doing inversions */
    private MatrixUtils matrixUtils;

    /** Request that the current drawing terminate immediately. App closing */
    private boolean terminate;

    /** The eye position.  Null if not set */
    private float[] eyePos;

    /** The len position.  Null if not set */
    private float[] lensPos;

    /** The screen orientation.  Null if not set */
    private double[] screenOrientation;

    /** The channel size.  Defaults to 1024x1024 */
    private int[] chanSize;

    /**
     * Construct handler for rendering objects to the main screen.
     */
    ElumensRenderingProcessor()
    {
        super();

        spi = new SPI();
        newNumChannels = 3;
        spiFormat = SPI.SPI_PF_3_CHAN | SPI.SPI_PF_BACKBUFFER;
        initSpi = true;

        chanSize = new int[] {1024,1024};
    }

    /**
     * Set the channel lens position.
     *
     * @param wall The walls to affect.  Defined in org.j3d.device.output.elumens.SPI
     * @param x The x position
     * @param y The y position
     * @param z The z position
     */
    public void setChanLensPosition(int wall, float x, float y, float z)
    {
        lensPos = new float[] {x,y,z};

        if (initComplete)
        {
            spi.setChanLensPosition(wall, lensPos[0], lensPos[1], lensPos[2]);
        }
    }

    /**
     * Set the channel eye position.
     *
     * @param wall The walls to affect.  Defined in org.j3d.device.output.elumens.SPI
     * @param x The x position
     * @param y The y position
     * @param z The z position
     */
    public void setChanEyePosition(int wall, float x, float y, float z)
    {
        eyePos = new float[] {x,y,z};

        if (initComplete)
        {
            spi.setChanEyePosition(wall, eyePos[0], eyePos[1], eyePos[2]);
        }
    }

    /**
     * Set the screen orientation.  Allows the project to rotated in software
     * for different hardware setups.
     *
     * @param r The roll
     * @param p The pitch
     * @param v The yaw
     */
    public void setScreenOrientation(double r, double p, double v)
    {
        screenOrientation = new double[] {r,p,v};

        if (initComplete)
        {
            spi.setScreenOrientation(screenOrientation[0], screenOrientation[1], screenOrientation[2]);
        }
    }

    /**
     * Set the channel size.
     *
     * @param wall The walls to affect.  Defined in org.j3d.device.output.elumens.SPI
     * @param h The height
     * @param w The width
     */
    public void setChanSize(int wall, int height, int width)
    {
        chanSize = new int[] {height,width};

        if (initComplete)
        {
            spi.setChanSize(wall,height, width);
        }
    }

    //---------------------------------------------------------------
    // Methods defined by GLEventListener
    //---------------------------------------------------------------

    /**
     * Called by the drawable immediately after the OpenGL context is
     * initialized; the GLContext has already been made current when
     * this method is called.
     *
     * @param drawable The display context to render to
     */
    public void init(GLDrawable drawable)
    {
        super.init(drawable);

        spi = new SPI();

        initializeSPI();
    }

    /**
     * Called by the drawable when the surface resizes itself. Used to
     * reset the viewport dimensions.
     *
     * @param drawable The display context to render to
     */
    public void reshape(GLDrawable drawable,
                        int x,
                        int y,
                        int width,
                        int height)
    {
        super.reshape(drawable, x,y,width,height);

        int w;
        int h;

        if (windowWidth > chanSize[0])
            w = chanSize[0];
        else
            w = windowWidth;

        if (windowHeight > chanSize[1])
            h = chanSize[1];
        else
            h = windowHeight;

        for(int i=0; i < numChannels; i++) {
            spi.setChanSize(getSpiToken(i), w, h);
        }

        initComplete = true;
    }

    /**
     * Called by the drawable to perform rendering by the client.
     *
     * @param drawable The display context to render to
     */
    public void display(GLDrawable drawable)
    {
        int token;

        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        if(numDeletions != 0)
        {
            for(int i = 0; i < numDeletions; i++)
            {
                deleteQueue[i].cleanup(gl, glu);
                deleteQueue[i] = null;
            }

            numDeletions = 0;
        }

        // Process the shader init calls and then log requests
        if(lastShaderInitIdx != 0)
        {
            for(int i = 0; i < lastShaderInitIdx; i++)
            {
                shaderInitList[i].initialize(gl, glu);
                shaderInitList[i] = null;
            }

            lastShaderInitIdx = 0;
        }

        if(lastShaderLogIdx != 0)
        {
            for(int i = 0; i < lastShaderLogIdx; i++)
            {
                shaderLogList[i].fetchLogInfo(gl, glu);
                shaderLogList[i] = null;
            }

            lastShaderLogIdx = 0;
        }

        if(terminate)
            return;

        if(initSpi)
            initializeSPI();

        if(terminate)
            return;

        renderViewpoint(gl, glu);

        if(terminate)
            return;

        spi.begin();

        if(terminate)
            return;

        for(int wallCnt = 0; wallCnt < numChannels && !terminate; wallCnt++)
        {
            token = getSpiToken(wallCnt);

            spi.preRender(token);

            if(terminate)
                return;

            if(effectsProcessor != null)
                effectsProcessor.preDraw(drawable, userData);

            if(terminate)
                return;

            if(currentBackground != null)
            {
                gl.glClear(GL.GL_DEPTH_BUFFER_BIT);

                gl.glMatrixMode(GL.GL_MODELVIEW);
                gl.glPushMatrix();
                gl.glLoadIdentity();
                gl.glMultMatrixf(backgroundMatrix);

                // TODO: Reseting the frustum messes up channels, so use old method that tears sometimes
                gl.glScaled(currentFrustum[5],currentFrustum[5],currentFrustum[5]);

                gl.glDepthMask(false);
                gl.glDisable(GL.GL_BLEND);
                gl.glDisable(GL.GL_DEPTH_TEST);

                currentBackground.render(gl, glu);
                currentBackground.postRender(gl, glu);

                gl.glDepthMask(true);
                gl.glEnable(GL.GL_DEPTH_TEST);
                gl.glEnable(GL.GL_BLEND);
                gl.glPopMatrix();


                gl.glMatrixMode(GL.GL_PROJECTION);
                gl.glPopMatrix();
                gl.glMatrixMode(GL.GL_MODELVIEW);

                renderViewpoint(gl, glu);
            }
            else
            {
                if(resetColor)
                {
                    gl.glClearColor(clearColor[0],
                                    clearColor[1],
                                    clearColor[2],
                                    clearColor[3]);
                }

                gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

                renderViewpoint(gl, glu);
            }

            if(terminate)
                return;

            boolean fog_active = false;

            if(currentFog != null)
            {
                fog_active = true;
                gl.glEnable(GL.GL_FOG);
                currentFog.render(gl, glu);
            }

            RenderableObject obj;
            RenderableComponent comp;

            for(int i = 0; i < numRenderables && !terminate; i++)
            {
                switch(operationList[i])
                {
                    case RenderOp.START_RENDER:
                        // load the matrix to render
                        gl.glPushMatrix();
                        gl.glMultMatrixf(renderableList[i].transform);
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.render(gl, glu);
                        break;

                    case RenderOp.STOP_RENDER:
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.postRender(gl, glu);
                        gl.glPopMatrix();
                        break;

                    case RenderOp.RENDER_GEOMETRY:
                        // load the matrix to render
                        gl.glPushMatrix();
                        gl.glMultMatrixf(renderableList[i].transform);
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.render(gl, glu);
                        gl.glPopMatrix();
                        break;

                    case RenderOp.RENDER_CUSTOM:
                        // load the matrix to render
                        gl.glPushMatrix();
                        gl.glMultMatrixf(renderableList[i].transform);

                        CustomRenderable cr =
                            (CustomRenderable)renderableList[i].renderable;
                        cr.render(gl, glu, renderableList[i].instructions);
                        gl.glPopMatrix();
                        break;

                    case RenderOp.START_STATE:
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.render(gl, glu);
                        break;

                    case RenderOp.STOP_STATE:
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.postRender(gl, glu);
                        break;

                    case RenderOp.START_LIGHT:
                        // Get the next available light ID

                        if(lastLightIdx >= availableLights.length)
                            continue;

                        Integer l_id = availableLights[lastLightIdx++];
                        lightIdMap.put(renderableList[i].id, l_id);

                        // load the matrix to render
                        gl.glPushMatrix();
                        gl.glMultMatrixf(renderableList[i].transform);
                        comp = (RenderableComponent)renderableList[i].renderable;
                        comp.render(gl, glu, l_id);
                        gl.glPopMatrix();
                        break;

                    case RenderOp.STOP_LIGHT:
                        if(lastLightIdx >= availableLights.length)
                            continue;

                        l_id = (Integer)lightIdMap.remove(renderableList[i].id);

                        comp = (RenderableComponent)renderableList[i].renderable;
                        comp.postRender(gl, glu, l_id);
                        availableLights[--lastLightIdx] = l_id;
                        break;

                    case RenderOp.START_CLIP_PLANE:
                        // Get the next available clip plane ID

                        if(lastClipIdx >= availableClips.length)
                            continue;

                        Integer c_id = availableClips[lastClipIdx++];
                        clipIdMap.put(renderableList[i].id, c_id);

                        // load the matrix to render
                        gl.glPushMatrix();
                        gl.glMultMatrixf(renderableList[i].transform);

                        comp = (RenderableComponent)renderableList[i].renderable;
                        comp.render(gl, glu, c_id);
                        gl.glPopMatrix();
                        break;

                    case RenderOp.STOP_CLIP_PLANE:
                        if(lastClipIdx >= availableClips.length)
                            continue;

                        c_id = (Integer)clipIdMap.remove(renderableList[i].id);

                        comp = (RenderableComponent)renderableList[i].renderable;
                        comp.postRender(gl, glu, c_id);

                        availableClips[--lastClipIdx] = c_id;
                        break;

                    case RenderOp.START_TRANSPARENT:
                        gl.glDepthMask(false);
                        gl.glEnable(GL.GL_BLEND);
                        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                        break;

                    case RenderOp.STOP_TRANSPARENT:
                        gl.glDisable(GL.GL_BLEND);
                        gl.glDepthMask(true);
                        break;

                    case RenderOp.START_FOG:
                        if(!fog_active)
                        {
                            gl.glEnable(GL.GL_FOG);
                            fog_active = true;
                        }

                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.render(gl, glu);
                        break;

                    case RenderOp.STOP_FOG:
                        if(currentFog != null)
                            currentFog.render(gl, glu);
                        else
                        {
                            obj = (RenderableObject)renderableList[i].renderable;
                            obj.postRender(gl, glu);
                            fog_active = false;
                            gl.glDisable(GL.GL_FOG);
                        }
                        break;

                    case RenderOp.START_SHADER_PROGRAM:
                        ShaderProgram prog =
                            (ShaderProgram)renderableList[i].renderable;

                        if(!prog.isValid(gl))
                        {
                            currentShaderProgramId = INVALID_SHADER;
                            continue;
                        }

// TODO: Optimise this to avoid the allocation. Use IntHashMap for lookup.
                        currentShaderProgramId = new Integer(prog.getProgramId(gl));
                        prog.render(gl, glu);
                        break;

                    case RenderOp.STOP_SHADER_PROGRAM:
                        if(currentShaderProgramId == INVALID_SHADER)
                            continue;

                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.postRender(gl, glu);

                        currentShaderProgramId = INVALID_SHADER;
                        break;

                    case RenderOp.SET_SHADER_ARGS:
                        if(currentShaderProgramId == INVALID_SHADER)
                            continue;

                        comp = (RenderableComponent)renderableList[i].renderable;
                        comp.render(gl, glu, currentShaderProgramId);
                        break;

                    case RenderOp.START_TEXTURE:
                        comp = (RenderableComponent)renderableList[i].renderable;
                        Integer id = (Integer)(renderableList[i].instructions);
                        comp.render(gl, glu, id);
                        break;

                    case RenderOp.STOP_TEXTURE:
                        comp = (RenderableComponent)renderableList[i].renderable;
                        id = (Integer)(renderableList[i].instructions);
                        comp.postRender(gl, glu, id);
                        break;
               }
            }

            if(terminate)
                return;

            if(currentFog != null)
            {
                currentFog.postRender(gl, glu);
                gl.glDisable(GL.GL_FOG);
            }

            if(terminate)
                return;

            if(effectsProcessor != null)
                effectsProcessor.postDraw(drawable, userData);

            if(terminate)
                return;

            gl.glFlush();

            if(terminate)
                return;

            spi.postRender(token);
        }

        if(terminate)
            return;

        if(currentViewpoint != null)
            currentViewpoint.postRender(gl, glu);

        if(terminate)
            return;

        spi.end();
        gl.glClearColor( 0.0f, 0.0f, 0.4f, 1.0f);
        gl.glClear (gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT);

        switch(numChannels)
        {
            case 1:
                spi.flush(SPI.SPI_1C_FRONT);
                break;
            case 2:
                spi.flush(SPI.SPI_ALL_2_CHAN);
                break;
            case 3:
                spi.flush(SPI.SPI_ALL_3_CHAN);
                break;
            case 4:
                spi.flush(SPI.SPI_ALL_4_CHAN);
                break;
        }
    }

    //---------------------------------------------------------------
    // Methods defined by Renderer
    //---------------------------------------------------------------

    /**
     * Convert a pixel location to surface coordinates.
     *
     * @param x The X coordinate
     * @param y The Y coordinate
     * @param position The converted position.  It must be preallocated.
     */
    public void getPixelLocationInSurface(float x, float y, Point3f position)
    {
        if (windowWidth == 0) {
            return;
        }

        position.x = (float) ((currentFrustum[1] - currentFrustum[0]) *
                     (x / windowWidth - 0.5f));
        position.y = (float) ((currentFrustum[3] - currentFrustum[2]) *
                     ((windowHeight - y) / windowHeight - 0.5f));
        position.z = (float) -currentFrustum[4];
    }

    /**
     * Get the Center Eye position in surface coordinates.
     *
     * @param position The current eye position.  It must be preallocated.
     */
    public void getCenterEyeInSurface(Point3f position)
    {
        position.set(eyePoint);
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Set the number of channels to display.  This will
     * happen the next render loop.
     */
    public void setNumberOfChannels(int num)
    {
        newNumChannels = num;
        initSpi = true;
    }

    /**
     * Get the spiToken based on the current wallCnt and number of channels.
     *
     * @return The token
     */
    private int getSpiToken(int wallCnt)
    {
        switch(numChannels)
        {
            case 1:
                return SPI.SPI_1C_FRONT;
            case 2:
                switch(wallCnt)
                {
                    case 0: return SPI.SPI_2C_LEFT;
                    case 1: return SPI.SPI_2C_RIGHT;
                    default: System.out.println("Unknown wall: " + wallCnt);
                }
            case 3:
                switch(wallCnt)
                {
                    case 0: return SPI.SPI_3C_LEFT;
                    case 1: return SPI.SPI_3C_RIGHT;
                    case 2: return SPI.SPI_3C_TOP;
                    default: System.out.println("Unknown wall: " + wallCnt);
                }
            case 4:
                switch(wallCnt)
                {
                    case 0: return SPI.SPI_4C_LEFT;
                    case 1: return SPI.SPI_4C_RIGHT;
                    case 2: return SPI.SPI_4C_TOP;
                    case 3: return SPI.SPI_4C_BOTTOM;
                    default: System.out.println("Unknown wall: " + wallCnt);
                }
            default:
                System.out.println("Unsupported number of channels in: " + this);
                return 0;
        }
    }

    /**
     * Get the SPI initialization format based on the
     * number of channels.
     *
     * @return the format
     */
    private int getSpiFormat()
    {
        int bufferType = SPI.SPI_PF_BACKBUFFER;

        switch(numChannels)
        {
            case 1:
                return (SPI.SPI_PF_1_CHAN | bufferType);
            case 2:
                return (SPI.SPI_PF_2_CHAN | bufferType);
            case 3:
                return (SPI.SPI_PF_3_CHAN | bufferType);
            case 4:
                return (SPI.SPI_PF_4_CHAN | bufferType);
            default:
                System.out.println("Unsupported number of channels in: " + this);
                return 0;
        }
    }

    /**
     * Initialize the SPI library
     */
    private void initializeSPI()
    {
        numChannels = newNumChannels;

        spiFormat = getSpiFormat();
        spi.initialize(spiFormat, numChannels);

        float near = 0.1f;
        float far = 3000;
        lastNear = near;
        lastFar = far;

        spi.setNearFar(near,far);

        int wall = 0;

        switch(numChannels)
        {
            case 1:
                wall = SPI.SPI_1C_FRONT;
                break;
            case 2:
                wall = SPI.SPI_2C_LEFT | SPI.SPI_2C_RIGHT;
                break;
            case 3:
                wall = SPI.SPI_3C_LEFT | SPI.SPI_3C_RIGHT | SPI.SPI_3C_TOP;
                break;
            case 4:
                wall = SPI.SPI_4C_LEFT | SPI.SPI_4C_RIGHT | SPI.SPI_4C_TOP | SPI.SPI_4C_BOTTOM;
                break;
        }

        if(initComplete)
        {
            int w = (windowWidth > chanSize[0]) ? chanSize[0] : windowWidth;
            int h = (windowHeight > chanSize[1]) ? chanSize[1] : windowHeight;

            switch(numChannels)
            {
                case 1:
                    spi.setChanSize(SPI.SPI_1C_FRONT, w, h);
                    break;
                case 2:
                    spi.setChanSize(SPI.SPI_ALL_2_CHAN, w, h);
                    break;
                case 3:
                    spi.setChanSize(SPI.SPI_ALL_3_CHAN, w, h);
                    break;
                case 4:
                    spi.setChanSize(SPI.SPI_ALL_4_CHAN, w, h);
                    break;
            }
        }

        if (lensPos != null) {
            System.out.println("Setting lensPosition: " + lensPos[0] + " " + lensPos[1] + " " + lensPos[2]);
            spi.setChanLensPosition(wall, lensPos[0], lensPos[1], lensPos[2]);
        }

        if (eyePos != null) {
            System.out.println("Setting eyePosition: " + eyePos[0] + " " + eyePos[1] + " " + eyePos[2]);
            spi.setChanEyePosition(wall, eyePos[0], eyePos[1], eyePos[2]);
        }
        if (screenOrientation != null) {
            System.out.println("Setting screenOrientation: " + screenOrientation[0] + " " + screenOrientation[1] + " " + screenOrientation[2]);
            spi.setScreenOrientation(screenOrientation[0], screenOrientation[1], screenOrientation[2]);
        }

        initSpi = false;
    }

    /**
     * Update the projection matrix.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    protected void updateProjectionMatrix(GL gl, GLU glu)
    {
        super.updateProjectionMatrix(gl, glu);

        spi.setNearFar((float)currentFrustum[4],(float)currentFrustum[5]);
    }

}
