/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.surface;

// External imports
import net.java.games.jogl.*;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.IntHashMap;
import org.j3d.util.MatrixUtils;

/**
 * Handles the rendering for a single output device using quad-buffered stereo
 * techniques.
 * <p>
 * The code expects that everything is set up before each call of the display()
 * callback. It does not handle any recursive rendering requests as that is
 * assumed to have been sorted out before calling this renderer.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.31 $
 */
class QuadBufferStereoRenderer extends BaseRenderingProcessor
    implements GLEventListener, StereoRenderer
{
    /** The current eye separation to use */
    private float eyeSeparation;

    /**
     * Flag indicating if the buffers are avialable for implementing
     * stereo as required by this implementation.
     */
    private boolean stereoAvailability;

    /**
     * Construct handler for rendering objects to the main screen.
     */
    QuadBufferStereoRenderer()
    {

        super();

        eyeSeparation = 0.005f;
        stereoAvailability = false;
    }

    //---------------------------------------------------------------
    // Methods defined by GLEventListener
    //---------------------------------------------------------------

    /**
     * Called by the drawable when the surface resizes itself. Used to
     * reset the viewport dimensions.
     *
     * @param drawable The display context to render to
     */
    public void reshape(GLDrawable drawable,
                        int x,
                        int y,
                        int width,
                        int height)
    {
        windowWidth = width;
        windowHeight = height;

        aspectRatio = (float)width / (float)height;
        if (recalcAspect)
        {
            viewEnvironment.setAspectRatio(aspectRatio);
        }

        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        gl.glViewport(0, 0, width, height);

//        updateProjectionMatrix(gl, glu);

        byte[] params = new byte[1];
        gl.glGetBooleanv(GL.GL_STEREO, params);

        stereoAvailability = (params[0] == GL.GL_TRUE);
    }

    /**
     * Called by the drawable to perform rendering by the client.
     *
     * @param drawable The display context to render to
     */
    public void display(GLDrawable drawable)
    {
        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        if(numDeletions != 0)
        {
            for(int i = 0; i < numDeletions; i++)
            {
                deleteQueue[i].cleanup(gl, glu);
                deleteQueue[i] = null;
            }

            numDeletions = 0;
        }

        // Process the shader init calls and then log requests
        if(lastShaderInitIdx != 0)
        {
            for(int i = 0; i < lastShaderInitIdx; i++)
            {
                shaderInitList[i].initialize(gl, glu);
                shaderInitList[i] = null;
            }

            lastShaderInitIdx = 0;
        }

        if(lastShaderLogIdx != 0)
        {
            for(int i = 0; i < lastShaderLogIdx; i++)
            {
                shaderLogList[i].fetchLogInfo(gl, glu);
                shaderLogList[i] = null;
            }

            lastShaderLogIdx = 0;
        }

        if(terminate)
            return;

        // Draw the left eye first, then right
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glDrawBuffer(GL.GL_BACK_LEFT);

        if(terminate)
            return;

        render(drawable, gl, glu, true);

        if(terminate)
            return;

        // Draw the left eye first, then right
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glDrawBuffer(GL.GL_BACK_RIGHT);

        if(terminate)
            return;

        render(drawable, gl, glu, false);
    }

    //---------------------------------------------------------------
    // Methods defined by StereoSurface
    //---------------------------------------------------------------

    /**
     * Check to see whether this surface supports stereo rendering. As this is
     * not known until after initialisation, this method will return false
     * until it can determine whether or not stereo is available.
     *
     * @return true Stereo support is currently available
     */
    public boolean isStereoAvailable()
    {
        return stereoAvailability;
    }

    /**
     * Set the eye separation value when rendering stereo. The default value is
     * 0.33 for most applications. The absolute value of the separation is
     * always used. Ignored for this implementation.
     *
     * @param sep The amount of eye separation
     */
    public void setStereoEyeSeparation(float sep)
    {
        eyeSeparation = (sep < 0) ? -sep : sep;
    }

    /**
     * Get the current eye separation value - always returns 0.
     *
     * @return sep The amount of eye separation
     */
    public float getStereoEyeSeparation()
    {
        return eyeSeparation;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Render the viewpoint setup.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    private void renderViewpoint(GL gl, GLU glu, boolean left)
    {
        // Will want to check here for the VP being non-live.
        if(currentViewpoint == null)
            return;

        updateProjectionMatrix(gl, glu, left);

        gl.glLoadIdentity();

        if(left)
            gl.glTranslated(0, 0, -eyeSeparation);
        else
            gl.glTranslated(0, 0, eyeSeparation);

        currentViewpoint.render(gl, glu);

        gl.glMultMatrixf(viewMatrix);
    }

    /**
     * Update the projection matrix.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     * @param left true if this is the left eye
     */
    private void updateProjectionMatrix(GL gl, GLU glu, boolean left)
    {
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();

        switch(viewEnvironment.getProjectionType())
        {
            case ViewEnvironment.PERSPECTIVE_PROJECTION:
                if(left)
                {
                    gl.glFrustum(currentFrustum[0] - eyeSeparation,
                                 currentFrustum[1],
                                 currentFrustum[2],
                                 currentFrustum[3] - eyeSeparation,
                                 currentFrustum[4],
                                 currentFrustum[5]);
                }
                else
                {
                    gl.glFrustum(currentFrustum[0] + eyeSeparation,
                                 currentFrustum[1],
                                 currentFrustum[2],
                                 currentFrustum[3] + eyeSeparation,
                                 currentFrustum[4],
                                 currentFrustum[5]);
                }
                break;

            case ViewEnvironment.ORTHOGRAPHIC_PROJECTION:
                // TODO:
                // EEK! Orthographic stereo projection. Ah.... doesn't sound
                // right, but not sure what to do about it here. Should we
                // change this and try to use the toe-in method for this style
                // or close our eyes and hope?
                double near_clip = viewEnvironment.getNearClipDistance();
                double far_clip = viewEnvironment.getFarClipDistance();

                gl.glOrtho(0,
                           windowWidth,
                           0,
                           windowHeight,
                           near_clip,
                           far_clip);
                break;

            default:
                System.out.println("unknown projection type");
        }

        gl.glMatrixMode(GL.GL_MODELVIEW);
    }

    /**
     * Perform the rendering loop for one of the two buffers, indicated by
     * the provided parameter.
     *
     * @param drawable The display context to render to
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     * @param left true if this is the left eye
     */
    private void render(GLDrawable drawable, GL gl, GLU glu, boolean left)
    {
        if(effectsProcessor != null)
            effectsProcessor.preDraw(drawable, userData);

        if(terminate)
            return;

        // TODO:
        // May want to put some optimisations here for systems that can clear
        // both back buffers at once with a single call to
        // glDrawBuffer(GL_BACK) before the glClear();

        if(currentBackground != null)
        {
            gl.glMatrixMode(GL.GL_PROJECTION);
            gl.glPushMatrix();
            gl.glLoadIdentity();

            setupBackgroundFrustum(gl, 0.01, 1.1, viewEnvironment.getFieldOfView());

            gl.glClear(GL.GL_DEPTH_BUFFER_BIT);

            gl.glMatrixMode(GL.GL_MODELVIEW);
            gl.glPushMatrix();

            gl.glLoadMatrixf(backgroundMatrix);

            gl.glDepthMask(false);
            gl.glDisable(GL.GL_BLEND);
            gl.glDisable(GL.GL_DEPTH_TEST);

            currentBackground.render(gl, glu);
            currentBackground.postRender(gl, glu);

            gl.glDepthMask(true);
            gl.glEnable(GL.GL_DEPTH_TEST);
            gl.glEnable(GL.GL_BLEND);
            gl.glPopMatrix();


            gl.glMatrixMode(GL.GL_PROJECTION);
            gl.glPopMatrix();
            gl.glMatrixMode(GL.GL_MODELVIEW);

            renderViewpoint(gl, glu, left);
        }
        else
        {
            if(resetColor)
            {
                gl.glClearColor(clearColor[0],
                                clearColor[1],
                                clearColor[2],
                                clearColor[3]);
            }

            gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

            renderViewpoint(gl, glu, left);
        }

        if(terminate)
            return;

        boolean fog_active = false;

        if(currentFog != null)
        {
            fog_active = true;
            gl.glEnable(GL.GL_FOG);
            currentFog.render(gl, glu);
        }

        RenderableObject obj;
        RenderableComponent comp;

        for(int i = 0; i < numRenderables && !terminate; i++)
        {
            switch(operationList[i])
            {
                case RenderOp.START_RENDER:
                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_RENDER:
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);
                    gl.glPopMatrix();
                    break;

                case RenderOp.RENDER_GEOMETRY:
                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    gl.glPopMatrix();
                    break;

                case RenderOp.RENDER_CUSTOM:
                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);

                    CustomRenderable cr =
                        (CustomRenderable)renderableList[i].renderable;
                    cr.render(gl, glu, renderableList[i].instructions);
                    gl.glPopMatrix();
                    break;

                case RenderOp.START_STATE:
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_STATE:
                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);
                    break;

                case RenderOp.START_LIGHT:
                    // Get the next available light ID

                    if(lastLightIdx >= availableLights.length)
                        continue;

                    Integer l_id = availableLights[lastLightIdx++];
                    lightIdMap.put(renderableList[i].id, l_id);

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);
                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, l_id);
                    gl.glPopMatrix();
                    break;

                case RenderOp.STOP_LIGHT:
                    if(lastLightIdx >= availableLights.length)
                        continue;

                    l_id = (Integer)lightIdMap.remove(renderableList[i].id);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.postRender(gl, glu, l_id);
                    availableLights[--lastLightIdx] = l_id;
                    break;

                case RenderOp.START_CLIP_PLANE:
                    // Get the next available clip plane ID

                    if(lastClipIdx >= availableClips.length)
                        continue;

                    Integer c_id = availableClips[lastClipIdx++];
                    clipIdMap.put(renderableList[i].id, c_id);

                    // load the matrix to render
                    gl.glPushMatrix();
                    gl.glMultMatrixf(renderableList[i].transform);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, c_id);
                    gl.glPopMatrix();
                    break;

                case RenderOp.STOP_CLIP_PLANE:
                    if(lastClipIdx >= availableClips.length)
                        continue;

                    c_id = (Integer)clipIdMap.remove(renderableList[i].id);

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.postRender(gl, glu, c_id);

                    availableClips[--lastClipIdx] = c_id;
                    break;

                case RenderOp.START_TRANSPARENT:
                    gl.glDepthMask(false);
                    gl.glEnable(GL.GL_BLEND);
                    gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                    break;

                case RenderOp.STOP_TRANSPARENT:
                    gl.glDisable(GL.GL_BLEND);
                    gl.glDepthMask(true);
                    break;

                case RenderOp.START_FOG:
                    if(!fog_active)
                    {
                        gl.glEnable(GL.GL_FOG);
                        fog_active = true;
                    }

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.render(gl, glu);
                    break;

                case RenderOp.STOP_FOG:
                    if(currentFog != null)
                        currentFog.render(gl, glu);
                    else
                    {
                        obj = (RenderableObject)renderableList[i].renderable;
                        obj.postRender(gl, glu);
                        fog_active = false;
                        gl.glDisable(GL.GL_FOG);
                    }
                    break;

                case RenderOp.START_SHADER_PROGRAM:
                    ShaderProgram prog =
                        (ShaderProgram)renderableList[i].renderable;

                    if(!prog.isValid(gl))
                    {
                        currentShaderProgramId = INVALID_SHADER;
                        continue;
                    }

// TODO: Optimise this to avoid the allocation. Use IntHashMap for lookup.
                    currentShaderProgramId = new Integer(prog.getProgramId(gl));
                    prog.render(gl, glu);
                    break;

                case RenderOp.STOP_SHADER_PROGRAM:
                    if(currentShaderProgramId == INVALID_SHADER)
                        continue;

                    obj = (RenderableObject)renderableList[i].renderable;
                    obj.postRender(gl, glu);

                    currentShaderProgramId = INVALID_SHADER;
                    break;

                case RenderOp.SET_SHADER_ARGS:
                    if(currentShaderProgramId == INVALID_SHADER)
                        continue;

                    comp = (RenderableComponent)renderableList[i].renderable;
                    comp.render(gl, glu, currentShaderProgramId);
                    break;

                 case RenderOp.START_TEXTURE:
                    comp = (RenderableComponent)renderableList[i].renderable;
                    Integer id = (Integer)(renderableList[i].instructions);
                    comp.render(gl, glu, id);
                    break;

                case RenderOp.STOP_TEXTURE:
                    comp = (RenderableComponent)renderableList[i].renderable;
                    id = (Integer)(renderableList[i].instructions);
                    comp.postRender(gl, glu, id);
                    break;
           }
        }

        if(terminate)
            return;

        if(currentFog != null)
        {
            currentFog.postRender(gl, glu);
            gl.glDisable(GL.GL_FOG);
        }

        if(terminate)
            return;

        if(effectsProcessor != null)
            effectsProcessor.postDraw(drawable, userData);

        if(terminate)
            return;

        if(currentViewpoint != null)
            currentViewpoint.postRender(gl, glu);

        if(terminate)
            return;

        gl.glFlush();
    }
}
