/*****************************************************************************
 *                        Copyright (c) 2001 Daniel Selman
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 ****************************************************************************/

package org.j3d.renderer.aviatrix3d.geom.volume;

// External imports
import javax.vecmath.Matrix4f;
import javax.vecmath.Vector4f;

// Local imports
import org.j3d.aviatrix3d.Cullable;
import org.j3d.aviatrix3d.CullInstructions;
import org.j3d.aviatrix3d.Node;
import org.j3d.aviatrix3d.InvalidWriteTimingException;

/**
 * Node that implements a simple OctTree-style data structure useful for volume
 * rendering of large datasets.
 * <p>
 *
 * The core implementation is a fairly simplistic distance-based algorithm to
 * determine when the level sets should be changed over. The class uses a single
 * Node object for the low-detail model, and allows the provision of a set of
 * nodes for the next detail level, allowing for nesting of the tree objects if
 * required.
 * <p>
 *
 * For describing the high-detail geometry, there is no specific geometric
 * representation implied by this node. There are just up to 8 children that can
 * be supplied, and that the user should make sure to place them in the correct
 * octant location using transformations as required.
 *
 * <p>
 * For bounds calculation, the bounds of the low-res model are used.
 *
 * @author Justin Couch
 * @version $Revision: 1.1 $
 */
public class OctTree extends Node implements Cullable
{
    /** Error message when the high detail geometry is out of range */
    private static final String HIGH_DETAIL_NUM_ERR =
        "numValid is out of range. It should be [0,8]";

    /** Error message when the user provides a negative range value */
    private static final String NEG_RANGE_MSG =
        "The range value provided was negative. It should be >= 0";

    /**
     * The range at which the geometry detail should switch over. This is actually
     * the range-squared value.
     */
    private float range;

    /** The low-detail model to use */
    private Node lowDetailGeom;

    /** The geometry for each of the high-res portions. */
    private Node[] highDetailGeom;

    /** The number of valid high-detail geometry children */
    private int numHighDetailGeom;

    /**
     * Create a new empty instance of the oct tree. The range is set to zero,
     * meaning it will always display the low res data, unless the user changes
     * the range.
     */
    public OctTree()
    {
        highDetailGeom = new Node[8];
    }

    //----------------------------------------------------------
    // Methods defined by Cullable
    //----------------------------------------------------------

    /**
     * Check this node for children to traverse. The angular resolution is
     * defined as Field Of View (in radians) / viewport width in pixels.
     *
     * @param output Fill in the child information here
     * @param vworldTx The transformation from the root of the scene to
     *    this node according to the current traversal path
     * @param viewTransform The transformation from the root of the scene
     *    graph to the active viewpoint
     * @param frustumPlanes Listing of frustum planes in the order: right,
     *    left, bottom, top, far, near
     * @param angularRes Angular resolution of the screen, or 0 if not
     *    calculable from the available data.
     */
    public void cullChildren(CullInstructions output,
                             Matrix4f vworldTx,
                             Matrix4f viewTransform,
                             Vector4f[] frustumPlanes,
                             float angularRes)
    {
        // if the distance from the viewTransform to vworldTx is less than
        // range, then return the high-detail, otherwise return the low detail.

        float x = vworldTx.m03 - viewTransform.m03;
        float y = vworldTx.m13 - viewTransform.m13;
        float z = vworldTx.m23 - viewTransform.m23;

        output.hasTransform = false;

        if(x * x + y * y + z * z < range)
        {
            output.numChildren = numHighDetailGeom;

            if(output.children.length < 8)
                output.resizeChildren(8);

            System.arraycopy(highDetailGeom,
                             0,
                             output.children,
                             0,
                             numHighDetailGeom);
        }
        else if(lowDetailGeom != null)
        {
            output.numChildren = 1;
            output.children[0] = lowDetailGeom;
        }
        else
            output.numChildren = 0;
    }

    //----------------------------------------------------------
    // Local Methods
    //----------------------------------------------------------

    /**
     * Set the range at which this geometry should change from low-detail to
     * high-detail models. The range should be a non-negative value. A range
     * value of zero will imply that the low detail geometry will always be
     * shown. Range is treated as a spherical distance from the center of
     * the local coordinate system of this geometry.
     *
     * @param distance The range at which the geometry should swap
     * @throws IllegalArgumentException The range value was negative
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setRange(float distance)
        throws IllegalArgumentException, InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isBoundsWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(distance < 0)
            throw new IllegalArgumentException(NEG_RANGE_MSG);

        range = distance * distance;
    }

    /**
     * Get the currently set range at which the high detail model will be
     * shown.
     *
     * @return A non-negative range value
     */
    public float getRange()
    {
        return (float)Math.sqrt(range);
    }

    /**
     * Set the low-detail geometry instance to use. Setting a null value will
     * clear the currently set geometry. The low detail can be any form of
     * scene graph desired.
     *
     * @param geom The sub scene graph to use for the low-detail geometry
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setLowDetail(Node geom)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isBoundsWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        lowDetailGeom = geom;
    }

    /**
     * Get the currently set low-detail geometry structure. If none is set, null
     * is returned.
     *
     * @return The current geometry structure or null
     */
    public Node getLowDetail()
    {
        return lowDetailGeom;
    }

    /**
     * Set the low-detail geometry instance to use. Setting a null value will
     * clear the currently set geometry. The low detail can be any form of
     * scene graph desired.
     *
     * @param geom The sub scene graph list to use for the high-detail geometry
     * @param numValid The number of valid bits of geometry to use from the array
     * @throws IllegalArgumentException numValid was outside [0, 8]
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setHighDetail(Node[] geom, int numValid)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isBoundsWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(numValid < 0 || numValid > 8)
            throw new IllegalArgumentException(HIGH_DETAIL_NUM_ERR);

        System.arraycopy(geom, 0, highDetailGeom, 0, numValid);
        numHighDetailGeom = numValid;
    }

    /**
     * Get the number of valid high-detail children in use by this node.
     *
     * @return A number between 0 and 8
     */
    public int numHighDetailGeom()
    {
        return numHighDetailGeom;
    }

    /**
     * Get the currently set low-detail geometry structure. If none is set, null
     * is returned.
     *
     * @param geom An array to copy the values into
     */
    public void getHighDetail(Node[] geom)
    {
        System.arraycopy(highDetailGeom, 0, geom, 0, numHighDetailGeom);
    }
}
