/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004-2005
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.util.ArrayList;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.MatrixUtils;

/**
 * The default internal implementation of the pick handling system.
 * tests.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.26.2.4 $
 */
public class DefaultPickingHandler
    implements PickingHandler
{
    /** The initial size of the children list */
    private static final int LIST_START_SIZE = 32;

    /** The increment size of the list if it gets overflowed */
    private static final int LIST_INCREMENT = 8;

    /** Handler for batch picking implementation. This class is big enough! */
    private DefaultBatchPickingHandler batchPicker;

    /** Path from the root of the scene graph to the current place */
    private Node[] pickPath;

    /** Path of transforms from the root of the scene to the current place */
    private Matrix4f[] transformPath;

    /** Flags to say with of the transforms are currently valid */
    private boolean[] validTransform;

    /** Last index of valid items on the path */
    private int lastPathIndex;

    /** Distance for the closest object found to date */
    private float closestDistance;

    /** The working vector for start and finish locations or direction */
    private float[] start;
    private float[] end;

    /** Place to fetch the intersection point data for the line picking */
    private float[] vertexPickData;

    /** The matrix to set everything in and then invert it. */
    private Matrix4f vworldMatrix;

    /** The matrix containing the local inverted form. */
    private Matrix4f invertedMatrix;

    /** Working vector for interacting with the world matrix */
    private Vector4f wkVec;

    /** Working vector for interacting with the world matrix */
    private Vector3f wkNormal;

    /** Temp value for holding the frustum planes */
    private Vector4f[] frustumPlanes;

    /** Matrix utility code for doing inversions */
    private MatrixUtils matrixUtils;

    /**
     * Initialise a new instance of the pick handler.
     */
    public DefaultPickingHandler()
    {
        start = new float[4];
        end = new float[4];
        start[3] = 1;
        end[3] = 1;

        vertexPickData = new float[3];
        pickPath = new Node[LIST_START_SIZE];
        transformPath = new Matrix4f[LIST_START_SIZE];
        validTransform = new boolean[LIST_START_SIZE];

        for(int i = 0; i < LIST_START_SIZE; i++)
            transformPath[i] = new Matrix4f();

        vworldMatrix = new Matrix4f();
        invertedMatrix = new Matrix4f();
        wkVec = new Vector4f();
        wkVec.w = 1;

        wkNormal = new Vector3f();

        frustumPlanes = new Vector4f[]
        {
            new Vector4f(), new Vector4f(), new Vector4f(),
            new Vector4f(), new Vector4f(), new Vector4f()
        };

        matrixUtils = new MatrixUtils();
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given set of requests.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @param numRequests The number of valid pick requests to process
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    public void pickBatch(PickableObject root, PickRequest[] req, int numRequests)
        throws NotPickableException
    {
        // Weed out the common cases first.
        if(numRequests == 0 || root == null || req == null || req.length == 0)
            return;

        if(numRequests == 1 || req.length == 1)
            pickSingle(root, req[0]);

        if(batchPicker == null)
            batchPicker = new DefaultBatchPickingHandler();

        batchPicker.processPick(root, req, numRequests);
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given set of requests.
     *
     * @param root The root point to start the pick processing from
     * @param req The details of the pick to be made
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    public void pickSingle(PickableObject root, PickRequest req)
        throws NotPickableException
    {
        if(root == null || req == null)
            return;

        if((root.getPickMask() & req.pickType) == 0)
        {
            req.pickCount = 0;
            return;
        }

        switch(req.pickGeometryType)
        {
            case PickRequest.PICK_POINT:
                pickPoint(root, req);
                break;

            case PickRequest.PICK_RAY:
                pickRay(root, req);
                break;

            case PickRequest.PICK_LINE_SEGMENT:
                pickLineSegment(root, req);
                break;

            case PickRequest.PICK_CYLINDER:
            case PickRequest.PICK_CYLINDER_SEGMENT:
                pickCylinder(root, req);
                break;

            case PickRequest.PICK_CONE:
            case PickRequest.PICK_CONE_SEGMENT:
                pickCone(root, req);
                break;

            case PickRequest.PICK_BOX:
                pickBox(root, req);
                break;

            case PickRequest.PICK_FRUSTUM:
                pickFrustum(root, req);
                break;

            case PickRequest.PICK_SPHERE:
                pickSphere(root, req);
                break;

            default:
System.out.println("No picking type has been specified");
        }
    }

    //---------------------------------------------------------------
    // Misc Internal methods
    //---------------------------------------------------------------

    // ----------------------- Point Picking ------------------------------

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickPoint(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
            case PickRequest.SORT_ORDERED:
                pickAllPoint(root, req);
                break;

            case PickRequest.SORT_ANY:
            case PickRequest.SORT_CLOSEST:
                pickSinglePoint(root, req);
                break;

            default:
System.out.println("No sort type has been specified");
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSinglePoint(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionPoint(req.origin))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSinglePoint((Group)kids[i],
                                                    req.pickType,
                                                    start,
                                                    output_path,
                                                    req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSinglePoint((SharedNode)kids[i],
                                                    req.pickType,
                                                    start,
                                                    output_path,
                                                    req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSinglePoint((Shape3D)kids[i],
                                                    req.pickType,
                                                    start,
                                                    output_path,
                                                    req.generateVWorldMatrix);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSinglePoint((SharedNode)target_node,
                                        req.pickType,
                                        start,
                                        output_path,
                                        req.generateVWorldMatrix);
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSinglePoint((Shape3D)target_node,
                                        req.pickType,
                                        start,
                                        output_path,
                                        req.generateVWorldMatrix);

            }
            else
            {
System.out.println("pickSinglePoint not handling non-shape/group root proxy");
            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSinglePoint not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param loc The location of the point to pick with in local coords
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSinglePoint(SharedNode root,
                                    int pickType,
                                    float[] loc,
                                    SceneGraphPath path,
                                    boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSinglePoint((Group)child,
                                   pickType,
                                   loc,
                                   path,
                                   needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickSinglePoint((SharedNode)child,
                                   pickType,
                                   loc,
                                   path,
                                   needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickSinglePoint((Shape3D)child,
                                   pickType,
                                   loc,
                                   path,
                                   needTransform);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param pickType flags to compare against for picking
     */
    private boolean pickSinglePoint(Group root,
                                    int pickType,
                                    float[] loc,
                                    SceneGraphPath path,
                                    boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionPoint(loc))
            return false;

        // Bounding check. If it has it, recurse on the bounding geometry
        // rather than the real geometry.
        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSinglePoint((Group)target_node,
                                       pickType,
                                       loc,
                                       path,
                                       needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSinglePoint((SharedNode)target_node,
                                       pickType,
                                       loc,
                                       path,
                                       needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSinglePoint((Shape3D)target_node,
                                       pickType,
                                       loc,
                                       path,
                                       needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickSinglePoint node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float old_x = loc[0];
            float old_y = loc[1];
            float old_z = loc[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, loc);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSinglePoint((Group)n,
                                            pickType,
                                            loc,
                                            path,
                                            needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSinglePoint((SharedNode)n,
                                            pickType,
                                            loc,
                                            path,
                                            needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSinglePoint((Shape3D)n,
                                            pickType,
                                            loc,
                                            path,
                                            needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            loc[0] = old_x;
            loc[1] = old_y;
            loc[2] = old_z;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param pickType flags to compare against for picking
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private boolean pickSinglePoint(Shape3D shape,
                                    int pickType,
                                    float[] loc,
                                    SceneGraphPath path,
                                    boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionPoint(loc))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();

            path.updatePath(pickPath,
                            lastPathIndex,
                            vworldMatrix,
                            invertedMatrix);

            found = true;
            lastPathIndex--;
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllPoint(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionPoint(req.origin))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found += pickAllPoint((Group)kids[i],
                                                  req.pickType,
                                                  start,
                                                  output_list,
                                                  found,
                                                  req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllPoint((SharedNode)kids[i],
                                                  req.pickType,
                                                  start,
                                                  output_list,
                                                  found,
                                                  req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllPoint((Shape3D)kids[i],
                                                  req.pickType,
                                                  start,
                                                  output_list,
                                                  found,
                                                  req.generateVWorldMatrix);
                        }
                    }

                    req.pickCount = found;
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllPoint((SharedNode)target_node,
                                      req.pickType,
                                      start,
                                      output_list,
                                      0,
                                      req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllPoint((Shape3D)target_node,
                                      req.pickType,
                                      start,
                                      output_list,
                                      0,
                                      req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else
            {
System.out.println("pickAllPoint not handling non-shape/group root proxy");
            }
        }
        else
        {
System.out.println("pickAllPoint not handling non-node root");
        }
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param pickType flags to compare against for picking
     */
    private int pickAllPoint(Group root,
                             int pickType,
                             float[] loc,
                             ArrayList paths,
                             int currentPath,
                             boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionPoint(loc))
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllPoint((Group)target_node,
                                    pickType,
                                    loc,
                                    paths,
                                    currentPath,
                                    needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllPoint((SharedNode)target_node,
                                    pickType,
                                    loc,
                                    paths,
                                    currentPath,
                                    needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllPoint((Shape3D)target_node,
                                    pickType,
                                    loc,
                                    paths,
                                    currentPath,
                                    needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickAllPoint node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float old_x = loc[0];
            float old_y = loc[1];
            float old_z = loc[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, loc);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllPoint((Group)n,
                                          pickType,
                                          loc,
                                          paths,
                                          currentPath + found,
                                          needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllPoint((SharedNode)n,
                                          pickType,
                                          loc,
                                          paths,
                                          currentPath + found,
                                          needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllPoint((Shape3D)n,
                                          pickType,
                                          loc,
                                          paths,
                                          currentPath + found,
                                          needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            loc[0] = old_x;
            loc[1] = old_y;
            loc[2] = old_z;
        }

        return found;
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param loc The location of the point to pick with in local coords
     * @param paths The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllPoint(SharedNode root,
                             int pickType,
                             float[] loc,
                             ArrayList paths,
                             int currentPath,
                             boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllPoint((Group)child,
                                pickType,
                                loc,
                                paths,
                                currentPath,
                                needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllPoint((SharedNode)child,
                                pickType,
                                loc,
                                paths,
                                currentPath,
                                needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllPoint((Shape3D)child,
                                pickType,
                                loc,
                                paths,
                                currentPath,
                                needTransform);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param pickType flags to compare against for picking
     * @param paths A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private int pickAllPoint(Shape3D shape,
                             int pickType,
                             float[] loc,
                             ArrayList paths,
                             int currentPath,
                             boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return 0;

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionPoint(loc))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            SceneGraphPath p;

            if(currentPath >= paths.size())
            {
                p = new SceneGraphPath();
                paths.add(p);
            }
            else
                p = (SceneGraphPath)paths.get(currentPath);

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();


            p.updatePath(pickPath,
                         lastPathIndex,
                         vworldMatrix,
                         invertedMatrix);

            found = 1;
            lastPathIndex--;
        }

        return found;
    }

    // ----------------------- Line Picking ------------------------------

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickLineSegment(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
                pickAllLineSegment(root, req);
                break;

            case PickRequest.SORT_ORDERED:
                pickAllLineSegmentSorted(root, req);
                break;

            case PickRequest.SORT_ANY:
                pickSingleLineSegment(root, req);
                break;

            case PickRequest.SORT_CLOSEST:
                pickSingleLineSegmentSorted(root, req);
                break;
            default:
System.out.println("No sort type has been specified");
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSingleLineSegment(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionSegment(req.origin, req.destination))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transform(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleLineSegment((Group)kids[i],
                                                          req.pickType,
                                                          start,
                                                          end,
                                                          output_path,
                                                          req.generateVWorldMatrix,
                                                          req.useGeometry);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleLineSegment((SharedNode)kids[i],
                                                          req.pickType,
                                                          start,
                                                          end,
                                                          output_path,
                                                          req.generateVWorldMatrix,
                                                          req.useGeometry);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleLineSegment((Shape3D)kids[i],
                                                          req.pickType,
                                                          start,
                                                          end,
                                                          output_path,
                                                          req.generateVWorldMatrix,
                                                          req.useGeometry,
                                                          false);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleLineSegment((SharedNode)target_node,
                                              req.pickType,
                                              start,
                                              end,
                                              output_path,
                                              req.generateVWorldMatrix,
                                              req.useGeometry);

            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleLineSegment((Shape3D)target_node,
                                              req.pickType,
                                              start,
                                              end,
                                              output_path,
                                              req.generateVWorldMatrix,
                                              req.useGeometry,
                                              false);

            }
            else
            {
System.out.println("pickSingleLineSegment not handling non-shape/group root proxy");
            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleLineSegment not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleLineSegment(SharedNode root,
                                          int pickType,
                                          float[] p1,
                                          float[] p2,
                                          SceneGraphPath path,
                                          boolean needTransform,
                                          boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleLineSegment((Group)child,
                                         pickType,
                                         p1,
                                         p2,
                                         path,
                                         needTransform,
                                         useGeom);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleLineSegment((SharedNode)child,
                                         pickType,
                                         p1,
                                         p2,
                                         path,
                                         needTransform,
                                         useGeom);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleLineSegment((Shape3D)child,
                                         pickType,
                                         p1,
                                         p2,
                                         path,
                                         needTransform,
                                         useGeom,
                                         false);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single line segment.
     *
     * @param root The now that is acting as the local root to work through
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleLineSegment(Group root,
                                          int pickType,
                                          float[] p1,
                                          float[] p2,
                                          SceneGraphPath path,
                                          boolean needTransform,
                                          boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionSegment(p1, p2))
            return false;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleLineSegment((Group)target_node,
                                             pickType,
                                             p1,
                                             p2,
                                             path,
                                             needTransform,
                                             useGeom);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleLineSegment((SharedNode)target_node,
                                             pickType,
                                             p1,
                                             p2,
                                             path,
                                             needTransform,
                                             useGeom);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleLineSegment((Shape3D)target_node,
                                             pickType,
                                             p1,
                                             p2,
                                             path,
                                             needTransform,
                                             useGeom,
                                             false);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleLineSegment node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float p1_x = p1[0];
            float p1_y = p1[1];
            float p1_z = p1[2];

            float p2_x = p2[0];
            float p2_y = p2[1];
            float p2_z = p2[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, p1);
                transform(invertedMatrix, p2);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleLineSegment((Group)n,
                                                  pickType,
                                                  p1,
                                                  p2,
                                                  path,
                                                  needTransform,
                                                  useGeom);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleLineSegment((SharedNode)n,
                                                  pickType,
                                                  p1,
                                                  p2,
                                                  path,
                                                  needTransform,
                                                  useGeom);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleLineSegment((Shape3D)n,
                                                  pickType,
                                                  p1,
                                                  p2,
                                                  path,
                                                  needTransform,
                                                  useGeom,
                                                  false);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            p1[0] = p1_x;
            p1[1] = p1_y;
            p1[2] = p1_z;

            p2[0] = p2_x;
            p2[1] = p2_y;
            p2[2] = p2_z;
        }

        return found;
    }

    /**
     * Look at a specific Shape instance for intersections with a single line
     * segment.
     *
     * @param shape The shape we are picking against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     * @param needClosest true if every triangle should be searched to find
     *    the real one that is closest. Only used if useGeom is true
     */
    private boolean pickSingleLineSegment(Shape3D shape,
                                          int pickType,
                                          float[] p1,
                                          float[] p2,
                                          SceneGraphPath path,
                                          boolean needTransform,
                                          boolean useGeom,
                                          boolean needClosest)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionSegment(p1, p2))
        {
            if(useGeom)
            {
                Geometry geom = shape.getGeometry();
                if(geom.pickLineSegment(p1, p2, needClosest, vertexPickData, 0))
                {
                    resizePath();
                    pickPath[lastPathIndex] = shape;
                    validTransform[lastPathIndex] = false;
                    lastPathIndex++;

                    if(needTransform)
                        buildVWorldTransform();
                    else
                        vworldMatrix.setIdentity();

                    path.updatePath(pickPath,
                                    lastPathIndex,
                                    vworldMatrix,
                                    invertedMatrix);

                    found = true;
                    lastPathIndex--;
                }
            }
            else
            {
                resizePath();
                pickPath[lastPathIndex] = shape;
                validTransform[lastPathIndex] = false;
                lastPathIndex++;

                if(needTransform)
                    buildVWorldTransform();
                else
                    vworldMatrix.setIdentity();

                path.updatePath(pickPath,
                                lastPathIndex,
                                vworldMatrix,
                                invertedMatrix);

                found = true;
                lastPathIndex--;
            }
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllLineSegment(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionSegment(req.origin, req.destination))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transform(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found += pickAllLineSegment((Group)kids[i],
                                                        req.pickType,
                                                        start,
                                                        end,
                                                        output_list,
                                                        found,
                                                        req.generateVWorldMatrix,
                                                        req.useGeometry);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllLineSegment((SharedNode)kids[i],
                                                        req.pickType,
                                                        start,
                                                        end,
                                                        output_list,
                                                        found,
                                                        req.generateVWorldMatrix,
                                                        req.useGeometry);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllLineSegment((Shape3D)kids[i],
                                                        req.pickType,
                                                        start,
                                                        end,
                                                        output_list,
                                                        found,
                                                        req.generateVWorldMatrix,
                                                        req.useGeometry,
                                                        false);
                        }
                    }

                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllLineSegment((SharedNode)target_node,
                                            req.pickType,
                                            start,
                                            end,
                                            output_list,
                                            0,
                                            req.generateVWorldMatrix,
                                            req.useGeometry);
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllLineSegment((Shape3D)target_node,
                                            req.pickType,
                                            start,
                                            end,
                                            output_list,
                                            0,
                                            req.generateVWorldMatrix,
                                            req.useGeometry,
                                            false);
            }
        }
        else
        {
System.out.println("pickAllLineSegment not handling non-node root");
        }

        req.pickCount = found;
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param paths Array of paths place to set the results in
     * @param currentPath Active index in the paths array
     * @param needTransform true if we should calc vworld information
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private int pickAllLineSegment(SharedNode root,
                                   int pickType,
                                   float[] p1,
                                   float[] p2,
                                   ArrayList paths,
                                   int currentPath,
                                   boolean needTransform,
                                   boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllLineSegment((Group)child,
                                      pickType,
                                      p1,
                                      p2,
                                      paths,
                                      currentPath,
                                      needTransform,
                                      useGeom);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllLineSegment((SharedNode)child,
                                      pickType,
                                      p1,
                                      p2,
                                      paths,
                                      currentPath,
                                      needTransform,
                                      useGeom);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllLineSegment((Shape3D)child,
                                      pickType,
                                      p1,
                                      p2,
                                      paths,
                                      currentPath,
                                      needTransform,
                                      useGeom,
                                      false);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The now that is acting as the local root to work through
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param paths Array of paths place to set the results in
     * @param currentPath Active index in the paths array
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private int pickAllLineSegment(Group root,
                                   int pickType,
                                   float[] p1,
                                   float[] p2,
                                   ArrayList paths,
                                   int currentPath,
                                   boolean needTransform,
                                   boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionSegment(p1, p2))
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllLineSegment((Group)target_node,
                                          pickType,
                                          p1,
                                          p2,
                                          paths,
                                          currentPath,
                                          needTransform,
                                          useGeom);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllLineSegment((SharedNode)target_node,
                                          pickType,
                                          p1,
                                          p2,
                                          paths,
                                          currentPath,
                                          needTransform,
                                          useGeom);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllLineSegment((Shape3D)target_node,
                                          pickType,
                                          p1,
                                          p2,
                                          paths,
                                          currentPath,
                                          needTransform,
                                          useGeom,
                                          false);
            }
            else if(target_node != null)
            {
System.out.println("pickAllLineSegment node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float p1_x = p1[0];
            float p1_y = p1[1];
            float p1_z = p1[2];

            float p2_x = p2[0];
            float p2_y = p2[1];
            float p2_z = p2[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, p1);
                transform(invertedMatrix, p2);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllLineSegment((Group)n,
                                                pickType,
                                                p1,
                                                p2,
                                                paths,
                                                currentPath + found,
                                                needTransform,
                                                useGeom);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllLineSegment((SharedNode)n,
                                                pickType,
                                                p1,
                                                p2,
                                                paths,
                                                currentPath + found,
                                                needTransform,
                                                useGeom);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllLineSegment((Shape3D)n,
                                                pickType,
                                                p1,
                                                p2,
                                                paths,
                                                currentPath + found,
                                                needTransform,
                                                useGeom,
                                                false);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            p1[0] = p1_x;
            p1[1] = p1_y;
            p1[2] = p1_z;

            p2[0] = p2_x;
            p2[1] = p2_y;
            p2[2] = p2_z;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param pickType flags to compare against for picking
     * @param paths A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private int pickAllLineSegment(Shape3D shape,
                                   int pickType,
                                   float[] p1,
                                   float[] p2,
                                   ArrayList paths,
                                   int currentPath,
                                   boolean needTransform,
                                   boolean useGeom,
                                   boolean needClosest)
    {
        if((shape.getPickMask() & pickType) == 0)
            return 0;

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionSegment(p1, p2))
        {
            if(useGeom)
            {
                Geometry geom = shape.getGeometry();
                if(geom.pickLineSegment(p1, p2, needClosest, vertexPickData, 0))
                {
                    resizePath();
                    pickPath[lastPathIndex] = shape;
                    validTransform[lastPathIndex] = false;
                    lastPathIndex++;

                    SceneGraphPath p;

                    if(currentPath >= paths.size())
                    {
                        p = new SceneGraphPath();
                        paths.add(p);
                    }
                    else
                        p = (SceneGraphPath)paths.get(currentPath);

                    if(needTransform)
                        buildVWorldTransform();
                    else
                        vworldMatrix.setIdentity();


                    p.updatePath(pickPath,
                                 lastPathIndex,
                                 vworldMatrix,
                                 invertedMatrix);

                    found = 1;
                    lastPathIndex--;
                }
            }
            else
            {
                resizePath();
                pickPath[lastPathIndex] = shape;
                validTransform[lastPathIndex] = false;
                lastPathIndex++;

                SceneGraphPath p;

                if(currentPath >= paths.size())
                {
                    p = new SceneGraphPath();
                    paths.add(p);
                }
                else
                    p = (SceneGraphPath)paths.get(currentPath);

                if(needTransform)
                    buildVWorldTransform();
                else
                    vworldMatrix.setIdentity();


                p.updatePath(pickPath,
                             lastPathIndex,
                             vworldMatrix,
                             invertedMatrix);

                found = 1;
                lastPathIndex--;
            }
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSingleLineSegmentSorted(PickableObject root,
                                             PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];


        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionSegment(req.origin, req.destination))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transform(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleLineSegmentSorted((Group)kids[i],
                                                          req.pickType,
                                                          start,
                                                          end,
                                                          output_path,
                                                          req.generateVWorldMatrix,
                                                          req.useGeometry);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleLineSegmentSorted((SharedNode)kids[i],
                                                                req.pickType,
                                                                start,
                                                                end,
                                                                output_path,
                                                                req.generateVWorldMatrix,
                                                                req.useGeometry);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleLineSegment((Shape3D)kids[i],
                                                          req.pickType,
                                                          start,
                                                          end,
                                                          output_path,
                                                          req.generateVWorldMatrix,
                                                          req.useGeometry,
                                                          true);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleLineSegmentSorted((SharedNode)target_node,
                                                    req.pickType,
                                                    start,
                                                    end,
                                                    output_path,
                                                    req.generateVWorldMatrix,
                                                    req.useGeometry);

            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleLineSegment((Shape3D)target_node,
                                              req.pickType,
                                              start,
                                              end,
                                              output_path,
                                              req.generateVWorldMatrix,
                                              req.useGeometry,
                                              false);

            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleLineSegment not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleLineSegmentSorted(SharedNode root,
                                                int pickType,
                                                float[] p1,
                                                float[] p2,
                                                SceneGraphPath path,
                                                boolean needTransform,
                                                boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleLineSegmentSorted((Group)child,
                                               pickType,
                                               p1,
                                               p2,
                                               path,
                                               needTransform,
                                               useGeom);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleLineSegmentSorted((SharedNode)child,
                                               pickType,
                                               p1,
                                               p2,
                                               path,
                                               needTransform,
                                               useGeom);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleLineSegment((Shape3D)child,
                                         pickType,
                                         p1,
                                         p2,
                                         path,
                                         needTransform,
                                         useGeom,
                                         true);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The now that is acting as the local root to work through
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleLineSegmentSorted(Group root,
                                                int pickType,
                                                float[] p1,
                                                float[] p2,
                                                SceneGraphPath path,
                                                boolean needTransform,
                                                boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionSegment(p1, p2))
            return false;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleLineSegmentSorted((Group)target_node,
                                                   pickType,
                                                   p1,
                                                   p2,
                                                   path,
                                                   needTransform,
                                                   useGeom);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleLineSegmentSorted((SharedNode)target_node,
                                                   pickType,
                                                   p1,
                                                   p2,
                                                   path,
                                                   needTransform,
                                                   useGeom);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleLineSegment((Shape3D)target_node,
                                             pickType,
                                             p1,
                                             p2,
                                             path,
                                             needTransform,
                                             useGeom,
                                             true);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleLineSegmentSorted node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float p1_x = p1[0];
            float p1_y = p1[1];
            float p1_z = p1[2];

            float p2_x = p2[0];
            float p2_y = p2[1];
            float p2_z = p2[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, p1);
                transform(invertedMatrix, p2);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleLineSegmentSorted((Group)n,
                                                        pickType,
                                                        p1,
                                                        p2,
                                                        path,
                                                        needTransform,
                                                        useGeom);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleLineSegmentSorted((SharedNode)n,
                                                        pickType,
                                                        p1,
                                                        p2,
                                                        path,
                                                        needTransform,
                                                        useGeom);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleLineSegment((Shape3D)n,
                                                  pickType,
                                                  p1,
                                                  p2,
                                                  path,
                                                  needTransform,
                                                  useGeom,
                                                  true);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            p1[0] = p1_x;
            p1[1] = p1_y;
            p1[2] = p1_z;

            p2[0] = p2_x;
            p2[1] = p2_y;
            p2[2] = p2_z;
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllLineSegmentSorted(PickableObject root, PickRequest req)
    {
    }

    // ----------------------- Ray Picking ------------------------------

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickRay(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
            case PickRequest.SORT_ORDERED:
                pickAllRay(root, req);
                break;

            case PickRequest.SORT_ANY:
            case PickRequest.SORT_CLOSEST:
                pickSingleRay(root, req);
                break;
            default:
System.out.println("No sort type has been specified");
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSingleRay(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];


        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionRay(req.origin, req.destination))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transformNormal(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleRay((Group)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  output_path,
                                                  req.generateVWorldMatrix,
                                                  req.useGeometry);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleRay((SharedNode)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  output_path,
                                                  req.generateVWorldMatrix,
                                                  req.useGeometry);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleRay((Shape3D)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  output_path,
                                                  req.generateVWorldMatrix,
                                                  req.useGeometry,
                                                  false);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleRay((SharedNode)target_node,
                                      req.pickType,
                                      start,
                                      end,
                                      output_path,
                                      req.generateVWorldMatrix,
                                      req.useGeometry);

            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleRay((Shape3D)target_node,
                                      req.pickType,
                                      start,
                                      end,
                                      output_path,
                                      req.generateVWorldMatrix,
                                      req.useGeometry,
                                      false);

            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleRay not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleRay(SharedNode root,
                                  int pickType,
                                  float[] p1,
                                  float[] p2,
                                  SceneGraphPath path,
                                  boolean needTransform,
                                  boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleRay((Group)child,
                                 pickType,
                                 p1,
                                 p2,
                                 path,
                                 needTransform,
                                 useGeom);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleRay((SharedNode)child,
                                 pickType,
                                 p1,
                                 p2,
                                 path,
                                 needTransform,
                                 useGeom);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleRay((Shape3D)child,
                                 pickType,
                                 p1,
                                 p2,
                                 path,
                                 needTransform,
                                 useGeom,
                                 true);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single line segment.
     *
     * @param root The now that is acting as the local root to work through
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleRay(Group root,
                                  int pickType,
                                  float[] p1,
                                  float[] p2,
                                  SceneGraphPath path,
                                  boolean needTransform,
                                  boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionRay(p1, p2))
            return false;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleRay((Group)target_node,
                                     pickType,
                                     p1,
                                     p2,
                                     path,
                                     needTransform,
                                     useGeom);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleRay((SharedNode)target_node,
                                     pickType,
                                     p1,
                                     p2,
                                     path,
                                     needTransform,
                                     useGeom);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleRay((Shape3D)target_node,
                                     pickType,
                                     p1,
                                     p2,
                                     path,
                                     needTransform,
                                     useGeom,
                                     false);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleRay node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float p1_x = p1[0];
            float p1_y = p1[1];
            float p1_z = p1[2];

            float p2_x = p2[0];
            float p2_y = p2[1];
            float p2_z = p2[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, p1);
                transformNormal(invertedMatrix, p2);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleRay((Group)n,
                                          pickType,
                                          p1,
                                          p2,
                                          path,
                                          needTransform,
                                          useGeom);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleRay((SharedNode)n,
                                          pickType,
                                          p1,
                                          p2,
                                          path,
                                          needTransform,
                                          useGeom);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleRay((Shape3D)n,
                                          pickType,
                                          p1,
                                          p2,
                                          path,
                                          needTransform,
                                          useGeom,
                                          false);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            p1[0] = p1_x;
            p1[1] = p1_y;
            p1[2] = p1_z;

            p2[0] = p2_x;
            p2[1] = p2_y;
            p2[2] = p2_z;
        }

        return found;
    }

    /**
     * Look at a specific Shape instance for intersections with a single line
     * segment.
     *
     * @param shape The shape we are picking against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     * @param needClosest true if every triangle should be searched to find
     *    the real one that is closest. Only used if useGeom is true
     */
    private boolean pickSingleRay(Shape3D shape,
                                  int pickType,
                                  float[] p1,
                                  float[] p2,
                                  SceneGraphPath path,
                                  boolean needTransform,
                                  boolean useGeom,
                                  boolean needClosest)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionRay(p1, p2))
        {
            if(useGeom)
            {
                Geometry geom = shape.getGeometry();
                if(geom.pickLineRay(p1, p2, needClosest, vertexPickData, 0))
                {
                    resizePath();
                    pickPath[lastPathIndex] = shape;
                    validTransform[lastPathIndex] = false;
                    lastPathIndex++;

                    if(needTransform)
                        buildVWorldTransform();
                    else
                        vworldMatrix.setIdentity();

                    path.updatePath(pickPath,
                                    lastPathIndex,
                                    vworldMatrix,
                                    invertedMatrix);

                    found = true;
                    lastPathIndex--;
                }
            }
            else
            {
                resizePath();
                pickPath[lastPathIndex] = shape;
                validTransform[lastPathIndex] = false;
                lastPathIndex++;

                if(needTransform)
                    buildVWorldTransform();
                else
                    vworldMatrix.setIdentity();

                path.updatePath(pickPath,
                                lastPathIndex,
                                vworldMatrix,
                                invertedMatrix);

                found = true;
                lastPathIndex--;
            }
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllRay(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionRay(req.origin, req.destination))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);

                        transform(invertedMatrix, start);
                        transformNormal(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;

                        if(kids[i] instanceof Group)
                        {
                            found += pickAllRay((Group)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix,
                                                req.useGeometry);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllRay((SharedNode)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix,
                                                req.useGeometry);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllRay((Shape3D)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix,
                                                req.useGeometry,
                                                false);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllRay((SharedNode)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix,
                                    req.useGeometry);
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllRay((Shape3D)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix,
                                    req.useGeometry,
                                    false);
            }
        }
        else
        {
System.out.println("pickAllRay not handling non-node root");
        }

        req.pickCount = found;
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param paths Array of paths place to set the results in
     * @param currentPath Active index in the paths array
     * @param needTransform true if we should calc vworld information
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private int pickAllRay(SharedNode root,
                           int pickType,
                           float[] p1,
                           float[] p2,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform,
                           boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllRay((Group)child,
                              pickType,
                              p1,
                              p2,
                              paths,
                              currentPath,
                              needTransform,
                              useGeom);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllRay((SharedNode)child,
                               pickType,
                               p1,
                               p2,
                               paths,
                               currentPath,
                               needTransform,
                               useGeom);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllRay((Shape3D)child,
                               pickType,
                               p1,
                               p2,
                               paths,
                               currentPath,
                               needTransform,
                               useGeom,
                               false);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The now that is acting as the local root to work through
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param paths Array of paths place to set the results in
     * @param currentPath Active index in the paths array
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private int pickAllRay(Group root,
                           int pickType,
                           float[] p1,
                           float[] p2,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform,
                           boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionRay(p1, p2))
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllRay((Group)target_node,
                                  pickType,
                                  p1,
                                  p2,
                                  paths,
                                  currentPath,
                                  needTransform,
                                  useGeom);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllRay((SharedNode)target_node,
                                  pickType,
                                  p1,
                                  p2,
                                  paths,
                                  currentPath,
                                  needTransform,
                                  useGeom);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllRay((Shape3D)target_node,
                                  pickType,
                                  p1,
                                  p2,
                                  paths,
                                  currentPath,
                                  needTransform,
                                  useGeom,
                                  false);
            }
            else if(target_node != null)
            {
System.out.println("pickAllRay node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float p1_x = p1[0];
            float p1_y = p1[1];
            float p1_z = p1[2];

            float p2_x = p2[0];
            float p2_y = p2[1];
            float p2_z = p2[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);

                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, p1);
                transformNormal(invertedMatrix, p2);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllRay((Group)n,
                                        pickType,
                                        p1,
                                        p2,
                                        paths,
                                        currentPath + found,
                                        needTransform,
                                        useGeom);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllRay((SharedNode)n,
                                        pickType,
                                        p1,
                                        p2,
                                        paths,
                                        currentPath + found,
                                        needTransform,
                                        useGeom);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllRay((Shape3D)n,
                                        pickType,
                                        p1,
                                        p2,
                                        paths,
                                        currentPath + found,
                                        needTransform,
                                        useGeom,
                                        false);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            p1[0] = p1_x;
            p1[1] = p1_y;
            p1[2] = p1_z;

            p2[0] = p2_x;
            p2[1] = p2_y;
            p2[2] = p2_z;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param pickType flags to compare against for picking
     * @param paths A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private int pickAllRay(Shape3D shape,
                           int pickType,
                           float[] p1,
                           float[] p2,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform,
                           boolean useGeom,
                           boolean needClosest)
    {
        if((shape.getPickMask() & pickType) == 0) {
            return 0;
        }

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionRay(p1, p2))
        {
            if(useGeom)
            {
                Geometry geom = shape.getGeometry();
                if(geom.pickLineRay(p1, p2, needClosest, vertexPickData, 0))
                {
                    resizePath();
                    pickPath[lastPathIndex] = shape;
                    validTransform[lastPathIndex] = false;
                    lastPathIndex++;
                    SceneGraphPath p;

                    if(currentPath >= paths.size())
                    {
                        p = new SceneGraphPath();
                        paths.add(p);
                    }
                    else
                        p = (SceneGraphPath)paths.get(currentPath);

                    if(needTransform)
                        buildVWorldTransform();
                    else
                        vworldMatrix.setIdentity();


                    p.updatePath(pickPath,
                                 lastPathIndex,
                                 vworldMatrix,
                                 invertedMatrix);

                    found = 1;
                    lastPathIndex--;
                }
            }
            else
            {
                resizePath();
                pickPath[lastPathIndex] = shape;
                validTransform[lastPathIndex] = false;
                lastPathIndex++;

                SceneGraphPath p;

                if(currentPath >= paths.size())
                {
                    p = new SceneGraphPath();
                    paths.add(p);
                }
                else
                    p = (SceneGraphPath)paths.get(currentPath);

                if(needTransform)
                    buildVWorldTransform();
                else
                    vworldMatrix.setIdentity();


                p.updatePath(pickPath,
                             lastPathIndex,
                             vworldMatrix,
                             invertedMatrix);

                found = 1;
                lastPathIndex--;
            }
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSingleRaySorted(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];


        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionRay(req.origin, req.destination))
                return;

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transform(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleRaySorted((Group)kids[i],
                                                        req.pickType,
                                                        start,
                                                        end,
                                                        output_path,
                                                        req.generateVWorldMatrix,
                                                        req.useGeometry);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleRaySorted((SharedNode)kids[i],
                                                        req.pickType,
                                                        start,
                                                        end,
                                                        output_path,
                                                        req.generateVWorldMatrix,
                                                        req.useGeometry);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleRay((Shape3D)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  output_path,
                                                  req.generateVWorldMatrix,
                                                  req.useGeometry,
                                                  true);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleRaySorted((SharedNode)target_node,
                                            req.pickType,
                                            start,
                                            end,
                                            output_path,
                                            req.generateVWorldMatrix,
                                            req.useGeometry);

            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleRay((Shape3D)target_node,
                                      req.pickType,
                                      start,
                                      end,
                                      output_path,
                                      req.generateVWorldMatrix,
                                      req.useGeometry,
                                      false);

            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleRay not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleRaySorted(SharedNode root,
                                        int pickType,
                                        float[] p1,
                                        float[] p2,
                                        SceneGraphPath path,
                                        boolean needTransform,
                                        boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleRaySorted((Group)child,
                                       pickType,
                                       p1,
                                       p2,
                                       path,
                                       needTransform,
                                       useGeom);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleRaySorted((SharedNode)child,
                                       pickType,
                                       p1,
                                       p2,
                                       path,
                                       needTransform,
                                       useGeom);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleRay((Shape3D)child,
                                 pickType,
                                 p1,
                                 p2,
                                 path,
                                 needTransform,
                                 useGeom,
                                 true);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The now that is acting as the local root to work through
     * @param pickType flags to compare against for picking
     * @param p1 The first point of the segment
     * @param p2 The second point of the segment
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     * @param useGeom true if this should check the results based on geometry
     *    rather than just bounds
     */
    private boolean pickSingleRaySorted(Group root,
                                        int pickType,
                                        float[] p1,
                                        float[] p2,
                                        SceneGraphPath path,
                                        boolean needTransform,
                                        boolean useGeom)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionRay(p1, p2))
            return false;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleRaySorted((Group)target_node,
                                           pickType,
                                           p1,
                                           p2,
                                           path,
                                           needTransform,
                                           useGeom);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleRaySorted((SharedNode)target_node,
                                           pickType,
                                           p1,
                                           p2,
                                           path,
                                           needTransform,
                                           useGeom);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleRay((Shape3D)target_node,
                                     pickType,
                                     p1,
                                     p2,
                                     path,
                                     needTransform,
                                     useGeom,
                                     true);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleRaySorted node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float p1_x = p1[0];
            float p1_y = p1[1];
            float p1_z = p1[2];

            float p2_x = p2[0];
            float p2_y = p2[1];
            float p2_z = p2[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, p1);
                transform(invertedMatrix, p2);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleRaySorted((Group)n,
                                                pickType,
                                                p1,
                                                p2,
                                                path,
                                                needTransform,
                                                useGeom);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleRaySorted((SharedNode)n,
                                                pickType,
                                                p1,
                                                p2,
                                                path,
                                                needTransform,
                                                useGeom);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleRay((Shape3D)n,
                                          pickType,
                                          p1,
                                          p2,
                                          path,
                                          needTransform,
                                          useGeom,
                                          true);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            p1[0] = p1_x;
            p1[1] = p1_y;
            p1[2] = p1_z;

            p2[0] = p2_x;
            p2[1] = p2_y;
            p2[2] = p2_z;
        }

        return found;
    }

    // ----------------------- Cylinder Picking ------------------------------

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickCylinder(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
            case PickRequest.SORT_ORDERED:
                pickAllCylinder(root, req);
                break;

            case PickRequest.SORT_ANY:
            case PickRequest.SORT_CLOSEST:
                pickSingleCylinder(root, req);
                break;
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickSingleCylinder(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            float x = req.origin[0] - req.destination[0];
            float y = req.origin[1] - req.destination[1];
            float z = req.origin[2] - req.destination[2];

            float height = (float)Math.sqrt(x * x + y * y + z * z);

            if(height == 0)
            {
                req.pickCount = 0;
                return;
            }

            // Start will be the center and end will be the axis vector
            start[0] = (req.origin[0] + req.destination[0]) * 0.5f;
            start[1] = (req.origin[1] + req.destination[0]) * 0.5f;
            start[2] = (req.origin[2] + req.destination[0]) * 0.5f;

            end[0] = x;
            end[1] = y;
            end[2] = z;

            float radius = req.additionalData;

            if(!bounds.checkIntersectionCylinder(start, end, radius, height))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transformNormal(invertedMatrix, end);

                        // need to scale the radius and height as well.
                        float scale = invertedMatrix.getScale();
                        radius *= scale;
                        height *= scale;
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleCylinder((Group)kids[i],
                                                    req.pickType,
                                                    start,
                                                    end,
                                                    radius,
                                                    height,
                                                    output_path,
                                                    req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleCylinder((SharedNode)kids[i],
                                                    req.pickType,
                                                    start,
                                                    end,
                                                    radius,
                                                    height,
                                                    output_path,
                                                    req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleCylinder((Shape3D)kids[i],
                                                    req.pickType,
                                                    start,
                                                    end,
                                                    radius,
                                                    height,
                                                    output_path,
                                                    req.generateVWorldMatrix);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleCylinder((SharedNode)target_node,
                                           req.pickType,
                                           start,
                                           end,
                                           radius,
                                           height,
                                           output_path,
                                           req.generateVWorldMatrix);
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleCylinder((Shape3D)target_node,
                                           req.pickType,
                                           start,
                                           end,
                                           radius,
                                           height,
                                           output_path,
                                           req.generateVWorldMatrix);

            }
            else
            {
System.out.println("pickSingleCylinder not handling non-shape/group root proxy");
            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleCylinder not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleCylinder(SharedNode root,
                                       int pickType,
                                       float[] center,
                                       float[] axis,
                                       float radius,
                                       float height,
                                       SceneGraphPath path,
                                       boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleCylinder((Group)child,
                                      pickType,
                                      center,
                                      axis,
                                      radius,
                                      height,
                                      path,
                                      needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleCylinder((SharedNode)child,
                                       pickType,
                                       center,
                                       axis,
                                       radius,
                                       height,
                                       path,
                                       needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleCylinder((Shape3D)child,
                                       pickType,
                                       center,
                                       axis,
                                       radius,
                                       height,
                                       path,
                                       needTransform);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and recurse into
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleCylinder(Group root,
                                       int pickType,
                                       float[] center,
                                       float[] axis,
                                       float radius,
                                       float height,
                                       SceneGraphPath path,
                                       boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionCylinder(center, axis, radius, height))
            return false;

        // Bounding check. If it has it, recurse on the bounding geometry
        // rather than the real geometry.
        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleCylinder((Group)target_node,
                                          pickType,
                                          center,
                                          axis,
                                          radius,
                                          height,
                                          path,
                                          needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleCylinder((SharedNode)target_node,
                                          pickType,
                                          center,
                                          axis,
                                          radius,
                                          height,
                                          path,
                                          needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleCylinder((Shape3D)target_node,
                                     pickType,
                                     center,
                                     axis,
                                     radius,
                                     height,
                                     path,
                                     needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleCylinder node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();
            float center_x = center[0];
            float center_y = center[1];
            float center_z = center[2];

            float axis_x = axis[0];
            float axis_y = axis[1];
            float axis_z = axis[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, center);
                transformNormal(invertedMatrix, axis);

                // need to scale the radius and height as well.
                float scale = invertedMatrix.getScale();
                radius *= scale;
                height *= scale;

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleCylinder((Group)n,
                                              pickType,
                                              center,
                                              axis,
                                              radius,
                                              height,
                                              path,
                                              needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleCylinder((SharedNode)n,
                                              pickType,
                                              center,
                                              axis,
                                              radius,
                                              height,
                                              path,
                                              needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleCylinder((Shape3D)n,
                                              pickType,
                                              center,
                                              axis,
                                              radius,
                                              height,
                                              path,
                                              needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            center[0] = center_x;
            center[1] = center_y;
            center[2] = center_z;

            axis[0] = axis_x;
            axis[1] = axis_y;
            axis[2] = axis_z;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private boolean pickSingleCylinder(Shape3D shape,
                                       int pickType,
                                       float[] center,
                                       float[] axis,
                                       float radius,
                                       float height,
                                       SceneGraphPath path,
                                       boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionCylinder(center, axis, radius, height))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();

            path.updatePath(pickPath,
                            lastPathIndex,
                            vworldMatrix,
                            invertedMatrix);

            found = true;
            lastPathIndex--;
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllCylinder(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            float x = req.origin[0] - req.destination[0];
            float y = req.origin[1] - req.destination[1];
            float z = req.origin[2] - req.destination[2];

            float height = (float)Math.sqrt(x * x + y * y + z * z);

            if(height == 0)
            {
                req.pickCount = 0;
                return;
            }

            // Start will be the center and end will be the axis vector
            start[0] = (req.origin[0] + req.destination[0]) * 0.5f;
            start[1] = (req.origin[1] + req.destination[0]) * 0.5f;
            start[2] = (req.origin[2] + req.destination[0]) * 0.5f;

            end[0] = x;
            end[1] = y;
            end[2] = z;

            float radius = req.additionalData;

            if(!bounds.checkIntersectionCylinder(start, end, radius, height))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {

                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transformNormal(invertedMatrix, end);

                        // need to scale the radius and height as well.
                        float scale = invertedMatrix.getScale();
                        radius *= scale;
                        height *= scale;
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found += pickAllCylinder((Group)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                radius,
                                                height,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllCylinder((SharedNode)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                radius,
                                                height,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllCylinder((Shape3D)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                radius,
                                                height,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                    }

                    req.pickCount = found;
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllCylinder((SharedNode)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    radius,
                                    height,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllCylinder((Shape3D)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    radius,
                                    height,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else
            {
System.out.println("pickAllCylinder not handling non-shape/group root proxy");
            }
        }
        else
        {
System.out.println("pickAllCylinder not handling non-node root");
        }
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and descend into
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllCylinder(Group root,
                               int pickType,
                               float[] center,
                               float[] axis,
                               float radius,
                               float height,
                               ArrayList paths,
                               int currentPath,
                               boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionCylinder(center, axis, radius, height))
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllCylinder((Group)target_node,
                                  pickType,
                                  center,
                                  axis,
                                  radius,
                                  height,
                                  paths,
                                  currentPath,
                                  needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllCylinder((SharedNode)target_node,
                                  pickType,
                                  center,
                                  axis,
                                  radius,
                                  height,
                                  paths,
                                  currentPath,
                                  needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllCylinder((Shape3D)target_node,
                                  pickType,
                                  center,
                                  axis,
                                  radius,
                                  height,
                                  paths,
                                  currentPath,
                                  needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickAllCylinder node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float center_x = center[0];
            float center_y = center[1];
            float center_z = center[2];

            float axis_x = axis[0];
            float axis_y = axis[1];
            float axis_z = axis[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, center);
                transformNormal(invertedMatrix, axis);

                // need to scale the radius and height as well.
                float scale = invertedMatrix.getScale();
                radius *= scale;
                height *= scale;

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllCylinder((Group)n,
                                        pickType,
                                        center,
                                        axis,
                                        radius,
                                        height,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllCylinder((SharedNode)n,
                                        pickType,
                                        center,
                                        axis,
                                        radius,
                                        height,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllCylinder((Shape3D)n,
                                        pickType,
                                        center,
                                        axis,
                                        radius,
                                        height,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            center[0] = center_x;
            center[1] = center_y;
            center[2] = center_z;

            axis[0] = axis_x;
            axis[1] = axis_y;
            axis[2] = axis_z;
        }

        return found;
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllCylinder(SharedNode root,
                                int pickType,
                                float[] center,
                                float[] axis,
                                float radius,
                                float height,
                                ArrayList paths,
                                int currentPath,
                                boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllCylinder((Group)child,
                                   pickType,
                                   center,
                                   axis,
                                   radius,
                                   height,
                                   paths,
                                   currentPath,
                                   needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllCylinder((SharedNode)child,
                                   pickType,
                                   center,
                                   axis,
                                   radius,
                                   height,
                                   paths,
                                   currentPath,
                                   needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllCylinder((Shape3D)child,
                                   pickType,
                                   center,
                                   axis,
                                   radius,
                                   height,
                                   paths,
                                   currentPath,
                                   needTransform);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param paths A place to set the results in
     * @param needTransform True if the minal to v-world transform needs
     *    calculating
     */
    private int pickAllCylinder(Shape3D shape,
                                int pickType,
                                float[] center,
                                float[] axis,
                                float radius,
                                float height,
                                ArrayList paths,
                                int currentPath,
                                boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return 0;

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionCylinder(center, axis, radius, height))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            SceneGraphPath p;

            if(currentPath >= paths.size())
            {
                p = new SceneGraphPath();
                paths.add(p);
            }
            else
                p = (SceneGraphPath)paths.get(currentPath);

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();


            p.updatePath(pickPath,
                         lastPathIndex,
                         vworldMatrix,
                         invertedMatrix);

            found = 1;
            lastPathIndex--;
        }

        return found;
    }

    // ----------------------- Cone Picking ------------------------------

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickCone(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
            case PickRequest.SORT_ORDERED:
                pickAllCone(root, req);
                break;

            case PickRequest.SORT_ANY:
            case PickRequest.SORT_CLOSEST:
                pickSingleCone(root, req);
                break;

            default:
System.out.println("No sort type has been specified");
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSingleCone(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];


        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();
            float angle = req.additionalData;

            if(!bounds.checkIntersectionCone(req.origin,
                                             req.destination,
                                             req.additionalData))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transformNormal(invertedMatrix, end);

                        // Angle does not need to be changed.
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleCone((Group)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  angle,
                                                  output_path,
                                                  req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleCone((SharedNode)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  angle,
                                                  output_path,
                                                  req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleCone((Shape3D)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  angle,
                                                  output_path,
                                                  req.generateVWorldMatrix);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleCone((SharedNode)target_node,
                                      req.pickType,
                                      start,
                                      end,
                                      angle,
                                      output_path,
                                      req.generateVWorldMatrix);
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleCone((Shape3D)target_node,
                                       req.pickType,
                                       start,
                                       end,
                                       angle,
                                       output_path,
                                       req.generateVWorldMatrix);

            }
            else
            {
System.out.println("pickSingleCone not handling non-shape/group root proxy");
            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleCone not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleCone(SharedNode root,
                                  int pickType,
                                  float[] vertex,
                                  float[] axis,
                                  float angle,
                                  SceneGraphPath path,
                                  boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleCone((Group)child,
                                  pickType,
                                  vertex,
                                  axis,
                                  angle,
                                  path,
                                  needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleCone((SharedNode)child,
                                 pickType,
                                 vertex,
                                 axis,
                                  angle,
                                 path,
                                 needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleCone((Shape3D)child,
                                 pickType,
                                 vertex,
                                 axis,
                                  angle,
                                 path,
                                 needTransform);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and recurse into
     * @param pickType flags to compare against for picking
     * @param vertex The verteximum extents of the box in local coordinate space
     * @param axis The axisimum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleCone(Group root,
                                  int pickType,
                                  float[] vertex,
                                  float[] axis,
                                  float angle,
                                  SceneGraphPath path,
                                  boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionCone(vertex, axis, angle))
            return false;

        // Bounding check. If it has it, recurse on the bounding geometry
        // rather than the real geometry.
        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleCone((Group)target_node,
                                      pickType,
                                      vertex,
                                      axis,
                                      angle,
                                      path,
                                      needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleCone((SharedNode)target_node,
                                      pickType,
                                      vertex,
                                      axis,
                                      angle,
                                      path,
                                      needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleCone((Shape3D)target_node,
                                      pickType,
                                      vertex,
                                      axis,
                                      angle,
                                      path,
                                      needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleCone node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float vertex_x = vertex[0];
            float vertex_y = vertex[1];
            float vertex_z = vertex[2];

            float axis_x = axis[0];
            float axis_y = axis[1];
            float axis_z = axis[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, vertex);
                transformNormal(invertedMatrix, axis);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleCone((Group)n,
                                          pickType,
                                          vertex,
                                          axis,
                                          angle,
                                          path,
                                          needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleCone((SharedNode)n,
                                          pickType,
                                          vertex,
                                          axis,
                                          angle,
                                          path,
                                          needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleCone((Shape3D)n,
                                          pickType,
                                          vertex,
                                          axis,
                                          angle,
                                          path,
                                          needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            vertex[0] = vertex_x;
            vertex[1] = vertex_y;
            vertex[2] = vertex_z;

            axis[0] = axis_x;
            axis[1] = axis_y;
            axis[2] = axis_z;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param vertex The verteximum extents of the box in local coordinate space
     * @param axis The axisimum extents of the box in local coordinate space
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private boolean pickSingleCone(Shape3D shape,
                                  int pickType,
                                  float[] vertex,
                                  float[] axis,
                                  float angle,
                                  SceneGraphPath path,
                                  boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionCone(vertex, axis, angle))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();

            path.updatePath(pickPath,
                            lastPathIndex,
                            vworldMatrix,
                            invertedMatrix);

            found = true;
            lastPathIndex--;
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllCone(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];


        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();
            float angle = req.additionalData;

            if(!bounds.checkIntersectionCone(req.origin,
                                             req.destination,
                                             req.additionalData))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transformNormal(invertedMatrix, end);

                        // keep angle unmodified.
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found += pickAllCone((Group)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                angle,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllCone((SharedNode)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                angle,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllCone((Shape3D)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                angle,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                    }

                    req.pickCount = found;
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllCone((SharedNode)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    angle,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllCone((Shape3D)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    angle,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else
            {
System.out.println("pickAllCone not handling non-shape/group root proxy");
            }
        }
        else
        {
System.out.println("pickAllCone not handling non-node root");
        }
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and descend into
     * @param pickType flags to compare against for picking
     * @param vertex The verteximum extents of the box in local coordinate space
     * @param axis The axisimum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllCone(Group root,
                           int pickType,
                           float[] vertex,
                           float[] axis,
                           float angle,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionCone(vertex, axis, angle))
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllCone((Group)target_node,
                                  pickType,
                                  vertex,
                                  axis,
                                  angle,
                                  paths,
                                  currentPath,
                                  needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllCone((SharedNode)target_node,
                                   pickType,
                                   vertex,
                                   axis,
                                   angle,
                                   paths,
                                   currentPath,
                                   needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllCone((Shape3D)target_node,
                                   pickType,
                                   vertex,
                                   axis,
                                   angle,
                                   paths,
                                   currentPath,
                                   needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickAllCone node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float vertex_x = vertex[0];
            float vertex_y = vertex[1];
            float vertex_z = vertex[2];

            float axis_x = axis[0];
            float axis_y = axis[1];
            float axis_z = axis[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, vertex);
                transformNormal(invertedMatrix, axis);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllCone((Group)n,
                                        pickType,
                                        vertex,
                                        axis,
                                        angle,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllCone((SharedNode)n,
                                        pickType,
                                        vertex,
                                        axis,
                                        angle,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllCone((Shape3D)n,
                                        pickType,
                                        vertex,
                                        axis,
                                        angle,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            vertex[0] = vertex_x;
            vertex[1] = vertex_y;
            vertex[2] = vertex_z;

            axis[0] = axis_x;
            axis[1] = axis_y;
            axis[2] = axis_z;
        }

        return found;
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param vertex The verteximum extents of the box in local coordinate space
     * @param axis The axisimum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllCone(SharedNode root,
                           int pickType,
                           float[] vertex,
                           float[] axis,
                           float angle,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllCone((Group)child,
                               pickType,
                               vertex,
                               axis,
                               angle,
                               paths,
                               currentPath,
                               needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllCone((SharedNode)child,
                               pickType,
                               vertex,
                               axis,
                               angle,
                               paths,
                               currentPath,
                               needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllCone((Shape3D)child,
                               pickType,
                               vertex,
                               axis,
                               angle,
                               paths,
                               currentPath,
                               needTransform);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param vertex The verteximum extents of the box in local coordinate space
     * @param axis The axisimum extents of the box in local coordinate space
     * @param paths A place to set the results in
     * @param needTransform True if the vertexal to v-world transform needs
     *    calculating
     */
    private int pickAllCone(Shape3D shape,
                           int pickType,
                           float[] vertex,
                           float[] axis,
                           float angle,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return 0;

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionCone(vertex, axis, angle))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            SceneGraphPath p;

            if(currentPath >= paths.size())
            {
                p = new SceneGraphPath();
                paths.add(p);
            }
            else
                p = (SceneGraphPath)paths.get(currentPath);

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();


            p.updatePath(pickPath,
                         lastPathIndex,
                         vworldMatrix,
                         invertedMatrix);

            found = 1;
            lastPathIndex--;
        }

        return found;
    }

    // ----------------------- Box Picking -------------------------------

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickBox(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
            case PickRequest.SORT_ORDERED:
                pickAllBox(root, req);
                break;

            case PickRequest.SORT_ANY:
            case PickRequest.SORT_CLOSEST:
                pickSingleBox(root, req);
                break;

            default:
System.out.println("No sort type has been specified");
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSingleBox(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];


        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionBox(req.origin, req.destination))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transform(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleBox((Group)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  output_path,
                                                  req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleBox((SharedNode)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  output_path,
                                                  req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleBox((Shape3D)kids[i],
                                                  req.pickType,
                                                  start,
                                                  end,
                                                  output_path,
                                                  req.generateVWorldMatrix);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleBox((SharedNode)target_node,
                                      req.pickType,
                                      start,
                                      end,
                                      output_path,
                                      req.generateVWorldMatrix);
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleBox((Shape3D)target_node,
                                       req.pickType,
                                       start,
                                       end,
                                       output_path,
                                       req.generateVWorldMatrix);

            }
            else
            {
System.out.println("pickSingleBox not handling non-shape/group root proxy");
            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleBox not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleBox(SharedNode root,
                                  int pickType,
                                  float[] min,
                                  float[] max,
                                  SceneGraphPath path,
                                  boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleBox((Group)child,
                                 pickType,
                                 min,
                                 max,
                                 path,
                                 needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleBox((SharedNode)child,
                                 pickType,
                                 min,
                                 max,
                                 path,
                                 needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleBox((Shape3D)child,
                                 pickType,
                                 min,
                                 max,
                                 path,
                                 needTransform);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and recurse into
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleBox(Group root,
                                  int pickType,
                                  float[] min,
                                  float[] max,
                                  SceneGraphPath path,
                                  boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionBox(min, max))
            return false;

        // Bounding check. If it has it, recurse on the bounding geometry
        // rather than the real geometry.
        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleBox((Group)target_node,
                                     pickType,
                                     min,
                                     max,
                                     path,
                                     needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleBox((SharedNode)target_node,
                                     pickType,
                                     min,
                                     max,
                                     path,
                                     needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleBox((Shape3D)target_node,
                                     pickType,
                                     min,
                                     max,
                                     path,
                                     needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleBox node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float min_x = min[0];
            float min_y = min[1];
            float min_z = min[2];

            float max_x = max[0];
            float max_y = max[1];
            float max_z = max[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, min);
                transform(invertedMatrix, max);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleBox((Group)n,
                                          pickType,
                                          min,
                                          max,
                                          path,
                                          needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleBox((SharedNode)n,
                                          pickType,
                                          min,
                                          max,
                                          path,
                                          needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleBox((Shape3D)n,
                                          pickType,
                                          min,
                                          max,
                                          path,
                                          needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            min[0] = min_x;
            min[1] = min_y;
            min[2] = min_z;

            max[0] = max_x;
            max[1] = max_y;
            max[2] = max_z;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private boolean pickSingleBox(Shape3D shape,
                                  int pickType,
                                  float[] min,
                                  float[] max,
                                  SceneGraphPath path,
                                  boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionBox(min, max))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();

            path.updatePath(pickPath,
                            lastPathIndex,
                            vworldMatrix,
                            invertedMatrix);

            found = true;
            lastPathIndex--;
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllBox(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        end[0] = req.destination[0];
        end[1] = req.destination[1];
        end[2] = req.destination[2];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionBox(req.origin, req.destination))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        transform(invertedMatrix, end);
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found += pickAllBox((Group)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllBox((SharedNode)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllBox((Shape3D)kids[i],
                                                req.pickType,
                                                start,
                                                end,
                                                output_list,
                                                found,
                                                req.generateVWorldMatrix);
                        }
                    }

                    req.pickCount = found;
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllBox((SharedNode)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllBox((Shape3D)target_node,
                                    req.pickType,
                                    start,
                                    end,
                                    output_list,
                                    0,
                                    req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else
            {
System.out.println("pickAllBox not handling non-shape/group root proxy");
            }
        }
        else
        {
System.out.println("pickAllBox not handling non-node root");
        }
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and descend into
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllBox(Group root,
                           int pickType,
                           float[] min,
                           float[] max,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionBox(min, max))
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllBox((Group)target_node,
                                  pickType,
                                  min,
                                  max,
                                  paths,
                                  currentPath,
                                  needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllBox((SharedNode)target_node,
                                  pickType,
                                  min,
                                  max,
                                  paths,
                                  currentPath,
                                  needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllBox((Shape3D)target_node,
                                  pickType,
                                  min,
                                  max,
                                  paths,
                                  currentPath,
                                  needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickAllBox node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float min_x = min[0];
            float min_y = min[1];
            float min_z = min[2];

            float max_x = max[0];
            float max_y = max[1];
            float max_z = max[2];

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, min);
                transform(invertedMatrix, max);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllBox((Group)n,
                                        pickType,
                                        min,
                                        max,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllBox((SharedNode)n,
                                        pickType,
                                        min,
                                        max,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllBox((Shape3D)n,
                                        pickType,
                                        min,
                                        max,
                                        paths,
                                        currentPath + found,
                                        needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            min[0] = min_x;
            min[1] = min_y;
            min[2] = min_z;

            max[0] = max_x;
            max[1] = max_y;
            max[2] = max_z;
        }

        return found;
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllBox(SharedNode root,
                           int pickType,
                           float[] min,
                           float[] max,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllBox((Group)child,
                              pickType,
                              min,
                              max,
                              paths,
                              currentPath,
                              needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllBox((SharedNode)child,
                              pickType,
                              min,
                              max,
                              paths,
                              currentPath,
                              needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllBox((Shape3D)child,
                              pickType,
                              min,
                              max,
                              paths,
                              currentPath,
                              needTransform);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param max The maximum extents of the box in local coordinate space
     * @param paths A place to set the results in
     * @param needTransform True if the minal to v-world transform needs
     *    calculating
     */
    private int pickAllBox(Shape3D shape,
                           int pickType,
                           float[] min,
                           float[] max,
                           ArrayList paths,
                           int currentPath,
                           boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return 0;

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionBox(min, max))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            SceneGraphPath p;

            if(currentPath >= paths.size())
            {
                p = new SceneGraphPath();
                paths.add(p);
            }
            else
                p = (SceneGraphPath)paths.get(currentPath);

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();


            p.updatePath(pickPath,
                         lastPathIndex,
                         vworldMatrix,
                         invertedMatrix);

            found = 1;
            lastPathIndex--;
        }

        return found;
    }

    // ----------------------- Frustum Picking ------------------------------

    // Implementation Notes:
    // Unlike the other forms of picking, it is more efficient for us to
    // provide the v-world to local transformation to the picked object rather
    // than the other way around. This saves on a lot of matrix-vector
    // multiplications. Because of this, the transformPath is used differently.
    // Instead of just holding the local transformation, what this is doing is
    // premultiplying the matrix for the world transform as we walk the stack.
    //
    // Also, the frustum planes are just referenced as the global variables
    // rather than using local parameter copies. Need to monitor if this causes
    // any multithreaded issues or not.

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickFrustum(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
            case PickRequest.SORT_ORDERED:
                pickAllFrustum(root, req);
                break;

            case PickRequest.SORT_ANY:
            case PickRequest.SORT_CLOSEST:
                pickSingleFrustum(root, req);
                break;

            default:
System.out.println("No sort type has been specified");
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickSingleFrustum(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        frustumPlanes[0].x = req.origin[0];
        frustumPlanes[0].y = req.origin[1];
        frustumPlanes[0].z = req.origin[2];
        frustumPlanes[0].w = req.origin[3];

        frustumPlanes[1].x = req.origin[4];
        frustumPlanes[1].y = req.origin[5];
        frustumPlanes[1].z = req.origin[6];
        frustumPlanes[1].w = req.origin[7];

        frustumPlanes[2].x = req.origin[8];
        frustumPlanes[2].y = req.origin[9];
        frustumPlanes[2].z = req.origin[10];
        frustumPlanes[2].w = req.origin[11];

        frustumPlanes[3].x = req.origin[12];
        frustumPlanes[3].y = req.origin[13];
        frustumPlanes[3].z = req.origin[14];
        frustumPlanes[3].w = req.origin[15];

        frustumPlanes[4].x = req.origin[16];
        frustumPlanes[4].y = req.origin[17];
        frustumPlanes[4].z = req.origin[18];
        frustumPlanes[4].w = req.origin[19];

        frustumPlanes[5].x = req.origin[20];
        frustumPlanes[5].y = req.origin[21];
        frustumPlanes[5].z = req.origin[22];
        frustumPlanes[5].w = req.origin[23];

        boolean found = false;

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();
            transformPath[0].setIdentity();

            if(bounds.checkIntersectionFrustum(frustumPlanes,
                                               transformPath[0]) ==
                                               BoundingVolume.FRUSTUM_ALLOUT)
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                        ((TransformGroup)target_node).getTransform(transformPath[0]);

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleFrustum((Group)kids[i],
                                                      req.pickType,
                                                      output_path,
                                                      req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleFrustum((SharedNode)kids[i],
                                                      req.pickType,
                                                      output_path,
                                                      req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleFrustum((Shape3D)kids[i],
                                                      req.pickType,
                                                      output_path,
                                                      req.generateVWorldMatrix);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleFrustum((SharedNode)target_node,
                                          req.pickType,
                                          output_path,
                                          req.generateVWorldMatrix);

            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleFrustum((Shape3D)target_node,
                                          req.pickType,
                                          output_path,
                                          req.generateVWorldMatrix);

            }
            else
            {
System.out.println("pickSingleFrustum not handling non-shape/group root proxy");
            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleFrustum not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a frustum. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleFrustum(SharedNode root,
                                      int pickType,
                                      SceneGraphPath path,
                                      boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleFrustum((Group)child,
                                     pickType,
                                     path,
                                     needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleFrustum((SharedNode)child,
                                     pickType,
                                     path,
                                     needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleFrustum((Shape3D)child,
                                     pickType,
                                     path,
                                     needTransform);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group to test against
     * @param pickType flags to compare against for picking
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleFrustum(Group root,
                                      int pickType,
                                      SceneGraphPath path,
                                      boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(bounds.checkIntersectionFrustum(frustumPlanes,
                                           transformPath[lastPathIndex]) ==
                                           BoundingVolume.FRUSTUM_ALLOUT)
            return false;

        // Bounding check. If it has it, recurse on the bounding geometry
        // rather than the real geometry.
        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleFrustum((Group)target_node,
                                         pickType,
                                         path,
                                         needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleFrustum((SharedNode)target_node,
                                         pickType,
                                         path,
                                         needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleFrustum((Shape3D)target_node,
                                         pickType,
                                         path,
                                         needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleFrustum node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                tx.mul(transformPath[lastPathIndex - 1], tx);

                validTransform[lastPathIndex] = true;
            }
            else
            {
                Matrix4f tx = transformPath[lastPathIndex - 1];
                transformPath[lastPathIndex].set(tx);
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;
            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleFrustum((Group)n,
                                              pickType,
                                              path,
                                              needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleFrustum((SharedNode)n,
                                              pickType,
                                              path,
                                              needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleFrustum((Shape3D)n,
                                              pickType,
                                              path,
                                              needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private boolean pickSingleFrustum(Shape3D shape,
                                      int pickType,
                                      SceneGraphPath path,
                                      boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionFrustum(frustumPlanes,
                                           transformPath[lastPathIndex-1]) !=
                                           BoundingVolume.FRUSTUM_ALLOUT)
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            // Note that the transform path is already pre-multiplied through
            // here. Just and invert is all that needs to be done.
            if(needTransform)
                matrixUtils.inverse(transformPath[lastPathIndex - 2],
                                    vworldMatrix);
            else
                vworldMatrix.setIdentity();

            path.updatePath(pickPath,
                            lastPathIndex,
                            transformPath[lastPathIndex - 2],
                            vworldMatrix);

            found = true;
            lastPathIndex--;
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickAllFrustum(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        frustumPlanes[0].x = req.origin[0];
        frustumPlanes[0].y = req.origin[1];
        frustumPlanes[0].z = req.origin[2];
        frustumPlanes[0].w = req.origin[3];

        frustumPlanes[1].x = req.origin[4];
        frustumPlanes[1].y = req.origin[5];
        frustumPlanes[1].z = req.origin[6];
        frustumPlanes[1].w = req.origin[7];

        frustumPlanes[2].x = req.origin[8];
        frustumPlanes[2].y = req.origin[9];
        frustumPlanes[2].z = req.origin[10];
        frustumPlanes[2].w = req.origin[11];

        frustumPlanes[3].x = req.origin[12];
        frustumPlanes[3].y = req.origin[13];
        frustumPlanes[3].z = req.origin[14];
        frustumPlanes[3].w = req.origin[15];

        frustumPlanes[4].x = req.origin[16];
        frustumPlanes[4].y = req.origin[17];
        frustumPlanes[4].z = req.origin[18];
        frustumPlanes[4].w = req.origin[19];

        frustumPlanes[5].x = req.origin[20];
        frustumPlanes[5].y = req.origin[21];
        frustumPlanes[5].z = req.origin[22];
        frustumPlanes[5].w = req.origin[23];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();
            transformPath[0].setIdentity();

            if(bounds.checkIntersectionFrustum(frustumPlanes,
                                               transformPath[0]) ==
                                               BoundingVolume.FRUSTUM_ALLOUT)
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                        ((TransformGroup)target_node).getTransform(transformPath[0]);

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;

                        if(kids[i] instanceof Group)
                        {
                            found += pickAllFrustum((Group)kids[i],
                                                    req.pickType,
                                                    output_list,
                                                    found,
                                                    req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllFrustum((SharedNode)kids[i],
                                                    req.pickType,
                                                    output_list,
                                                    found,
                                                    req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllFrustum((Shape3D)kids[i],
                                                    req.pickType,
                                                    output_list,
                                                    found,
                                                    req.generateVWorldMatrix);
                        }
                    }

                    req.pickCount = found;
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllFrustum((SharedNode)target_node,
                                        req.pickType,
                                        output_list,
                                        0,
                                        req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllFrustum((Shape3D)target_node,
                                        req.pickType,
                                        output_list,
                                        0,
                                        req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else
            {
System.out.println("pickAllFrustum not handling non-shape/group root proxy");
            }
        }
        else
        {
System.out.println("pickAllFrustum not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a frustum. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param paths Array of paths place to set the results in
     * @param currentPath Active index in the paths array
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllFrustum(SharedNode root,
                               int pickType,
                               ArrayList paths,
                               int currentPath,
                               boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllFrustum((Group)child,
                                  pickType,
                                  paths,
                                  currentPath,
                                  needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllFrustum((SharedNode)child,
                                  pickType,
                                  paths,
                                  currentPath,
                                  needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllFrustum((Shape3D)child,
                                  pickType,
                                  paths,
                                  currentPath,
                                  needTransform);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param paths The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllFrustum(Group root,
                               int pickType,
                               ArrayList paths,
                               int currentPath,
                               boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(bounds.checkIntersectionFrustum(frustumPlanes,
                                           transformPath[lastPathIndex-1]) ==
                                           BoundingVolume.FRUSTUM_ALLOUT)
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllFrustum((Group)target_node,
                                      pickType,
                                      paths,
                                      currentPath,
                                      needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllFrustum((SharedNode)target_node,
                                      pickType,
                                      paths,
                                      currentPath,
                                      needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllFrustum((Shape3D)target_node,
                                      pickType,
                                      paths,
                                      currentPath,
                                      needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickAllFrustum node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                tx.mul(transformPath[lastPathIndex - 1], tx);
                validTransform[lastPathIndex] = true;
            }
            else
            {
                Matrix4f tx = transformPath[lastPathIndex - 1];
                transformPath[lastPathIndex].set(tx);
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;


            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllFrustum((Group)n,
                                            pickType,
                                            paths,
                                            currentPath + found,
                                            needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllFrustum((SharedNode)n,
                                            pickType,
                                            paths,
                                            currentPath + found,
                                            needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllFrustum((Shape3D)n,
                                            pickType,
                                            paths,
                                            currentPath + found,
                                            needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param pickType flags to compare against for picking
     * @param paths Array of paths place to set the results in
     * @param currentPath Active index in the paths array
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private int pickAllFrustum(Shape3D shape,
                               int pickType,
                               ArrayList paths,
                               int currentPath,
                               boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return 0;

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionFrustum(frustumPlanes,
                                           transformPath[lastPathIndex-1]) !=
                                           BoundingVolume.FRUSTUM_ALLOUT)
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            SceneGraphPath p;

            if(currentPath >= paths.size())
            {
                p = new SceneGraphPath();
                paths.add(p);
            }
            else
                p = (SceneGraphPath)paths.get(currentPath);

            // Note that the transform path is already pre-multiplied through
            // here. Just and invert is all that needs to be done.
            if(needTransform)
                matrixUtils.inverse(transformPath[lastPathIndex - 2],
                                    vworldMatrix);
            else
                vworldMatrix.setIdentity();

            p.updatePath(pickPath,
                         lastPathIndex,
                         transformPath[lastPathIndex - 2],
                         vworldMatrix);

            found = 1;
            lastPathIndex--;
        }

        return found;
    }

    // ----------------------- Sphere Picking -------------------------------

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     * @return The number of intersections found
     */
    private void pickSphere(PickableObject root, PickRequest req)
    {
        switch(req.pickSortType)
        {
            case PickRequest.SORT_ALL:
            case PickRequest.SORT_ORDERED:
                pickAllSphere(root, req);
                break;

            case PickRequest.SORT_ANY:
            case PickRequest.SORT_CLOSEST:
                pickSingleSphere(root, req);
                break;

            default:
System.out.println("No sort type has been specified");
        }
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickSingleSphere(PickableObject root, PickRequest req)
    {
        SceneGraphPath output_path = null;

        if(req.foundPaths instanceof SceneGraphPath)
            output_path = (SceneGraphPath)req.foundPaths;
        else
        {
            output_path = new SceneGraphPath();
            req.foundPaths = output_path;
        }

        boolean found = false;

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];


        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionSphere(req.origin, req.additionalData))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    float radius = req.additionalData;

                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        radius *= invertedMatrix.getScale();
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids && !found; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found = pickSingleSphere((Group)kids[i],
                                                     req.pickType,
                                                     start,
                                                     radius,
                                                     output_path,
                                                     req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found = pickSingleSphere((SharedNode)kids[i],
                                                     req.pickType,
                                                     start,
                                                     radius,
                                                     output_path,
                                                     req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found = pickSingleSphere((Shape3D)kids[i],
                                                     req.pickType,
                                                     start,
                                                     radius,
                                                     output_path,
                                                     req.generateVWorldMatrix);
                        }
                    }
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found = pickSingleSphere((SharedNode)target_node,
                                         req.pickType,
                                         start,
                                         req.additionalData,
                                         output_path,
                                         req.generateVWorldMatrix);
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found = pickSingleSphere((Shape3D)target_node,
                                         req.pickType,
                                         start,
                                         req.additionalData,
                                         output_path,
                                         req.generateVWorldMatrix);

            }
            else
            {
System.out.println("pickSingleSphere not handling non-shape/group root proxy");
            }

            req.pickCount = found ? 1 : 0;
        }
        else
        {
System.out.println("pickSingleSphere not handling non-node root");
        }
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param radius The radius of the sphere
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleSphere(SharedNode root,
                                     int pickType,
                                     float[] min,
                                     float radius,
                                     SceneGraphPath path,
                                     boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickSingleSphere((Group)child,
                                    pickType,
                                    min,
                                    radius,
                                    path,
                                    needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickSingleSphere((SharedNode)child,
                                    pickType,
                                    min,
                                    radius,
                                    path,
                                    needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickSingleSphere((Shape3D)child,
                                    pickType,
                                    min,
                                    radius,
                                    path,
                                    needTransform);
        }

        return false;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and recurse into
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param radius The radius of the sphere
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private boolean pickSingleSphere(Group root,
                                     int pickType,
                                     float[] min,
                                     float radius,
                                     SceneGraphPath path,
                                     boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return false;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionSphere(min, radius))
            return false;

        // Bounding check. If it has it, recurse on the bounding geometry
        // rather than the real geometry.
        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickSingleSphere((Group)target_node,
                                        pickType,
                                        min,
                                        radius,
                                        path,
                                        needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickSingleSphere((SharedNode)target_node,
                                        pickType,
                                        min,
                                        radius,
                                        path,
                                        needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickSingleSphere((Shape3D)target_node,
                                        pickType,
                                        min,
                                        radius,
                                        path,
                                        needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickSingleSphere node handling oddball proxy geom");
            }
        }

        boolean found = false;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float min_x = min[0];
            float min_y = min[1];
            float min_z = min[2];

            float old_radius = radius;

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, min);

                radius *= invertedMatrix.getScale();

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;

            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids && !found; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found = pickSingleSphere((Group)n,
                                             pickType,
                                             min,
                                             radius,
                                             path,
                                             needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found = pickSingleSphere((SharedNode)n,
                                             pickType,
                                             min,
                                             radius,
                                             path,
                                             needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found = pickSingleSphere((Shape3D)n,
                                             pickType,
                                             min,
                                             radius,
                                             path,
                                             needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            min[0] = min_x;
            min[1] = min_y;
            min[2] = min_z;

            radius = old_radius;
        }

        return found;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param radius The radius of the sphere
     * @param path A place to set the results in
     * @param needTransform True if the local to v-world transform needs
     *    calculating
     */
    private boolean pickSingleSphere(Shape3D shape,
                                     int pickType,
                                     float[] min,
                                     float radius,
                                     SceneGraphPath path,
                                     boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return false;

        boolean found = false;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionSphere(min, radius))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();

            path.updatePath(pickPath,
                            lastPathIndex,
                            vworldMatrix,
                            invertedMatrix);

            lastPathIndex--;
            found = true;
        }

        return found;
    }

    /**
     * Check for all intersections against this geometry and it's children to
     * see if there is an intersection with the given ray. Return the first
     * found.
     *
     * @param root The root point to start the pick processing from
     * @param req The list of picks to be made, starting at this object
     */
    private void pickAllSphere(PickableObject root, PickRequest req)
    {
        int found = 0;
        ArrayList output_list = null;

        if(req.foundPaths instanceof ArrayList)
            output_list = (ArrayList)req.foundPaths;
        else
        {
            output_list = new ArrayList();
            req.foundPaths = output_list;
        }

        start[0] = req.origin[0];
        start[1] = req.origin[1];
        start[2] = req.origin[2];

        if(root instanceof Node)
        {
            Node target_node = (Node)root;
            BoundingVolume bounds = target_node.getBounds();

            if(!bounds.checkIntersectionSphere(req.origin, req.additionalData))
            {
                req.pickCount = 0;
                return;
            }

            if(bounds instanceof BoundingGeometry)
            {
                BoundingGeometry bg = (BoundingGeometry)bounds;
                target_node = bg.getProxyGeometry();
            }

            if(target_node instanceof Group)
            {
                Group g = (Group)target_node;
                int num_kids = g.numRenderableChildren();

                if(num_kids != 0)
                {
                    float radius = req.additionalData;

                    // reset the transform at the top of the stack
                    if(target_node instanceof TransformGroup)
                    {
                        ((TransformGroup)target_node).getTransform(transformPath[0]);
                        matrixUtils.inverse(transformPath[0], invertedMatrix);
                        transform(invertedMatrix, start);
                        radius *= invertedMatrix.getScale();
                    }
                    else
                        transformPath[0].setIdentity();

                    validTransform[0] = true;
                    pickPath[0] = g;
                    Node[] kids = g.getRenderableChild();

                    // now that the setup is done, walk down the tree.
                    for(int i = 0; i < num_kids; i++)
                    {
                        lastPathIndex = 1;
                        if(kids[i] instanceof Group)
                        {
                            found += pickAllSphere((Group)kids[i],
                                                   req.pickType,
                                                   start,
                                                   radius,
                                                   output_list,
                                                   found,
                                                   req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof SharedNode)
                        {
                            found += pickAllSphere((SharedNode)kids[i],
                                                   req.pickType,
                                                   start,
                                                   radius,
                                                   output_list,
                                                   found,
                                                   req.generateVWorldMatrix);
                        }
                        else if(kids[i] instanceof Shape3D)
                        {
                            found += pickAllSphere((Shape3D)kids[i],
                                                   req.pickType,
                                                   start,
                                                   radius,
                                                   output_list,
                                                   found,
                                                   req.generateVWorldMatrix);
                        }
                    }

                    req.pickCount = found;
                }
            }
            else if(target_node instanceof SharedNode)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (SharedNode)target_node;
                lastPathIndex = 1;

                found += pickAllSphere((SharedNode)target_node,
                                       req.pickType,
                                       start,
                                       req.additionalData,
                                       output_list,
                                       0,
                                       req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else if(target_node instanceof Shape3D)
            {
                transformPath[0].setIdentity();
                validTransform[0] = true;
                pickPath[0] = (Shape3D)target_node;
                lastPathIndex = 1;

                found += pickAllSphere((Shape3D)target_node,
                                       req.pickType,
                                       start,
                                       req.additionalData,
                                       output_list,
                                       0,
                                       req.generateVWorldMatrix);

                req.pickCount = found;
            }
            else
            {
System.out.println("pickAllSphere not handling non-shape/group root proxy");
            }
        }
        else
        {
System.out.println("pickAllSphere not handling non-node root");
        }
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The group node to test against and descend into
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param radius The radius of the sphere
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllSphere(Group root,
                              int pickType,
                              float[] min,
                              float radius,
                              ArrayList paths,
                              int currentPath,
                              boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        BoundingVolume bounds = root.getBounds();

        if(!bounds.checkIntersectionSphere(min, radius))
            return 0;

        if(bounds instanceof BoundingGeometry)
        {
            BoundingGeometry bg = (BoundingGeometry)bounds;
            Node target_node = bg.getProxyGeometry();

            if(target_node instanceof Group)
            {
                return pickAllSphere((Group)target_node,
                                     pickType,
                                     min,
                                     radius,
                                     paths,
                                     currentPath,
                                     needTransform);
            }
            else if(target_node instanceof SharedNode)
            {
                return pickAllSphere((SharedNode)target_node,
                                     pickType,
                                     min,
                                     radius,
                                     paths,
                                     currentPath,
                                     needTransform);
            }
            else if(target_node instanceof Shape3D)
            {
                return pickAllSphere((Shape3D)target_node,
                                     pickType,
                                     min,
                                     radius,
                                     paths,
                                     currentPath,
                                     needTransform);
            }
            else if(target_node != null)
            {
System.out.println("pickAllSphere node handling oddball proxy geom");
            }
        }

        int found = 0;
        int num_kids = root.numRenderableChildren();

        if(num_kids != 0)
        {
            resizePath();

            float min_x = min[0];
            float min_y = min[1];
            float min_z = min[2];

            float old_radius = radius;

            // reset the transform at the top of the stack
            if(root instanceof TransformGroup)
            {
                Matrix4f tx = transformPath[lastPathIndex];
                ((TransformGroup)root).getTransform(tx);
                matrixUtils.inverse(tx, invertedMatrix);
                transform(invertedMatrix, min);
                radius *= invertedMatrix.getScale();

                validTransform[lastPathIndex] = true;
            }
            else
            {
                // don't bother setting to identity, just mark it as non-valid.
                // Assignment is a waste of CPU cycles.
                validTransform[lastPathIndex] = false;
            }

            pickPath[lastPathIndex] = root;
            lastPathIndex++;

            Node[] kids = root.getRenderableChild();

            for(int i = 0; i < num_kids; i++)
            {
                Node n = kids[i];

                if(n == null)
                    continue;

                if(n instanceof Group)
                {
                    found += pickAllSphere((Group)n,
                                           pickType,
                                           min,
                                           radius,
                                           paths,
                                           currentPath + found,
                                           needTransform);
                }
                else if(n instanceof SharedNode)
                {
                    found += pickAllSphere((SharedNode)n,
                                           pickType,
                                           min,
                                           radius,
                                           paths,
                                           currentPath + found,
                                           needTransform);
                }
                else if(n instanceof Shape3D)
                {
                    found += pickAllSphere((Shape3D)n,
                                           pickType,
                                           min,
                                           radius,
                                           paths,
                                           currentPath + found,
                                           needTransform);
                }
            }

            lastPathIndex--;
            pickPath[lastPathIndex] = null;

            min[0] = min_x;
            min[1] = min_y;
            min[2] = min_z;

            radius = old_radius;
        }

        return found;
    }

    /**
     * Recurse a single-child shared node tree looking for
     * intersections with a single point. Shared nodes are different to normal
     * because we don't bother doing bounds tests as the bounds will be
     * identical to their child. Skip the test and just head straight to the
     * child node to test against.
     *
     * @param root The shared node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param radius The radius of the sphere
     * @param path The place to put the results in
     * @param needTransform true if we should calc vworld information
     */
    private int pickAllSphere(SharedNode root,
                              int pickType,
                              float[] min,
                              float radius,
                              ArrayList paths,
                              int currentPath,
                              boolean needTransform)
    {
        if((root.getPickMask() & pickType) == 0)
            return 0;

        Node child = root.getChild();

        if(child instanceof Group)
        {
            return pickAllSphere((Group)child,
                                 pickType,
                                 min,
                                 radius,
                                 paths,
                                 currentPath,
                                 needTransform);
        }
        else if(child instanceof SharedNode)
        {
            return pickAllSphere((SharedNode)child,
                                 pickType,
                                 min,
                                 radius,
                                 paths,
                                 currentPath,
                                 needTransform);
        }
        else if(child instanceof Shape3D)
        {
            return pickAllSphere((Shape3D)child,
                                 pickType,
                                 min,
                                 radius,
                                 paths,
                                 currentPath,
                                 needTransform);
        }

        return 0;
    }

    /**
     * Recurse the tree looking for intersections with a single point.
     *
     * @param root The shape node to test against
     * @param pickType flags to compare against for picking
     * @param min The minimum extents of the box in local coordinate space
     * @param radius The radius of the sphere
     * @param paths A place to set the results in
     * @param needTransform True if the minal to v-world transform needs
     *    calculating
     */
    private int pickAllSphere(Shape3D shape,
                              int pickType,
                              float[] min,
                              float radius,
                              ArrayList paths,
                              int currentPath,
                              boolean needTransform)
    {
        if((shape.getPickMask() & pickType) == 0)
            return 0;

        int found = 0;
        BoundingVolume bounds = shape.getBounds();

        if(bounds.checkIntersectionSphere(min, radius))
        {
            resizePath();
            pickPath[lastPathIndex] = shape;
            validTransform[lastPathIndex] = false;
            lastPathIndex++;

            SceneGraphPath p;

            if(currentPath >= paths.size())
            {
                p = new SceneGraphPath();
                paths.add(p);
            }
            else
                p = (SceneGraphPath)paths.get(currentPath);

            if(needTransform)
                buildVWorldTransform();
            else
                vworldMatrix.setIdentity();


            p.updatePath(pickPath,
                         lastPathIndex,
                         vworldMatrix,
                         invertedMatrix);

            lastPathIndex--;
            found = 1;
        }

        return found;
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizePath()
    {
        if(lastPathIndex == pickPath.length)
        {
            int old_size = pickPath.length;
            int new_size = old_size + LIST_INCREMENT;

            Node[] tmp_nodes = new Node[new_size];

            System.arraycopy(pickPath, 0, tmp_nodes, 0, old_size);

            pickPath = tmp_nodes;

            Matrix4f[] tmp_tx = new Matrix4f[new_size];
            System.arraycopy(transformPath, 0, tmp_tx, 0, old_size);
            transformPath = tmp_tx;

            for(int i = old_size; i < new_size; i++)
                transformPath[i] = new Matrix4f();

            boolean[] tmp_flags = new boolean[new_size];
            System.arraycopy(validTransform, 0, tmp_flags, 0, old_size);
            validTransform = tmp_flags;
        }
    }

    /**
     * Transform the location vector by the matrix.
     *
     * @param mat The matrix to do the transformation with
     * @param vec The vector to be changed
     */
    private void transform(Matrix4f mat, float[] vec)
    {
        wkVec.set(vec);
        mat.transform(wkVec);
        wkVec.get(vec);
    }

    /**
     * Transform the normal vector by the matrix.
     *
     * @param mat The matrix to do the transformation with
     * @param vec The vector to be changed
     */
    private void transformNormal(Matrix4f mat, float[] vec)
    {
        wkNormal.set(vec);
        mat.transform(wkNormal);
        wkNormal.get(vec);
    }

    /**
     * Walk the transform stack and compute the total transform to get from
     * the world root to here. The forward matrix is placed in vworldMatrix
     * and the inverse is placed in invertedMatrix. Both are class-scope
     * variables.
     */
    private void buildVWorldTransform()
    {
        int last_valid = 0;
        vworldMatrix.set(transformPath[0]);

        for(int i = 1; i < lastPathIndex; i++)
        {
            if(!validTransform[i])
                continue;

            vworldMatrix.mul(transformPath[i]);
        }

        matrixUtils.inverse(vworldMatrix, invertedMatrix);
    }
}
